/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.bridge;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.hook.AtlasHook;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.notification.HookNotification;
import org.apache.atlas.type.AtlasType;
import org.apache.commons.collections.CollectionUtils;
import org.apache.hadoop.security.UserGroupInformation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Collections;
import java.util.List;

import static org.apache.atlas.repository.Constants.AWSS3_SOURCE;

/**
 * Aws S3 hook to facilitate communication with Atlas through Kafka channel.
 */
public class AwsS3Hook extends AtlasHook {
    private static final Logger    LOG      = LoggerFactory.getLogger(AwsS3Hook.class);
    private static final AwsS3Hook INSTANCE = new AwsS3Hook();

    private UserGroupInformation userGroupInformation;

    public static AwsS3Hook getInstance() {
        return INSTANCE;
    }

    private AwsS3Hook() {
    }

    @Override
    public String getMessageSource() {
        return AWSS3_SOURCE;
    }

    public void initialize() throws AtlasBaseException {
        try {
            userGroupInformation = UserGroupInformation.getCurrentUser();
        } catch (IOException e) {
            String msg = "Error encountered in getting the user information.";

            LOG.error(msg, e);

            throw new AtlasBaseException(msg, e);
        }

        LOG.info("Initialization completed successfully.");
    }

    public void publishEvent(AtlasAwsS3Event s3Event) throws Exception {
        switch (s3Event.getEventType()) {
            case CREATE:
                publishCreateEventNotification((AtlasAwsS3CreateEvent) s3Event);
                break;
            case DELETE:
                publishDeleteEventNotification((AtlasAwsS3DeleteEvent) s3Event);
                break;
            default:
                throw new AtlasBaseException("Unknown event type received. EventType: " + s3Event.getEventType());
        }
    }

    private void publishDeleteEventNotification(AtlasAwsS3DeleteEvent deleteEvent) {
        List<HookNotification> notificationMessages = null;
        List<AtlasObjectId>    objectIdsForDeletion = deleteEvent.getObjectIdsForDeletion();

        if (objectIdsForDeletion != null && CollectionUtils.isNotEmpty(objectIdsForDeletion)) {
            notificationMessages = Collections.singletonList(new HookNotification.EntityDeleteRequestV2(deleteEvent.getUserName(), objectIdsForDeletion));
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("Message sent is {}", AtlasType.toJson(notificationMessages));
        }

        super.notifyEntities(notificationMessages, userGroupInformation);
    }

    private void publishCreateEventNotification(AtlasAwsS3CreateEvent createEvent) {
        List<HookNotification>               notificationMessages = null;
        AtlasEntity.AtlasEntitiesWithExtInfo entitiesWithExtInfo  = createEvent.getEntitiesWithExtInfo();

        if (entitiesWithExtInfo != null && CollectionUtils.isNotEmpty(entitiesWithExtInfo.getEntities())) {
            notificationMessages = Collections.singletonList(new HookNotification.EntityCreateRequestV2(createEvent.getUserName(), entitiesWithExtInfo));
        }

        super.notifyEntities(notificationMessages, userGroupInformation);
    }
}