/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.s3.bridge;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.AtlasBaseClient;
import org.apache.atlas.s3.bridge.models.Credentials;
import org.apache.atlas.utils.AuthenticationUtil;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.HttpMethod;
import javax.ws.rs.core.Response;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.io.Serializable;

public class IDBrokerHandler implements Serializable {
    private static final long   serialVersionUID                  = 1L;
    private static final Logger LOG                               = LoggerFactory.getLogger(IDBrokerHandler.class);
    private static final String CONFIG_CAB_ADDRESS_FORMAT         = "fs.%s.ext.cab.address"; // http://host:8444/gateway
    private static final String CONFIG_CAB_ADDRESS_PATH_FORMAT    = "fs.%s.ext.cab.path"; //aws-cab
    private static final String API_CAB_CREDENTIALS_GLOBAL        = "/cab/api/v1/credentials";
    private static final char   CONFIG_PATH_SEPARATOR             = ',';

    private final Properties coreSiteConfiguration;
    private final String     fileSystemScheme;

    private List<AtlasIDBrokerClient> idBrokerClients = new LinkedList<>();
    private Credentials               awsS3Credential;

    public IDBrokerHandler(String fileSystemScheme, Properties coreSiteConfiguration) throws AtlasBaseException {
        this.fileSystemScheme      = fileSystemScheme;
        this.coreSiteConfiguration = coreSiteConfiguration;

        String idBrokerAddresses = getConfigCabAddress();
        if (StringUtils.isNotBlank(idBrokerAddresses)) {
            if (!AuthenticationUtil.isKerberosAuthenticationEnabled()) {
                throw new AtlasBaseException("No Kerberos authentication enabled");
            }

            try {
                for (String eachIDBrokerAddress : StringUtils.split(idBrokerAddresses, CONFIG_PATH_SEPARATOR)) {
                    eachIDBrokerAddress = StringUtils.trim(eachIDBrokerAddress);
                    idBrokerClients.add(new AtlasIDBrokerClient(eachIDBrokerAddress));
                }
            } catch (Exception e) {
                throw new AtlasBaseException("Failed to initialize IdBroker Client", e);
            }
        }
    }

    public void fetchAwsS3CredentialIfNeeded() throws AtlasBaseException {
        for (AtlasIDBrokerClient idBrokerClient : idBrokerClients) {
            try {
                if (isCredentialExpired()) {
                    AtlasBaseClient.API getAwsS3CredentialAPI = new AtlasBaseClient.API(getConfigCabAddressPath() + API_CAB_CREDENTIALS_GLOBAL, HttpMethod.GET, Response.Status.OK);

                    awsS3Credential = idBrokerClient.getAwsS3Credential(getAwsS3CredentialAPI);
                    LOG.info("Successfully fetched AWS credential from IDBroker {}", idBrokerClient.getResource().getURI().toString());
                }

                validateAwsS3Credential();

                return;
            } catch (Exception e) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Error occurred for IDBroker {}", idBrokerClient.getResource().getURI().toString(), e);
                } else {
                    LOG.warn("Error occurred for IDBroker {} : {}", idBrokerClient.getResource().getURI().toString(), e.getMessage());
                }
            }
        }

        throw new AtlasBaseException("Invalid Aws S3 Credential or Unable to fetch Credential from IDBroker(s)");
    }

    public String getAwsS3AccessKey() {
        return awsS3Credential.getAccessKeyId();
    }

    public String getAwsS3SecretKey() {
        return awsS3Credential.getSecretAccessKey();
    }

    public String getAwsS3SessionToken() {
        return awsS3Credential.getSessionToken();
    }

    public boolean isIDBrokerEnabled() {
        return StringUtils.isNotEmpty(getConfigCabAddress());
    }

    private boolean isCredentialExpired() {
        return (awsS3Credential == null || awsS3Credential.getExpiration() < System.currentTimeMillis());
    }

    private String getConfigCabAddress() {
        String configParameter = String.format(CONFIG_CAB_ADDRESS_FORMAT, fileSystemScheme);

        return coreSiteConfiguration.getProperty(configParameter);
    }

    private String getConfigCabAddressPath() {
        String configParameter = String.format(CONFIG_CAB_ADDRESS_PATH_FORMAT, fileSystemScheme);

        return coreSiteConfiguration.getProperty(configParameter);
    }

    private void validateAwsS3Credential() throws AtlasBaseException {
        if (isCredentialExpired() || StringUtils.isEmpty(awsS3Credential.getAccessKeyId()) ||
                                     StringUtils.isEmpty(awsS3Credential.getSecretAccessKey()) ||
                                     StringUtils.isEmpty(awsS3Credential.getSessionToken())) {
            throw new AtlasBaseException("Invalid Aws S3 Credential or Unable to fetch Credential from IDBroker(s)");
        }
    }
}
