/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.bridge.models;

import org.apache.atlas.s3.bridge.AwsS3DataTypes;
import org.apache.atlas.s3.bridge.AwsS3Hook;
import org.apache.atlas.s3.extractors.ExtractorContext;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.codehaus.jettison.json.JSONException;
import org.codehaus.jettison.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.regions.Region;
import software.amazon.awssdk.services.s3.model.Bucket;
import software.amazon.awssdk.services.s3.model.GetBucketAclResponse;
import software.amazon.awssdk.services.s3.model.Grant;
import software.amazon.awssdk.services.s3.model.Tag;

import java.util.List;

public class AtlasAwsS3BucketEntity implements AtlasAwsS3Entity {
    private static final Logger LOG = LoggerFactory.getLogger(AtlasAwsS3BucketEntity.class);

    private static final String PROPERTY_PERMISSIONS   = "permissions";
    private static final String PROPERTY_CREATION_TIME = "creationTime";

    private final Bucket bucket;
    private final GetBucketAclResponse bucketAcl;
    private final AtlasEntity atlasEntity;
    private final ExtractorContext extractorContext;
    private final List<Tag> awsTag;
    private final String encryption;

    public AtlasAwsS3BucketEntity(ExtractorContext extractorContext,
                                  Bucket bucket,
                                  GetBucketAclResponse bucketAcl,
                                  List<Tag> awsTag,
                                  String encryption) {
        if (bucket == null || StringUtils.isEmpty(bucket.name()) ||
                bucket.creationDate() == null) {
            String msg = "Illegal bucket object received.";
            throw new IllegalArgumentException(msg);
        }

        this.extractorContext = extractorContext;
        this.bucket = bucket;
        this.bucketAcl = bucketAcl;
        this.awsTag = awsTag;
        this.encryption = encryption;
        this.atlasEntity = constructAtlasEntity();
    }

    @Override
    public String getQualifiedName() {
        return new StringBuilder(extractorContext.getFileSystemScheme().toLowerCase())
                .append(SCHEME_SEPARATOR)
                .append(bucket.name())
                .append(QNAME_SEP_METADATA_NAMESPACE)
                .append((AwsS3Hook.getInstance() != null?
                        AwsS3Hook.getInstance().getMetadataNamespace() : DEFAULT_METADATA_NAMESPACE))
                .toString();
    }

    public Region getAwsRegion() {
        return extractorContext.getAWSRegion();
    }

    @Override
    public String name() {
        return bucket.name();
    }

    @Override
    public AtlasEntity getAtlasEntity() {
        return atlasEntity;
    }

    @Override
    public AtlasEntity.AtlasEntitiesWithExtInfo getAtlasEntitiesWithExtInfo() {
        AtlasEntity.AtlasEntitiesWithExtInfo entitiesWithExtInfo = new AtlasEntity.AtlasEntitiesWithExtInfo();
        entitiesWithExtInfo.addEntity(this.getAtlasEntity());
        return entitiesWithExtInfo;
    }

    private AtlasEntity constructAtlasEntity() {
        AtlasEntity entity = new AtlasEntity(AwsS3DataTypes.AWS_S3_BUCKET.getEntityType());

        entity.setAttribute(PROPERTY_QUALIFIED_NAME, getQualifiedName());
        entity.setAttribute(PROPERTY_NAME, bucket.name());
        entity.setAttribute(PROPERTY_REGION, extractorContext.getAWSRegion().toString());
        entity.setAttribute(PROPERTY_AWS_TAG, convertToAtlasAwsTags(awsTag));
        entity.setAttribute(PROPERTY_ENCRYPTION, encryption);

        if (bucket.creationDate() != null) {
            entity.setAttribute(PROPERTY_CREATION_TIME, bucket.creationDate().toEpochMilli());
        }

        if (bucketAcl != null) {
            entity.setAttribute(PROPERTY_OWNER_NAME, bucketAcl.owner().displayName());
            entity.setAttribute(PROPERTY_OWNER_ID, bucketAcl.owner().id());

            List<Grant> grants = bucketAcl.grants();
            JSONObject permissions = new JSONObject();

            if (CollectionUtils.isNotEmpty(grants)) {
                for (Grant grant : grants) {
                    try {
                        permissions.put("Grantee Display Name", grant.grantee().displayName());
                        permissions.put("Grantee Id", grant.grantee().id());
                        permissions.put("Grantee Type", grant.grantee().type() == null ? StringUtils.EMPTY : grant.grantee().type().name());
                        permissions.put("Permission", grant.permissionAsString() == null ? StringUtils.EMPTY : grant.permissionAsString());
                    } catch (JSONException e) {
                        LOG.warn("Ignoring the Aws S3 permission for Atlas entity {}.", getQualifiedName(), e);
                    }
                }
            }
            entity.setAttribute(PROPERTY_PERMISSIONS, permissions.toString());
        }

        if(LOG.isDebugEnabled()) {
            LOG.debug("Constructed Atlas entity with name {} and qualified name {}", bucket.name(), getQualifiedName());
        }

        return entity;
    }
}
