/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.bridge.models;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.bridge.AwsS3DataTypes;
import org.apache.atlas.s3.bridge.AwsS3Hook;
import org.apache.atlas.s3.extractors.ExtractorContext;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.services.s3.model.HeadObjectResponse;
import software.amazon.awssdk.services.s3.model.S3Object;
import software.amazon.awssdk.services.s3.model.Tag;

import java.util.List;

public class AtlasAwsS3ObjectEntity implements AtlasAwsS3Entity {
    private static final Logger LOG = LoggerFactory.getLogger(AtlasAwsS3BucketEntity.class);

    private static final String PROPERTY_BUCKET_NAME = "bucketName";
    private static final String PROPERTY_ETAG = "eTag";
    private static final String PROPERTY_SIZE = "size";
    private static final String PROPERTY_STORAGE_CLASS = "storageClass";
    private static final String PROPERTY_LAST_MODIFICATION_TIME = "lastModifiedTime";
    private static final String PROPERTY_OBJECT_PREFIX = "objectPrefix";

    private final ExtractorContext extractorContext;
    private final String bucketName;
    private final S3Object s3Object;
    private final AtlasEntity entity;
    private final String entityType;
    private final List<Tag> objectTags;
    private final String encryption;

    public AtlasAwsS3ObjectEntity(ExtractorContext extractorContext,
                                  String bucketName,
                                  S3Object s3Object,
                                  List<Tag> objectTags,
                                  HeadObjectResponse headObjectResponse
                                  ) throws AtlasBaseException {
        this(extractorContext,
                bucketName,
                s3Object,
                headObjectResponse,
                objectTags,
                AwsS3DataTypes.AWS_S3_OBJECT.getEntityType());
    }

    public AtlasAwsS3ObjectEntity(ExtractorContext extractorContext,
                                  String bucketName,
                                  S3Object s3Object,
                                  HeadObjectResponse headObjectResponse,
                                  List<Tag> objectTags,
                                  String entityType) throws AtlasBaseException {
        if (StringUtils.isEmpty(bucketName)) {
            String msg = "Illegal S3 Bucket received.";
            throw new IllegalArgumentException(msg);
        }

        if(s3Object == null || StringUtils.isEmpty(s3Object.key())) {
            String msg = "Illegal S3 Object received.";
            throw new IllegalArgumentException(msg);
        }

        this.bucketName = bucketName;
        this.extractorContext = extractorContext;
        this.s3Object = getUpdatedS3ObjectFromHeadObject(s3Object, headObjectResponse);
        this.entityType = entityType;
        this.objectTags = objectTags;

        if(headObjectResponse != null) {
            this.encryption = headObjectResponse.serverSideEncryption() != null?
                    headObjectResponse.serverSideEncryption().toString() : null;
        } else {
            this.encryption = null;
        }

        this.entity = constructAtlasEntity();
    }

    public String getQualifiedName() {
        return new StringBuilder(extractorContext.getFileSystemScheme().toLowerCase())
                .append(SCHEME_SEPARATOR)
                .append(bucketName)
                .append(extractorContext.getPrefixDelimiter())
                .append(s3Object.key())
                .append(QNAME_SEP_METADATA_NAMESPACE)
                .append(AwsS3Hook.getInstance() != null ?
                        AwsS3Hook.getInstance().getMetadataNamespace() : DEFAULT_METADATA_NAMESPACE)
                .toString();
    }

    @Override
    public String name() {
        return s3Object.key();
    }

    @Override
    public AtlasEntity getAtlasEntity() {
        return entity;
    }

    @Override
    public AtlasEntity.AtlasEntitiesWithExtInfo getAtlasEntitiesWithExtInfo() {
        AtlasEntity.AtlasEntitiesWithExtInfo entitiesWithExtInfo = new AtlasEntity.AtlasEntitiesWithExtInfo();
        entitiesWithExtInfo.addEntity(this.getAtlasEntity());
        return entitiesWithExtInfo;
    }

    protected AtlasEntity constructAtlasEntity() throws AtlasBaseException {
        AtlasEntity entity = new AtlasEntity(entityType);

        entity.setAttribute(PROPERTY_QUALIFIED_NAME, getQualifiedName());
        entity.setAttribute(PROPERTY_NAME, getEntityName(s3Object.key()));
        entity.setAttribute(PROPERTY_BUCKET_NAME, bucketName);
        entity.setAttribute(PROPERTY_REGION, extractorContext.getAWSRegion().toString());
        entity.setAttribute(PROPERTY_LAST_MODIFICATION_TIME,
                s3Object.lastModified() != null? s3Object.lastModified().toEpochMilli() : null);
        entity.setAttribute(PROPERTY_ETAG,
                s3Object.eTag() != null ? s3Object.eTag().replace("\"", "") : null);
        entity.setAttribute(PROPERTY_SIZE, s3Object.size());
        entity.setAttribute(PROPERTY_ENCRYPTION, encryption);
        entity.setAttribute(PROPERTY_STORAGE_CLASS,
                s3Object.storageClass() != null ? s3Object.storageClass().toString() : null);
        entity.setAttribute(PROPERTY_AWS_TAG, convertToAtlasAwsTags(objectTags));
        entity.setAttribute(PROPERTY_OWNER_NAME, s3Object.owner() == null? "" : s3Object.owner().displayName());
        entity.setAttribute(PROPERTY_OWNER_ID, s3Object.owner() == null? "" : s3Object.owner().id());

        if(isS3DirectoryEntityType()) {
            entity.setAttribute(PROPERTY_OBJECT_PREFIX, extractorContext.getPrefixDelimiter() + getParentDirectory());
        }

        if(LOG.isDebugEnabled()) {
            LOG.debug("Constructed Atlas entity with name {} and qualified name {}",
                    getEntityName(s3Object.key()), getQualifiedName());
        }

        return entity;
    }

    private S3Object getUpdatedS3ObjectFromHeadObject(S3Object s3Object, HeadObjectResponse headObjectResponse) {
        S3Object ret = s3Object;
        if(headObjectResponse != null) {
            ret = s3Object.toBuilder().copy()
                    .lastModified(headObjectResponse.lastModified())
                    .size(headObjectResponse.contentLength())
                    .eTag(headObjectResponse.eTag())
                    .build();
        }
        return ret;
    }

    private boolean isS3DirectoryEntityType() {
        return entityType.equalsIgnoreCase(AwsS3DataTypes.AWS_S3_DIRECTORY.getEntityType());
    }

    private String getEntityName(String objectKey) throws AtlasBaseException {
        if(StringUtils.isEmpty(objectKey)) {
            throw new AtlasBaseException("Invalid Aws S3 Object key.");
        }

        int endPosWithoutLastDirectorySeparator = objectKey.length() - 1;
        if(doesObjectKeyEndsWithDirectorySeparator(objectKey)) {
            endPosWithoutLastDirectorySeparator--;
        }

        int nextDirectorySeparatorPositionFromLast = objectKey.lastIndexOf(extractorContext.getPrefixDelimiter(), endPosWithoutLastDirectorySeparator);
        return objectKey.substring(nextDirectorySeparatorPositionFromLast + 1, endPosWithoutLastDirectorySeparator + 1);
    }

    private String getParentDirectory() {
        String objectKey = s3Object.key();

        if (doesObjectKeyEndsWithDirectorySeparator(objectKey)) {
            objectKey = removeLastPrefixDelimiterFromDirectory(objectKey);
        }

        int    lastPrefixDelimiterPos = objectKey.lastIndexOf(extractorContext.getPrefixDelimiter());
        return objectKey.substring(0, lastPrefixDelimiterPos + 1);
    }

    private boolean doesObjectKeyEndsWithDirectorySeparator(String objectKey) {
        return objectKey.endsWith(extractorContext.getPrefixDelimiter());
    }

    private String removeLastPrefixDelimiterFromDirectory(String directoryPath) {
        return StringUtils.chop(directoryPath);
    }
}
