/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.extractors;

import org.apache.atlas.s3.bridge.AtlasAwsS3CreateEvent;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.bridge.AwsS3Hook;
import org.apache.atlas.s3.bridge.models.AtlasAwsS3BucketEntity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.awscore.exception.AwsServiceException;
import software.amazon.awssdk.core.exception.SdkClientException;
import software.amazon.awssdk.services.s3.S3Client;
import software.amazon.awssdk.services.s3.model.*;

import java.io.Serializable;
import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This class handles the S3 Bucket specific extractions.
 */
public class AwsS3BucketExtractor implements Serializable {

    private static final long   serialVersionUID = 1L;
    private static final Logger LOG              = LoggerFactory.getLogger(AwsS3BucketExtractor.class);

    private final ExtractorContext extractorContext;

    public AwsS3BucketExtractor(ExtractorContext extractorContext) {
        this.extractorContext = extractorContext;
    }

    public List<Bucket> getAllRequiredBuckets() throws AtlasBaseException {
        List<Bucket> ret                    = new ArrayList<>();
        boolean      isListBucketsSucceeded = false;
        int          retry                  = 0;

        while(retry < extractorContext.getMaxReconnectCount()) {
            try {
                S3Client s3Client = AwsS3Client.getInstance().getS3Client();
                ListBucketsResponse listBucketsResponse = s3Client.listBuckets();

                if (listBucketsResponse != null) {
                    ret = listBucketsResponse.buckets();
                    isListBucketsSucceeded = true;
                    break;
                }
            } catch (AwsServiceException | SdkClientException e) {
                if (!AwsExceptionHandler.handleException(e, AwsS3Client.getInstance(), LOG,  "Aws S3 exception occurred while listing all buckets")) {
                    break;
                }

                retry++;
            }
        }

        if(!isListBucketsSucceeded && extractorContext.isWhiteListingConfigurationPresent()) {
            for(Map.Entry<String, Set<String>> each : extractorContext.getWhitelistingBucketAndKey().entrySet()) {
                ret.add(Bucket.builder()
                        .name(each.getKey())
                        .build());
            }
        }

        return ret;
    }

    public AtlasAwsS3BucketEntity extractWhiteListedBucket(Bucket bucket) throws AtlasBaseException {
        int                    retry          = 0;
        String                 errMsg         = String.format("Aws S3 exception occurred while extracting bucket %s", bucket.name());

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                AtlasAwsS3BucketEntity s3BucketEntity = null;
                if (isWhiteListedBucket(bucket)) {
                    s3BucketEntity = extractBucket(bucket);
                } else if (LOG.isDebugEnabled()) {
                    LOG.debug("Skipping the extraction of bucket {} from region {} as it is not whitelisted.", bucket.name(), extractorContext.getAWSRegion());
                }

                return s3BucketEntity;
            } catch (AwsServiceException | SdkClientException e) {
                if (!AwsExceptionHandler.handleException(e, AwsS3Client.getInstance(), LOG, errMsg)) {
                    break;
                }

                retry++;
            }
        }

        throw new AtlasBaseException(errMsg);
    }

    public void publishCreateBucketToAtlas(AtlasAwsS3BucketEntity awsS3BucketEntity) throws AtlasBaseException {
        if (awsS3BucketEntity != null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Publish create bucket entity event to atlas.");
            }

            try {
                AtlasAwsS3CreateEvent awsS3CreateEvent = new AtlasAwsS3CreateEvent(awsS3BucketEntity.getAtlasEntitiesWithExtInfo());

                AwsS3Hook.getInstance().publishEvent(awsS3CreateEvent);
            } catch (Exception e) {
                String msg = String.format("Error encountered while publishing create bucket entity for bucket{%s} and region{%s}." +
                        "Extraction process will continue with remaining buckets.", awsS3BucketEntity.name(), extractorContext.getAWSRegion());

                LOG.error(msg, e);
                throw new AtlasBaseException(msg, e);
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("Published Aws S3 Bucket {} entity create event to Atlas.", awsS3BucketEntity.name());
            }
        }
    }

    public boolean isWhiteListedBucket(Bucket bucket) {
        return extractorContext.isBucketWhiteListed(bucket.name());
    }

    private AtlasAwsS3BucketEntity extractBucket(Bucket bucket) throws AtlasBaseException {
        LOG.info("Extracting bucket {} from region {} as it is whitelisted.", bucket.name(), extractorContext.getAWSRegion());

        GetBucketAclResponse   bucketAclResponse = getAwsS3BucketOwnerAndGrants(bucket.name());
        List<Tag>              awsTag            = getAwsS3BucketTagging(bucket.name());
        String                 encryption        = getAwsS3BucketEncryption(bucket.name());
        AtlasAwsS3BucketEntity s3BucketEntity    = new AtlasAwsS3BucketEntity(extractorContext, bucket, bucketAclResponse, awsTag, encryption);

        return s3BucketEntity;
    }

    private GetBucketAclResponse getAwsS3BucketOwnerAndGrants(String bucketName) throws AtlasBaseException {
        S3Client             s3                   = AwsS3Client.getInstance().getS3Client();
        GetBucketAclResponse getBucketAclResponse = null;
        int                  retry                = 0;
        String               errMsg               = String.format("Error occurred while fetching bucket{%s} owner and grant", bucketName);

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                getBucketAclResponse = s3.getBucketAcl(GetBucketAclRequest.builder().bucket(bucketName).build());

                break;
            } catch (AwsServiceException | SdkClientException e) {
                if (!AwsExceptionHandler.handleException(e, AwsS3Client.getInstance(), LOG, errMsg)) {
                    break;
                }

                retry++;
            }
        }

        return getBucketAclResponse;
    }

    private List<Tag> getAwsS3BucketTagging(String bucketName) throws AtlasBaseException {
        S3Client  s3Client = AwsS3Client.getInstance().getS3Client();
        List<Tag> ret      = null;
        int       retry    = 0;
        String    errMsg   = String.format("Error occurred while fetching bucket{%s} tagging", bucketName);

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                GetBucketTaggingRequest  getBucketTaggingRequest  = GetBucketTaggingRequest.builder().bucket(bucketName).build();
                GetBucketTaggingResponse getBucketTaggingResponse = s3Client.getBucketTagging(getBucketTaggingRequest);
                ret                                               = getBucketTaggingResponse.tagSet();

                break;
            } catch (AwsServiceException | SdkClientException e) {
                if (!AwsExceptionHandler.handleException(e, AwsS3Client.getInstance(), LOG, errMsg)) {
                    break;
                }

                retry++;
            }
        }

        return ret;
    }

    private String getAwsS3BucketEncryption(String bucketName) throws AtlasBaseException {
        S3Client s3Client          = AwsS3Client.getInstance().getS3Client();
        String   defaultEncryption = null;
        int      retry             = 0;
        String   errMsg            = String.format("Error occurred while fetching bucket{%s} encryption", bucketName);

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                GetBucketEncryptionRequest getBucketEncryptionRequest   = GetBucketEncryptionRequest.builder().bucket(bucketName).build();
                GetBucketEncryptionResponse getBucketEncryptionResponse = s3Client.getBucketEncryption(getBucketEncryptionRequest);
                defaultEncryption = getBucketEncryptionResponse.serverSideEncryptionConfiguration().rules().get(0).applyServerSideEncryptionByDefault().sseAlgorithmAsString();

                break;
            } catch (AwsServiceException | SdkClientException e) {
                if (!AwsExceptionHandler.handleException(e, AwsS3Client.getInstance(), LOG, errMsg)) {
                    break;
                }

                retry++;
            }
        }

        return defaultEncryption;
    }
}
