/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.extractors;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.auth.credentials.AwsBasicCredentials;
import software.amazon.awssdk.auth.credentials.AwsCredentials;
import software.amazon.awssdk.auth.credentials.AwsCredentialsProvider;
import software.amazon.awssdk.auth.credentials.AwsSessionCredentials;

/**
 * An implementation class for AWS Credentials provider.
 * It reads the credentials details from atlas configuration.
 */
public class AwsS3CredentialsProvider implements AwsCredentialsProvider {
    private static final Logger LOG = LoggerFactory.getLogger(AwsS3CredentialsProvider.class);
    private ExtractorContext    extractorContext;

    public AwsS3CredentialsProvider(ExtractorContext extractorContext) throws AtlasBaseException {
        this.extractorContext = extractorContext;

        if (StringUtils.isEmpty(extractorContext.getAwsS3AccessKey()) || StringUtils.isEmpty(extractorContext.getAwsS3SecretKey())) {
            String msg = "Invalid AWS access credentials are provided.";

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    @Override
    public AwsCredentials resolveCredentials() {
        AwsCredentials awsCredentials = null;

        try {
            if (StringUtils.isNotEmpty(extractorContext.getAwsS3SessionToken())) {
                awsCredentials = AwsSessionCredentials.create(extractorContext.getAwsS3AccessKey(),
                                                              extractorContext.getAwsS3SecretKey(),
                                                              extractorContext.getAwsS3SessionToken());
            } else {
                awsCredentials = AwsBasicCredentials.create(extractorContext.getAwsS3AccessKey(),
                                                            extractorContext.getAwsS3SecretKey());
            }
        } catch (AtlasBaseException e) {
            LOG.error("Error occurred while creating Aws Credential", e);
        }

        return awsCredentials;
    }
}
