/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.extractors;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.instance.AtlasRelatedObjectId;
import org.apache.atlas.s3.bridge.AtlasAwsS3CreateEvent;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.bridge.AwsS3Hook;
import org.apache.atlas.s3.bridge.AtlasAwsS3DeleteEvent;
import org.apache.atlas.s3.bridge.models.AtlasAwsS3BucketEntity;
import org.apache.atlas.s3.bridge.models.AtlasAwsS3DirectoryEntity;
import org.apache.atlas.s3.bridge.models.AtlasAwsS3Entity;
import org.apache.atlas.s3.bridge.models.AtlasAwsS3ObjectEntity;
import org.apache.atlas.type.AtlasTypeUtil;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.awscore.exception.AwsServiceException;
import software.amazon.awssdk.core.exception.SdkClientException;
import software.amazon.awssdk.services.s3.S3Client;
import software.amazon.awssdk.services.s3.model.*;
import software.amazon.awssdk.services.s3.paginators.ListObjectsV2Iterable;

import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.io.Serializable;

public class AwsS3ObjectExtractor implements Serializable {
    private static final long   serialVersionUID                     = 1L;
    private static final Logger LOG                                  = LoggerFactory.getLogger(AwsS3ObjectExtractor.class);
    private static final String AWS_S3_RELATIONSHIP_DEF              = "aws_s3_v2_container_contained";
    private static final String CONTAINER_END_DEF2                   = "container";
    private static final String PROPERTY_NEXT_TOKEN_AT_PROGRESS_FILE = "_NEXT_TOKEN_FOR_OBJECT_FETCH";

    private final ExtractorContext extractorContext;

    public AwsS3ObjectExtractor(ExtractorContext extractorContext) {
        this.extractorContext = extractorContext;
    }

    public void extractAllRequiredObjects(AtlasAwsS3BucketEntity bucketEntity) throws AtlasBaseException {
        if (bucketEntity != null) {
            Set<String> prefixes = new TreeSet<>();
            prefixes.add("");

            if(extractorContext.isWhiteListingConfigurationPresent()) {
                prefixes = extractorContext.getWhitelistingBucketAndKey().get(bucketEntity.name());
            }

            for(String prefix : prefixes) {
                LOG.info("Using prefix {} to get all objects", prefix);
                ListObjectsV2Iterable listObjectsV2Iterable = getAwsS3ListObjectIterable(bucketEntity.name(), prefix);

                for (ListObjectsV2Response listObjectsV2Response : listObjectsV2Iterable) {
                    AtlasAwsS3Entity lastContainerEntity = bucketEntity;

                    for (S3Object s3Object : listObjectsV2Response.contents()) {
                        try {
                            lastContainerEntity = createAwsS3ObjectEntitiesAtAtlas(s3Object, bucketEntity, lastContainerEntity);
                        } catch (AtlasBaseException e) {
                            LOG.info("Unable to extract s3 object {}", s3Object.key(), e);

                            if (extractorContext.getFailOnErrorFlag()) {
                                throw new AtlasBaseException(e.getMessage(), e);
                            }
                        }
                    }

                    writeNextTokenToProgressFile(bucketEntity.name(), prefix, listObjectsV2Response.nextContinuationToken());
                }

                writeNextTokenToProgressFile(bucketEntity.name(), prefix, null);
            }
        }
    }

    public AtlasAwsS3Entity createAwsS3ObjectEntitiesAtAtlas(S3Object s3Object, AtlasAwsS3BucketEntity bucketEntity) throws AtlasBaseException {
        return createAwsS3ObjectEntitiesAtAtlas(s3Object, bucketEntity, bucketEntity);
    }

    //Returns the container entity for the Object
    public AtlasAwsS3Entity createAwsS3ObjectEntitiesAtAtlas(S3Object s3Object, AtlasAwsS3BucketEntity bucketEntity, AtlasAwsS3Entity containerEntity) throws AtlasBaseException {
        AtlasAwsS3Entity lastContainerEntity = containerEntity;

        if (isEntityCreationNeeded(bucketEntity.name(), s3Object)) {
            int              retry               = 0;
            String           errMsg              = String.format("Aws S3 exception occurred while creating Atlas entity with key{%s}", s3Object.key());

            while (retry < extractorContext.getMaxReconnectCount()) {
                try {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Atlas entity will be created for Aws S3 Object {} as the key is white listed", s3Object.key());
                    }

                    if (isDirectlyInsideBucket(s3Object, lastContainerEntity)) {
                        lastContainerEntity = bucketEntity;
                    }

                    lastContainerEntity = createDirectoryHierarchyAtAtlas(bucketEntity.name(), s3Object, lastContainerEntity);

                    if (isS3Directory(s3Object)) {
                        lastContainerEntity = createAwsS3DirectoryAtAtlas(bucketEntity.name(), s3Object, lastContainerEntity);
                    } else {
                        createAwsS3ObjectAtAtlas(bucketEntity.name(), s3Object, lastContainerEntity);
                    }

                    return lastContainerEntity;
                } catch (AwsServiceException | SdkClientException e) {
                    if (!AwsExceptionHandler.handleException(e, AwsS3Client.getInstance(), LOG, errMsg)) {
                        break;
                    }

                    retry++;
                }
            }

            throw new AtlasBaseException(errMsg);
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("No Atlas entity will be created for Aws S3 Object {} as either the directory is not white listed or it is blacklisted",
                        s3Object.key());
            }
        }

        return lastContainerEntity;
    }

    public void publishCreateObjectToAtlas(AtlasAwsS3Entity s3ObjectEntity) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("publish create object entity event to atlas.");
        }

        try {
            AtlasAwsS3CreateEvent s3CreateEvent = new AtlasAwsS3CreateEvent(s3ObjectEntity.getAtlasEntitiesWithExtInfo());

            AwsS3Hook.getInstance().publishEvent(s3CreateEvent);
        } catch (Exception e) {
            String msg = String.format("Error encountered while publishing create object entity for object{%s} and region{%s}.", s3ObjectEntity.name(), extractorContext.getAWSRegion());

            LOG.error(msg, e);

            throw new AtlasBaseException(msg, e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("Published Aws S3 Object entity create event to Atlas.");
        }
    }

    public void publishDeleteObjectToAtlas(AtlasAwsS3Entity s3ObjectEntity) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("publish delete object entity event to atlas.");
        }

        try {
            AtlasAwsS3DeleteEvent s3DeleteEvent = new AtlasAwsS3DeleteEvent(s3ObjectEntity);

            AwsS3Hook.getInstance().publishEvent(s3DeleteEvent);
        } catch (Exception e) {
            String msg = String.format("Error encountered while publishing delete object entity for object{%s} and region{%s}.", s3ObjectEntity.name(), extractorContext.getAWSRegion());

            LOG.error(msg, e);

            throw new AtlasBaseException(msg, e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("Published Aws S3 Object entity delete event to Atlas.");
        }
    }

    public HeadObjectResponse getObjectMoreInfo(String bucketName, S3Object s3Object) throws AtlasBaseException {
        S3Client           s3Client           = AwsS3Client.getInstance().getS3Client();
        HeadObjectRequest  headObjectRequest  = HeadObjectRequest.builder().bucket(bucketName).key(s3Object.key()).build();
        HeadObjectResponse headObjectResponse = null;

        try {
            headObjectResponse = s3Client.headObject(headObjectRequest);
        } catch (AwsServiceException | SdkClientException e) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Unable to fetch object {} meta data: {}", s3Object.key(), e.getMessage());
            }
        }

        return headObjectResponse;
    }

    public S3Object getAwsS3ObjectDetailsIfAvailable(String bucketName, S3Object s3Object) {
        S3Object ret = s3Object;

        try {
            S3Client             s3Client           = AwsS3Client.getInstance().getS3Client();
            ListObjectsV2Request listObjectsRequest = ListObjectsV2Request.builder().bucket(bucketName)
                                                                                    .fetchOwner(true)
                                                                                    .prefix(s3Object.key())
                                                                                    .maxKeys(1)
                                                                                    .build();

            ListObjectsV2Response listObjectsV2Response = s3Client.listObjectsV2(listObjectsRequest);

            if (listObjectsV2Response != null && listObjectsV2Response.contents() != null) {
                ret = listObjectsV2Response.contents().get(0);
            }
        } catch (Exception e) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Unable to fetch object{} from AWS S3", s3Object.key());
            }
        }

        return ret;
    }

    private ListObjectsV2Iterable getAwsS3ListObjectIterable(String bucketName, String prefix) throws AtlasBaseException {
        S3Client              s3Client = AwsS3Client.getInstance().getS3Client();
        ListObjectsV2Iterable listObjectsV2Iterable;
        int                   retry    = 0;
        String                errMsg   = String.format("Aws S3 exception occurred while listing objects for bucket{%s}", bucketName);

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                ListObjectsV2Request listObjectsRequest = ListObjectsV2Request.builder().bucket(bucketName)
                                                                                        .fetchOwner(true)
                                                                                        .maxKeys(extractorContext.getMaxNumberOfObjectFetchPerCall())
                                                                                        .prefix(prefix)
                                                                                        .continuationToken(getNextTokenForObjectExtraction(bucketName, prefix))
                                                                                        .build();

                listObjectsV2Iterable = s3Client.listObjectsV2Paginator(listObjectsRequest);

                return listObjectsV2Iterable;
            } catch (AwsServiceException | SdkClientException e) {
                if (!AwsExceptionHandler.handleException(e, AwsS3Client.getInstance(), LOG, errMsg)) {
                    break;
                }
                retry++;
            }
        }

        throw new AtlasBaseException(errMsg);
    }

    private String getNextTokenForObjectExtraction(String bucketName, String prefix) {
        String ret = null;

        if (extractorContext.isResumeFromProgressFileNeeded()) {
            ret = extractorContext.getPropertyFromProgressFile(bucketName + prefix + PROPERTY_NEXT_TOKEN_AT_PROGRESS_FILE);
        }

        return ret;
    }

    private boolean isEntityCreationNeeded(String bucketName, S3Object object) {
        return extractorContext.doesKeyNeedExtraction(bucketName, object.key());
    }

    private boolean isDirectlyInsideBucket(S3Object s3Object, AtlasAwsS3Entity lastContainerEntity) {
        boolean isInsideBucket = true;
        String  directoryKey   = getDirectory(s3Object);

        if (!directoryKey.isEmpty()) {
            isInsideBucket = !(directoryKey.contains(lastContainerEntity.name()));
        }

        return isInsideBucket;
    }

    private String getDirectory(S3Object s3Object) {
        String objectKey = s3Object.key();

        if (isS3Directory(s3Object)) {
            objectKey = removeLastPrefixDelimiterFromDirectory(objectKey);
        }

        int    lastPrefixDelimiterPos = objectKey.lastIndexOf(extractorContext.getPrefixDelimiter());
        return objectKey.substring(0, lastPrefixDelimiterPos + 1);
    }

    private boolean isS3Directory(S3Object s3Object) {
        return s3Object.key().endsWith(extractorContext.getPrefixDelimiter());
    }

    private String removeLastPrefixDelimiterFromDirectory(String directoryPath) {
        return StringUtils.chop(directoryPath);
    }

    private AtlasAwsS3Entity createDirectoryHierarchyAtAtlas(String bucketName, S3Object s3Object,
                                                             AtlasAwsS3Entity lastContainerEntity) throws AtlasBaseException {
        String directoriesToBeCreated = getDirectory(s3Object);

        if (!StringUtils.isEmpty(directoriesToBeCreated)) {
            String directory = getRemainingDirectoryNeedsTobeCreated(directoriesToBeCreated, lastContainerEntity.name());

            if (!StringUtils.isEmpty(directory)) {
                String directoryKey = s3Object.key().contains(lastContainerEntity.name()) ? lastContainerEntity.name() + directory : directory;

                S3Object s3DirectoryObject = S3Object.builder()
                                                     .key(directoryKey)
                                                     .owner(s3Object.owner())
                                                     .lastModified(s3Object.lastModified())
                                                     .build();

                lastContainerEntity = createAwsS3DirectoryAtAtlas(bucketName, s3DirectoryObject, lastContainerEntity);
                lastContainerEntity = createDirectoryHierarchyAtAtlas(bucketName, s3Object, lastContainerEntity);
            }
        }

        return lastContainerEntity;
    }

    private String getRemainingDirectoryNeedsTobeCreated(String directoryKey, String directoryAlreadyCreated) {
        int positionAfterAlreadyCreatedDirectory = 0;

        if (!StringUtils.isEmpty(directoryAlreadyCreated) && directoryKey.contains(directoryAlreadyCreated)) {
            positionAfterAlreadyCreatedDirectory = directoryKey.indexOf(directoryAlreadyCreated) + directoryAlreadyCreated.length();
        }

        int positionTillNextDelimiter = directoryKey.indexOf(extractorContext.getPrefixDelimiter(), positionAfterAlreadyCreatedDirectory) + 1;

        String directoryToBeCreated = "";

        if (positionTillNextDelimiter > positionAfterAlreadyCreatedDirectory) {
            directoryToBeCreated = directoryKey.substring(positionAfterAlreadyCreatedDirectory, positionTillNextDelimiter);
        }

        return directoryToBeCreated;
    }

    private AtlasAwsS3Entity createAwsS3ObjectAtAtlas(String bucketName, S3Object s3Object,
                                                      AtlasAwsS3Entity containerEntity) throws AtlasBaseException {
        AtlasAwsS3Entity s3Entity;

        try {
            List<Tag>          tag                = getAwsS3ObjectTags(bucketName, s3Object.key());
            HeadObjectResponse headObjectResponse = getObjectMoreInfo(bucketName, s3Object);
            s3Entity                              = new AtlasAwsS3ObjectEntity(extractorContext, bucketName, s3Object, tag, headObjectResponse);

            linkAtlasAwsS3ContainerToContained(containerEntity.getAtlasEntity(), s3Entity.getAtlasEntity());

            publishCreateObjectToAtlas(s3Entity);
        } catch (IllegalArgumentException | AtlasBaseException e) {
            LOG.error("Exception occurred while creating an S3 Object entity: " + e.getMessage());
            throw new AtlasBaseException("Exception occurred while creating an S3 Object entity: " + e.getMessage());
        }

        return s3Entity;
    }

    private List<Tag> getAwsS3ObjectTags(String bucketName, String objectKey) throws AtlasBaseException {
        List<Tag>               ret               = null;
        S3Client                s3Client          = AwsS3Client.getInstance().getS3Client();
        GetObjectTaggingRequest getTaggingRequest = GetObjectTaggingRequest.builder()
                                                                           .bucket(bucketName)
                                                                           .key(objectKey)
                                                                           .build();

        try {
            GetObjectTaggingResponse objectTaggingResponse = s3Client.getObjectTagging(getTaggingRequest);

            ret = objectTaggingResponse.tagSet();
        } catch (AwsServiceException e) {
            LOG.info("No Tag information found for object [{}]: {}", objectKey, e.getMessage());
        }

        return ret;
    }

    private AtlasAwsS3Entity createAwsS3DirectoryAtAtlas(String bucketName, S3Object s3Object, AtlasAwsS3Entity containerEntity) throws AtlasBaseException {
        AtlasAwsS3Entity s3Entity;

        try {
            HeadObjectResponse headObjectResponse = getObjectMoreInfo(bucketName, s3Object);
            s3Entity                              = new AtlasAwsS3DirectoryEntity(extractorContext, bucketName, s3Object, headObjectResponse);

            linkAtlasAwsS3ContainerToContained(containerEntity.getAtlasEntity(), s3Entity.getAtlasEntity());

            publishCreateObjectToAtlas(s3Entity);
        } catch (IllegalArgumentException | AtlasBaseException e) {
            LOG.error("Exception occurred while creating an S3 directory entity: " + e.getMessage());
            throw new AtlasBaseException("Exception occurred while creating an S3 directory entity: " + e.getMessage());
        }

        return s3Entity;
    }

    private void linkAtlasAwsS3ContainerToContained(AtlasEntity containerAtlasEntity, AtlasEntity containedAtlasEntity) {
        containerAtlasEntity.setGuid(null);

        AtlasObjectId        containerEntityId = AtlasTypeUtil.getObjectId(containerAtlasEntity);
        AtlasRelatedObjectId relatedObjectId   = new AtlasRelatedObjectId(containerEntityId, AWS_S3_RELATIONSHIP_DEF);

        containedAtlasEntity.setRelationshipAttribute(CONTAINER_END_DEF2, relatedObjectId);
    }

    private void writeNextTokenToProgressFile(String bucketName, String prefix, String nextToken) {
        extractorContext.writePropertyToProgressFile(bucketName + prefix + PROPERTY_NEXT_TOKEN_AT_PROGRESS_FILE, nextToken);
    }
}