/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.s3.extractors;

import org.apache.atlas.exception.AtlasBaseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.services.sqs.SqsClient;
import software.amazon.awssdk.http.urlconnection.UrlConnectionHttpClient;

import java.util.concurrent.atomic.AtomicBoolean;
import java.io.Serializable;

public class AwsSQSClient implements AwsClient,Serializable {
    private static final long             serialVersionUID = 1L;
    private static final Logger           LOG              = LoggerFactory.getLogger(AwsSQSClient.class);
    private static final AwsSQSClient     INSTANCE         = new AwsSQSClient();
    private              SqsClient        queueClient;
    private              ExtractorContext extractorContext;
    private        final AtomicBoolean    isInitialized;

    public static AwsSQSClient getInstance() {
        return INSTANCE;
    }

    @Override
    public void initialize(ExtractorContext extractorContext) throws AtlasBaseException {
        if (!isInitialized.get()) {
            if (extractorContext != null) {
                this.queueClient = SqsClient.builder()
                                            .region(extractorContext.getAWSRegion())
                                            .credentialsProvider(new AwsS3CredentialsProvider(extractorContext))
                                            .httpClient(UrlConnectionHttpClient.builder().build())
                                            .build();

                this.extractorContext = extractorContext;

                isInitialized.set(true);
            } else {
                throw new AtlasBaseException("Invalid extractor context");
            }
        }
    }

    @Override
    public void reinitialize() throws AtlasBaseException {
        LOG.info("Reconnecting to Aws SQS");

        isInitialized.set(false);

        initialize(extractorContext);
    }

    public SqsClient getSQSClient() throws AtlasBaseException {
        if (isInitialized.get()) {
            return this.queueClient;
        }

        throw new AtlasBaseException("Aws S3 client is not initialized");
    }

    private AwsSQSClient() {
        isInitialized = new AtomicBoolean(false);
    }
}