/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.extractors;

import org.apache.atlas.s3.bridge.IDBrokerHandler;
import org.apache.atlas.utils.XMLConfigLoader;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.bridge.AwsS3Hook;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import software.amazon.awssdk.regions.Region;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;

import static org.apache.atlas.s3.bridge.models.AtlasAwsS3Entity.SCHEME_SEPARATOR;

/**
 * A context class that contains shared information for extraction
 */
public abstract class ExtractorContext implements Serializable {
    private static final long   serialVersionUID = 1L;

    private static final Logger LOG = LoggerFactory.getLogger(ExtractorContext.class);

    private       String                           awsRegion;
    private       IDBrokerHandler                  idBrokerHandler;

    private final Map<String, Set<String>>         whitelistingBucketAndKey;
    private final Map<String, Set<String>>         blacklistingBucketAndKey;
    private final Properties                       progressPropertyFile;
    private final Properties                       coreSiteConfiguration;
    private       boolean                          isCoreConfigPresent;
    private final boolean                          failOnError;

    protected static final String    AWS_S3_SQS_QUEUE_NAME_CONFIG = Constants.AWS_S3_EXTRACTION_CONFIG_PREFIX + "incremental.queueName";
    protected        final Properties extractionConfiguration;

    public ExtractorContext(Properties extractionConfiguration) throws AtlasBaseException {

        this.extractionConfiguration  = extractionConfiguration;
        this.progressPropertyFile     = new Properties();
        this.coreSiteConfiguration    = new Properties();
        this.whitelistingBucketAndKey = new HashMap<>();
        this.blacklistingBucketAndKey = new HashMap<>();
        this.isCoreConfigPresent      = false;
        this.failOnError              = Boolean.parseBoolean(extractionConfiguration.getProperty(Constants.FAIL_ON_ERROR_FLAG));
    }

    public void initialize() throws AtlasBaseException {
        loadCoreSiteConfiguration();

        validateConfiguration();

        AwsS3Hook.getInstance().initialize();

        loadWhitelisting();

        loadBlacklisting();

        loadProgressFile();
    }

    public Map<String, Set<String>> getWhitelistingBucketAndKey() {
        return whitelistingBucketAndKey;
    }

    public Region getAWSRegion() {
        return Region.of(awsRegion);
    }

    public String getSQSQueueName() {
        return extractionConfiguration.getProperty(AWS_S3_SQS_QUEUE_NAME_CONFIG);
    }

    public String getPrefixDelimiter() {
        String s3PrefixDelimiter = extractionConfiguration.getProperty(Constants.CONFIG_PREFIX_DELIMITER);

        if (StringUtils.isEmpty(s3PrefixDelimiter)) {
            s3PrefixDelimiter = Constants.DEFAULT_S3_PREFIX_DELIMITER;
        }

        return s3PrefixDelimiter;
    }

    public String getPropertyFromProgressFile(String property) {
        return progressPropertyFile.getProperty(property);
    }

    public int getMaxReconnectCount() {
        String stringValue = extractionConfiguration.getProperty(Constants.CONFIG_MAX_RECONNECT_COUNT);

        if (StringUtils.isEmpty(stringValue)) {
            return Constants.DEFAULT_MAX_RECONNECT_COUNT;
        } else {
            return Integer.parseInt(stringValue);
        }
    }

    public String getFileSystemScheme() {
        String stringValue = extractionConfiguration.getProperty(Constants.CONFIG_FILE_SYSTEM_SCHEME);

        if (StringUtils.isEmpty(stringValue)) {
            stringValue = Constants.DEFAULT_FILE_SYSTEM_SCHEME;
        }

        return stringValue;
    }

    public void writePropertyToProgressFile(String property, String value) {
        try {
            File         progressFile    = new File(getExtractorProgressFile());
            OutputStream progressFileOut = new FileOutputStream(progressFile);

            if (StringUtils.isEmpty(value)) {
                LOG.info("Removing the property {} from progress file", property);
                progressPropertyFile.remove(property);
            } else {
                LOG.info("Adding the property {} and value {} to progress file", property, value);
                progressPropertyFile.setProperty(property, value);
            }

            progressPropertyFile.store(progressFileOut, "");
        } catch (IOException e) {
            LOG.warn("Error while writing property {} to extraction progress file: " + e.getMessage());
        }
    }

    public String getAwsS3AccessKey() throws AtlasBaseException {
        String ret = getAwsAccessKeyConfig();

        if (isAwsS3CredentialFetchNeededFromIDBroker()) {
            refreshAwsS3CredentialFromIDBroker();
            ret = idBrokerHandler.getAwsS3AccessKey();
        }

        return ret;
    }

    public String getAwsS3SecretKey() throws AtlasBaseException {
        String ret = getAwsSecretKeyConfig();

        if (isAwsS3CredentialFetchNeededFromIDBroker()) {
            refreshAwsS3CredentialFromIDBroker();

            ret = idBrokerHandler.getAwsS3SecretKey();
        }

        return ret;
    }

    public String getAwsS3SessionToken() throws AtlasBaseException {
        String ret = getAwsSessionTokenConfig();

        if (isAwsS3CredentialFetchNeededFromIDBroker()) {
            refreshAwsS3CredentialFromIDBroker();

            ret = idBrokerHandler.getAwsS3SessionToken();
        }

        return ret;
    }

    public void refreshAwsS3CredentialFromIDBroker() throws AtlasBaseException {
        if (idBrokerHandler.isIDBrokerEnabled()) {
            idBrokerHandler.fetchAwsS3CredentialIfNeeded();
        }
    }

    public boolean getFailOnErrorFlag() {
        return this.failOnError;
    }

    boolean isResumeFromProgressFileNeeded() {
        boolean ret                          = Constants.DEFAULT_RESUME_FROM_PROGRESS_FILE;
        String  resumeFromProgressFileConfig = extractionConfiguration.getProperty(Constants.CONFIG_RESUME_FROM_PROGRESS_FILE);

        if (StringUtils.isNotEmpty(resumeFromProgressFileConfig)
                && resumeFromProgressFileConfig.equalsIgnoreCase("true")) {
            ret = true;
        }

        return ret;
    }

    int getMaxNumberOfObjectFetchPerCall() {
        int    maxObjectFetchPerCall = Constants.DEFAULT_MAX_NUMBER_OF_OBJECT_PER_CALL;
        String maxObjectPerCall      = extractionConfiguration.getProperty(Constants.CONFIG_EXTRACTOR_MAX_OBJECT_PER_CALL);

        if (StringUtils.isNotEmpty(maxObjectPerCall)) {
            maxObjectFetchPerCall = Integer.parseInt(maxObjectPerCall);
        }

        return maxObjectFetchPerCall;
    }

    boolean isBucketWhiteListed(String name) {
        return (isWhiteListingConfigurationPresent() && whitelistingBucketAndKey.containsKey(name)) ||
                (!isWhiteListingConfigurationPresent() && !(blacklistingBucketAndKey.containsKey(name) && blacklistingBucketAndKey.get(name).size() == 1 && StringUtils.isEmpty(blacklistingBucketAndKey.get(name).iterator().next())));
    }

    boolean doesKeyNeedExtraction(String bucketName, String key) {
        boolean ret = false;

        if(!isWhiteListingConfigurationPresent()) {
            ret = true;
        } else if (whitelistingBucketAndKey.containsKey(bucketName)) {
            //Check white listing first
            for (String keyFromList : whitelistingBucketAndKey.get(bucketName)) {
                if (key.indexOf(keyFromList) == 0 || keyFromList.indexOf(key) == 0) {
                    ret = true;
                    break;
                }
            }
        }

        //Now check if it is also black listed
        if (blacklistingBucketAndKey.containsKey(bucketName)) {
            for (String keyFromList : blacklistingBucketAndKey.get(bucketName)) {
                if (key.indexOf(keyFromList) == 0) {
                    ret = false;
                    break;
                }
            }
        }


        return ret;
    }

    boolean isWhiteListingConfigurationPresent() {
        return StringUtils.isNotEmpty(extractionConfiguration.getProperty(Constants.CONFIG_WHITELIST_PATHS));
    }

    protected void loadCoreSiteConfiguration() {
        try {
            XMLConfigLoader.loadConfig(Constants.CORE_SITE_PROPERTIES, coreSiteConfiguration);

            isCoreConfigPresent = true;
        } catch (AtlasBaseException e) {
            LOG.warn("Configuration file {} is missing and IDBroker will not be used", Constants.CORE_SITE_PROPERTIES);
        }
    }

    protected void validateConfiguration() throws AtlasBaseException {
        String awsRegion = getAwsRegionConfig();

        checkEmptinessForConfigurationParameter(Constants.CONFIG_EXTRACTION_AWS_REGION, awsRegion);

        if(isValidRegion(awsRegion)) {
            this.awsRegion = awsRegion;
        } else {
            throw new AtlasBaseException("Invalid aws Region");
        }

        if (isAwsS3CredentialFetchNeededFromIDBroker()) {
            this.idBrokerHandler = new IDBrokerHandler(getFileSystemScheme(), coreSiteConfiguration);
        }

        checkEmptinessForConfigurationParameter(Constants.CONFIG_ACCESS_KEY, getAwsS3AccessKey());
        checkEmptinessForConfigurationParameter(Constants.CONFIG_SECRET_KEY, getAwsS3SecretKey());
    }

    protected void checkEmptinessForConfigurationParameter(String configurationParameter, String configuredValue) throws AtlasBaseException {
        if (StringUtils.isEmpty(configuredValue)) {
            String msg = generateMissingConfigurationMessage(configurationParameter);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    protected String generateMissingConfigurationMessage(String extractionConfiguraitonName) {
        return String.format("There is no value specified for configuration parameter %s", extractionConfiguraitonName);
    }

    protected String generateInvalidConfigurationMessage(String extractionConfigurationName) {
        return String.format("Invalid value specified for configuration parameter %s", extractionConfigurationName);
    }

    private String getAwsAccessKeyConfig() {
        return extractionConfiguration.getProperty(Constants.CONFIG_ACCESS_KEY);
    }

    private void loadWhitelisting() throws AtlasBaseException {
        if (isWhiteListingConfigurationPresent()) {
            for (String eachPath : StringUtils.split(extractionConfiguration.getProperty(Constants.CONFIG_WHITELIST_PATHS), Constants.CONFIG_PATH_SEPARATOR)) {
                if (StringUtils.isNotBlank(eachPath)) {
                    eachPath      = getPathWithoutScheme(eachPath);
                    String bucket = getBucketNameFromPath(eachPath);
                    String key    = getKeyFromPath(eachPath);

                    if (whitelistingBucketAndKey.containsKey(bucket)) {
                        whitelistingBucketAndKey.get(bucket).add(key);

                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Adding bucket " + bucket + " and key " + key + " to whitelist");
                        }
                    } else {
                        Set<String> dirSet = new TreeSet<>();

                        dirSet.add(key);

                        whitelistingBucketAndKey.put(bucket, dirSet);

                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Adding bucket " + bucket + " and key " + dirSet + " to whitelist");
                        }
                    }
                }
            }

            squeezeBucketAndKeyEntries(whitelistingBucketAndKey);
        }
    }

    private void loadBlacklisting() throws AtlasBaseException {
        String configBlackListing = extractionConfiguration.getProperty(Constants.CONFIG_BLACKLIST_PATHS);

        if (StringUtils.isNotEmpty(configBlackListing)) {
            for (String eachPath : StringUtils.split(configBlackListing, Constants.CONFIG_PATH_SEPARATOR)) {
                if (StringUtils.isNotBlank(eachPath)) {
                    boolean doesNeedToAddToBlackList = true;
                    eachPath                         = getPathWithoutScheme(eachPath);
                    String   bucket                  = getBucketNameFromPath(eachPath);
                    String   key                     = getKeyFromPath(eachPath);

                    if (whitelistingBucketAndKey.containsKey(bucket)) {
                        removeAllKeysStartsWithBlacklistKey(whitelistingBucketAndKey.get(bucket), key);

                        if (whitelistingBucketAndKey.get(bucket).isEmpty()) {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Removing bucket " + bucket + " from whitelist");
                            }

                            whitelistingBucketAndKey.remove(bucket);

                            doesNeedToAddToBlackList = false;
                        }
                    }

                    if (doesNeedToAddToBlackList) {
                        if (blacklistingBucketAndKey.containsKey(bucket)) {
                            blacklistingBucketAndKey.get(bucket).add(key);

                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Adding bucket " + bucket + " and key " + key + " to blacklist");
                            }
                        } else {
                            Set<String> dirSet = new TreeSet<>();

                            dirSet.add(key);

                            blacklistingBucketAndKey.put(bucket, dirSet);

                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Adding bucket " + bucket + " and key " + dirSet + " to blacklist");
                            }
                        }
                    }
                }
            }

            squeezeBucketAndKeyEntries(blacklistingBucketAndKey);
        }
    }

    private void squeezeBucketAndKeyEntries(Map<String, Set<String>> bucketAndKeyEntries) {
        Iterator<Map.Entry<String, Set<String>>> bucketItr = bucketAndKeyEntries.entrySet().iterator();
        while(bucketItr.hasNext()) {
            Map.Entry<String, Set<String>> eachEntry = bucketItr.next();
            squeezeKeyEntries(eachEntry.getValue());
            if(CollectionUtils.isEmpty(eachEntry.getValue())) {
                bucketItr.remove();
            }
        }
    }

    private void squeezeKeyEntries(Set<String> keys) {
        Iterator<String> itr = keys.iterator();

        String lastParentKey = null;
        while (itr.hasNext()) {
            String currKey = itr.next();
            if(lastParentKey != null && currKey.indexOf(lastParentKey) == 0) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Removing key " + currKey + " from whitelist");
                }

                itr.remove();
            } else {
                lastParentKey = currKey;
            }
        }
    }

    private String getPathWithoutScheme(String path) throws AtlasBaseException {
        String pathWithoutScheme                    = StringUtils.trim(path);
        String fileSystemSchemeWithSeparator        = getFileSystemScheme() + SCHEME_SEPARATOR;
        int    indexOfFileSystemSchemeWithSeparator = pathWithoutScheme.indexOf(fileSystemSchemeWithSeparator);

        if (indexOfFileSystemSchemeWithSeparator == 0) {
            pathWithoutScheme = pathWithoutScheme.substring(indexOfFileSystemSchemeWithSeparator + fileSystemSchemeWithSeparator.length());
        } else {
            throw new AtlasBaseException("Invalid configuration for white or black listing: \""
                    + pathWithoutScheme + "\". It should start with: " + fileSystemSchemeWithSeparator);
        }

        return pathWithoutScheme;
    }

    private void removeAllKeysStartsWithBlacklistKey(Set<String> whiteListSet, String blacklistKey) {
        Iterator<String> itr = whiteListSet.iterator();

        while (itr.hasNext()) {
            String eachWhiteList = itr.next();

            if (eachWhiteList.indexOf(blacklistKey) == 0) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Removing key " + eachWhiteList + " from whitelist");
                }

                itr.remove();
            }
        }
    }

    private String getBucketNameFromPath(String bucketAndKey) {
        String bucket                  = bucketAndKey;
        int    firstPrefixDelimiterPos = bucketAndKey.indexOf(getPrefixDelimiter());

        if (firstPrefixDelimiterPos > -1) {
            bucket = bucketAndKey.substring(0, firstPrefixDelimiterPos);
        }

        return bucket;
    }

    private String getKeyFromPath(String bucketAndKey) {
        String key                     = "";
        int    firstPrefixDelimiterPos = bucketAndKey.indexOf(getPrefixDelimiter());

        if (firstPrefixDelimiterPos > -1) {
            key = bucketAndKey.substring(firstPrefixDelimiterPos + 1);
        }

        return key;
    }

    private void loadProgressFile() {
        String progressFileName = getExtractorProgressFile();
        File   progressFile     = new File(progressFileName);

        if (progressFile.exists() && progressFile.isFile() && progressFile.canRead()) {
            LOG.info("The progress file " + progressFileName + " is available.");

            try {
                this.progressPropertyFile.load(new FileInputStream(progressFile));
            } catch (IOException e) {
                LOG.error("Error while loading the extractor progress file: " + progressFileName);
            }
        }
    }

    private String getAwsSecretKeyConfig() {
        return extractionConfiguration.getProperty(Constants.CONFIG_SECRET_KEY);
    }

    private String getAwsSessionTokenConfig() {
        return extractionConfiguration.getProperty(Constants.CONFIG_SESSION_TOKEN);
    }

    private String getExtractorProgressFile() {
        String s3ExtractorProgressFile = extractionConfiguration.getProperty(Constants.CONFIG_EXTRACTOR_PROGRESS_FILE);

        if (StringUtils.isEmpty(s3ExtractorProgressFile)) {
            s3ExtractorProgressFile = Constants.DEFAULT_EXTRACTOR_PROGRESS_FILE;
        }

        return System.getProperty(Constants.ATLAS_LOG_DIR) + File.separator + s3ExtractorProgressFile;
    }

    private String getAwsRegionConfig() {
        String awsRegion = extractionConfiguration.getProperty(Constants.CONFIG_EXTRACTION_AWS_REGION);

        if (StringUtils.isEmpty(awsRegion) && isCoreConfigPresent) {
            String configParameter = String.format(Constants.CONFIG_EXTRACTION_AWS_REGION_FROM_CORE_SITE, getFileSystemScheme());
            awsRegion              = coreSiteConfiguration.getProperty(configParameter);
        }

        return awsRegion;
    }

    private boolean isAwsS3CredentialFetchNeededFromIDBroker() {
        return isCoreConfigPresent && (StringUtils.isEmpty(getAwsAccessKeyConfig()) || StringUtils.isEmpty(getAwsSecretKeyConfig()));
    }

    private boolean isValidRegion(String inputRegion) {
        List<Region>   listOfValidRegions        = Region.regions();
        List<String>   listOfValidRegionStrings  = new ArrayList<>();

        listOfValidRegions.forEach(region -> {
            listOfValidRegionStrings.add(region.toString());
        });

        return listOfValidRegionStrings.contains(inputRegion);
    }
}