/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.extractors;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.Serializable;
import java.util.Properties;

import org.apache.atlas.AtlasException;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.cli.CliContext;
import org.apache.atlas.s3.extractors.bulk.BulkExtractorTask;
import org.apache.atlas.s3.extractors.incremental.IncrementalExtractorTask;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This factory class for all extractors.
 */
public class ExtractorTaskFactory implements Serializable {
    private static final long serialVersionUID = 1L;
    private static final Logger LOG            = LoggerFactory.getLogger(ExtractorTaskFactory.class);

    protected ExtractorTaskFactory() {
    }

    /**
     * Return the task that needs to be executed
     *
     * @return The ExtractorTask for extraction
     */
    public static ExtractorTask getExtractorTask(CliContext cliContext) throws AtlasException, AtlasBaseException {
        File       configFile              = cliContext.getConfigFile();
        Properties extractionConfiguration = loadConfigFile(configFile);
        Boolean    failOnError             = cliContext.isFailOnError();

        extractionConfiguration.setProperty(Constants.FAIL_ON_ERROR_FLAG, failOnError.toString());

        switch (cliContext.getExtractionType()) {
            case BULK:
                return new BulkExtractorTask(extractionConfiguration);
            case INCREMENTAL:
                return new IncrementalExtractorTask(extractionConfiguration);
        }

        String msg = String.format("Received unknown/unsupported extraction type {}.", cliContext.getExtractionType().name());

        LOG.error(msg);

        throw new AtlasException(msg);
    }

    private static Properties loadConfigFile(File configFile) throws AtlasBaseException {
        Properties configuration = new Properties();

        try (InputStream inputStream = new FileInputStream(configFile);) {
            configuration.load(inputStream);
        } catch (Exception e) {
            String msg = String.format("Error encountered in reading the configuration file %s. Reason: %s", configFile, e.getMessage());

            LOG.error(msg, e);

            throw new AtlasBaseException(msg);
        }

        return configuration;
    }
}