/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.s3.extractors.bulk;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.bridge.models.AtlasAwsS3BucketEntity;
import org.apache.atlas.s3.cli.CliContext;
import org.apache.atlas.s3.extractors.AwsS3BucketExtractor;
import org.apache.atlas.s3.extractors.AwsS3Client;
import org.apache.atlas.s3.extractors.AwsS3ObjectExtractor;
import org.apache.atlas.s3.extractors.Constants;
import org.apache.atlas.s3.extractors.ExtractorContext;
import org.apache.atlas.s3.extractors.ExtractorTask;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.services.s3.model.Bucket;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.Serializable;
import java.util.List;
import java.util.Properties;

public class BulkExtractorTask implements ExtractorTask, Serializable {
    private static final long                 serialVersionUID = 1L;
    private static final Logger               LOG              = LoggerFactory.getLogger(BulkExtractorTask.class);
    private static final String               TASK_NAME        = "Aws S3 Bulk Extractor";
    private        final ExtractorContext     extractorContext;
    private              AwsS3BucketExtractor s3BucketHandler;
    private              AwsS3ObjectExtractor awsS3ObjectExtractor;

    private long awsS3ExtractionStartTime;

    public BulkExtractorTask(Properties extractionConfiguration) throws AtlasBaseException {
        this.extractorContext     = new BulkExtractorContext(extractionConfiguration);
        this.s3BucketHandler      = new AwsS3BucketExtractor(extractorContext);
        this.awsS3ObjectExtractor = new AwsS3ObjectExtractor(extractorContext);
    }

    @Override
    public String getName() {
        return TASK_NAME;
    }

    @Override
    public void initialize() throws AtlasBaseException {
        extractorContext.initialize();

        AwsS3Client.getInstance().initialize(extractorContext);
    }

    @Override
    public void execute() throws AtlasBaseException {
        try {
            LOG.info("Starting bulk extraction for Aws S3.");

            initializeExtractionStartTime();

            List<Bucket> buckets = s3BucketHandler.getAllRequiredBuckets();

            for (Bucket bucket : buckets) {
                try {
                    AtlasAwsS3BucketEntity awsS3BucketEntity = s3BucketHandler.extractWhiteListedBucket(bucket);

                    s3BucketHandler.publishCreateBucketToAtlas(awsS3BucketEntity);

                    awsS3ObjectExtractor.extractAllRequiredObjects(awsS3BucketEntity);
                } catch (AtlasBaseException e) {
                    LOG.info("Unable to extract bucket {} completely", bucket.name(), e);

                    if (extractorContext.getFailOnErrorFlag()) {
                        throw new AtlasBaseException(e.getMessage(), e);
                    }
                }
            }

            writeExtractionStartTimeToProgressFile();

            LOG.info("Completed bulk extraction for S3.");
        } catch (AtlasBaseException e) {
            throw new AtlasBaseException("Error encountered while doing S3 Bulk extraction.", e);
        }
    }

    @Override
    public ExtractorContext getExtractorContext() {
        return extractorContext;
    }

    private void initializeExtractionStartTime() {
        awsS3ExtractionStartTime = System.currentTimeMillis();
    }

    private void writeExtractionStartTimeToProgressFile() {
        extractorContext.writePropertyToProgressFile(Constants.PROPERTY_EXTRACTION_START_TIME_AT_PROGRESS_FILE,
                String.valueOf(awsS3ExtractionStartTime));
    }
}