/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.extractors.incremental;

import java.io.Serializable;
import java.util.Properties;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.extractors.ExtractorContext;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A context class that contains shared information for extraction
 */
public class IncrementalExtractorContext extends ExtractorContext implements Serializable {
    private static final long serialVersionUID                       = 1L;
    private static final Logger LOG                                  = LoggerFactory.getLogger(IncrementalExtractorContext.class);
    private static final String AWS_S3_EXTRACTION_CONFIG_PREFIX      = "atlas.s3.extraction.";
    private static final String CONFIG_SQS_MAX_MESSAGES_PER_REQUEST  = AWS_S3_EXTRACTION_CONFIG_PREFIX + "incremental.messagesPerRequest";
    private static final String CONFIG_SQS_REQUEST_WAIT_TIME         = AWS_S3_EXTRACTION_CONFIG_PREFIX + "incremental.requestWaitTime";
    private static final String CONFIG_MAX_RETRY                     = AWS_S3_EXTRACTION_CONFIG_PREFIX + "incremental.max.retry";
    private static final int    DEFAULT_SQS_MAX_MESSAGES_PER_REQUEST = 10;
    private static final int    DEFAULT_SQS_REQUEST_WAIT_TIME        = 20;
    private static final int    DEFAULT_MAX_RETRY                    = 20;

    public IncrementalExtractorContext(Properties extractionConfiguration) throws AtlasBaseException {
        super(extractionConfiguration);
    }

    public void validateConfiguration() throws AtlasBaseException {
        super.validateConfiguration();

        checkEmptinessForConfigurationParameter(AWS_S3_SQS_QUEUE_NAME_CONFIG, getSQSQueueName());

        validateMaxNumberOfMsgPerRequestConfig();

        validateSqsRequestWaitTimeConfig();

        validateMaxRetryConfig();
    }

    public Integer getMaxNumberOfMessagesPerRequest() {
        String stringValue = extractionConfiguration.getProperty(CONFIG_SQS_MAX_MESSAGES_PER_REQUEST);

        if (stringValue == null) {
            return DEFAULT_SQS_MAX_MESSAGES_PER_REQUEST;
        } else {
            return Integer.parseInt(stringValue);
        }
    }

    public Integer getWaitTimeInSeconds() {
        String stringValue = extractionConfiguration.getProperty(CONFIG_SQS_REQUEST_WAIT_TIME);

        if (stringValue == null) {
            return DEFAULT_SQS_REQUEST_WAIT_TIME;
        } else {
            return Integer.parseInt(stringValue);
        }
    }

    public int getMaxRetryCount() {
        String stringValue = extractionConfiguration.getProperty(CONFIG_MAX_RETRY);

        if (StringUtils.isEmpty(stringValue)) {
            return DEFAULT_MAX_RETRY;
        } else {
            return Integer.parseInt(stringValue);
        }
    }

    private void validateMaxNumberOfMsgPerRequestConfig() throws AtlasBaseException {
        try {
            getMaxNumberOfMessagesPerRequest();
        } catch (NumberFormatException e) {
            String msg = generateInvalidConfigurationMessage(CONFIG_SQS_MAX_MESSAGES_PER_REQUEST);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    private void validateSqsRequestWaitTimeConfig() throws AtlasBaseException {
        try {
            getWaitTimeInSeconds();
        } catch (NumberFormatException e) {
            String msg = generateInvalidConfigurationMessage(CONFIG_SQS_REQUEST_WAIT_TIME);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    private void validateMaxRetryConfig() throws AtlasBaseException {
        try {
            getMaxRetryCount();
        } catch (NumberFormatException e) {
            String msg = generateInvalidConfigurationMessage(CONFIG_MAX_RETRY);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }
}
