/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.s3.extractors.incremental;

import org.apache.atlas.s3.extractors.AwsExceptionHandler;
import org.apache.atlas.s3.extractors.ExtractorTask;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.extractors.ExtractorContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.core.exception.SdkException;
import software.amazon.awssdk.services.sqs.model.Message;

import java.util.List;
import java.util.Properties;

public class IncrementalExtractorTask implements ExtractorTask {
    private static final long serialVersionUID = 1L;
    private static final Logger LOG            = LoggerFactory.getLogger(IncrementalExtractorTask.class);
    private static final String TASK_NAME      = "Aws S3 Incremental Extractor";

    private final IncrementalExtractorContext s3IncrementalExtractorContext;
    private final SQSQueueMsgHandler          sqsQueueMsgHandler;

    public IncrementalExtractorTask(Properties extractionConfiguration) throws AtlasBaseException {
        this.s3IncrementalExtractorContext = new IncrementalExtractorContext(extractionConfiguration);
        this.sqsQueueMsgHandler            = new SQSQueueMsgHandler(s3IncrementalExtractorContext);
    }

    @Override
    public String getName() {
        return TASK_NAME;
    }

    @Override
    public void initialize() throws AtlasBaseException {
        s3IncrementalExtractorContext.initialize();

        sqsQueueMsgHandler.initialize();
    }

    @Override
    public void execute() throws AtlasBaseException {
        LOG.info("Starting incremental extraction for S3.");

        int idleOrFailureAttempt = 0;

        while (idleOrFailureAttempt < s3IncrementalExtractorContext.getMaxRetryCount()) {
            idleOrFailureAttempt++;

            if (LOG.isDebugEnabled()) {
                LOG.debug("Performing attempt #{} to read messages from SQS.", idleOrFailureAttempt);
            }

            try {
                List<Message> receivedMessages = sqsQueueMsgHandler.getQueueMessages();

                if (receivedMessages.size() > 0) {
                    for (Message message : receivedMessages) {
                        sqsQueueMsgHandler.processQueueMessage(message);
                        sqsQueueMsgHandler.deleteQueueMessage(message.receiptHandle());
                    }

                    idleOrFailureAttempt = 0;
                } else {
                    LOG.info("No messages received from the queue.");
                }
            } catch (SdkException e) {
                if (AwsExceptionHandler.isAwsS3ExpiredTokenException(e)) {
                    sqsQueueMsgHandler.reinitializeAwsQueueClient();
                } else if (!e.retryable()) {
                    LOG.warn("Exception occurred: " + e.getMessage());

                    throw new AtlasBaseException(e.getMessage(), e);
                }
            }
        }

        LOG.info("Stopping the extraction after receiving no messages in {} attempts.", idleOrFailureAttempt);
    }

    @Override
    public ExtractorContext getExtractorContext() {
        return s3IncrementalExtractorContext;
    }

    public SQSQueueMsgHandler getSQSQueueMsgHandler() {
        return sqsQueueMsgHandler;
    }
}