/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.extractors.incremental;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.extractors.AwsS3BucketExtractor;
import org.apache.atlas.s3.extractors.AwsS3Client;
import org.apache.atlas.s3.extractors.AwsS3ObjectExtractor;
import org.apache.atlas.s3.extractors.AwsSQSClient;
import org.apache.atlas.s3.extractors.Constants;
import org.apache.atlas.s3.extractors.incremental.models.Record;
import org.apache.atlas.s3.extractors.incremental.models.SQSBucket;
import org.apache.atlas.s3.extractors.incremental.models.SQSEvent;
import org.apache.atlas.s3.extractors.incremental.models.SQSObject;
import org.apache.atlas.s3.bridge.models.AtlasAwsS3BucketEntity;
import org.apache.atlas.s3.bridge.models.AtlasAwsS3ObjectEntity;
import org.apache.atlas.type.AtlasType;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import software.amazon.awssdk.services.s3.model.Bucket;
import software.amazon.awssdk.services.s3.model.HeadObjectResponse;
import software.amazon.awssdk.services.s3.model.S3Object;
import software.amazon.awssdk.services.sqs.model.*;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.io.Serializable;

public class SQSQueueMsgHandler implements Serializable {
    private static final long serialVersionUID       = 1L;
    private static final Logger LOG                  = LoggerFactory.getLogger(SQSQueueMsgHandler.class);
    private static final String OBJECT_CREATED_EVENT = "ObjectCreated";
    private static final String OBJECT_REMOVED_EVENT = "ObjectRemoved";
    private static final String PLUS_SIGN            = "+";
    private static final String SPACE                = " ";

    private final IncrementalExtractorContext         s3ExtractorContext;
    private final AwsS3BucketExtractor                awsS3BucketExtractor;
    private final AwsS3ObjectExtractor                awsS3ObjectExtractor;
    private final SimpleDateFormat                    dateFormatter;

    private String queueUrl;

    public SQSQueueMsgHandler(IncrementalExtractorContext s3ExtractorContext) throws AtlasBaseException {
        this.s3ExtractorContext   = s3ExtractorContext;
        this.awsS3BucketExtractor = new AwsS3BucketExtractor(s3ExtractorContext);
        this.awsS3ObjectExtractor = new AwsS3ObjectExtractor(s3ExtractorContext);
        this.dateFormatter        = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'");
    }

    public void initialize() throws AtlasBaseException {
        AwsS3Client.getInstance().initialize(s3ExtractorContext);

        AwsSQSClient.getInstance().initialize(s3ExtractorContext);

        this.queueUrl = getQueueUrl();

        loadAwsS3BucketsIntoCache();
    }

    public List<Message> getQueueMessages() throws AtlasBaseException {
        String queueName = s3ExtractorContext.getSQSQueueName();

        LOG.info("Read messages using queue url {} for queue {}.", queueUrl, queueName);

        ReceiveMessageRequest receiveMessageRequest = ReceiveMessageRequest.builder()
                .queueUrl(queueUrl)
                .maxNumberOfMessages(s3ExtractorContext.getMaxNumberOfMessagesPerRequest())
                .waitTimeSeconds(s3ExtractorContext.getWaitTimeInSeconds())
                .build();

        return AwsSQSClient.getInstance().getSQSClient().receiveMessage(receiveMessageRequest).messages();
    }

    public void processQueueMessage(Message message) throws AtlasBaseException {
        try {
            SQSEvent event = getEventFromSqsMessage(message);

            if (doesEventContainRecords(event)) {
                for (Record record : event.getRecords()) {
                    if (isEventRecordStructureValid(record)
                            && isAtlasSupportedEventType(record.getEventName())
                            && (isObjectRemoveEvent(record.getEventName())
                            || isEventOccurredAfterBulkExtractionStart(getEventTime(record).toEpochMilli()))) {
                        processEventRecord(record);
                    }
                }
            }
        } catch (AtlasBaseException e) {
            String msg = String.format("Error encountered while processing AWS S3 SQS message. Reason: %s", e.getMessage());

            LOG.warn(msg);

            if (s3ExtractorContext.getFailOnErrorFlag()) {
                throw new AtlasBaseException(msg, e);
            }
        }
    }

    public void deleteQueueMessage(String receiptHandle) throws AtlasBaseException {
        DeleteMessageRequest deleteRequest = DeleteMessageRequest.builder()
                                                                 .queueUrl(queueUrl)
                                                                 .receiptHandle(receiptHandle)
                                                                 .build();

        AwsSQSClient.getInstance().getSQSClient().deleteMessage(deleteRequest);
    }

    public void reinitializeAwsQueueClient() throws AtlasBaseException {
        LOG.info("Aws S3 token has expired. Trying to reconnect");

        AwsS3Client.getInstance().reinitialize();

        AwsSQSClient.getInstance().reinitialize();

        this.queueUrl = getQueueUrl();
    }

    public SQSEvent getEventFromSqsMessage(Message message) {
        return AtlasType.fromJson(message.body(), SQSEvent.class);
    }

    public Instant getEventTime(Record record) throws AtlasBaseException {
        Instant retTime  = null;
        String eventTime = record.getEventTime();

        if (StringUtils.isNotEmpty(eventTime)) {
            try {
                retTime = Instant.ofEpochMilli(dateFormatter.parse(eventTime).getTime());
            } catch (ParseException e) {
                LOG.warn("Unable to get event time from SQS. Considering event time to current time.", e);

                throw new AtlasBaseException("Unable to get event time from SQS. Considering event time to current time.", e);
            }
        }

        return retTime;
    }

    public AtlasAwsS3BucketEntity getBucketEntity(Record record) throws AtlasBaseException {
        AtlasAwsS3BucketEntity s3BucketEntity = getAtlasS3BucketEntityFromCache(record.getS3Body().getBucket());

        if (s3BucketEntity == null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Received event for the bucket {} which is no more present at AWS S3.", record.getS3Body().getBucket().getName());
            }

            s3BucketEntity = createBucketEntityFromEvent(record);
        }

        return s3BucketEntity;
    }

    public S3Object createAwsS3Object(Record record) throws AtlasBaseException {
        SQSObject sqsObject      = record.getS3Body().getSqsObject();
        String    awsS3ObjectKey = convertSqsMsgObjectKeyToAwsS3ObjectKey(sqsObject.getKey());

        return S3Object.builder()
                .key(awsS3ObjectKey)
                .lastModified(getEventTime(record))
                .eTag(sqsObject.geteTag())
                .size(sqsObject.getSize() != null ? Long.valueOf(sqsObject.getSize()) : null)
                .build();
    }

    public boolean isObjectRemoveEvent(String eventName) {
        return eventName.startsWith(OBJECT_REMOVED_EVENT);
    }

    public boolean isObjectCreateEvent(String eventName) {
        return eventName.startsWith(OBJECT_CREATED_EVENT);
    }

    public void createEntitiesAtAtlas(S3Object s3Object, AtlasAwsS3BucketEntity atlasAwsS3BucketEntity) throws AtlasBaseException {
        s3Object = awsS3ObjectExtractor.getAwsS3ObjectDetailsIfAvailable(atlasAwsS3BucketEntity.name(), s3Object);

        awsS3BucketExtractor.publishCreateBucketToAtlas(atlasAwsS3BucketEntity);

        awsS3ObjectExtractor.createAwsS3ObjectEntitiesAtAtlas(s3Object, atlasAwsS3BucketEntity);
    }

    public boolean isNoMoreVersionOfObjectExistAtAWS(String bucketName, S3Object s3Object) {
        boolean isNotPresent = true;

        try {
            HeadObjectResponse headObjectResponse = awsS3ObjectExtractor.getObjectMoreInfo(bucketName, s3Object);

            if (headObjectResponse != null) {
                isNotPresent = false;
            }
        } catch (AtlasBaseException e) {
            e.printStackTrace();
        }

        return isNotPresent;
    }

    public void deleteEntityAtAtlas(S3Object s3Object, AtlasAwsS3BucketEntity atlasAwsS3BucketEntity) throws AtlasBaseException {
        AtlasAwsS3ObjectEntity atlasAwsS3Entity = new AtlasAwsS3ObjectEntity(s3ExtractorContext, atlasAwsS3BucketEntity.name(), s3Object, null, null);

        awsS3ObjectExtractor.publishDeleteObjectToAtlas(atlasAwsS3Entity);
    }

    public AwsS3BucketExtractor getBucketExtractor() {
        return awsS3BucketExtractor;
    }

    public AwsS3ObjectExtractor getObjectExtractor() {
        return awsS3ObjectExtractor;
    }

    private String getQueueUrl() throws AtlasBaseException {
        String queueName = s3ExtractorContext.getSQSQueueName();
        String queueUrl  = AwsSQSClient.getInstance().getSQSClient().getQueueUrl(GetQueueUrlRequest.builder().queueName(queueName).build()).queueUrl();

        if (StringUtils.isEmpty(queueUrl)) {
            String msg = String.format("Error encountered in retrieving the queue URL from AWS Queue: %s", queueName);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }

        return queueUrl;
    }

    private void loadAwsS3BucketsIntoCache() throws AtlasBaseException {
        AtlasS3BucketCache.getInstance().getAtlasS3BucketCache().clear();

        List<Bucket> buckets = awsS3BucketExtractor.getAllRequiredBuckets();

        for (Bucket bucket : buckets) {
            try {
                AtlasAwsS3BucketEntity awsS3BucketEntity = awsS3BucketExtractor.extractWhiteListedBucket(bucket);

                if (awsS3BucketEntity != null) {
                    AtlasS3BucketCache.getInstance().getAtlasS3BucketCache().put(formKeyForCache(awsS3BucketEntity.name(), awsS3BucketEntity.getAwsRegion().toString()), awsS3BucketEntity);
                }
            } catch (AtlasBaseException e) {
                LOG.info("Unable to load bucket{} to bucket cache.", bucket.name());
            }
        }
    }

    private String formKeyForCache(String bucketName, String region) {
        return bucketName + s3ExtractorContext.getPrefixDelimiter() + region;
    }

    private boolean doesEventContainRecords(SQSEvent event) {
        boolean ret = false;

        if (event != null && event.getRecords() != null && event.getRecords().size() > 0) {
            ret = true;
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("No event record found at SQS Message");
        }

        return ret;
    }

    private boolean isEventRecordStructureValid(Record record) {
        boolean ret = false;

        if (record.getS3Body() != null && record.getS3Body().getBucket() != null && StringUtils.isNotEmpty(record.getS3Body().getBucket().getName())
                && record.getS3Body().getSqsObject() != null && StringUtils.isNotEmpty(record.getS3Body().getSqsObject().getKey())) {
            ret = true;
        } else {
            LOG.info("Skipping unsupported event record structure with event name {} and event time {}", record.getEventName(), record.getEventTime());
        }

        return ret;
    }

    private boolean isAtlasSupportedEventType(String eventName) {
        boolean ret = false;

        if (StringUtils.isNotEmpty(eventName) && (eventName.startsWith(OBJECT_CREATED_EVENT) || eventName.startsWith(OBJECT_REMOVED_EVENT))) {
            ret = true;
        } else {
            LOG.info("Skipping unsupported event type {}.", eventName);
        }

        return ret;
    }

    private boolean isEventOccurredAfterBulkExtractionStart(long eventTime) {
        boolean isEventOccurredAfterBulk = true;
        String  lastExtractionTime       = s3ExtractorContext.getPropertyFromProgressFile(Constants.PROPERTY_EXTRACTION_START_TIME_AT_PROGRESS_FILE);

        if (StringUtils.isNotEmpty(lastExtractionTime)) {
            if (eventTime < Long.parseLong(lastExtractionTime)) {
                isEventOccurredAfterBulk = false;

                LOG.info("Skipping event with event time {} which has been reported before last successful bulk extraction start time {}.",
                        eventTime, lastExtractionTime);
            }
        }

        return isEventOccurredAfterBulk;
    }

    private void processEventRecord(Record record) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Processing event {} for object key {} and bucket {} ", record.getEventName(), record.getS3Body().getSqsObject().getKey(), record.getS3Body().getBucket());
        }

        AtlasAwsS3BucketEntity atlasAwsS3BucketEntity = getBucketEntity(record);

        if (atlasAwsS3BucketEntity != null) {
            S3Object s3Object  = createAwsS3Object(record);
            String   eventName = record.getEventName();

            if (isObjectCreateEvent(eventName)) {
                createEntitiesAtAtlas(s3Object, atlasAwsS3BucketEntity);
            } else if (isObjectRemoveEvent(eventName)) {
                if (isNoMoreVersionOfObjectExistAtAWS(atlasAwsS3BucketEntity.name(), s3Object)) {
                    deleteEntityAtAtlas(s3Object, atlasAwsS3BucketEntity);
                } else {
                    createEntitiesAtAtlas(s3Object, atlasAwsS3BucketEntity);
                }
            }
        } else {
            LOG.info("Skipping the event with Object Key {} generated for non white listed bucket {}.",
                    record.getS3Body().getSqsObject().getKey(),
                    record.getS3Body().getBucket().getName());
        }
    }

    private AtlasAwsS3BucketEntity getAtlasS3BucketEntityFromCache(SQSBucket bucketInfoFromEvent) throws AtlasBaseException {
        String bucketNameForCacheLookup = formKeyForCache(bucketInfoFromEvent.getName(), s3ExtractorContext.getAWSRegion().toString());

        return AtlasS3BucketCache.getInstance().getAtlasS3BucketCache().get(bucketNameForCacheLookup);
    }

    private AtlasAwsS3BucketEntity createBucketEntityFromEvent(Record record) throws AtlasBaseException {
        Bucket bucket = Bucket.builder()
                .name(record.getS3Body().getBucket().getName())
                .creationDate(getEventTime(record))
                .build();

        AtlasAwsS3BucketEntity s3BucketEntity = null;

        if (awsS3BucketExtractor.isWhiteListedBucket(bucket)) {
            s3BucketEntity = new AtlasAwsS3BucketEntity(s3ExtractorContext, bucket, null, null, null);
        }

        return s3BucketEntity;
    }

    private String convertSqsMsgObjectKeyToAwsS3ObjectKey(String key) {
        //E.g. "My Object" S3 Object key comes "My+Object" at SQS Nsg
        return key.replace(PLUS_SIGN, SPACE);
    }
}
