/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.s3.cli;

import org.apache.atlas.ApplicationProperties;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.s3.extractors.ExtractorTask;
import org.apache.atlas.s3.extractors.bulk.BulkExtractorTask;
import org.apache.atlas.s3.extractors.incremental.IncrementalExtractorContext;
import org.apache.atlas.s3.extractors.incremental.IncrementalExtractorTask;
import org.testng.Assert;
import org.testng.annotations.*;

import java.io.File;

public class S3ExtractorMainTest {
    @BeforeSuite
    public void initOneTime() {
        System.setProperty(ApplicationProperties.ATLAS_CONFIGURATION_DIRECTORY_PROPERTY, "src/test/resources/");
    }

    @BeforeMethod
    public void init() {
        S3ExtractorMain.exitCode = S3ExtractorMain.EXIT_CODE_FAILED;
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: The configuration file aws-s3.conf is not a file or it does not exist or it is not readable.")
    public void testCreateCliContextWithNullArgs() throws AtlasBaseException {
        S3ExtractorMain.createCliContext(null);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: The configuration file aws-s3.conf is not a file or it does not exist or it is not readable.")
    public void testCreateCliContextWithEmptyArray() throws AtlasBaseException {
        S3ExtractorMain.createCliContext(new String[0]);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: Unrecognized arguments.")
    public void testCreateCliContextWithMissingArgs() throws AtlasBaseException {
        S3ExtractorMain.createCliContext(new String[]{""});
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: Received unsupported extraction type testType.")
    public void testCreateCliContextWithUnsupportedExtractionType() throws AtlasBaseException {
        S3ExtractorMain.createCliContext(new String[]{"-e", "testType"});
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: The configuration file aws-s3.conf is not a file or it does not exist or it is not readable.")
    public void testCreateCliContextWithSupportedExtractionType1() throws AtlasBaseException {
        S3ExtractorMain.createCliContext(new String[]{"-e", "incremental"});
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: The configuration file dummyConfigFile.txt is not a file or it does not exist or it is not readable.")
    public void testCreateCliContextWithNonExistingConfigurationFile() throws AtlasBaseException {
        S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental", "-c", "dummyConfigFile.txt"});
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: Unrecognized arguments.")
    public void testCreateCliContextWithInvalidParameter() throws AtlasBaseException {
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{
                "-e", "Incremental",
                "-c", "src/test/resources/TestExtractionConfigFile.txt",
                "-g"});
        Assert.assertEquals(cliContext.getExtractionType(), CliContext.ExtractionType.INCREMENTAL);
        Assert.assertEquals(cliContext.getSourceType(), CliContext.SourceType.S3);
        Assert.assertEquals(cliContext.isFailOnError(), true);
    }

    @Test
    public void testCreateCliContextWithValidConfigurationFile() throws AtlasBaseException {
        String configFilePath = "src/test/resources/TestExtractionConfigFile.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.s3.extraction.aws.region")
    public void testInitializeBulkExtractorTaskFactoryWithMissingAWSRegion() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/configFileWithMissingAWSRegion.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);
        S3ExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.s3.extraction.access.key")
    public void testInitializeBulkExtractorTaskFactoryWithMissingAccessKey() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/configFileWithMissingAccessKey.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);
        S3ExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.s3.extraction.secret.key")
    public void testInitializeBulkExtractorTaskFactoryWithMissingSecretKey() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/configFileWithMissingSecretKey.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);
        S3ExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test
    public void testInitializeBulkExtractorTaskFactoryWithCorrectWhiteListingFromat() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/configFileWithCorrectBucketWhiteBlackListing.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);
        S3ExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test
    public void testInitializeBulkExtractorTaskFactoryValidConfigPath() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/bulkExtractionConfig.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);

        ExtractorTask extractorTask = S3ExtractorMain.getInitializedExtractorTask(cliContext);
        Assert.assertEquals(S3ExtractorMain.exitCode, S3ExtractorMain.EXIT_CODE_FAILED);
        Assert.assertNotNull(extractorTask);
        Assert.assertEquals(extractorTask.getClass(), BulkExtractorTask.class);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.s3.extraction.aws.region")
    public void testInitializeExtractorIncrementalTaskFactoryWithMissingAWSRegion() throws AtlasBaseException {
        String configFilePath = "src/test/resources/incremental/configFileWithMissingAWSRegion.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);

        ExtractorTask extractorTask = S3ExtractorMain.getInitializedExtractorTask(cliContext);
        Assert.assertEquals(extractorTask.getClass(), IncrementalExtractorTask.class);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.s3.extraction.access.key")
    public void testInitializeExtractorIncrementalTaskFactoryWithMissingAccessKey() throws AtlasBaseException {
        String configFilePath = "src/test/resources/incremental/configFileWithMissingAccessKey.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);
        S3ExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.s3.extraction.secret.key")
    public void testInitializeExtractorIncrementalTaskFactoryWithMissingSecretKey() throws AtlasBaseException {
        String configFilePath = "src/test/resources/incremental/configFileWithMissingSecretKey.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);
        S3ExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: Invalid configuration for white or black listing: \"S3a:/test\". It should start with: s3a://")
    public void testInitializeExtractorIncrementalTaskFactoryWithWrongBlackListing() throws AtlasBaseException {
        String configFilePath = "src/test/resources/incremental/configFileWithWrongBlackListing.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);
        S3ExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: Invalid configuration for white or black listing: \"someWhiteListing\". It should start with: s3a://")
    public void testInitializeExtractorIncrementalTaskFactoryWithWrongWhiteListingFormat() throws AtlasBaseException {
        String configFilePath = "src/test/resources/incremental/configFileWithWrongWhiteListingFormat.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);

        S3ExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.s3.extraction.incremental.queueName")
    public void testInitializeExtractorIncrementalTaskFactoryMissingQueueName() throws AtlasBaseException {
        String configFilePath = "src/test/resources/incremental/configFileWithMissingQueueName.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);

        S3ExtractorMain.getInitializedExtractorTask(cliContext);
        Assert.assertEquals(S3ExtractorMain.exitCode, S3ExtractorMain.EXIT_CODE_FAILED);
    }

    @Test(expectedExceptions = AtlasBaseException.class)
    public void testInitializeExtractorIncrementalTaskFactoryMissingMessagesPerRequest() throws AtlasBaseException {
        String configFilePath = "src/test/resources/incremental/configFileWithMissingMessagesPerRequest.txt";
        CliContext cliContext = S3ExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);

        ExtractorTask extractorTask = S3ExtractorMain.getInitializedExtractorTask(cliContext);
        Assert.assertEquals(S3ExtractorMain.exitCode, S3ExtractorMain.EXIT_CODE_FAILED);
        Assert.assertNotNull(extractorTask);
        Assert.assertEquals(extractorTask.getClass(), IncrementalExtractorTask.class);

        IncrementalExtractorContext s3ExtractorContext = (IncrementalExtractorContext) extractorTask.getExtractorContext();
        Assert.assertNotNull(s3ExtractorContext);
    }

    private void assertCheckForCliContext(CliContext cliContext, String configFilePath, CliContext.ExtractionType extractionType) {
        Assert.assertEquals(cliContext.getExtractionType(), extractionType);
        Assert.assertEquals(cliContext.getSourceType(), CliContext.SourceType.S3);
        Assert.assertEquals(cliContext.isFailOnError(), true);
        Assert.assertEquals(cliContext.getConfigFile(), new File(configFilePath));
        Assert.assertEquals(S3ExtractorMain.exitCode, S3ExtractorMain.EXIT_CODE_FAILED);
    }
}
