/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.bridge;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.hook.AtlasHook;
import org.apache.atlas.model.instance.AtlasEntity.AtlasEntitiesWithExtInfo;
import org.apache.atlas.model.instance.AtlasEntity.AtlasEntityWithExtInfo;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.notification.HookNotification;
import org.apache.atlas.model.notification.HookNotification.EntityCreateRequestV2;
import org.apache.atlas.model.notification.HookNotification.EntityDeleteRequestV2;
import org.apache.atlas.model.notification.HookNotification.EntityPartialUpdateRequestV2;
import org.apache.atlas.model.notification.HookNotification.EntityUpdateRequestV2;
import org.apache.atlas.type.AtlasType;
import org.apache.commons.collections.CollectionUtils;
import org.apache.hadoop.security.UserGroupInformation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Collections;
import java.util.List;

import static org.apache.atlas.repository.Constants.ADLS_SOURCE;

/**
 * Azure Data Lake Storage hook to facilitate communication with Atlas through Kafka channel.
 */
public class ADLSHook extends AtlasHook {
    private static final Logger   LOG      = LoggerFactory.getLogger(ADLSHook.class);
    private static final ADLSHook INSTANCE = new ADLSHook();

    private UserGroupInformation ugi;

    public static ADLSHook getInstance() {
        return INSTANCE;
    }

    private ADLSHook() {
    }

    @Override
    public String getMessageSource() {
        return ADLS_SOURCE;
    }

    public void initialize() throws AtlasBaseException {
        try {
            ugi = UserGroupInformation.getCurrentUser();
        } catch (IOException e) {
            String msg = "Error encountered in getting the user information.";

            LOG.error(msg, e);

            throw new AtlasBaseException(msg, e);
        }

        LOG.info("Initialization completed successfully.");
    }

    public void publishEvent(AtlasADLSEvent event) throws Exception {
        switch (event.getEventType()) {
            case CREATE:
                publishCreateEventNotification(event);
                break;

            case DELETE:
                publishDeleteEventNotification(event);
                break;

            case PARTIAL_UPDATE:
                publishPartialUpdateEventNotification(event);
                break;

            case UPDATE:
                publishUpdateEventNotification(event);
                break;

            default:
                throw new AtlasBaseException("Unknown event type received. EventType: " + event.getEventType());
        }
    }

    private void publishCreateEventNotification(AtlasADLSEvent event) {
        if (event instanceof AtlasADLSCreateEvent) {
            AtlasADLSCreateEvent createEvent = (AtlasADLSCreateEvent) event;

            if (LOG.isDebugEnabled()) {
                LOG.debug("Sending create event to Atlas: {}", createEvent.getEntitiesWithExtInfo().toString());
            }

            List<HookNotification>   notificationMessages = null;
            AtlasEntitiesWithExtInfo entitiesWithExtInfo  = createEvent.getEntitiesWithExtInfo();

            if (entitiesWithExtInfo != null && CollectionUtils.isNotEmpty(entitiesWithExtInfo.getEntities())) {
                notificationMessages = Collections.singletonList(new EntityCreateRequestV2(createEvent.getUserName(), entitiesWithExtInfo));
            }

            super.notifyEntities(notificationMessages, ugi);

            if (LOG.isDebugEnabled()) {
                LOG.debug("Message sent is {}", AtlasType.toJson(notificationMessages));
            }
        }
    }

    private void publishDeleteEventNotification(AtlasADLSEvent event) {
        if (event instanceof AtlasADLSDeleteEvent) {
            AtlasADLSDeleteEvent deleteEvent = (AtlasADLSDeleteEvent) event;

            if (LOG.isDebugEnabled()) {
                LOG.debug("Sending delete event to Atlas: {}", deleteEvent.getObjectIdsForDeletion().toString());
            }

            List<HookNotification> notificationMessages = null;
            List<AtlasObjectId>    objectIdsForDeletion = deleteEvent.getObjectIdsForDeletion();

            if (CollectionUtils.isNotEmpty(objectIdsForDeletion)) {
                notificationMessages = Collections.singletonList(new EntityDeleteRequestV2(deleteEvent.getUserName(), objectIdsForDeletion));
            }

            super.notifyEntities(notificationMessages, ugi);

            if (LOG.isDebugEnabled()) {
                LOG.debug("Message sent is {}", AtlasType.toJson(notificationMessages));
            }
        }
    }

    private void publishPartialUpdateEventNotification(AtlasADLSEvent event) {
        if (event instanceof AtlasADLSPartialUpdateEvent) {
            AtlasADLSPartialUpdateEvent partialUpdateEvent = (AtlasADLSPartialUpdateEvent) event;

            if (LOG.isDebugEnabled()) {
                LOG.debug("Sending partial update event to Atlas: {}", partialUpdateEvent.getEntityExtInfo().toString());
            }

            List<HookNotification> notificationMessages = null;
            AtlasEntityWithExtInfo entityExtInfo        = partialUpdateEvent.getEntityExtInfo();

            if (entityExtInfo != null) {
                notificationMessages = Collections.singletonList(new EntityPartialUpdateRequestV2(partialUpdateEvent.getUserName(), partialUpdateEvent.getOldObjectId(), entityExtInfo));
            }

            super.notifyEntities(notificationMessages, ugi);

            if (LOG.isDebugEnabled()) {
                LOG.debug("Message sent is {}", AtlasType.toJson(notificationMessages));
            }
        }
    }

    private void publishUpdateEventNotification(AtlasADLSEvent event) {
        if (event instanceof AtlasADLSUpdateEvent) {
            AtlasADLSUpdateEvent updateEvent = (AtlasADLSUpdateEvent) event;

            if (LOG.isDebugEnabled()) {
                LOG.debug("The update event sending to Atlas: {}", updateEvent.getEntitiesWithExtInfo().toString());
            }

            List<HookNotification>   notificationMessages = null;
            AtlasEntitiesWithExtInfo entitiesWithExtInfo  = updateEvent.getEntitiesWithExtInfo();

            if (entitiesWithExtInfo != null && CollectionUtils.isNotEmpty(entitiesWithExtInfo.getEntities())) {
                notificationMessages = Collections.singletonList(new EntityUpdateRequestV2(updateEvent.getUserName(), entitiesWithExtInfo));
            }

            super.notifyEntities(notificationMessages, ugi);

            if (LOG.isDebugEnabled()) {
                LOG.debug("Message sent is {}", AtlasType.toJson(notificationMessages));
            }
        }
    }
}