/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.adls.bridge;

import org.apache.atlas.adls.extractors.Constants;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.AtlasBaseClient;
import org.apache.atlas.adls.bridge.models.IDBrokerAzureCredential;
import org.apache.atlas.utils.AuthenticationUtil;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.HttpMethod;
import javax.ws.rs.core.Response;
import java.util.LinkedList;
import java.util.List;
import java.time.LocalDateTime;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.Properties;
import java.io.Serializable;

public class AzureIDBrokerHandler implements Serializable {
    private static final long serialVersionUID                    = 1L;
    private static final Logger LOG                               = LoggerFactory.getLogger(AzureIDBrokerHandler.class);
    private static final String CONFIG_CAB_ADDRESS_FORMAT         = "fs.%s.ext.cab.address"; // http://host:8444/gateway
    private static final String CONFIG_CAB_ADDRESS_PATH_FORMAT    = "fs.%s.ext.cab.path"; //azure-cab
    private static final String API_CAB_CREDENTIALS_GLOBAL        = "/cab/api/v1/credentials";
    private static final char   CONFIG_PATH_SEPARATOR             = ',';

    private final Properties              coreSiteConfiguration;
    private final String                  fileSystemScheme;

    private List<AtlasIDBrokerClient> idBrokerClients = new LinkedList<>();
    private IDBrokerAzureCredential   azureCredential;

    public AzureIDBrokerHandler(Properties coreSiteConfiguration) throws AtlasBaseException {
        this.fileSystemScheme = Constants.FILE_SYSTEM_SCHEME_FOR_ID_BROKER;
        this.coreSiteConfiguration = coreSiteConfiguration;

        String idBrokerAddresses = getConfigCabAddress();
        if (StringUtils.isNotBlank(idBrokerAddresses)) {
            if (!AuthenticationUtil.isKerberosAuthenticationEnabled()) {
                throw new AtlasBaseException("No Kerberos authentication enabled");
            }

            try {
                for (String eachIDBrokerAddress : StringUtils.split(idBrokerAddresses, CONFIG_PATH_SEPARATOR)) {
                    eachIDBrokerAddress = StringUtils.trim(eachIDBrokerAddress);
                    idBrokerClients.add(new AtlasIDBrokerClient(eachIDBrokerAddress));
                }
            } catch (Exception e) {
                throw new AtlasBaseException("Failed to initialize IdBroker Client", e);
            }
        }
    }

    public void fetchADLSCredentialIfNeeded() throws AtlasBaseException {
        for (AtlasIDBrokerClient idBrokerClient : idBrokerClients) {
            try {
                if (isCredentialExpired()) {
                    AtlasBaseClient.API getADLSCredentialAPI = new AtlasBaseClient.API(getConfigCabAddressPath() + API_CAB_CREDENTIALS_GLOBAL, HttpMethod.GET, Response.Status.OK);

                    azureCredential = idBrokerClient.getADLSCredential(getADLSCredentialAPI);
                    LOG.info("Successfully fetched Azure credential from IDBroker {}", idBrokerClient.getResource().getURI().toString());
                }

                validateADLSCredential();

                return;
            } catch (Exception e) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Error occurred for IDBroker {}", idBrokerClient.getResource().getURI().toString(), e);
                } else {
                    LOG.warn("Error occurred for IDBroker {} : {}", idBrokerClient.getResource().getURI().toString(), e.getMessage());
                }
            }
        }

        throw new AtlasBaseException("Invalid Azure Credential or Unable to fetch Credential from IDBroker(s)");
    }

    public String getADLSAccessToken() {
        return azureCredential.getAccess_token();
    }

    public OffsetDateTime getADLSAccessTokenExpiresAt() {
        return OffsetDateTime.of(LocalDateTime.ofEpochSecond(azureCredential.getExpires_on(), 0, ZoneOffset.UTC), ZoneOffset.UTC);
    }

    public boolean isIDBrokerEnabled() {
        return StringUtils.isNotEmpty(getConfigCabAddress());
    }

    private boolean isCredentialExpired() {
        return azureCredential == null || azureCredential.getExpires_on() < (System.currentTimeMillis() / 1000);
    }

    private String getConfigCabAddress() {
        String configParameter = String.format(CONFIG_CAB_ADDRESS_FORMAT, fileSystemScheme);

        return coreSiteConfiguration.getProperty(configParameter);
    }

    private String getConfigCabAddressPath() {
        String configParameter = String.format(CONFIG_CAB_ADDRESS_PATH_FORMAT, fileSystemScheme);

        return coreSiteConfiguration.getProperty(configParameter);
    }

    private void validateADLSCredential() throws AtlasBaseException {
        if (isCredentialExpired() || StringUtils.isEmpty(azureCredential.getAccess_token())) {
            throw new AtlasBaseException("Invalid Azure Data Lake Storage Credential from IDBroker");
        }
    }
}
