/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.bridge.models;

import com.azure.storage.blob.BlobContainerClient;
import com.azure.storage.blob.models.StorageAccountInfo;
import org.apache.atlas.adls.bridge.ADLSDataTypes;
import org.apache.atlas.adls.bridge.ADLSHook;
import org.apache.atlas.adls.extractors.Constants;
import org.apache.atlas.adls.extractors.ADLSExtractorContext;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Duration;

public class AtlasADLSAccountEntity implements AtlasADLSEntity {
    private static final Logger LOG                   = LoggerFactory.getLogger(AtlasADLSAccountEntity.class);
    private static final String PROPERTY_ACCOUNT_KIND = "accountKind";
    private static final String PROPERTY_SKU_NAME     = "skuName";

    private final BlobContainerClient  blobContainerClient;
    private final AtlasEntity          entity;
    private final ADLSExtractorContext extractorContext;
    private final Duration             timeout;

    public AtlasADLSAccountEntity(ADLSExtractorContext extractorContext, BlobContainerClient blobContainerClient) {
        this.timeout = extractorContext.getTimeoutInSecPerCall();

        if (isInvalidAccount(blobContainerClient)) {
            String msg = "Illegal Account received.";

            throw new IllegalArgumentException(msg);
        }

        this.extractorContext    = extractorContext;
        this.blobContainerClient = blobContainerClient;
        this.entity              = constructAtlasEntity();
    }

    @Override
    public String getQualifiedName() {
        String fileSystemScheme  = extractorContext.getFileSystemScheme().toLowerCase();
        String accountName       = blobContainerClient.getAccountName().toLowerCase();
        String metadataNamespace = ADLSHook.getInstance() != null ?
                                   ADLSHook.getInstance().getMetadataNamespace() : DEFAULT_METADATA_NAMESPACE;

        return fileSystemScheme + SCHEME_SEPARATOR + accountName + QNAME_SEP_METADATA_NAMESPACE + metadataNamespace;
    }

    @Override
    public String name() {
        String blobContainerName = blobContainerClient.getBlobContainerName();
        String accountName       = blobContainerClient.getAccountName();

        return blobContainerName + Constants.DEFAULT_CONTAINER_ACCOUNT_DELIMITER + accountName;
    }

    @Override
    public String getParentName() {
        return "";
    }

    @Override
    public AtlasEntity getEntity() {
        return entity;
    }

    @Override
    public AtlasEntity.AtlasEntitiesWithExtInfo getAtlasEntitiesWithExtInfo() {
        return new AtlasEntity.AtlasEntitiesWithExtInfo(this.getEntity());
    }

    private boolean isInvalidAccount(BlobContainerClient blobContainerClient) {
        return blobContainerClient == null ||
               StringUtils.isEmpty(blobContainerClient.getAccountName());
    }

    private AtlasEntity constructAtlasEntity() {
        AtlasEntity entity = new AtlasEntity(ADLSDataTypes.ADLS_GEN2_ACCOUNT.getEntityType());

        entity.setAttribute(PROPERTY_QUALIFIED_NAME, getQualifiedName());
        entity.setAttribute(PROPERTY_NAME, blobContainerClient.getAccountName());
        entity.setAttribute(PROPERTY_ACCOUNT_KIND, getAccountKind());
        entity.setAttribute(PROPERTY_SKU_NAME, getSKUName());

        if (LOG.isDebugEnabled()) {
            LOG.debug("Constructed Atlas entity with name {} and qualified name {}", blobContainerClient.getBlobContainerName(), getQualifiedName());
        }

        return entity;
    }

    private String getAccountKind() {
        String ret                     = "";
        StorageAccountInfo accountInfo = getStorageAccountInfo();

        if (accountInfo != null) {
            ret = accountInfo.getAccountKind().toString();
        }

        return ret;
    }

    private String getSKUName() {
        String             ret         = "";
        StorageAccountInfo accountInfo = getStorageAccountInfo();

        if (accountInfo != null) {
            ret = accountInfo.getSkuName().toString();
        }

        return ret;
    }

    private StorageAccountInfo getStorageAccountInfo() {
        StorageAccountInfo ret = null;

        try {
            ret = blobContainerClient.getAccountInfo(this.timeout);
        } catch (Exception e) {
            LOG.debug("Unable to get Account info for the account{}", blobContainerClient.getAccountName(), e);
        }

        return ret;
    }
}