/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.bridge.models;

import com.azure.storage.blob.BlobContainerClient;
import org.apache.atlas.adls.bridge.ADLSDataTypes;
import org.apache.atlas.adls.bridge.ADLSHook;
import org.apache.atlas.adls.extractors.Constants;
import org.apache.atlas.adls.extractors.ADLSExtractorContext;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class AtlasADLSContainerEntity implements AtlasADLSEntity {
    private static final Logger LOG = LoggerFactory.getLogger(AtlasADLSContainerEntity.class);

    private static final String PROPERTY_CREATION_TIME         = "lastModifiedTime";
    private static final String PROPERTY_ETAG                  = "eTag";
    private static final String PROPERTY_ADDITIONAL_PROPERTIES = "additionalProperties";
    private static final String PROPERTY_PUBLIC_ACCESS         = "publicAccess";

    private final BlobContainerClient  blobContainerClient;
    private final AtlasEntity          entity;
    private final ADLSExtractorContext extractorContext;

    public AtlasADLSContainerEntity(ADLSExtractorContext extractorContext, BlobContainerClient blobContainerClient) {
        if (isInvalidContainer(blobContainerClient)) {
            String msg = "Illegal container received.";

            throw new IllegalArgumentException(msg);
        }

        this.extractorContext    = extractorContext;
        this.blobContainerClient = blobContainerClient;
        this.entity              = constructAtlasEntity();
    }

    @Override
    public String getQualifiedName() {
        String fileSystemScheme  = extractorContext.getFileSystemScheme().toLowerCase();
        String blobContainerName = blobContainerClient.getBlobContainerName().toLowerCase();
        String accountName       = blobContainerClient.getAccountName().toLowerCase();
        String metadataNamespace = ADLSHook.getInstance() != null ?
                                   ADLSHook.getInstance().getMetadataNamespace() : DEFAULT_METADATA_NAMESPACE;

        return fileSystemScheme + SCHEME_SEPARATOR
                + blobContainerName + QNAME_SEP_METADATA_NAMESPACE
                + accountName + QNAME_SEP_METADATA_NAMESPACE
                + metadataNamespace;
    }

    @Override
    public String name() {
        return blobContainerClient.getBlobContainerName() + Constants.DEFAULT_CONTAINER_ACCOUNT_DELIMITER + blobContainerClient.getAccountName();
    }

    @Override
    public String getParentName() {
        return blobContainerClient.getAccountName();
    }

    @Override
    public AtlasEntity getEntity() {
        return entity;
    }

    @Override
    public AtlasEntity.AtlasEntitiesWithExtInfo getAtlasEntitiesWithExtInfo() {
        return new AtlasEntity.AtlasEntitiesWithExtInfo(this.getEntity());
    }

    private boolean isInvalidContainer(BlobContainerClient blobContainerClient) {
        return blobContainerClient == null ||
               StringUtils.isEmpty(blobContainerClient.getBlobContainerName()) ||
               blobContainerClient.getProperties() == null;
    }

    private AtlasEntity constructAtlasEntity() {
        AtlasEntity entity = new AtlasEntity(ADLSDataTypes.ADLS_GEN2_CONTAINER.getEntityType());

        entity.setAttribute(PROPERTY_QUALIFIED_NAME, getQualifiedName());
        entity.setAttribute(PROPERTY_NAME, blobContainerClient.getBlobContainerName());

        if (blobContainerClient.getBlobContainerUrl() != null) {
            entity.setAttribute(PROPERTY_URL, blobContainerClient.getBlobContainerUrl());
        }

        if (blobContainerClient.getProperties().getLastModified() != null) {
            entity.setAttribute(PROPERTY_CREATION_TIME, blobContainerClient.getProperties().getLastModified().toInstant().toEpochMilli());
        }

        if (blobContainerClient.getProperties().getETag() != null) {
            entity.setAttribute(PROPERTY_ETAG, blobContainerClient.getProperties().getETag());
        }

        if (blobContainerClient.getProperties().getBlobPublicAccess() != null) {
            entity.setAttribute(PROPERTY_PUBLIC_ACCESS, blobContainerClient.getProperties().getBlobPublicAccess());
        }

        if (blobContainerClient.getProperties().getMetadata() != null) {
            entity.setAttribute(PROPERTY_ADDITIONAL_PROPERTIES, blobContainerClient.getProperties().getMetadata());
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("Constructed Atlas entity with name {} and qualified name {}", blobContainerClient.getBlobContainerName(), getQualifiedName());
        }

        return entity;
    }
}