/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.extractors;

import com.azure.core.http.rest.PagedIterable;
import com.azure.core.http.rest.PagedResponse;
import com.azure.storage.blob.BlobClient;
import com.azure.storage.blob.BlobContainerClient;
import com.azure.storage.blob.models.*;
import org.apache.atlas.adls.bridge.AtlasADLSPartialUpdateEvent;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.instance.AtlasRelatedObjectId;
import org.apache.atlas.adls.bridge.ADLSHook;
import org.apache.atlas.adls.bridge.AtlasADLSCreateEvent;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.adls.bridge.AtlasADLSDeleteEvent;
import org.apache.atlas.adls.bridge.models.AtlasADLSContainerEntity;
import org.apache.atlas.adls.bridge.models.AtlasADLSDirectoryEntity;
import org.apache.atlas.adls.bridge.models.AtlasADLSEntity;
import org.apache.atlas.adls.bridge.models.AtlasADLSBlobEntity;
import org.apache.atlas.type.AtlasTypeUtil;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

public class ADLSBlobExtractor implements Serializable {
    private static final long serialVersionUID                       = 1L;
    private static final Logger LOG                                  = LoggerFactory.getLogger(ADLSBlobExtractor.class);
    private static final String AZURE_PARENT_CHILD_RELATIONSHIP_DEF  = "adls_gen2_parent_children";
    private static final String PARENT_END_DEF                       = "parent";
    private static final String PROPERTY_NEXT_TOKEN_AT_PROGRESS_FILE = "_NEXT_TOKEN_FOR_BLOB_FETCH";

    private final ADLSExtractorContext extractorContext;

    public ADLSBlobExtractor(ADLSExtractorContext extractorContext) {
        this.extractorContext = extractorContext;
    }

    public void extractAllRequiredBlobs(BlobContainerClient blobContainerClient, AtlasADLSContainerEntity containerEntity) throws AtlasBaseException {
        if (blobContainerClient != null && containerEntity != null) {
            Set<String> prefixes = new TreeSet<>();

            if (extractorContext.isAllowListingConfigurationPresent() && MapUtils.isNotEmpty(extractorContext.getAllowableContainerPaths())) {
                prefixes = extractorContext.getAllowableContainerPaths().get(containerEntity.name());
            } else {
                prefixes.add("");
            }

            for (String prefix : prefixes) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Listing blobs using Prefix{} for container{}.", prefix, blobContainerClient.getBlobContainerName());
                }

                int retry = 0;
                while (retry < extractorContext.getMaxReconnectCount()) {
                    try {
                        PagedIterable<BlobItem> blobItems = getADLSListBlobsIterable(blobContainerClient, containerEntity.name(), prefix);


                        for (PagedResponse<BlobItem> eachPage : blobItems.iterableByPage()) {
                            AtlasADLSEntity lastParentEntity = containerEntity;

                            for (BlobItem eachBlobItem : eachPage.getElements()) {
                                try {
                                    lastParentEntity = createADLSDirOrBlobEntitiesAtAtlas(eachBlobItem, containerEntity, lastParentEntity);
                                } catch (AtlasBaseException e) {
                                    LOG.error("Unable to extract blob {}", eachBlobItem.getName(), e);

                                    if (extractorContext.getFailOnErrorFlag()) {
                                        throw new AtlasBaseException(e.getMessage(), e);
                                    }
                                }
                            }

                            writeNextTokenToProgressFile(containerEntity.name(), prefix, eachPage.getContinuationToken());
                        }

                        writeNextTokenToProgressFile(containerEntity.name(), prefix, null);
                        break;
                    } catch (Exception e) {
                        LOG.warn("Exception occurred while fetching blobs for container {}", blobContainerClient.getBlobContainerName(), e);
                        retry++;
                    }
                }
            }
        }
    }

    public void createADLSDirOrBlobEntitiesAtAtlas(BlobItem blobItem, AtlasADLSContainerEntity containerEntity) throws AtlasBaseException {
        createADLSDirOrBlobEntitiesAtAtlas(blobItem, containerEntity, containerEntity);
    }

    public BlobItem getBlobWithMoreDetailsIfPresent(BlobContainerClient blobContainerClient, String blobName, boolean isDirectory) throws AtlasBaseException {
        BlobItem ret = new BlobItem();

        ret.setName(blobName);

        if (isDirectory) {
            Map<String, String> metadata = new HashMap<>();

            metadata.put(Constants.IS_DIRECTORY_METADATA_KEY, Boolean.TRUE.toString());

            ret.setMetadata(metadata);
        }

        int retry = 0;

        if (checkIfBlobExists(blobContainerClient, blobName)) {
            while (retry < extractorContext.getMaxReconnectCount()) {
                try {
                    BlobClient         blobClient            = blobContainerClient.getBlobClient(blobName);
                    BlobProperties     fromBlobProperties    = blobClient.getProperties();
                    BlobItemProperties retBlobItemProperties = new BlobItemProperties();

                    retBlobItemProperties.setCreationTime(fromBlobProperties.getCreationTime())
                            .setETag(fromBlobProperties.getETag())
                            .setLastModified(fromBlobProperties.getLastModified())
                            .setCreationTime(fromBlobProperties.getCreationTime())
                            .setContentLength(fromBlobProperties.getBlobSize())
                            .setContentType(fromBlobProperties.getContentType())
                            .setContentMd5(fromBlobProperties.getContentMd5())
                            .setContentEncoding(fromBlobProperties.getContentEncoding())
                            .setContentDisposition(fromBlobProperties.getContentDisposition())
                            .setCacheControl(fromBlobProperties.getCacheControl())
                            .setBlobType(fromBlobProperties.getBlobType())
                            .setAccessTier(fromBlobProperties.getAccessTier())
                            .setServerEncrypted(fromBlobProperties.isServerEncrypted());

                    ret.setMetadata(fromBlobProperties.getMetadata()).setProperties(retBlobItemProperties);

                    return ret;
                } catch (Exception e) {
                    retry++;
                }
            }

            LOG.info(String.format("Azure Data Lake Storage exception occurred while checking blob {%s} for container{%s} and account{%s}",
                    blobName, blobContainerClient.getBlobContainerName(), blobContainerClient.getAccountName()));
        }

        return ret;
    }

    public boolean checkIfBlobExists(BlobContainerClient blobContainerClient, String blobName) throws AtlasBaseException {
        int retry  = 0;

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                BlobClient blobClient = blobContainerClient.getBlobClient(blobName);

                return blobClient.exists();
            } catch (Exception e) {
                retry++;
            }
        }

        throw new AtlasBaseException(String.format("Azure Data Lake Storage exception occurred while checking existence of blob {%s} for container{%s} and account{%s}",
                blobName, blobContainerClient.getBlobContainerName(), blobContainerClient.getAccountName()));
    }

    private AtlasADLSEntity createDirectoryHierarchyAtAtlas(String containerName, BlobItem blobItem,
                                                            AtlasADLSEntity parentEntity) throws AtlasBaseException {
        String directoriesToBeCreated = getDirectory(blobItem);

        if (!StringUtils.isEmpty(directoriesToBeCreated)) {
            String directory = getRemainingDirectoryNeedsTobeCreated(directoriesToBeCreated, parentEntity.name());

            if (!StringUtils.isEmpty(directory)) {
                String              directoryKey = blobItem.getName().contains(parentEntity.name()) ? parentEntity.name() + directory : directory;
                BlobItem            newBlobItem  = new BlobItem().setName(directoryKey);
                Map<String, String> metadata     = new HashMap<>();

                metadata.put(Constants.IS_DIRECTORY_METADATA_KEY, Boolean.TRUE.toString());
                newBlobItem.setMetadata(metadata);

                parentEntity = createADLSEntityAtAtlas(containerName, newBlobItem, parentEntity);
                parentEntity = createDirectoryHierarchyAtAtlas(containerName, blobItem, parentEntity);
            }
        }

        return parentEntity;
    }

    /**
     * Returns the parent entity for the child
     */
    public AtlasADLSEntity createADLSDirOrBlobEntitiesAtAtlas(BlobItem blobItem, AtlasADLSContainerEntity containerEntity, AtlasADLSEntity parentEntity) throws AtlasBaseException {
        if (isEntityChangeNeeded(containerEntity, blobItem)) {
            int    retry  = 0;
            String errMsg = String.format("Azure Data Lake Storage exception occurred while creating Atlas entity with name{%s}", blobItem.getName());

            while (retry < extractorContext.getMaxReconnectCount()) {
                try {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Atlas entity will be created for Azure Data Lake Storage Blob {} as the path is allow listed", blobItem.getName());
                    }

                    if (isDirectlyInsideContainer(blobItem, parentEntity)) {
                        parentEntity = containerEntity;
                    }

                    parentEntity = createDirectoryHierarchyAtAtlas(containerEntity.name(), blobItem, parentEntity);
                    parentEntity = createADLSEntityAtAtlas(containerEntity.name(), blobItem, parentEntity);

                    return parentEntity;
                } catch (IllegalArgumentException | AtlasBaseException e) {
                    LOG.error("Exception occurred while creating an Atlas entity: " + e.getMessage());
                    throw new AtlasBaseException("Exception occurred while creating an Atlas entity: " + e.getMessage());
                } catch (Exception e) {
                    LOG.error("Exception occurred while creating an Atlas entity: " + e.getMessage());
                    //Still unable to find more information about the retryable exception. So for now retry for all exceptions
                    retry++;
                }
            }

            throw new AtlasBaseException(errMsg);
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("No Atlas entity will be created for Azure Data Lake Storage Blob {} as either the directory is not in allow list or it is in deny list",
                        blobItem.getName());
            }
        }

        return parentEntity;
    }

    public void updateEntireDirectoryContentAtAtlas(BlobContainerClient blobContainerClient, BlobItem oldBlob, BlobItem newBlob, AtlasADLSContainerEntity containerEntity) throws AtlasBaseException {
        AtlasADLSEntity parentEntity = updateEntityAtAtlas(oldBlob, newBlob, containerEntity, containerEntity);

        renameAllRequiredBlobsForRename(blobContainerClient, containerEntity, oldBlob.getName(), newBlob.getName(), parentEntity);
    }

    public AtlasADLSEntity updateEntityAtAtlas(BlobItem oldBlob, BlobItem newBlob, AtlasADLSEntity parentEntity, AtlasADLSContainerEntity containerEntity) throws AtlasBaseException {
        AtlasADLSEntity newAtlasADLSEntity, oldAtlasADLSEntity;

        if (isADLSDirectory(newBlob)) {
            newAtlasADLSEntity = new AtlasADLSDirectoryEntity(extractorContext, containerEntity.name(), newBlob);
            oldAtlasADLSEntity = new AtlasADLSDirectoryEntity(extractorContext, containerEntity.name(), oldBlob);
        } else {
            newAtlasADLSEntity = new AtlasADLSBlobEntity(extractorContext, containerEntity.name(), newBlob);
            oldAtlasADLSEntity = new AtlasADLSBlobEntity(extractorContext, containerEntity.name(), oldBlob);
        }

        if (isEntityChangeNeeded(containerEntity, newBlob)) {
            // Update qualifiedName and other attributes (like params - which include lastModifiedTime, lastModifiedBy) of the table
            publishPartialUpdateEventToAtlas(oldAtlasADLSEntity, newAtlasADLSEntity);

            // Send create entity to handle if the entity not present at Atlas
            parentEntity = createADLSDirOrBlobEntitiesAtAtlas(newBlob, containerEntity, parentEntity);
        } else if (isEntityChangeNeeded(containerEntity, oldBlob) || extractorContext.isDeleteNeededForRenameIfNewBlobNotAllowListed()){
            publishDeleteEventToAtlas(oldAtlasADLSEntity);
        }

        return parentEntity;
    }

    public void deleteEntityAtAtlas(BlobItem blobItem, AtlasADLSContainerEntity containerEntity) throws AtlasBaseException {
        AtlasADLSEntity entity;

        if (isADLSDirectory(blobItem)) {
            entity = new AtlasADLSDirectoryEntity(extractorContext, containerEntity.name(), blobItem);
        } else {
            entity = new AtlasADLSBlobEntity(extractorContext, containerEntity.name(), blobItem);
        }

        publishDeleteEventToAtlas(entity);
    }

    private void renameAllRequiredBlobsForRename(BlobContainerClient blobContainerClient, AtlasADLSContainerEntity containerEntity,
                                                 String oldParent, String newParent, AtlasADLSEntity parentEntity) throws AtlasBaseException {
        if (blobContainerClient != null && containerEntity != null) {
            PagedIterable<BlobItem> blobItems = getADLSListBlobsIterable(blobContainerClient, containerEntity.name(), newParent);

            for (PagedResponse<BlobItem> eachPage : blobItems.iterableByPage()) {
                for (BlobItem eachBlobItem : eachPage.getElements()) {
                    try {
                        BlobItem oldBlob = getOldBlobItemFromNew(eachBlobItem, newParent, oldParent);
                        parentEntity     = updateEntityAtAtlas(oldBlob, eachBlobItem, parentEntity, containerEntity);
                    } catch (AtlasBaseException e) {
                        LOG.error("Unable to extract blob {}", eachBlobItem.getName(), e);

                        if (extractorContext.getFailOnErrorFlag()) {
                            throw new AtlasBaseException(e.getMessage(), e);
                        }
                    }
                }
            }
        }
    }

    private BlobItem getOldBlobItemFromNew(BlobItem newBlobItem, String newParent, String oldParent) {
        BlobItem oldBlob = new BlobItem();

        oldBlob.setName(newBlobItem.getName().replace(newParent, oldParent));
        oldBlob.setMetadata(newBlobItem.getMetadata());

        return oldBlob;
    }

    public void publishCreateEventToAtlas(AtlasADLSEntity entity) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("publish create blob entity event to atlas.");
        }

        try {
            AtlasADLSCreateEvent createEvent = new AtlasADLSCreateEvent(entity.getAtlasEntitiesWithExtInfo());

            ADLSHook.getInstance().publishEvent(createEvent);
        } catch (Exception e) {
            String msg = String.format("Error encountered while publishing create blob entity for blob{%s} and container{%s}.", entity.name(), entity.getParentName());

            LOG.error(msg, e);

            throw new AtlasBaseException(msg, e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("Published Azure Data Lake Storage Object entity create event to Atlas.");
        }
    }

    public void publishDeleteEventToAtlas(AtlasADLSEntity entity) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("publish delete blob entity event to atlas.");
        }

        try {
            AtlasADLSDeleteEvent deleteEvent = new AtlasADLSDeleteEvent(entity);

            ADLSHook.getInstance().publishEvent(deleteEvent);
        } catch (Exception e) {
            String msg = String.format("Error encountered while publishing delete blob entity for blob{%s} and container{%s}.", entity.name(), entity.getParentName());

            LOG.error(msg, e);

            throw new AtlasBaseException(msg, e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("Published Azure Data Lake Storage Blob entity delete event to Atlas.");
        }
    }

    public void publishPartialUpdateEventToAtlas(AtlasADLSEntity oldAtlasADLSEntity, AtlasADLSEntity entity) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("publish partial update blob entity event to atlas.");
        }

        try {
            AtlasADLSPartialUpdateEvent partialUpdateEvent = new AtlasADLSPartialUpdateEvent(oldAtlasADLSEntity, entity);

            ADLSHook.getInstance().publishEvent(partialUpdateEvent);
        } catch (Exception e) {
            String msg = String.format("Error encountered while publishing delete blob entity for blob{%s} and container{%s}.", entity.name(), entity.getParentName());

            LOG.error(msg, e);

            throw new AtlasBaseException(msg, e);
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("Published Azure Data Lake Storage Blob entity delete event to Atlas.");
        }
    }

    private PagedIterable<BlobItem> getADLSListBlobsIterable(BlobContainerClient blobContainerClient, String containerAndAccountName, String prefix) throws AtlasBaseException {
        String                errMsg   = String.format("Azure Data Lake Storage exception occurred while listing blobs for container@accountName{%s}", containerAndAccountName);
        int                   retry    = 0;

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                ListBlobsOptions listBlobsOptions = new ListBlobsOptions()
                        .setPrefix(prefix)
                        .setMaxResultsPerPage(extractorContext.getMaxNumberOfBlobFetchPerCall())
                        .setDetails(new BlobListDetails().setRetrieveDeletedBlobs(true).setRetrieveSnapshots(true).setRetrieveMetadata(true));

                String continuationToken = getNextTokenForObjectExtraction(containerAndAccountName, prefix);

                return blobContainerClient.listBlobs(listBlobsOptions, continuationToken, extractorContext.getTimeoutInSecPerCall());
            } catch (Exception e) {
                LOG.warn("Exception occurred while listing blobs for container{}", blobContainerClient.getBlobContainerName(), e);
                retry++;
            }
        }

        throw new AtlasBaseException(errMsg);
    }

    private String getNextTokenForObjectExtraction(String containerAndAccountName, String prefix) {
        String ret = null;

        if (extractorContext.isResumeFromProgressFileNeeded()) {
            ret = extractorContext.getPropertyFromProgressFile(containerAndAccountName + prefix + PROPERTY_NEXT_TOKEN_AT_PROGRESS_FILE);
        }

        return ret;
    }

    public boolean isEntityChangeNeeded(AtlasADLSContainerEntity containerEntity, BlobItem blobItem) {
        String keyWithPrefix = blobItem.getName();

        if (isADLSDirectory(blobItem) && !doesBlobKeyEndsWithDirectorySeparator(keyWithPrefix)) {
            keyWithPrefix += extractorContext.getPrefixDelimiter();
        }

        return extractorContext.doesEntityInAllowListAndNotInDenyList(containerEntity.name(), keyWithPrefix);
    }

    private boolean doesBlobKeyEndsWithDirectorySeparator(String blobName) {
        return blobName.endsWith(extractorContext.getPrefixDelimiter());
    }

    private boolean isDirectlyInsideContainer(BlobItem blobItem, AtlasADLSEntity lastParentEntity) {
        boolean isInsideContainer = true;
        String  directoryPath     = getDirectory(blobItem);

        if (!directoryPath.isEmpty()) {
            isInsideContainer = !(directoryPath.contains(lastParentEntity.name()));
        }

        return isInsideContainer;
    }

    private String getDirectory(BlobItem blobItem) {
        String blobFullPath = blobItem.getName();

        if (isADLSDirectory(blobItem)) {
            blobFullPath = removeLastPrefixDelimiterFromDirectory(blobItem.getName());
        }

        int lastPrefixDelimiterPos = blobFullPath.lastIndexOf(extractorContext.getPrefixDelimiter());

        return blobFullPath.substring(0, lastPrefixDelimiterPos + 1);
    }

    private boolean isADLSDirectory(BlobItem blobItem) {
        return MapUtils.isNotEmpty(blobItem.getMetadata()) &&
               blobItem.getMetadata().containsKey(Constants.IS_DIRECTORY_METADATA_KEY) &&
               blobItem.getMetadata().get(Constants.IS_DIRECTORY_METADATA_KEY).equalsIgnoreCase(Boolean.TRUE.toString());
    }

    private String getRemainingDirectoryNeedsTobeCreated(String directoryKey, String directoryAlreadyCreated) {
        int positionAfterAlreadyCreatedDirectory = 0;

        if (!StringUtils.isEmpty(directoryAlreadyCreated) && directoryKey.contains(directoryAlreadyCreated)) {
            positionAfterAlreadyCreatedDirectory = directoryKey.indexOf(directoryAlreadyCreated) + directoryAlreadyCreated.length();
        }

        int positionTillNextDelimiter = directoryKey.indexOf(extractorContext.getPrefixDelimiter(), positionAfterAlreadyCreatedDirectory) + 1;

        String directoryToBeCreated = "";

        if (positionTillNextDelimiter > positionAfterAlreadyCreatedDirectory) {
            directoryToBeCreated = directoryKey.substring(positionAfterAlreadyCreatedDirectory, positionTillNextDelimiter);
        }

        return directoryToBeCreated;
    }

    private void linkAtlasADLSContainerToContained(AtlasEntity parentAtlasEntity, AtlasEntity childAtlasEntity) {
        parentAtlasEntity.setGuid(null);

        AtlasObjectId        parentEntityId  = AtlasTypeUtil.getObjectId(parentAtlasEntity);
        AtlasRelatedObjectId relatedObjectId = new AtlasRelatedObjectId(parentEntityId, AZURE_PARENT_CHILD_RELATIONSHIP_DEF);

        childAtlasEntity.setRelationshipAttribute(PARENT_END_DEF, relatedObjectId);
    }

    private void writeNextTokenToProgressFile(String containerAndStorageAccountName, String prefix, String nextToken) {
        extractorContext.writePropertyToProgressFile(containerAndStorageAccountName + prefix + PROPERTY_NEXT_TOKEN_AT_PROGRESS_FILE, nextToken);
    }

    private String removeLastPrefixDelimiterFromDirectory(String directoryPath) {
        return StringUtils.chop(directoryPath);
    }

    /**
     * Returns the parent entity of the created entity if its a Blob type or entity itself if its a directory
     */
    private AtlasADLSEntity createADLSEntityAtAtlas(String containerAndAccountName, BlobItem blobItem, AtlasADLSEntity parentEntity) throws AtlasBaseException {
        AtlasADLSEntity newToBeParentEntity = parentEntity;
        AtlasADLSEntity entity;

        try {
            if (isADLSDirectory(blobItem)) {
                entity              = new AtlasADLSDirectoryEntity(extractorContext, containerAndAccountName, blobItem);
                newToBeParentEntity = entity;
            } else {
                entity = new AtlasADLSBlobEntity(extractorContext, containerAndAccountName, blobItem);
            }

            linkAtlasADLSContainerToContained(parentEntity.getEntity(), entity.getEntity());

            publishCreateEventToAtlas(entity);
        } catch (IllegalArgumentException | AtlasBaseException e) {
            LOG.error("Exception occurred while creating an ADLS directory entity: " + e.getMessage());

            throw new AtlasBaseException("Exception occurred while creating an ADLS directory entity: " + e.getMessage());
        }

        return newToBeParentEntity;
    }
}
