/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.extractors;

import com.azure.core.credential.AccessToken;
import com.azure.storage.blob.BlobServiceClientBuilder;
import com.azure.storage.common.StorageSharedKeyCredential;
import com.azure.storage.queue.QueueServiceClientBuilder;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import reactor.core.publisher.Mono;

/**
 * An implementation class for AWS IDBrokerAzureCredential provider.
 * It reads the credentials details from atlas configuration.
 */
public class ADLSCredentialsProvider {
    private static final Logger LOG = LoggerFactory.getLogger(ADLSCredentialsProvider.class);

    private final ADLSExtractorContext extractorContext;

    public ADLSCredentialsProvider(ADLSExtractorContext extractorContext) {
        this.extractorContext = extractorContext;
    }

    public void setCredential(BlobServiceClientBuilder blobServiceClientBuilder) {
        try {
            if (StringUtils.isNotEmpty(extractorContext.getADLSAccessToken())) {
                AccessToken accessToken = getAccessTokenForAzure();
                blobServiceClientBuilder.credential(request -> Mono.just(accessToken));
            } else {
                StorageSharedKeyCredential credential = getStorageSharedKeyCredentialForAzure();
                blobServiceClientBuilder.credential(credential);
            }
        } catch (AtlasBaseException e) {
            LOG.error("Error occurred while creating Azure Data Lake Storage Credential for Blob Service Client", e);
        }
    }

    public void setCredential(QueueServiceClientBuilder queueServiceClientBuilder) {
        try {
            if (StringUtils.isNotEmpty(extractorContext.getADLSAccessToken())) {
                AccessToken accessToken = getAccessTokenForAzure();
                queueServiceClientBuilder.credential(request -> Mono.just(accessToken));
            } else {
                StorageSharedKeyCredential credential = getStorageSharedKeyCredentialForAzure();
                queueServiceClientBuilder.credential(credential);
            }
        } catch (AtlasBaseException e) {
            LOG.error("Error occurred while creating Azure Data Lake Storage Credential for Queue Service Client", e);
        }
    }

    private AccessToken getAccessTokenForAzure() throws AtlasBaseException {
        return new AccessToken(extractorContext.getADLSAccessToken(), extractorContext.getADLSAccessTokenExpireAt());
    }

    private StorageSharedKeyCredential getStorageSharedKeyCredentialForAzure() {
        return new StorageSharedKeyCredential(extractorContext.getAzureAccountNameConfig(),
                extractorContext.getAzureAccountKeyConfig());
    }
}
