/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.extractors;

import org.apache.atlas.adls.bridge.AzureIDBrokerHandler;
import org.apache.atlas.utils.XMLConfigLoader;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.adls.bridge.ADLSHook;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.ObjectUtils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.time.Duration;
import java.time.LocalDateTime;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;

import static org.apache.atlas.adls.bridge.models.AtlasADLSEntity.SCHEME_SEPARATOR;

/**
 * A context class that contains shared information for extraction
 */
public abstract class ADLSExtractorContext implements Serializable {

    private static final long   serialVersionUID        = 1L;
    private static final Logger LOG                     = LoggerFactory.getLogger(ADLSExtractorContext.class);

    private         AzureIDBrokerHandler       idBrokerHandler;
    private final   Map<String, Set<String>>   allowableContainerPaths;
    private final   Map<String, Set<String>>   allowingAccountAndContainers;
    private final   Map<String, Set<String>>   deniableContainerPaths;
    private final   Properties                 progressPropertyFile;
    private final   Properties                 coreSiteConfiguration;
    private         boolean                    isCoreConfigPresent;
    private final   boolean                    failOnError;
    protected final Properties                 extractionConfiguration;
    protected final Map<String, Set<String>>   accountAndQueueNames;

    public ADLSExtractorContext(Properties extractionConfiguration) throws AtlasBaseException {
        this.extractionConfiguration      = extractionConfiguration;
        this.progressPropertyFile         = new Properties();
        this.coreSiteConfiguration        = new Properties();
        this.allowableContainerPaths      = new HashMap<>();
        this.deniableContainerPaths       = new HashMap<>();
        this.allowingAccountAndContainers = new HashMap<>();
        this.accountAndQueueNames         = new HashMap<>();
        this.isCoreConfigPresent          = false;
        this.failOnError                  = Boolean.parseBoolean(extractionConfiguration.getProperty(Constants.FAIL_ON_ERROR_FLAG));
    }

    public void initialize() throws AtlasBaseException {
        loadCoreSiteConfiguration();

        validateConfiguration();

        ADLSHook.getInstance().initialize();

        loadAllowable();

        loadDenyListing();

        loadProgressFile();
    }

    public Map<String, Set<String>> getAllowableContainerPaths() {
        return allowableContainerPaths;
    }

    public Map<String, Set<String>> getAllowableAccountAndContainers() {
        return allowingAccountAndContainers;
    }

    public Map<String, Set<String>> getAccountAndQueueNames() {
        return accountAndQueueNames;
    }

    public Set<String> getAllowableAccountNames() {
        return allowingAccountAndContainers.keySet();
    }

    public String getQueueName() {
        return extractionConfiguration.getProperty(Constants.CONFIG_QUEUE_NAMES);
    }

    public String getPrefixDelimiter() {
        String prefixDelimiter = extractionConfiguration.getProperty(Constants.CONFIG_PREFIX_DELIMITER);

        if (StringUtils.isEmpty(prefixDelimiter)) {
            prefixDelimiter = Constants.DEFAULT_PREFIX_DELIMITER;
        }

        return prefixDelimiter;
    }

    public String getPropertyFromProgressFile(String property) {
        return progressPropertyFile.getProperty(property);
    }

    public int getMaxReconnectCount() {
        String maxReconnectCount = extractionConfiguration.getProperty(Constants.CONFIG_MAX_RECONNECT_COUNT);

        if (StringUtils.isEmpty(maxReconnectCount)) {
            return Constants.DEFAULT_MAX_RECONNECT_COUNT;
        } else {
            return Integer.parseInt(maxReconnectCount);
        }
    }

    public String getFileSystemScheme() {
        String fileSystemScheme = extractionConfiguration.getProperty(Constants.CONFIG_ADLS_FILE_SYSTEM);

        if (StringUtils.isEmpty(fileSystemScheme)) {
            fileSystemScheme = Constants.DEFAULT_FILE_SYSTEM_SCHEME;
        }

        return fileSystemScheme;
    }

    public void writePropertyToProgressFile(final String property, final String value) {
        try {
            File         progressFile    = new File(getExtractorProgressFile());
            OutputStream progressFileOut = new FileOutputStream(progressFile);

            if (StringUtils.isEmpty(value)) {
                LOG.info("Removing the property {} from progress file", property);

                progressPropertyFile.remove(property);
            } else {
                LOG.info("Adding the property {} and value {} to progress file", property, value);

                progressPropertyFile.setProperty(property, value);
            }

            progressPropertyFile.store(progressFileOut, "");
        } catch (IOException e) {
            LOG.warn("Error while writing property {} to extraction progress file: " + e.getMessage());
        }
    }

    public String getAzureAccountNameConfig() {
        return extractionConfiguration.getProperty(Constants.CONFIG_ACCOUNT_NAME);
    }

    public String getAzureAccountKeyConfig() {
        return extractionConfiguration.getProperty(Constants.CONFIG_ACCOUNT_KEY);
    }

    public String getADLSAccessToken() throws AtlasBaseException {
        String ret = getAzureAccessTokenConfig();

        if (isADLSCredentialFetchNeededFromIDBroker()) {
            refreshADLSCredentialFromIDBroker();

            ret = idBrokerHandler.getADLSAccessToken();
        }

        return ret;
    }

    public OffsetDateTime getADLSAccessTokenExpireAt() throws AtlasBaseException {
        OffsetDateTime ret = getAzureAccessTokenExpiresAtConfig();

        if (isADLSCredentialFetchNeededFromIDBroker()) {
            refreshADLSCredentialFromIDBroker();

            ret = idBrokerHandler.getADLSAccessTokenExpiresAt();
        }

        return ret;
    }

    public void refreshADLSCredentialFromIDBroker() throws AtlasBaseException {
        if (idBrokerHandler.isIDBrokerEnabled()) {
            idBrokerHandler.fetchADLSCredentialIfNeeded();
        }
    }

    public Duration getTimeoutInSecPerCall() {
        Duration maxObjectFetchPerCall = Constants.DEFAULT_TIMEOUT_DURATION_IN_SEC;
        String   maxObjectPerCall      = extractionConfiguration.getProperty(Constants.CONFIG_TIMEOUT_PER_CALL_IN_SEC);

        if (StringUtils.isNotEmpty(maxObjectPerCall)) {
            maxObjectFetchPerCall = Duration.ofSeconds(Long.parseLong(maxObjectPerCall));
        }

        return maxObjectFetchPerCall;
    }

    public boolean getFailOnErrorFlag() {
        return this.failOnError;
    }

    boolean isResumeFromProgressFileNeeded() {
        boolean ret                          = Constants.DEFAULT_RESUME_FROM_PROGRESS_FILE;
        String  resumeFromProgressFileConfig = extractionConfiguration.getProperty(Constants.CONFIG_RESUME_FROM_PROGRESS_FILE);

        if (StringUtils.isNotEmpty(resumeFromProgressFileConfig) && resumeFromProgressFileConfig.equalsIgnoreCase("true")) {
            ret = true;
        }

        return ret;
    }

    public boolean isDeleteNeededForRenameIfNewBlobNotAllowListed() {
        boolean ret                          = Constants.DEFAULT_DELETE_NEEDED_FOR_RENAME;
        String  resumeFromProgressFileConfig = extractionConfiguration.getProperty(Constants.CONFIG_DELETE_NEEDED_FOR_RENAME);

        if (StringUtils.isNotEmpty(resumeFromProgressFileConfig) && resumeFromProgressFileConfig.equalsIgnoreCase("false")) {
            ret = false;
        }

        return ret;
    }

    int getMaxNumberOfBlobFetchPerCall() {
        int    maxObjectFetchPerCall = Constants.DEFAULT_MAX_NUMBER_OF_BLOB_PER_CALL;
        String maxObjectPerCall      = extractionConfiguration.getProperty(Constants.CONFIG_EXTRACTOR_MAX_BLOB_PER_CALL);

        if (StringUtils.isNotEmpty(maxObjectPerCall)) {
            maxObjectFetchPerCall = Integer.parseInt(maxObjectPerCall);
        }

        return maxObjectFetchPerCall;
    }

    public boolean isContainerAllowListed(final String containerAndAccountName) {
        return (isAllowListingConfigurationPresent() && allowableContainerPaths.containsKey(containerAndAccountName)) ||
                (!isAllowListingConfigurationPresent() && !(deniableContainerPaths.containsKey(containerAndAccountName)
                        && deniableContainerPaths.get(containerAndAccountName).size() == 1
                        && StringUtils.isEmpty(deniableContainerPaths.get(containerAndAccountName).iterator().next())));
    }

    boolean doesEntityInAllowListAndNotInDenyList(String containerAndAccountName, String prefix) {
        boolean ret = false;

        if (!isAllowListingConfigurationPresent()) {
            ret = true;
        } else if (allowableContainerPaths.containsKey(containerAndAccountName)) {
            //Check allow listing first
            for (String keyFromList : allowableContainerPaths.get(containerAndAccountName)) {
                if (prefix.indexOf(keyFromList) == 0 || keyFromList.indexOf(prefix) == 0) {
                    ret = true;

                    break;
                }
            }
        }

        //Now check if it is also deny listed
        if (deniableContainerPaths.containsKey(containerAndAccountName)) {
            for (String keyFromList : deniableContainerPaths.get(containerAndAccountName)) {
                if (prefix.indexOf(keyFromList) == 0) {
                    ret = false;

                    break;
                }
            }
        }

        return ret;
    }

    public boolean isAllowListingConfigurationPresent() {
        return StringUtils.isNotEmpty(extractionConfiguration.getProperty(Constants.CONFIG_WHITELIST_PATHS));
    }

    protected void loadCoreSiteConfiguration() {
        try {
            XMLConfigLoader.loadConfig(Constants.CORE_SITE_PROPERTIES, coreSiteConfiguration);

            isCoreConfigPresent = true;
        } catch (AtlasBaseException e) {
            LOG.warn("Configuration file {} is missing and IDBroker will not be used", Constants.CORE_SITE_PROPERTIES);
        }
    }

    protected void validateConfiguration() throws AtlasBaseException {
        if (isADLSCredentialFetchNeededFromIDBroker()) {
            this.idBrokerHandler = new AzureIDBrokerHandler(coreSiteConfiguration);
        }

        if (!isAllowListingConfigurationPresent()) {
            checkEmptinessForConfigurationParameter(Constants.CONFIG_ACCOUNT_NAME, getAzureAccountNameConfig());
        }

        try {
            checkEmptinessForConfigurationParameter(Constants.CONFIG_ACCOUNT_KEY, getAzureAccountKeyConfig());
        } catch (AtlasBaseException e) {
            checkEmptinessForConfigurationParameter(Constants.CONFIG_ACCESS_TOKEN, getADLSAccessToken());
            checkEmptinessForConfigurationParameter(Constants.CONFIG_ACCESS_TOKEN_EXPIRES_AT, getADLSAccessTokenExpireAt());
        }
    }

    protected void checkEmptinessForConfigurationParameter(String configurationParameter, Object configuredValue) throws AtlasBaseException {
        if (ObjectUtils.isEmpty(configuredValue)) {
            String msg = generateMissingConfigurationMessage(configurationParameter);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    protected String generateMissingConfigurationMessage(String extractionConfigurationName) {
        return String.format("There is no value specified for configuration parameter %s", extractionConfigurationName);
    }

    protected String generateInvalidConfigurationMessage(String extractionConfigurationName) {
        return String.format("Invalid value specified for configuration parameter %s", extractionConfigurationName);
    }

    private void loadAllowable() throws AtlasBaseException {
        if (isAllowListingConfigurationPresent()) {
            for (String eachPath : StringUtils.split(extractionConfiguration.getProperty(Constants.CONFIG_WHITELIST_PATHS), Constants.CONFIG_PATH_SEPARATOR)) {
                if (StringUtils.isNotBlank(eachPath)) {
                    eachPath                       = getPathWithoutScheme(eachPath);
                    String containerAndAccountName = getContainerAndAccountNameFromPath(eachPath);
                    String pathPrefix              = getPrefixFromPath(eachPath);

                    if (allowableContainerPaths.containsKey(containerAndAccountName)) {
                        allowableContainerPaths.get(containerAndAccountName).add(pathPrefix);

                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Adding container@Account Name " + containerAndAccountName + " and prefix " + pathPrefix + " to allow list");
                        }
                    } else {
                        Set<String> dirSet = new TreeSet<>();

                        dirSet.add(pathPrefix);

                        allowableContainerPaths.put(containerAndAccountName, dirSet);

                        if (LOG.isDebugEnabled()) {
                            LOG.debug("Adding container@Account Name " + containerAndAccountName + " and prefix " + dirSet + " to allow list");
                        }
                    }
                }
            }

            squeezeContainerAndPrefixEntries(allowableContainerPaths);

            populateAccountContainersMap();
        }
    }

    private void populateAccountContainersMap() {
        if (allowableContainerPaths != null) {
            for (String eachContainerAccount : allowableContainerPaths.keySet()) {
                if (StringUtils.isNotBlank(eachContainerAccount)) {
                    String[] containerAccount = StringUtils.split(eachContainerAccount, Constants.DEFAULT_CONTAINER_ACCOUNT_DELIMITER);

                    if (containerAccount.length == 2) {
                        if (!allowingAccountAndContainers.containsKey(containerAccount[1])) {
                            allowingAccountAndContainers.put(containerAccount[1], new HashSet<>());
                        }

                        allowingAccountAndContainers.get(containerAccount[1]).add(containerAccount[0]);
                    }
                }
            }
        }
    }

    private void loadDenyListing() throws AtlasBaseException {
        String configDenyListing = extractionConfiguration.getProperty(Constants.CONFIG_BLACKLIST_PATHS);

        if (StringUtils.isNotEmpty(configDenyListing)) {
            for (String eachPath : StringUtils.split(configDenyListing, Constants.CONFIG_PATH_SEPARATOR)) {
                if (StringUtils.isNotBlank(eachPath)) {
                    boolean  doesNeedToAddToDenyList = true;
                    eachPath                         = getPathWithoutScheme(eachPath);
                    String   containerAndAccountName = getContainerAndAccountNameFromPath(eachPath);
                    String   pathPrefix              = getPrefixFromPath(eachPath);

                    if (allowableContainerPaths.containsKey(containerAndAccountName)) {
                        removeAllKeysStartsWithDenyListKey(allowableContainerPaths.get(containerAndAccountName), pathPrefix);

                        if (allowableContainerPaths.get(containerAndAccountName).isEmpty()) {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Removing conatiner@Acoount Name " + containerAndAccountName + " from allow list");
                            }

                            allowableContainerPaths.remove(containerAndAccountName);

                            doesNeedToAddToDenyList = false;
                        }
                    }

                    if (doesNeedToAddToDenyList) {
                        if (deniableContainerPaths.containsKey(containerAndAccountName)) {
                            deniableContainerPaths.get(containerAndAccountName).add(pathPrefix);

                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Adding conatiner@Acoount Name " + containerAndAccountName + " and prefix " + pathPrefix + " to deny list");
                            }
                        } else {
                            Set<String> dirSet = new TreeSet<>();

                            dirSet.add(pathPrefix);

                            deniableContainerPaths.put(containerAndAccountName, dirSet);

                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Adding conatiner@Acoount Name " + containerAndAccountName + " and prefix " + dirSet + " to deny list");
                            }
                        }
                    }
                }
            }

            squeezeContainerAndPrefixEntries(deniableContainerPaths);
        }
    }

    private void squeezeContainerAndPrefixEntries(Map<String, Set<String>> containerAndPrefixEntries) {
        Iterator<Map.Entry<String, Set<String>>> containerItr = containerAndPrefixEntries.entrySet().iterator();

        while(containerItr.hasNext()) {
            Map.Entry<String, Set<String>> eachEntry = containerItr.next();

            squeezePrefixEntries(eachEntry.getValue());

            if (CollectionUtils.isEmpty(eachEntry.getValue())) {
                containerItr.remove();
            }
        }
    }

    private void squeezePrefixEntries(Set<String> keys) {
        Iterator<String> itr           = keys.iterator();
        String           lastParentKey = null;

        while (itr.hasNext()) {
            String currKey = itr.next();

            if (lastParentKey != null && currKey.indexOf(lastParentKey) == 0) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Removing prefix " + currKey + " from allow list");
                }

                itr.remove();
            } else {
                lastParentKey = currKey;
            }
        }
    }

    private String getPathWithoutScheme(String path) throws AtlasBaseException {
        String pathWithoutScheme                    = StringUtils.trim(path);
        String fileSystemSchemeWithSeparator        = getFileSystemScheme() + SCHEME_SEPARATOR;
        int    indexOfFileSystemSchemeWithSeparator = pathWithoutScheme.indexOf(fileSystemSchemeWithSeparator);

        if (indexOfFileSystemSchemeWithSeparator == 0) {
            pathWithoutScheme = pathWithoutScheme.substring(indexOfFileSystemSchemeWithSeparator + fileSystemSchemeWithSeparator.length());
        } else {
            throw new AtlasBaseException("Invalid configuration for allow or deny listing: \""
                    + pathWithoutScheme + "\". It should start with: " + fileSystemSchemeWithSeparator);
        }

        return pathWithoutScheme;
    }

    private void removeAllKeysStartsWithDenyListKey(Set<String> allowListSet, String denyListKey) {
        Iterator<String> itr = allowListSet.iterator();

        while (itr.hasNext()) {
            String eachAllowList = itr.next();

            if (eachAllowList.indexOf(denyListKey) == 0) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Removing key " + eachAllowList + " from allow list");
                }

                itr.remove();
            }
        }
    }

    private String getContainerAndAccountNameFromPath(String containerAccountPath) throws AtlasBaseException {
        String containerAndAccountName;
        int    endPosForAccountName = containerAccountPath.indexOf(Constants.DEFAULT_ADLS_GEN2_ACCOUNT_HOST_SUFFIX);

        if (endPosForAccountName > -1) {
            containerAndAccountName = containerAccountPath.substring(0, endPosForAccountName);
        } else {
            throw new AtlasBaseException("Invalid configuration for allow or deny listing: \""
                    + containerAccountPath + "\". It should contain <containerName>@<accountName>"
                    + Constants.DEFAULT_ADLS_GEN2_ACCOUNT_HOST_SUFFIX);
        }

        return containerAndAccountName;
    }

    private String getPrefixFromPath(String path) {
        String prefix                     = "";
        String intermediatePatternForPath = Constants.DEFAULT_ADLS_GEN2_ACCOUNT_HOST_SUFFIX + Constants.DEFAULT_PREFIX_DELIMITER;
        int    startPosForPrefix          = path.indexOf(intermediatePatternForPath);

        if (startPosForPrefix > -1) {
            startPosForPrefix = startPosForPrefix + intermediatePatternForPath.length();

            if (startPosForPrefix < path.length()) {
                prefix = path.substring(startPosForPrefix);
            }
        }

        return prefix;
    }

    private void loadProgressFile() {
        String progressFileName = getExtractorProgressFile();
        File   progressFile     = new File(progressFileName);

        if (progressFile.exists() && progressFile.isFile() && progressFile.canRead()) {
            LOG.info("The progress file " + progressFileName + " is available.");

            try {
                this.progressPropertyFile.load(new FileInputStream(progressFile));
            } catch (IOException e) {
                LOG.error("Error while loading the extractor progress file: " + progressFileName);
            }
        }
    }

    private String getAzureAccessTokenConfig() {
        return this.extractionConfiguration.getProperty(Constants.CONFIG_ACCESS_TOKEN);
    }

    private OffsetDateTime getAzureAccessTokenExpiresAtConfig() {
        String expiresAtStr = this.extractionConfiguration.getProperty(Constants.CONFIG_ACCESS_TOKEN_EXPIRES_AT);
        long expiresAt = 0L;

        if (StringUtils.isNotEmpty(expiresAtStr)) {
            expiresAt = Long.parseLong(expiresAtStr);
        }
        return OffsetDateTime.of(LocalDateTime.ofEpochSecond(expiresAt, 0, ZoneOffset.UTC), ZoneOffset.UTC);
    }

    private String getExtractorProgressFile() {
        String extractorProgressFile = extractionConfiguration.getProperty(Constants.CONFIG_EXTRACTOR_PROGRESS_FILE);

        if (StringUtils.isEmpty(extractorProgressFile)) {
            extractorProgressFile = Constants.DEFAULT_ADLS_EXTRACTOR_PROGRESS_FILE;
        }

        return System.getProperty(Constants.ATLAS_LOG_DIR) + File.separator + extractorProgressFile;
    }

    private boolean isADLSCredentialFetchNeededFromIDBroker() {
        return isCoreConfigPresent &&
                (StringUtils.isEmpty(getAzureAccountNameConfig()) || StringUtils.isEmpty(getAzureAccountKeyConfig())) &&
                StringUtils.isEmpty(getAzureAccessTokenConfig());
    }
}
