/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.adls.extractors.bulk;

import com.azure.storage.blob.BlobContainerClient;
import org.apache.atlas.adls.bridge.models.AtlasADLSAccountEntity;
import org.apache.atlas.adls.bridge.models.AtlasADLSContainerEntity;
import org.apache.atlas.adls.extractors.ADLSBlobExtractor;
import org.apache.atlas.adls.extractors.ADLSClient;
import org.apache.atlas.adls.extractors.Constants;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.adls.extractors.ADLSContainerExtractor;
import org.apache.atlas.adls.extractors.ADLSExtractorContext;
import org.apache.atlas.adls.extractors.ExtractorTask;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

public class ADLSBulkExtractorTask implements ExtractorTask {
    private static final Logger LOG       = LoggerFactory.getLogger(ADLSBulkExtractorTask.class);
    private static final String TASK_NAME = "Azure Data Lake Storage Bulk Extractor";

    private final ADLSExtractorContext   extractorContext;
    private final ADLSContainerExtractor containerHandler;
    private final ADLSBlobExtractor      blobExtractor;
    private       long                   extractionStartTime;

    public ADLSBulkExtractorTask(Properties extractionConfiguration) throws AtlasBaseException {
        this.extractorContext = new ADLSBulkExtractorContext(extractionConfiguration);
        this.containerHandler = new ADLSContainerExtractor(extractorContext);
        this.blobExtractor    = new ADLSBlobExtractor(extractorContext);
    }

    @Override
    public String getName() {
        return TASK_NAME;
    }

    @Override
    public void initialize() throws AtlasBaseException {
        extractorContext.initialize();

        ADLSClient.getInstance().initialize(extractorContext);
    }

    @Override
    public void execute() throws AtlasBaseException {
        try {
            LOG.info("Starting bulk extraction for Azure Data Lake Storage.");

            initializeExtractionStartTime();

            Map<String, Set<String>> accountContainerList;

            if (extractorContext.isAllowListingConfigurationPresent()) {
                accountContainerList = extractorContext.getAllowableAccountAndContainers();
            } else {
                accountContainerList = new HashMap<>();

                String      adlsAccountName   = extractorContext.getAzureAccountNameConfig();
                Set<String> allContainerNames = containerHandler.getAllContainerNames(adlsAccountName);

                accountContainerList.put(adlsAccountName, allContainerNames);
            }

            for (Map.Entry<String, Set<String>> eachEntry : accountContainerList.entrySet()) {
                for (String eachContainer : eachEntry.getValue()) {
                    extractContainerWithBlobs(containerHandler.getContainerClient(eachEntry.getKey(), eachContainer));
                }
            }

            writeExtractionStartTimeToProgressFile();

            LOG.info("Completed bulk meta data extraction for Azure Data Lake Storage.");
        } catch (AtlasBaseException e) {
            throw new AtlasBaseException("Error encountered while performing bulk meta data extraction for Azure Data Lake Storage.", e);
        }
    }

    @Override
    public ADLSExtractorContext getExtractorContext() {
        return extractorContext;
    }

    private void extractContainerWithBlobs(BlobContainerClient blobContainerClient) throws AtlasBaseException {
        try {
            AtlasADLSAccountEntity accountEntity = containerHandler.extractAccount(blobContainerClient);

            AtlasADLSContainerEntity azureADLSContainerEntity = containerHandler.extractContainer(accountEntity, blobContainerClient);

            blobExtractor.extractAllRequiredBlobs(blobContainerClient, azureADLSContainerEntity);
        } catch (AtlasBaseException e) {
            LOG.info("Unable to extract container {} completely", blobContainerClient.getBlobContainerName(), e);

            if (extractorContext.getFailOnErrorFlag()) {
                throw new AtlasBaseException(e.getMessage(), e);
            }
        }
    }

    private void initializeExtractionStartTime() {
        extractionStartTime = System.currentTimeMillis();
    }

    private void writeExtractionStartTimeToProgressFile() {
        extractorContext.writePropertyToProgressFile(Constants.PROPERTY_EXTRACTION_START_TIME_AT_PROGRESS_FILE,
                String.valueOf(extractionStartTime));
    }
}