/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.adls.extractors.incremental;

import org.apache.atlas.adls.extractors.ExtractorTask;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.adls.extractors.ADLSExtractorContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.Set;
import java.util.Properties;
import java.io.Serializable;

public class ADLSIncrementalExtractorTask implements ExtractorTask, Serializable{
    private static final long serialVersionUID = 1L;
    private static final Logger LOG       = LoggerFactory.getLogger(ADLSIncrementalExtractorTask.class);
    private static final String TASK_NAME = "Azure Data Lake Storage Incremental Extractor";

    private final ADLSIncrementalExtractorContext incrementalExtractorContext;
    private final QueueMsgHandler                 queueMsgHandler;

    public ADLSIncrementalExtractorTask(Properties extractionConfiguration) throws AtlasBaseException {
        this.incrementalExtractorContext = new ADLSIncrementalExtractorContext(extractionConfiguration);
        this.queueMsgHandler             = new QueueMsgHandler(incrementalExtractorContext);
    }

    @Override
    public String getName() {
        return TASK_NAME;
    }

    @Override
    public void initialize() throws AtlasBaseException {
        incrementalExtractorContext.initialize();
        queueMsgHandler.initialize();
    }

    @Override
    public void execute() {
        LOG.info("Starting incremental extraction for Azure Data Lake Storage.");

        Map<String, Set<String>> accountQueuesList = incrementalExtractorContext.getAccountAndQueueNames();

        for (Map.Entry<String, Set<String>> eachAccountQueueNames : accountQueuesList.entrySet()) {
            for (String eachQueue : eachAccountQueueNames.getValue()) {
                int idleOrFailureAttempt = 0;

                while (idleOrFailureAttempt < incrementalExtractorContext.getMaxRetryCount()) {
                    idleOrFailureAttempt++;

                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Performing attempt #{} to read messages from Data Lake Storage Queue.", idleOrFailureAttempt);
                    }

                    try {
                        int totalMessagesProcessed = queueMsgHandler.processQueueMessages(eachAccountQueueNames.getKey(), eachQueue);

                        if (totalMessagesProcessed > 0) {
                            idleOrFailureAttempt = 0;

                            if (LOG.isDebugEnabled()) {
                                LOG.debug("Total {} messages processed from queue {} for account {}", totalMessagesProcessed, eachQueue, eachAccountQueueNames);
                            }
                        } else {
                            if (LOG.isDebugEnabled()) {
                                LOG.debug("No message received from queue {} for account {}", eachQueue, eachAccountQueueNames);
                            }
                        }
                    } catch (Exception e) {
                        LOG.error("Exception occurred while processing queue messages from queue {} for account {}", eachQueue, eachAccountQueueNames, e);

                        idleOrFailureAttempt++;
                    }
                }

                LOG.info("Stopping the extraction after {} attempts.", idleOrFailureAttempt);
            }
        }
    }

    @Override
    public ADLSExtractorContext getExtractorContext() {
        return incrementalExtractorContext;
    }

    public QueueMsgHandler getQueueMsgHandler() {
        return queueMsgHandler;
    }
}