/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.cli;

import org.apache.atlas.ApplicationProperties;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.adls.extractors.ExtractorTask;
import org.apache.atlas.adls.extractors.bulk.ADLSBulkExtractorTask;
import org.testng.Assert;
import org.testng.annotations.*;

import java.io.File;

public class ADLSExtractorMainTest {
    @BeforeSuite
    public void initOneTime() {
        System.setProperty(ApplicationProperties.ATLAS_CONFIGURATION_DIRECTORY_PROPERTY, "src/test/resources/");
    }

    @BeforeMethod
    public void init() {
        ADLSExtractorMain.exitCode = ADLSExtractorMain.EXIT_CODE_FAILED;
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: The configuration file adls.conf is not a file or it does not exist or it is not readable.")
    public void testCreateCliContextWithNullArgs() throws AtlasBaseException {
        ADLSExtractorMain.createCliContext(null);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: The configuration file adls.conf is not a file or it does not exist or it is not readable.")
    public void testCreateCliContextWithEmptyArray() throws AtlasBaseException {
        ADLSExtractorMain.createCliContext(new String[0]);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: Unrecognized arguments.")
    public void testCreateCliContextWithMissingArgs() throws AtlasBaseException {
        ADLSExtractorMain.createCliContext(new String[]{""});
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: Received unsupported extraction type testType.")
    public void testCreateCliContextWithUnsupportedExtractionType() throws AtlasBaseException {
        ADLSExtractorMain.createCliContext(new String[]{"-e", "testType"});
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: The configuration file adls.conf is not a file or it does not exist or it is not readable.")
    public void testCreateCliContextWithSupportedExtractionType1() throws AtlasBaseException {
        ADLSExtractorMain.createCliContext(new String[]{"-e", "incremental"});
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: The configuration file dummyConfigFile.txt is not a file or it does not exist or it is not readable.")
    public void testCreateCliContextWithNonExistingConfigurationFile() throws AtlasBaseException {
        ADLSExtractorMain.createCliContext(new String[]{"-e", "Incremental", "-c", "dummyConfigFile.txt"});
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Invalid arguments. Reason: Unrecognized arguments.")
    public void testCreateCliContextWithInvalidParameter() throws AtlasBaseException {
        CliContext cliContext = ADLSExtractorMain.createCliContext(new String[]{
                "-e", "Incremental",
                "-c", "src/test/resources/TestExtractionConfigFile.txt",
                "-g"});
        Assert.assertEquals(cliContext.getExtractionType(), CliContext.ExtractionType.INCREMENTAL);
        Assert.assertEquals(cliContext.getSourceType(), CliContext.SourceType.ADLS);
        Assert.assertEquals(cliContext.isFailOnError(), true);
    }

    @Test
    public void testCreateCliContextWithValidConfigurationFile() throws AtlasBaseException {
        String configFilePath = "src/test/resources/TestExtractionConfigFile.txt";
        CliContext cliContext = ADLSExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.adls.extraction.account.name")
    public void testInitializeBulkExtractorTaskFactoryWithMissingAccountName() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/configFileWithMissingAccountName.txt";
        CliContext cliContext = ADLSExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);
        ADLSExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.adls.extraction.access.token")
    public void testInitializeBulkExtractorTaskFactoryWithMissingAccountKey() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/configFileWithMissingAccountKey.txt";
        CliContext cliContext = ADLSExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);
        ADLSExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test
    public void testInitializeBulkExtractorTaskFactoryWithCorrectAllowListingFromat() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/configFileWithCorrectAllowDenyListing.txt";
        CliContext cliContext = ADLSExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);
        ADLSExtractorMain.getInitializedExtractorTask(cliContext);
    }

    @Test
    public void testInitializeBulkExtractorTaskFactoryValidConfigPath() throws AtlasBaseException {
        String configFilePath = "src/test/resources/bulk/bulkExtractionConfig.txt";
        CliContext cliContext = ADLSExtractorMain.createCliContext(new String[]{"-e", "BULK",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.BULK);

        ExtractorTask extractorTask = ADLSExtractorMain.getInitializedExtractorTask(cliContext);
        Assert.assertEquals(ADLSExtractorMain.exitCode, ADLSExtractorMain.EXIT_CODE_FAILED);
        Assert.assertNotNull(extractorTask);
        Assert.assertEquals(extractorTask.getClass(), ADLSBulkExtractorTask.class);
    }

    @Test(expectedExceptions = AtlasBaseException.class,
            expectedExceptionsMessageRegExp = "Error encountered in initializing task: There is no value specified for configuration parameter atlas.adls.extraction.incremental.queueNames")
    public void testInitializeExtractorIncrementalTaskFactory() throws AtlasBaseException {
        String configFilePath = "src/test/resources/incremental/configFileWithMissingQueueName.txt";
        CliContext cliContext = ADLSExtractorMain.createCliContext(new String[]{"-e", "Incremental",
                "-c", configFilePath,
                "-failOnError"});
        assertCheckForCliContext(cliContext, configFilePath, CliContext.ExtractionType.INCREMENTAL);

        ADLSExtractorMain.getInitializedExtractorTask(cliContext);
        Assert.assertEquals(ADLSExtractorMain.exitCode, ADLSExtractorMain.EXIT_CODE_FAILED);
    }

    private void assertCheckForCliContext(CliContext cliContext, String configFilePath, CliContext.ExtractionType extractionType) {
        Assert.assertEquals(cliContext.getExtractionType(), extractionType);
        Assert.assertEquals(cliContext.getSourceType(), CliContext.SourceType.ADLS);
        Assert.assertEquals(cliContext.isFailOnError(), true);
        Assert.assertEquals(cliContext.getConfigFile(), new File(configFilePath));
        Assert.assertEquals(ADLSExtractorMain.exitCode, ADLSExtractorMain.EXIT_CODE_FAILED);
    }
}
