/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.hdfs.cli;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.net.URL;

/**
 * This is the class that provides the context representing
 * the CLI command line options/parameters.
 */
public class CliContext {
    private static final Logger LOG = LoggerFactory.getLogger(CliContext.class);

    static final String OPTION_LINEAGE_CONFIGURATION = "conf";
    static final String OPTION_LINEAGE_CONFIGURATION_LONG = "config";
    static final String OPTION_LINEAGE_HELP = "h";
    static final String OPTION_LINEAGE_HELP_LONG = "help";
    static final String DEFAULT_HDFS_CONFIG_FILE_NAME = "hdfs.conf";


    private final File configFile;

    public CliContext(CommandLine cmd) throws AtlasBaseException {
        this.configFile = getConfigFile(cmd);

    }

    public static CliContext createInstance(CommandLine cmd) throws AtlasBaseException {
        return new CliContext(cmd);
    }

    public File getConfigFile() {
        return configFile;
    }

    private File getConfigFile(CommandLine cmd) throws AtlasBaseException {
        String configFileName = cmd.getOptionValue(OPTION_LINEAGE_CONFIGURATION);

        if (StringUtils.isEmpty(configFileName)) {
            String msg = "Missing configuration file parameter. Using the default file: " + DEFAULT_HDFS_CONFIG_FILE_NAME;

            LOG.info(msg);

            configFileName = DEFAULT_HDFS_CONFIG_FILE_NAME;

            LOG.info("Looking for {} in classpath", configFileName);

            URL url = CliContext.class.getClassLoader().getResource(configFileName);

            if (url == null) {
                LOG.info("Looking for /{} in classpath", configFileName);

                url = CliContext.class.getClassLoader().getResource("/" + configFileName);
            }

            if (url != null) {
                configFileName = url.getPath();
            }
        }

        File file = new File(configFileName);

        if (!file.exists() || !file.isFile() || !file.canRead()) {
            String msg = "The configuration file " + configFileName + " is not a file or it does not exist or it is not readable. Please provide absolute path.";

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }

        return file;
    }


}
