/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.hdfs.cli;

import com.google.common.annotations.VisibleForTesting;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.hdfs.lineage.Constants;
import org.apache.atlas.hdfs.lineage.HDFSLineageTask;
import org.apache.atlas.hdfs.lineage.LineageTask;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;


import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


import java.util.Arrays;
import java.util.List;

import static org.apache.atlas.hdfs.cli.CliContext.OPTION_LINEAGE_CONFIGURATION;


public class HDFSLineageMain {
    private static final Logger LOG = LoggerFactory.getLogger(HDFSLineageMain.class);


    @VisibleForTesting
    static int exitCode = Constants.EXIT_CODE_FAILED;
    private static List<String> hdfsArgs;

    public static void main(String[] args) {

        LineageTask lineageTask = null;
        try {

            CliContext cliContext = createCliContext(args);
            if (cliContext != null) {

                lineageTask = getInitializedLineageTask(cliContext);

                exitCode = runLineageTasks(lineageTask, hdfsArgs);

            } else {
                exitCode = Constants.EXIT_CODE_HELP;
            }

        } catch (Exception e) {
            LOG.error("Error encountered.", e);
            System.out.println(e.getMessage());
        }

        System.exit(exitCode);
    }

    @VisibleForTesting
    static LineageTask getInitializedLineageTask(CliContext cliContext) throws AtlasBaseException {
        LineageTask lineageTask;

        try {
            LOG.info("Initializing the lineage task.");

            lineageTask = new HDFSLineageTask(cliContext);

            lineageTask.initialize();

            LOG.info("Successfully initialized the lineage task.");

            return lineageTask;
        } catch (Throwable e) {
            throw new AtlasBaseException("Error encountered in initializing task" + (StringUtils.isNotBlank(e.getMessage()) ? ": " + e.getMessage() : ""), e);
        }
    }

    @VisibleForTesting
    static int runLineageTasks(LineageTask lineageTask, List<String> args) throws AtlasBaseException {
        int exitCode = Constants.EXIT_CODE_FAILED;

        if (lineageTask != null) {
            LOG.info("Running the lineage task.");

            exitCode =  lineageTask.execute(args);

            LOG.info("Completed the execution of task.");
        } else {
            LOG.info("Lineage task is null. Nothing to be done.");
        }
        return exitCode;
    }


    @VisibleForTesting
    static CliContext createCliContext(String[] args) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Initializing the cli context.");
        }

        Options acceptedCliOptions = prepareCommandLineOptions();

        try {
            CommandLine cmd = new DefaultParser().parse(acceptedCliOptions, args, true);

            hdfsArgs = cmd.getArgList();
            if (LOG.isDebugEnabled()) {
                LOG.debug("Successfully initialized the cli context.");
            }

            CliContext ret = null;
            if (cmd.hasOption(CliContext.OPTION_LINEAGE_HELP)) {
                printUsage(acceptedCliOptions);
            } else {
                ret = CliContext.createInstance(cmd);
            }

            return ret;
        } catch (ParseException | AtlasBaseException e) {
            printUsage(acceptedCliOptions);

            throw new AtlasBaseException("Invalid arguments. Reason: " + e.getMessage(), e);
        }
    }

    private static Options prepareCommandLineOptions() {
        Options options = new Options();

        return options
                .addOption(OPTION_LINEAGE_CONFIGURATION, CliContext.OPTION_LINEAGE_CONFIGURATION_LONG, true,
                        "Configuration file for the HDFS lineage")
                .addOption(CliContext.OPTION_LINEAGE_HELP, CliContext.OPTION_LINEAGE_HELP_LONG, false,
                        "Print this help message");
    }

    private static void printUsage(Options options) {
        HelpFormatter formatter = new HelpFormatter();

        formatter.printHelp(HDFSLineageMain.class.getName(), options);
    }

}