/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.hdfs.lineage;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.hdfs.bridge.AtlasHDFSCreateEvent;
import org.apache.atlas.hdfs.bridge.AtlasHDFSDeleteEvent;
import org.apache.atlas.hdfs.bridge.AtlasHDFSHook;
import org.apache.atlas.hdfs.bridge.HDFSDataTypes;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.instance.AtlasRelatedObjectId;
import org.apache.atlas.type.AtlasTypeUtil;
import org.apache.atlas.utils.AtlasPathExtractorUtil;
import org.apache.atlas.utils.PathExtractorContext;

import org.apache.commons.lang3.StringUtils;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.security.UserGroupInformation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.List;

public class HDFSEventHandler {

    private static final Logger LOG = LoggerFactory.getLogger(HDFSEventHandler.class);

    private final HDFSLineageContext lineageContext;

    public HDFSEventHandler(HDFSLineageContext lineageContext) {
        this.lineageContext = lineageContext;
    }

    public int processEvent(List<String> args) throws AtlasBaseException, IOException {

        int exitCode = Constants.EXIT_CODE_FAILED;

        String command = args.get(0);
        String srcPath = args.get(1);
        String destPath = args.get(2);
        String timeStamp = args.get(3);
        URI srcPathURI;
        URI destPathURI;

        switch (command) {
            case "-copyFromLocal":
            case "-put":
                destPathURI = validateAndGetPathURI(destPath);
                if (lineageContext.doesPathNeedExtraction(destPathURI.getPath())) {

                    put(destPathURI, timeStamp);

                    exitCode = Constants.EXIT_CODE_SUCCESS;
                } else {
                    exitCode = Constants.EXIT_CODE_NOT_WHITELISTED;
                }
                break;
            case "-cp":

                srcPathURI = validateAndGetPathURI(srcPath);
                destPathURI = validateAndGetPathURI(destPath);
                if (lineageContext.doesPathNeedExtraction(srcPathURI.getPath()) || lineageContext.doesPathNeedExtraction(destPathURI.getPath())) {

                    copy(srcPathURI, destPathURI, timeStamp, Constants.PROCESS_TYPE_VAL_HDFS_COPY);

                    exitCode = Constants.EXIT_CODE_SUCCESS;
                } else {
                    exitCode = Constants.EXIT_CODE_NOT_WHITELISTED;
                }
                break;
            case "-rm":
                srcPathURI = validateAndGetPathURI(srcPath);
                if (lineageContext.doesPathNeedExtraction(srcPathURI.getPath())) {

                    delete(srcPathURI, timeStamp);

                    exitCode = Constants.EXIT_CODE_SUCCESS;
                } else {
                    exitCode = Constants.EXIT_CODE_NOT_WHITELISTED;
                }
                break;
            case "-mv":
                srcPathURI = validateAndGetPathURI(srcPath);
                destPathURI = validateAndGetPathURI(destPath);
                if (lineageContext.doesPathNeedExtraction(srcPathURI.getPath()) || lineageContext.doesPathNeedExtraction(destPathURI.getPath())) {
                    copy(srcPathURI, destPathURI, timeStamp, Constants.PROCESS_TYPE_VAL_HDFS_RENAME);
                    delete(srcPathURI, timeStamp);
                    exitCode = Constants.EXIT_CODE_SUCCESS;
                } else {
                    exitCode = Constants.EXIT_CODE_NOT_WHITELISTED;
                }

                break;
            default:
                break;
        }
        return exitCode;
    }

    private void put(URI path, String timeStamp) throws AtlasBaseException, IOException {

        if (LOG.isDebugEnabled()) {
            LOG.debug("HDFS PUT event: ");
            LOG.debug("  path = " + path);
            LOG.debug("  timestamp = " + timeStamp);
        }

        // get path entity
        AtlasEntity.AtlasEntityWithExtInfo pathEntity = getAtlasEntityWithExtInfo(path, timeStamp);

        //publish path entity to atlas
        publishCreateEventsToAtlas(new AtlasEntity.AtlasEntitiesWithExtInfo(pathEntity));

    }

    private AtlasEntity.AtlasEntityWithExtInfo getAtlasEntityWithExtInfo(URI path, String timeStamp) throws AtlasBaseException, IOException {
        //create NameNode
        AtlasEntity.AtlasEntityWithExtInfo nameNode = getAtlasEntityWithExtInfo(lineageContext.getNameNodeURI(), timeStamp, Constants.HDFS_PATH_NAMENODE);
        publishCreateEventsToAtlas(new AtlasEntity.AtlasEntitiesWithExtInfo(nameNode));

        //create directory hierarchy in atlas and publish
        AtlasEntity parentEntity = createParent(Paths.get(path.getPath()), nameNode.getEntity(), timeStamp);

        //create path
        AtlasEntity.AtlasEntityWithExtInfo pathEntity = getAtlasEntityWithExtInfo(path, timeStamp, Constants.HDFS_PATH_V2);

        //link path to parentEntity
        linkAtlasPathContainerToContained(parentEntity, pathEntity.getEntity());
        return pathEntity;
    }


    private void delete(URI path, String timeStamp) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("HDFS DELETE event: ");
            LOG.debug("  path = " + path);
            LOG.debug("  timestamp = " + timeStamp);
        }

        AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo = getAtlasEntityWithExtInfo(path, timeStamp, Constants.HDFS_PATH_V2);
        AtlasObjectId objectId = new AtlasObjectId(Constants.HDFS_PATH_V2, Constants.ENTITY_ATTR_QNAME, entityWithExtInfo.getEntity().getAttribute(Constants.ENTITY_ATTR_QNAME));

        publishDeleteEventsToAtlas(Collections.singletonList(objectId));

    }

    private void copy(URI srcPath, URI destPath, String timeStamp, String processTypeValue) throws AtlasBaseException, IOException {

        if (LOG.isDebugEnabled()) {
            LOG.debug(processTypeValue + " event: ");
            LOG.debug("  srcPath = " + srcPath);
            LOG.debug("  dstPath = " + destPath);
            LOG.debug("  timestamp = " + timeStamp);
        }

        // get srcPath entity
        AtlasEntity.AtlasEntityWithExtInfo srcEntityWithExtInfo = getAtlasEntityWithExtInfo(srcPath, timeStamp);

        // get destPath entity
        AtlasEntity.AtlasEntityWithExtInfo dstEntityWithExtInfo = getAtlasEntityWithExtInfo(destPath, timeStamp);

        // get process entity
        AtlasEntity processEntity = getProcessEntity(srcEntityWithExtInfo.getEntity(), dstEntityWithExtInfo.getEntity(), processTypeValue, timeStamp);

        AtlasEntity.AtlasEntitiesWithExtInfo createHDFSProcess = new AtlasEntity.AtlasEntitiesWithExtInfo();
        createHDFSProcess.addEntity(processEntity);
        createHDFSProcess.addEntity(srcEntityWithExtInfo.getEntity());
        createHDFSProcess.addEntity(dstEntityWithExtInfo.getEntity());

        // publish src, dest and process entities to atlas
        publishCreateEventsToAtlas(createHDFSProcess);

    }

    private AtlasEntity getProcessEntity(AtlasEntity srcEntity, AtlasEntity dstEntity, String processTypeValue, String timeStamp) {
        String metadataNamespace = AtlasHDFSHook.getInstance() != null ? AtlasHDFSHook.getInstance().getMetadataNamespace() : "";
        AtlasEntity processEntity = new AtlasEntity(HDFSDataTypes.HDFS_PROCESS.getEntityType());
        String srcPath = srcEntity.getAttribute(Constants.ATTRIBUTE_PATH).toString();
        String destPath = dstEntity.getAttribute(Constants.ATTRIBUTE_PATH).toString();

        String processName = getPathDifferenceOfFirst(srcPath, destPath) + Constants.MOVE_CMD_INDICATOR + getPathDifferenceOfFirst(destPath, srcPath);
        processEntity.setAttribute(Constants.ATTRIBUTE_NAME, processName);
        processEntity.setAttribute(Constants.ENTITY_ATTR_QNAME, processName + Constants.QNAME_SEP_METADATA_NAMESPACE + metadataNamespace + Constants.HOST_PORT_DELIMITER + timeStamp);
        processEntity.setAttribute(Constants.PROCESS_TYPE_ATTR, processTypeValue);
        processEntity.setAttribute(Constants.CREATION_TIME_ATTR, timeStamp);
        processEntity.setRelationshipAttribute(Constants.ATTRIBUTE_INPUTS, Collections.singletonList(AtlasTypeUtil.getAtlasRelatedObjectId(srcEntity, Constants.RELATIONSHIP_DATASET_PROCESS_INPUTS)));
        processEntity.setRelationshipAttribute(Constants.ATTRIBUTE_OUTPUTS, Collections.singletonList(AtlasTypeUtil.getAtlasRelatedObjectId(dstEntity, Constants.RELATIONSHIP_PROCESS_DATASET_OUTPUTS)));

        return processEntity;
    }

    public void publishDeleteEventsToAtlas( List<AtlasObjectId> entityObjectIds) throws AtlasBaseException {
        if (entityObjectIds != null) {
            if (LOG.isDebugEnabled()) {
                LOG.info("Publish delete hdfs_path entity event to Atlas.");
            }

            try {
                AtlasHDFSDeleteEvent atlasHDFSDeleteEvent = new AtlasHDFSDeleteEvent(entityObjectIds);

                AtlasHDFSHook.getInstance().publishEvent(atlasHDFSDeleteEvent);

            } catch (Exception e) {
                String msg = String.format("Error encountered while publishing delete entities with Qualified Name.");
                LOG.error(msg, e);
                throw new AtlasBaseException(msg, e);
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("Published HDFS entities delete event to Atlas.");
            }
        }
    }

    public void publishCreateEventsToAtlas(AtlasEntity.AtlasEntitiesWithExtInfo entitiesWithExtInfo) throws AtlasBaseException {

        if (entitiesWithExtInfo != null) {
            if (LOG.isDebugEnabled()) {
                LOG.info("Publish create Account entity event to Atlas.");
            }
            try {
                AtlasHDFSCreateEvent atlasHDFSCreateEvent = new AtlasHDFSCreateEvent(entitiesWithExtInfo);
                AtlasHDFSHook.getInstance().publishEvent(atlasHDFSCreateEvent);

            } catch (Exception e) {
                String msg = String.format("Error encountered while publishing create entities.");
                LOG.error(msg, e);
                throw new AtlasBaseException(msg, e);
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("Published HDFS entities create event to Atlas.");
            }
        }
    }

    private AtlasEntity.AtlasEntityWithExtInfo getAtlasEntityWithExtInfo(URI destPath, String timeStamp, String type) {

        String metadataNamespace = AtlasHDFSHook.getInstance() != null ? AtlasHDFSHook.getInstance().getMetadataNamespace() : "";
        PathExtractorContext pathExtractorContext = new PathExtractorContext(metadataNamespace, false, null);

        AtlasEntity.AtlasEntityWithExtInfo pathEntityWithExtInfo = AtlasPathExtractorUtil.addHDFSPathSubEntity(new Path(destPath.toString()), pathExtractorContext, type);
        pathEntityWithExtInfo.getEntity().setAttribute(Constants.CREATE_TIME_ATTR, timeStamp);
        return pathEntityWithExtInfo;
    }


    private AtlasEntity createParent(java.nio.file.Path path, AtlasEntity previousEntity, String timeStamp) throws AtlasBaseException {
        java.nio.file.Path parent = path.getParent();

        if (parent.getParent() != null) {
            previousEntity = createParent(parent, previousEntity, timeStamp);
            previousEntity = createPathAtAtlas(validateAndGetPathURI(parent.toString()), previousEntity, timeStamp);
        }
        return previousEntity;
    }

    private AtlasEntity createPathAtAtlas(URI path, AtlasEntity previousEntity, String timeStamp) throws AtlasBaseException {
        AtlasEntity.AtlasEntityWithExtInfo pathEntityWithExtInfo = getAtlasEntityWithExtInfo(path, timeStamp, Constants.HDFS_PATH_V2);

        linkAtlasPathContainerToContained(previousEntity, pathEntityWithExtInfo.getEntity());
        publishCreateEventsToAtlas(new AtlasEntity.AtlasEntitiesWithExtInfo(pathEntityWithExtInfo));

        previousEntity = pathEntityWithExtInfo.getEntity();

        return previousEntity;
    }

    private void linkAtlasPathContainerToContained(AtlasEntity containerAtlasEntity, AtlasEntity containedAtlasEntity) {
        containerAtlasEntity.setGuid(null);

        AtlasObjectId containerEntityId = AtlasTypeUtil.getObjectId(containerAtlasEntity);
        AtlasRelatedObjectId relatedObjectId = new AtlasRelatedObjectId(containerEntityId, Constants.HDFS_CONTAINER_CONTAINED);

        containedAtlasEntity.setRelationshipAttribute(Constants.CONTAINER, relatedObjectId);
    }


    private URI validateAndGetPathURI(String path) throws AtlasBaseException {
        URI pathURI;
        if (StringUtils.isNotBlank(path)) {
            if (StringUtils.startsWith(path, Constants.NAME_NODE_URI_PREFIX)) {
                pathURI = URI.create(path).normalize();
                if (!StringUtils.equals(pathURI.getAuthority(), lineageContext.getNameNodeURI().getAuthority())) {
                    String msg = String.format("Provided NameNode URI in path does not match NameNode URI set in conf.");
                    LOG.error(msg);
                    throw new AtlasBaseException(msg);
                }

            } else {
                pathURI = URI.create(lineageContext.getNameNodeURIConfig() + "/" + path).normalize();
            }

        } else {
            String msg = String.format("Path is blank.");
            LOG.error(msg);
            throw new AtlasBaseException(msg);
        }

        return pathURI;
    }


    private String getPathDifferenceOfFirst(String first, String second) {
        String ret = "";
        int indexOfDifference = StringUtils.indexOfDifference(first, second);

        if (indexOfDifference != -1) {
            int indexOfLastPathSeparator = first.lastIndexOf(File.separator, indexOfDifference);
            ret = indexOfLastPathSeparator != -1 ? first.substring(indexOfLastPathSeparator + 1) : ret;
        }

        return ret;
    }

}
