/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.hdfs.lineage;


import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.commons.lang3.StringUtils;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Paths;
import java.time.Instant;
import java.util.Arrays;
import java.util.List;

import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;


public class HDFSEventHandlerTest {

    public static final String QUALIFIED_NAME   = "qualifiedName";
    public static final String MEDATA_NAMESPACE = "@primary";
    public static final String SCHEME           = "hdfs://";
    private static      URI    NAME_NODE_URI    = URI.create(SCHEME + "localhost:21000");

    @Mock
    private HDFSLineageContext hdfsLineageContext;

    @Captor
    private ArgumentCaptor<AtlasEntity.AtlasEntitiesWithExtInfo> entitiesWithExtInfoCaptor;

    @Captor
    private ArgumentCaptor<List<AtlasObjectId>> listArgumentCaptor;

    @BeforeMethod
    public void init() {
        MockitoAnnotations.initMocks(this);

        when(hdfsLineageContext.getNameNodeURIConfig()).thenReturn(NAME_NODE_URI.toString());
        when(hdfsLineageContext.getNameNodeURI()).thenReturn(NAME_NODE_URI);
        when(hdfsLineageContext.doesPathNeedExtraction(anyString())).thenReturn(true);
    }

    @Test(dataProvider = "put-data")
    public void testProcessEventPut(String cmd, String srcPath, String destPath) throws AtlasBaseException, IOException {
        String timestamp = String.valueOf(Instant.now().toEpochMilli());
        String destPathParent = Paths.get(destPath).getParent().toString();

        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));
        hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));

        verify(hdfsEventHandler, times(3)).publishCreateEventsToAtlas(entitiesWithExtInfoCaptor.capture());
        List<AtlasEntity.AtlasEntitiesWithExtInfo> entitiesWithExtInfos = entitiesWithExtInfoCaptor.getAllValues();
        assertEquals(entitiesWithExtInfos.get(0).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(1).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + destPathParent + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(2).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + destPath + MEDATA_NAMESPACE);

    }

    @Test(dataProvider = "put-data-2")
    public void testProcessEventPutWithNameNodeURI(String cmd, String srcPath, String destPath) throws AtlasBaseException, IOException {
        String timestamp = String.valueOf(Instant.now().toEpochMilli());
        String destPathParent = Paths.get(destPath).getParent().toString().replace("hdfs:/", SCHEME);

        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));
        hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));

        verify(hdfsEventHandler, times(3)).publishCreateEventsToAtlas(entitiesWithExtInfoCaptor.capture());
        List<AtlasEntity.AtlasEntitiesWithExtInfo> entitiesWithExtInfos = entitiesWithExtInfoCaptor.getAllValues();
        assertEquals(entitiesWithExtInfos.get(0).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(1).getEntities().get(0).getAttribute(QUALIFIED_NAME), destPathParent + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(2).getEntities().get(0).getAttribute(QUALIFIED_NAME), destPath + MEDATA_NAMESPACE);

    }

    @Test(dataProvider = "copy-data")
    public void testProcessEventCopy(String cmd, String srcPath, String destPath) throws AtlasBaseException, IOException {
        String timestamp = String.valueOf(Instant.now().toEpochMilli());
        String srcPathParent = Paths.get(srcPath).getParent().toString();
        String destPathParent = Paths.get(destPath).getParent().toString();


        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));

        hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));

        verify(hdfsEventHandler, times(5)).publishCreateEventsToAtlas(entitiesWithExtInfoCaptor.capture());

        List<AtlasEntity.AtlasEntitiesWithExtInfo> entitiesWithExtInfos = entitiesWithExtInfoCaptor.getAllValues();

        assertEquals(entitiesWithExtInfos.get(0).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(1).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + srcPathParent + MEDATA_NAMESPACE);

        assertEquals(entitiesWithExtInfos.get(2).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(3).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + destPathParent + MEDATA_NAMESPACE);

        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(0).getAttribute(QUALIFIED_NAME), getProcessQualifiedName(srcPath, destPath, timestamp));
        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(1).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + srcPath + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(2).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + destPath + MEDATA_NAMESPACE);

    }


    @Test(dataProvider = "copy-data-2")
    public void testProcessEventCopyWithNameNode(String cmd, String srcPath, String destPath) throws AtlasBaseException, IOException {
        String timestamp = String.valueOf(Instant.now().toEpochMilli());
        String srcPathParent = Paths.get(srcPath).getParent().toString().replace("hdfs:/", SCHEME);
        String destPathParent = Paths.get(destPath).getParent().toString().replace("hdfs:/", SCHEME);

        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));

        hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));

        verify(hdfsEventHandler, times(5)).publishCreateEventsToAtlas(entitiesWithExtInfoCaptor.capture());

        List<AtlasEntity.AtlasEntitiesWithExtInfo> entitiesWithExtInfos = entitiesWithExtInfoCaptor.getAllValues();

        assertEquals(entitiesWithExtInfos.get(0).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(1).getEntities().get(0).getAttribute(QUALIFIED_NAME), srcPathParent + MEDATA_NAMESPACE);

        assertEquals(entitiesWithExtInfos.get(2).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(3).getEntities().get(0).getAttribute(QUALIFIED_NAME), destPathParent + MEDATA_NAMESPACE);

        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(0).getAttribute(QUALIFIED_NAME), getProcessQualifiedName(srcPath, destPath, timestamp));
        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(1).getAttribute(QUALIFIED_NAME), srcPath + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(2).getAttribute(QUALIFIED_NAME), destPath + MEDATA_NAMESPACE);

    }


    @Test(dataProvider = "delete-data")
    public void testProcessEventDelete(String cmd, String srcPath, String destPath) throws AtlasBaseException, IOException {
        String timestamp = String.valueOf(Instant.now().toEpochMilli());

        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));
        hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));

        verify(hdfsEventHandler, times(1)).publishDeleteEventsToAtlas(listArgumentCaptor.capture());

        List<List<AtlasObjectId>> values = listArgumentCaptor.getAllValues();
        assertEquals(values.get(0).get(0).getUniqueAttributes().get(QUALIFIED_NAME), NAME_NODE_URI + srcPath + MEDATA_NAMESPACE);

    }

    @Test(dataProvider = "delete-data-2")
    public void testProcessEventDeleteWithNameNode(String cmd, String srcPath, String destPath) throws AtlasBaseException, IOException {
        String timestamp = String.valueOf(Instant.now().toEpochMilli());

        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));
        hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));

        verify(hdfsEventHandler, times(1)).publishDeleteEventsToAtlas(listArgumentCaptor.capture());

        List<List<AtlasObjectId>> values = listArgumentCaptor.getAllValues();
        assertEquals(values.get(0).get(0).getUniqueAttributes().get(QUALIFIED_NAME), srcPath + MEDATA_NAMESPACE);

    }

    @Test(dataProvider = "move-data")
    public void testProcessEventMove(String cmd, String srcPath, String destPath) throws AtlasBaseException, IOException {
        String timestamp = String.valueOf(Instant.now().toEpochMilli());
        String srcPathParent = Paths.get(srcPath).getParent().toString();
        String destPathParent = Paths.get(destPath).getParent().toString();

        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));

        hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));

        verify(hdfsEventHandler, times(5)).publishCreateEventsToAtlas(entitiesWithExtInfoCaptor.capture());
        verify(hdfsEventHandler, times(1)).publishDeleteEventsToAtlas(listArgumentCaptor.capture());

        List<AtlasEntity.AtlasEntitiesWithExtInfo> entitiesWithExtInfos = entitiesWithExtInfoCaptor.getAllValues();

        assertEquals(entitiesWithExtInfos.get(0).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(1).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + srcPathParent + MEDATA_NAMESPACE);

        assertEquals(entitiesWithExtInfos.get(2).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(3).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + destPathParent + MEDATA_NAMESPACE);

        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(0).getAttribute(QUALIFIED_NAME), getProcessQualifiedName(srcPath, destPath, timestamp));
        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(1).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + srcPath + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(2).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + destPath + MEDATA_NAMESPACE);

        List<List<AtlasObjectId>> values = listArgumentCaptor.getAllValues();
        assertEquals(values.get(0).get(0).getUniqueAttributes().get(QUALIFIED_NAME), NAME_NODE_URI + srcPath + MEDATA_NAMESPACE);

    }

    @Test(dataProvider = "move-data-2")
    public void testProcessEventMoveWithNameNode(String cmd, String srcPath, String destPath) throws AtlasBaseException, IOException {
        String timestamp = String.valueOf(Instant.now().toEpochMilli());
        String srcPathParent = Paths.get(srcPath).getParent().toString().replace("hdfs:/", SCHEME);
        String destPathParent = Paths.get(destPath).getParent().toString().replace("hdfs:/", SCHEME);

        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));

        hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));

        verify(hdfsEventHandler, times(5)).publishCreateEventsToAtlas(entitiesWithExtInfoCaptor.capture());
        verify(hdfsEventHandler, times(1)).publishDeleteEventsToAtlas(listArgumentCaptor.capture());

        List<AtlasEntity.AtlasEntitiesWithExtInfo> entitiesWithExtInfos = entitiesWithExtInfoCaptor.getAllValues();

        assertEquals(entitiesWithExtInfos.get(0).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(1).getEntities().get(0).getAttribute(QUALIFIED_NAME), srcPathParent + MEDATA_NAMESPACE);

        assertEquals(entitiesWithExtInfos.get(2).getEntities().get(0).getAttribute(QUALIFIED_NAME), NAME_NODE_URI + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(3).getEntities().get(0).getAttribute(QUALIFIED_NAME), destPathParent + MEDATA_NAMESPACE);

        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(0).getAttribute(QUALIFIED_NAME), getProcessQualifiedName(srcPath, destPath, timestamp));
        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(1).getAttribute(QUALIFIED_NAME), srcPath + MEDATA_NAMESPACE);
        assertEquals(entitiesWithExtInfos.get(4).getEntities().get(2).getAttribute(QUALIFIED_NAME), destPath + MEDATA_NAMESPACE);

        List<List<AtlasObjectId>> values = listArgumentCaptor.getAllValues();
        assertEquals(values.get(0).get(0).getUniqueAttributes().get(QUALIFIED_NAME), srcPath + MEDATA_NAMESPACE);
    }


    @Test
    public void testProcessEventPutWithException() throws AtlasBaseException, IOException {
        String cmd = "-put";
        String srcPath = "hdfs://somehost:8020/test/sample.txt";
        String destPath = "hdfs://somehost:8020/test2/sample2.txt";
        String timestamp = String.valueOf(Instant.now().toEpochMilli());

        HDFSEventHandler hdfsEventHandler = spy(new HDFSEventHandler(hdfsLineageContext));

        try {
            hdfsEventHandler.processEvent(Arrays.asList(cmd, srcPath, destPath, timestamp));
            Assert.fail("No exception thrown!");
        } catch (AtlasBaseException e) {
            assertEquals(e.getMessage(), "Provided NameNode URI in path does not match NameNode URI set in conf.");
        }

    }


    @DataProvider(name = "put-data")
    public Object[][] putData() {
        return new Object[][]{
                {"-put", "/test/sample.txt", "/test2/sample2.txt"},
                {"-put", "/dir/test5", "/dir2/test6"},
                {"-put", "/dir/test5", "/test6/sample.txt"},
                {"-copyFromLocal", "/test/sample.txt", "/test2/sample2.txt"},
                {"-copyFromLocal", "/dir/test5", "/dir2/test6"},
                {"-copyFromLocal", "/dir/test5", "/test6/sample.txt"}
        };
    }

    @DataProvider(name = "copy-data")
    public Object[][] copyData() {
        return new Object[][]{
                {"-cp", "/test/sample.txt", "/test2/sample2.txt"},
                {"-cp", "/dir/test5", "/dir2/test6"},
                {"-cp", "/dir/test5", "/test6/sample.txt"}
        };
    }

    @DataProvider(name = "put-data-2")
    public Object[][] putData2() {
        return new Object[][]{
                {"-put", NAME_NODE_URI + "/test/sample.txt", NAME_NODE_URI + "/test2/sample2.txt"},
                {"-put", NAME_NODE_URI + "/dir/test5", NAME_NODE_URI + "/dir2/test6"},
                {"-put", NAME_NODE_URI + "/dir/test5", NAME_NODE_URI + "/test6/sample.txt"},
                {"-copyFromLocal", NAME_NODE_URI + "/test/sample.txt", NAME_NODE_URI + "/test2/sample2.txt"},
                {"-copyFromLocal", NAME_NODE_URI + "/dir/test5", NAME_NODE_URI + "/dir2/test6"},
                {"-copyFromLocal", NAME_NODE_URI + "/dir/test5", NAME_NODE_URI + "/test6/sample.txt"}
        };
    }


    @DataProvider(name = "copy-data-2")
    public Object[][] copyData2() {
        return new Object[][]{
                {"-cp", NAME_NODE_URI + "/test/sample.txt", NAME_NODE_URI + "/test2/sample2.txt"},
                {"-cp", NAME_NODE_URI + "/dir/test5", NAME_NODE_URI + "/dir2/test6"},
                {"-cp", NAME_NODE_URI + "/dir/test5", NAME_NODE_URI + "/test6/sample.txt"}
        };
    }


    @DataProvider(name = "delete-data")
    public Object[][] deleteData() {
        return new Object[][]{
                {"-rm", "/test/sample.txt", ""},
                {"-rm", "/dir/test5", ""}
        };
    }

    @DataProvider(name = "delete-data-2")
    public Object[][] deleteData2() {
        return new Object[][]{
                {"-rm", NAME_NODE_URI + "/test/sample.txt", ""},
                {"-rm", NAME_NODE_URI + "/dir/test5", ""}
        };
    }

    @DataProvider(name = "move-data")
    public Object[][] moveData() {
        return new Object[][]{
                {"-mv", "/test/sample.txt", "/test2/sample2.txt"},
                {"-mv", "/dir/test5", "/dir2/test6"},
                {"-mv", "/dir/test5", "/test6/sample.txt"}
        };
    }

    @DataProvider(name = "move-data-2")
    public Object[][] moveData2() {
        return new Object[][]{
                {"-mv", NAME_NODE_URI + "/test/sample.txt", NAME_NODE_URI + "/test2/sample2.txt"},
                {"-mv", NAME_NODE_URI + "/dir/test5", NAME_NODE_URI + "/dir2/test6"},
                {"-mv", NAME_NODE_URI + "/dir/test5", NAME_NODE_URI + "/test6/sample.txt"}
        };
    }

    private String getProcessQualifiedName(String srcPath, String destPath, String timeStamp) {
        String processName = getPathDifferenceOfFirst(srcPath, destPath) + Constants.MOVE_CMD_INDICATOR + getPathDifferenceOfFirst(destPath, srcPath);
        return processName + MEDATA_NAMESPACE + Constants.HOST_PORT_DELIMITER + timeStamp;
    }


    private String getPathDifferenceOfFirst(String first, String second) {
        String ret = "";
        int indexOfDifference = StringUtils.indexOfDifference(first, second);

        if (indexOfDifference != -1) {
            int indexOfLastPathSeparator = first.lastIndexOf(File.separator, indexOfDifference);
            ret = indexOfLastPathSeparator != -1 ? first.substring(indexOfLastPathSeparator + 1) : ret;
        }

        return ret;
    }


}
