/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.hive.hook;

import org.apache.atlas.hive.hook.events.*;
import org.apache.atlas.hook.AtlasHook;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hive.metastore.MetaStoreEventListener;
import org.apache.hadoop.hive.metastore.api.EnvironmentContext;
import org.apache.hadoop.hive.metastore.api.FieldSchema;
import org.apache.hadoop.hive.metastore.events.*;
import org.apache.hadoop.hive.metastore.utils.SecurityUtils;
import org.apache.hadoop.hive.ql.ddl.table.AlterTableType;
import org.apache.hadoop.hive.ql.metadata.Table;
import org.apache.hadoop.hive.ql.plan.HiveOperation;
import org.apache.hadoop.hive.shims.Utils;
import org.apache.hadoop.security.UserGroupInformation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


import static org.apache.atlas.hive.hook.events.AlterTableRenameCol.findRenamedColumn;
import static org.apache.atlas.hive.hook.events.BaseHiveEvent.*;
import static org.apache.hadoop.hive.metastore.HiveMetaHook.ALTER_TABLE_OPERATION_TYPE;
import static org.apache.hadoop.hive.ql.ddl.table.AlterTableType.*;
import static org.apache.atlas.hive.hook.events.BaseHiveEvent.toTable;
import static org.apache.atlas.repository.Constants.HMS_SOURCE;
import static org.apache.hadoop.hive.ql.plan.HiveOperation.*;

public class HiveMetastoreHookImpl extends MetaStoreEventListener {
    private static final Logger            LOG = LoggerFactory.getLogger(HiveMetastoreHookImpl.class);
    private        final HiveHook          hiveHook;
    private        final HiveMetastoreHook hook;

    public HiveMetastoreHookImpl(Configuration config) {
        super(config);

        this.hiveHook = new HiveHook(this.getClass().getSimpleName());
        this.hook     = new HiveMetastoreHook();
    }

    @Override
    public void onCreateDatabase(CreateDatabaseEvent dbEvent) {
        HiveOperationContext context = new HiveOperationContext(CREATEDATABASE, dbEvent);

        hook.handleEvent(context);
    }

    @Override
    public void onDropDatabase(DropDatabaseEvent dbEvent) {
        HiveOperationContext context = new HiveOperationContext(DROPDATABASE, dbEvent);

        hook.handleEvent(context);
    }

    @Override
    public void onAlterDatabase(AlterDatabaseEvent dbEvent) {
        HiveOperationContext context = new HiveOperationContext(ALTERDATABASE, dbEvent);

        hook.handleEvent(context);
    }

    @Override
    public void onCreateTable(CreateTableEvent tableEvent) {
        HiveOperationContext context = new HiveOperationContext(CREATETABLE, tableEvent);

        hook.handleEvent(context);
    }

    @Override
    public void onDropTable(DropTableEvent tableEvent) {
        HiveOperationContext context = new HiveOperationContext(DROPTABLE, tableEvent);

        hook.handleEvent(context);
    }

    @Override
    public void onAlterTable(AlterTableEvent tableEvent) {
        HiveOperation operation = getAlterTableOperation(tableEvent);

        if (operation != null) {
            HiveOperationContext context = new HiveOperationContext(operation, tableEvent);

            hook.handleEvent(context);
        }
    }

    public class HiveMetastoreHook extends AtlasHook {
        public HiveMetastoreHook() {
        }

        @Override
        public String getMessageSource() {
            return HMS_SOURCE;
        }

        public void handleEvent(HiveOperationContext operContext) {
            ListenerEvent listenerEvent = operContext.getEvent();

            if (!listenerEvent.getStatus()) {
                return;
            }

            try {
                HiveOperation        oper    = operContext.getOperation();
                AtlasHiveHookContext context = new AtlasHiveHookContext(hiveHook, oper, hiveHook.getKnownObjects(), this, listenerEvent, hiveHook.isSkipTempTables());
                BaseHiveEvent        event   = null;

                switch (oper) {
                    case CREATEDATABASE:
                        event = new CreateDatabase(context);
                        break;

                    case DROPDATABASE:
                        event = new DropDatabase(context);
                        break;

                    case ALTERDATABASE:
                        event = new AlterDatabase(context);
                        break;

                    case CREATETABLE:
                        event = new CreateTable(context);
                        break;

                    case DROPTABLE:
                        event = new DropTable(context);
                        break;

                    case ALTERTABLE_REPLACECOLS:
                    case ALTERTABLE_ADDCOLS:
                    case ALTERTABLE_SERIALIZER:
                    case ALTERTABLE_FILEFORMAT:
                    case ALTERTABLE_CLUSTER_SORT:
                        event = new AlterTable(context);
                        break;

                    case ALTERTABLE_PROPERTIES:
                    case ALTERTABLE_OWNER:
                        event = new AlterTableProperties(context);
                        break;

                    case ALTERTABLE_RENAME:
                        event = new AlterTableRename(context);
                        break;

                    case ALTERTABLE_RENAMECOL:
                        FieldSchema columnOld = operContext.getColumnOld();
                        FieldSchema columnNew = operContext.getColumnNew();

                        event = new AlterTableRenameCol(columnOld, columnNew, context);
                        break;

                    default:
                        if (LOG.isDebugEnabled()) {
                            LOG.debug("HiveMetastoreHook.handleEvent({}): operation ignored.", listenerEvent);
                        }
                        break;
                }

                if (event != null) {
                    final UserGroupInformation ugi = SecurityUtils.getUGI() == null ? Utils.getUGI() : SecurityUtils.getUGI();

                    super.notifyEntities(event.getNotificationMessages(), ugi);
                }
            } catch (Throwable t) {
                LOG.error("HiveMetastoreHook.handleEvent({}): failed to process operation {}", listenerEvent, t);
            }
        }
    }

    private HiveOperation getAlterTableOperation(AlterTableEvent alterTableEvent) {
        HiveOperation  ret;
        AlterTableType alterTableType = getAlterTableType(alterTableEvent);

        if (alterTableType == RENAME) {
            ret = ALTERTABLE_RENAME;
        } else if (alterTableType == RENAME_COLUMN) {
            ret = ALTERTABLE_RENAMECOL;
        } else if (alterTableType == REPLACE_COLUMNS) {
            ret = ALTERTABLE_REPLACECOLS;
        } else if (alterTableType == ADDPROPS || alterTableType == DROPPROPS) {
            ret = ALTERTABLE_PROPERTIES;
        } else if (alterTableType == ADDCOLS || alterTableType == UPDATE_COLUMNS){
            ret = ALTERTABLE_ADDCOLS;
        } else if (alterTableType == OWNER) {
            ret = ALTERTABLE_OWNER;
        } else if (alterTableType == SET_SERDE || alterTableType == SET_SERDE_PROPS) {
            ret = ALTERTABLE_SERIALIZER;
        } else if (alterTableType == SET_FILE_FORMAT) {
            ret = ALTERTABLE_FILEFORMAT;
        } else if (alterTableType == CLUSTERED_BY) {
            ret = ALTERTABLE_CLUSTER_SORT;
        } else{
            ret = null;
        }

        return ret;
    }

    private AlterTableType getAlterTableType(AlterTableEvent alterTableEvent) {
        AlterTableType     ret        = null;
        EnvironmentContext envContext = alterTableEvent.getEnvironmentContext();
        Table              oldTable   = toTable(alterTableEvent.getOldTable());
        Table              newTable   = toTable(alterTableEvent.getNewTable());

        if (hasEnvContextProperties(envContext)) {
            if (StringUtils.isNotEmpty(envContext.getProperties().get(ALTER_TABLE_OPERATION_TYPE))) {
                String alterTableOperationType = envContext.getProperties().get(ALTER_TABLE_OPERATION_TYPE);

                try {
                    ret = AlterTableType.valueOf(alterTableOperationType.trim().toUpperCase());
                } catch(IllegalArgumentException ex) {
                    // do nothing
                }
            } else if (isAlterTableProperty(oldTable, newTable, HIVE_TABLE_NUMROWS_PROPERTY)) {
                ret = ADDPROPS;
            }

        } else {
            // if EnvironmentContext is not available (from Impala or Spark)
		    if (isTableRename(oldTable, newTable)) {
                ret = RENAME;
            } else if (isColumnRename(oldTable, newTable)) {
                ret = RENAME_COLUMN;
            } else if (isAlterTableProperty(oldTable, newTable, HIVE_TABLE_LAST_MODIFIED_TIME_PROPERTY) ||
                       isAlterTableProperty(oldTable, newTable, HIVE_TABLE_LAST_DDLTIME_PROPERTY) ||
                       isAlterTableRewriteAttribute(oldTable, newTable)) {
                ret = ADDCOLS;
            }
        }

        return ret;
    }

    private static boolean isTableRename(Table oldTable, Table newTable) {
        return !StringUtils.equalsIgnoreCase(oldTable.getTableName(), newTable.getTableName());
    }

    private static boolean isColumnRename(Table oldTable, Table newTable) {
        FieldSchema columnOld = findRenamedColumn(oldTable, newTable);
        FieldSchema columnNew = findRenamedColumn(newTable, oldTable);

        return columnOld != null && columnNew != null;
    }

    private boolean isAlterTableProperty(Table oldTable, Table newTable, String paramterToCheck) {
        final boolean ret;
        String        oldTableModifiedParameterValue = oldTable.getTTable().getParameters().get(paramterToCheck);
        String        newTableModifiedParameterValue = newTable.getTTable().getParameters().get(paramterToCheck);

	    if (oldTableModifiedParameterValue == null) {
            ret = paramterToCheck.equals(HIVE_TABLE_NUMROWS_PROPERTY)
                    ? newTableModifiedParameterValue != null && Integer.parseInt(newTableModifiedParameterValue) > 0
                    : newTableModifiedParameterValue != null;
        } else {
            ret = !oldTableModifiedParameterValue.equals(newTableModifiedParameterValue);
        }

        return ret;
    }

    private boolean hasEnvContextProperties(EnvironmentContext envContext) {
        return envContext != null && envContext.getProperties() != null;
    }

    private boolean isAlterTableRewriteAttribute(Table oldTable, Table newTable) {
        return oldTable.isRewriteEnabled() != newTable.isRewriteEnabled();
    }

}
