/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p/>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p/>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.hive.hook.events;

import org.apache.atlas.hive.hook.AtlasHiveHookContext;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.instance.AtlasEntity.AtlasEntityWithExtInfo;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.notification.HookNotification;
import org.apache.atlas.model.notification.HookNotification.EntityPartialUpdateRequestV2;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.hive.metastore.events.AlterTableEvent;
import org.apache.hadoop.hive.ql.metadata.Table;
import org.apache.hadoop.hive.ql.plan.HiveOperation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

public class AlterTableProperties extends BaseHiveEvent {
    private static final Logger LOG = LoggerFactory.getLogger(AlterTableProperties.class);

    public AlterTableProperties(AtlasHiveHookContext context) {
        super(context);
    }

    @Override
    public List<HookNotification> getNotificationMessages() throws Exception {
        return context.isMetastoreHook() ? getHiveMetastoreMessages() : getHiveMessages();
    }

    public List<HookNotification> getHiveMetastoreMessages() throws Exception {
        List<HookNotification> ret      = new ArrayList<>();
        AlterTableEvent        tblEvent = (AlterTableEvent) context.getMetastoreEvent();
        Table                  oldTable = toTable(tblEvent.getOldTable());
        Table                  newTable = toTable(tblEvent.getNewTable());

        if (tblEvent.isReplicated()) {
            return ret;
        }

        processProperties(oldTable, newTable, ret);

        return ret;
    }

    public List<HookNotification> getHiveMessages() throws Exception {
        List<HookNotification> ret      = new ArrayList<>();

        if (CollectionUtils.isEmpty(getInputs())) {
            LOG.error("AlterTableProperties: old-table not found in inputs list");

            return null;
        }

        if (CollectionUtils.isEmpty(getOutputs())) {
            LOG.error("AlterTableProperties: new-table not found in outputs list");

            return null;
        }

        Table oldTable = getInputs().iterator().next().getTable();
        Table newTable = getOutputs().iterator().next().getTable();

        if (newTable != null) {
            newTable = getHive().getTable(newTable.getDbName(), newTable.getTableName());
        }

        processProperties(oldTable, newTable, ret);

        return ret;
    }

    private void processProperties(Table oldTable, Table newTable, List<HookNotification> ret) {
        if (oldTable == null || newTable == null) {
            return;
        }

        String        tableQualifiedName = context.getQualifiedName(oldTable);
        AtlasObjectId tableObjId         = new AtlasObjectId(HIVE_TYPE_TABLE, ATTRIBUTE_QUALIFIED_NAME, tableQualifiedName);
        AtlasEntity   updatedTableEntity = new AtlasEntity(HIVE_TYPE_TABLE);

        updatedTableEntity.setGuid(null);
        updatedTableEntity.setRelationshipAttributes(null);

        switch (context.getHiveOperation()) {
            case ALTERTABLE_PROPERTIES:
                if (MapUtils.isNotEmpty(newTable.getParameters())) {
                    updatedTableEntity.setAttribute(ATTRIBUTE_PARAMETERS, newTable.getParameters());
                }
                break;

            case ALTERTABLE_OWNER:
                updatedTableEntity.setAttribute(ATTRIBUTE_OWNER, newTable.getOwner());
                break;

            default:
                LOG.error("AlterTableProperties: no change detected in properties or owner");
                return;
        }

        ret.add(new EntityPartialUpdateRequestV2(getUserName(), tableObjId, new AtlasEntityWithExtInfo(updatedTableEntity)));
    }
}