/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.kafka.hook;

import org.apache.atlas.kafka.hook.entities.TopicEntity;
import org.apache.kafka.common.resource.ResourceType;
import org.apache.kafka.server.auditor.Auditor;
import org.apache.kafka.server.auditor.TopicEvent;

import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class EntitySelector {
    private final Predicate<Auditor.AuditInformation> topic =
            t -> t.resource().resourceType() == ResourceType.TOPIC;
    private final Predicate<Auditor.AuditInformation> allowed = t -> t.error() == 0;

    public List<TopicEntity> selectTopicEntitiesToBeCreated(TopicEvent event,
                                                            Set<Auditor.AuditInformation> topicCreationAuthInfo) {
        return topicCreationAuthInfo.stream()
                .filter(topic.and(allowed))
                .flatMap(auditInfo -> {
                    String topicName = auditInfo.resource().name();
                    Optional<TopicEvent.AuditedTopic> topic = event.topics()
                        .stream()
                        .filter(t -> t.name().equals(topicName))
                        .findFirst();
                    return topic.map(creatableTopic ->
                        Stream.of(new TopicEntity(creatableTopic.name(), creatableTopic.numPartitions(), creatableTopic.replicationFactor())))
                            .orElse(Stream.empty());
                }).collect(Collectors.toList());
    }

    public List<TopicEntity> selectEntitiesToBeDeleted(TopicEvent event,
                                                       Set<Auditor.AuditInformation> topicDeleteInfo) {
        return topicDeleteInfo.stream()
                .filter(topic.and(allowed))
                .flatMap(auditInfo -> {
                    String topicName = auditInfo.resource().name();
                    Set<String> eventTopics = event.topics().stream().map(TopicEvent.AuditedTopic::name).collect(Collectors.toSet());
                        return eventTopics.contains(topicName)
                                ? Stream.of(new TopicEntity(topicName))
                                : Stream.empty();

                })
                .collect(Collectors.toList());
    }

    public List<TopicEntity> selectAllowedTopics(Set<Auditor.AuditInformation> produceInfo) {
        return produceInfo.stream()
                .filter(topic.and(allowed))
                .map(auditInfo -> new TopicEntity(auditInfo.resource().name()))
                .collect(Collectors.toList());
    }

    public List<TopicEntity> selectTopicsForCreatePartition(TopicEvent event,
                                                            Set<Auditor.AuditInformation> alterTopicsInfo) {
        Set<String> allowedTopics = alterTopicsInfo.stream()
                .filter(topic.and(allowed))
                .map(auditInfo -> auditInfo.resource().name())
                .collect(Collectors.toSet());

        return event.topics().stream()
                .filter(topic -> allowedTopics.contains(topic.name()))
                .map(topic -> TopicEntity.withPartitionCount(topic.name(), topic.numPartitions()))
                .collect(Collectors.toList());
    }

    public List<TopicEntity> selectTopicsToUpdateReplicationFactor(TopicEvent event, Set<Auditor.AuditInformation> auditInformation) {
        Set<String> allowedTopics = auditInformation.stream()
                .filter(topic.and(allowed))
                .map(auditInfo -> auditInfo.resource().name())
                .collect(Collectors.toSet());

        return event.topics().stream()
                .filter(topic -> allowedTopics.contains(topic.name()))
                .map(topic -> TopicEntity.withReplicationFactor(topic.name(), topic.replicationFactor())).collect(Collectors.toList());
    }
}
