/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.kafka.hook;

import com.google.common.annotations.VisibleForTesting;
import org.apache.atlas.kafka.hook.entities.AtlasKafkaClientEntity;
import org.apache.atlas.kafka.hook.entities.AtlasKafkaEntity;
import org.apache.atlas.kafka.hook.entities.ClientLineageEntity;
import org.apache.atlas.kafka.hook.entities.ClientRelationship;
import org.apache.atlas.kafka.hook.entities.ConsumerEntity;
import org.apache.atlas.kafka.hook.entities.ConsumerGroupEntity;
import org.apache.atlas.kafka.hook.entities.ProducerEntity;
import org.apache.atlas.kafka.hook.entities.TopicEntity;
import org.apache.kafka.common.acl.AclOperation;
import org.apache.kafka.server.auditor.AuditEvent;
import org.apache.kafka.server.auditor.Auditor;
import org.apache.kafka.server.auditor.FetchEvent;
import org.apache.kafka.server.auditor.ProduceEvent;
import org.apache.kafka.server.auditor.SyncGroupEvent;
import org.apache.kafka.server.auditor.TopicEvent;
import org.apache.kafka.server.authorizer.AuthorizableRequestContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * This class is the main implementation of the Atlas-Kafka hook. It is able to handle topic creation, topic deletion,
 * topic replication factor and partition count updates.
 */
public class KafkaHookImpl implements Auditor {

    private static final String AUDITORS_CONFIG_PREFIX = "cloudera.auditors.";
    private static final String ATLAS_CONFIG_PREFIX = "atlas.";
    private static final String HOOK_CONNECTOR_CONFIG = "hook.connector";
    private static final String REST_CONNECTOR_CONFIG = "rest";
    private static final String KAFKA_CONNECTOR_CONFIG = "kafka";

    private final Logger LOG = LoggerFactory.getLogger(KafkaHookImpl.class.getName());
    private final Executor asyncTaskExecutor;
    private final EntitySelector entitySelector;

    private ClientRelationshipCache<ProducerEntity, TopicEntity> producerCache;
    private ClientRelationshipCache<ConsumerEntity, TopicEntity> consumerCache;
    private ClientRelationshipCache<ConsumerEntity, ConsumerGroupEntity> consumerGroupCache;
    private AtlasConnector atlasConnector;

    /**
     * The plugin is instantiated with this constructor by org.apache.atlas.kafka.hook.KafkaHook through reflection.
     */
    public KafkaHookImpl() {
        this(new EntitySelector());
    }

    private KafkaHookImpl(EntitySelector entitySelector) {
        this.entitySelector = entitySelector;
        asyncTaskExecutor = Executors.newSingleThreadExecutor(r -> {
            Thread thread = new Thread(r);
            thread.setName("KafkaHookImpl_worker");
            thread.setUncaughtExceptionHandler((t, e) -> LOG.error("Uncaught exception on thread {}", t, e));
            return thread;
        });
    }

    @VisibleForTesting
    KafkaHookImpl(AtlasConnector atlasConnector,
                  ClientRelationshipCache<ProducerEntity, TopicEntity> producerCache,
                  ClientRelationshipCache<ConsumerEntity, TopicEntity> consumerCache,
                  ClientRelationshipCache<ConsumerEntity, ConsumerGroupEntity> consumerGroupCache,
                  EntitySelector entitySelector) {
        this(entitySelector);
        this.atlasConnector = atlasConnector;
        this.producerCache = producerCache;
        this.consumerCache = consumerCache;
        this.consumerGroupCache = consumerGroupCache;
    }

    @Override
    public void close() {
        if (atlasConnector != null) {
            atlasConnector.close();
        }
    }

    @Override
    public void audit(AuditEvent event, AuthorizableRequestContext requestContext, Map<AclOperation, Set<AuditInformation>> auditInformation) {
        asyncTaskExecutor.execute(() -> auditInternal(event, requestContext, auditInformation));
    }

    @VisibleForTesting
    void auditInternal(AuditEvent event, AuthorizableRequestContext requestContext, Map<AclOperation,
            Set<AuditInformation>> auditInformation) {
        if (event instanceof TopicEvent) {
            switch (((TopicEvent) event).eventType()) {
                case CREATE:
                    createTopics((TopicEvent) event, auditInformation);
                    break;
                case DELETE:
                    deleteTopics((TopicEvent) event, auditInformation);
                    break;
                case PARTITION_CHANGE:
                    handleCreatePartition((TopicEvent) event, auditInformation);
                    break;
                case REPLICATION_FACTOR_CHANGE:
                    handleReplicationFactorChange((TopicEvent) event, auditInformation);
                    break;
            }
        } else if (event instanceof ProduceEvent) {
            handleProduce(requestContext, auditInformation);
        } else if (event instanceof SyncGroupEvent) {
            handleSyncGroup((SyncGroupEvent) event, requestContext);
        } else if (event instanceof FetchEvent) {
            handleFetch((FetchEvent) event, auditInformation);
        }
    }

    @VisibleForTesting
    void createTopics(TopicEvent event, Map<AclOperation, Set<AuditInformation>> auditInformation) {
        Set<Auditor.AuditInformation> topicCreateInfo = auditInformation.get(AclOperation.CREATE);
        List<TopicEntity> topicsToBeCreated = entitySelector.selectTopicEntitiesToBeCreated(event, topicCreateInfo);
        if (topicsToBeCreated.isEmpty()) {
            return;
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("Creating topics {}", topicsToBeCreated.stream().map(AtlasKafkaEntity::toString).collect(Collectors.joining(", ")));
        }
        atlasConnector.create(topicsToBeCreated);
    }

    @VisibleForTesting
    void deleteTopics(TopicEvent event, Map<AclOperation, Set<AuditInformation>> auditInformation) {
        Set<Auditor.AuditInformation> topicDeleteInfo = auditInformation.get(AclOperation.DELETE);
        List<TopicEntity> topicsToBeDeleted = entitySelector.selectEntitiesToBeDeleted(event, topicDeleteInfo);
        if (topicsToBeDeleted.isEmpty()) {
            return;
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("Deleting topics {}", topicsToBeDeleted.stream().map(AtlasKafkaEntity::toString).collect(Collectors.joining(", ")));
        }
        atlasConnector.delete(topicsToBeDeleted);
    }

    @VisibleForTesting
    void handleProduce(AuthorizableRequestContext requestContext, Map<AclOperation,
            Set<AuditInformation>> auditInformation) {
        if (requestContext.clientId() == null || "".equals(requestContext.clientId())) {
            return;
        }

        Set<Auditor.AuditInformation> produceInfo = auditInformation.get(AclOperation.WRITE);
        List<TopicEntity> allowedTopics = entitySelector.selectAllowedTopics(produceInfo);
        ProducerEntity producerEntity = new ProducerEntity(requestContext.clientId());
        boolean producerExists = producerCache.clientExists(producerEntity);
        List<ClientRelationship<ProducerEntity, TopicEntity>> newRelationships =
                createMissingRelationships(producerEntity, allowedTopics, producerCache);
        if (!newRelationships.isEmpty()) {
            try {
                ClientLineageEntity producerLineageEntity = ClientLineageEntity.of(producerEntity, allowedTopics);
                if (producerExists) {
                    atlasConnector.update(producerLineageEntity);
                    // no need to update the producer as we currently store its client id only which isn't updateable
                } else {
                    createClientInAtlas(producerEntity);
                    atlasConnector.create(producerLineageEntity);
                }
            } catch (Exception e) {
                LOG.warn("Couldn't add producer " + producerEntity.getQualifiedName(), e);
                producerCache.invalidateAll(newRelationships);
            }
        }
    }

    @VisibleForTesting
    void handleSyncGroup(SyncGroupEvent event, AuthorizableRequestContext requestContext) {
        if (requestContext.clientId() == null || "".equals(requestContext.clientId())) {
            return;
        }

        ConsumerGroupEntity consumerGroupEntity = new ConsumerGroupEntity(event.groupId());
        ConsumerEntity consumerEntity = new ConsumerEntity(requestContext.clientId());
        boolean clientExistsInGroupCache = consumerGroupCache.clientExists(consumerEntity);
        Optional<ClientRelationship<ConsumerEntity, ConsumerGroupEntity>> relationship = consumerGroupCache
                .putRelationship(consumerEntity, consumerGroupEntity);
        consumerEntity.setGroupRelationship(consumerGroupEntity.toAtlasEntity());
        if (relationship.isPresent()) {
            try {
                atlasConnector.create(consumerGroupEntity);
                if (consumerCache.clientExists(consumerEntity) || clientExistsInGroupCache) {
                    updateClientInAtlas(consumerEntity);
                } else {
                    createClientInAtlas(consumerEntity);
                }
            } catch (Exception e) {
                LOG.warn("Couldn't add consumer group " + consumerGroupEntity.getQualifiedName(), e);
                consumerGroupCache.invalidateAll(Collections.singletonList(relationship.get()));
            }
        }
    }

    @VisibleForTesting
    void handleFetch(FetchEvent event, Map<AclOperation, Set<AuditInformation>> auditInformation) {
        if (event.clientId() == null || "".equals(event.clientId())) {
            return;
        }

        List<TopicEntity> assignedTopics = entitySelector.selectAllowedTopics(auditInformation.get(AclOperation.READ));
        ConsumerEntity consumerEntity = new ConsumerEntity(event.clientId());
        boolean clientExistsInConsumerCache = consumerCache.clientExists(consumerEntity);
        List<ClientRelationship<ConsumerEntity, TopicEntity>> newRelationships = createMissingRelationships(consumerEntity, assignedTopics, consumerCache);
        ClientLineageEntity consumerLineageEntity = ClientLineageEntity.of(consumerEntity, assignedTopics);
        if (!newRelationships.isEmpty()) {
            try {
                if (!consumerGroupCache.clientExists(consumerEntity) && !clientExistsInConsumerCache) {
                    createClientInAtlas(consumerEntity);
                }
                if (clientExistsInConsumerCache) {
                    atlasConnector.update(consumerLineageEntity);
                } else {
                    atlasConnector.create(consumerLineageEntity);
                }
            } catch (Exception e) {
                LOG.warn("Couldn't add consumer " + consumerEntity.getQualifiedName(), e);
                consumerCache.invalidateAll(newRelationships);
            }
        }
    }

    @VisibleForTesting
    void handleCreatePartition(TopicEvent event,
                               Map<AclOperation, Set<AuditInformation>> auditInformation) {
        Set<Auditor.AuditInformation> alterAuditInfo = auditInformation.get(AclOperation.ALTER);
        List<TopicEntity> allowedTopics =
                entitySelector.selectTopicsForCreatePartition(event, alterAuditInfo);
        if (allowedTopics.isEmpty()) {
            return;
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("Handling partition update {}", allowedTopics.stream().map(AtlasKafkaEntity::toString).collect(Collectors.joining(", ")));
        }
        atlasConnector.update(allowedTopics);
    }

    @VisibleForTesting
    void handleReplicationFactorChange(TopicEvent event, Map<AclOperation, Set<AuditInformation>> auditInformation) {
        Set<Auditor.AuditInformation> alterAuditInfo = auditInformation.get(AclOperation.ALTER);
        List<TopicEntity> updatedTopics = entitySelector
                .selectTopicsToUpdateReplicationFactor(event, alterAuditInfo);
        if (LOG.isDebugEnabled()) {
            LOG.debug("Handling replication factor update {}", updatedTopics.stream()
                    .map(AtlasKafkaEntity::toString).collect(Collectors.joining(", ")));
        }
        if (updatedTopics.isEmpty()) {
            return;
        }
        atlasConnector.update(updatedTopics);
    }

    private void createClientInAtlas(AtlasKafkaClientEntity clientEntity) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Creating client {}", clientEntity);
        }
        atlasConnector.create(clientEntity);
    }

    private void updateClientInAtlas(AtlasKafkaClientEntity clientEntity) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Partially updating existing client {}", clientEntity);
        }
        atlasConnector.update(clientEntity);
    }

    @Override
    public void configure(Map<String, ?> configs) {
        // On startup we asynchronously synchronize Atlas with the topics in Kafka
        Map<String, Object> auditorConfig = getConfigsOf(AUDITORS_CONFIG_PREFIX, configs);
        consumerCache = new ClientRelationshipCache<>(20000);
        producerCache = new ClientRelationshipCache<>(20000);
        consumerGroupCache = new ClientRelationshipCache<>(20000);
        //If you want to test with REST client you'll need this
        Map<String, Object> atlasAuditorRelatedKafkaConfigs = getConfigsOf(ATLAS_CONFIG_PREFIX, auditorConfig);
        String hookConnectorConfig = (String) atlasAuditorRelatedKafkaConfigs.get(HOOK_CONNECTOR_CONFIG);
        if (REST_CONNECTOR_CONFIG.equals(hookConnectorConfig)) {
            this.atlasConnector = new RestApiConnector(atlasAuditorRelatedKafkaConfigs);
        } else if (hookConnectorConfig == null || KAFKA_CONNECTOR_CONFIG.equals(hookConnectorConfig)) {
            this.atlasConnector = new KafkaHookNotification();
        } else {
            throw new RuntimeException(String.format("Invalid configuration for %s: %s", HOOK_CONNECTOR_CONFIG, hookConnectorConfig));
        }
    }

    private <T extends AtlasKafkaClientEntity, U extends AtlasKafkaEntity> List<ClientRelationship<T, U>> createMissingRelationships(
            T client,
            List<U> qualifiedTopicNames,
            ClientRelationshipCache<T, U> cache) {
        List<ClientRelationship<T, U>> newlyAddedRelationships = qualifiedTopicNames.stream()
                .flatMap(topic -> {
                    Optional<ClientRelationship<T, U>> relationship = cache
                            .putRelationship(client, topic);
                    return relationship.map(Stream::of).orElseGet(Stream::empty);
                })
                .collect(Collectors.toList());

        if (newlyAddedRelationships.isEmpty()) {
            return Collections.emptyList();
        }

        return newlyAddedRelationships;
    }

    static Map<String, Object> getConfigsOf(String prefix, Map<String, ?> configMap) {
        Map<String, Object> props = new HashMap<>();
        for (Map.Entry<String, ?> entry : configMap.entrySet()) {
            if (entry.getKey().startsWith(prefix)) {
                props.put(entry.getKey().replace(prefix, ""), entry.getValue());
            }
        }
        return props;
    }
}
