/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.kafka.hook;

import org.apache.atlas.kafka.hook.entities.AtlasQualifiedName;
import org.apache.atlas.kafka.hook.entities.LazyAtlasConfigStore;
import org.apache.atlas.kafka.hook.entities.TopicEntity;
import org.apache.kafka.common.resource.PatternType;
import org.apache.kafka.common.resource.ResourcePattern;
import org.apache.kafka.common.resource.ResourceType;
import org.apache.kafka.server.auditor.Auditor;
import org.apache.kafka.server.auditor.TopicEvent;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

public class EntitySelectorTest {

    private static EntitySelector entitySelector;

    private static final String TEST_TOPIC_NAME1 = "test1";
    private static final String TEST_TOPIC_NAME2 = "test2";

    @BeforeClass
    public static void setup() {
        entitySelector = new EntitySelector();
    }


    @Test
    public void testCreateTopicSelectEntities() {
        TopicEvent event = new TopicEvent(Collections.singleton(new TopicEvent.AuditedTopic(TEST_TOPIC_NAME1, 3, 3)),
                TopicEvent.EventType.CREATE);

        Set<Auditor.AuditInformation> auditInformation = createSampleAuditInformation();

        List<TopicEntity> entitiesToBeCreated =
                entitySelector.selectTopicEntitiesToBeCreated(event, auditInformation);

        Assert.assertEquals(entitiesToBeCreated.size(), 1);

        AtlasQualifiedName createdTopicName = entitiesToBeCreated.get(0).getQualifiedName();
        Assert.assertEquals(createdTopicName.getUniqueIdentifier(), TEST_TOPIC_NAME1);

    }

    @Test
    public void testCreateTopicSelectEntities_NothingToCreate() {
        TopicEvent event = new TopicEvent(Collections.singleton(new TopicEvent.AuditedTopic("test3", 3, 3)),
                TopicEvent.EventType.CREATE);

        Set<Auditor.AuditInformation> auditInformation = createSampleAuditInformation();

        List<TopicEntity> entitiesToBeCreated =
                entitySelector.selectTopicEntitiesToBeCreated(event, auditInformation);

        Assert.assertTrue(entitiesToBeCreated.isEmpty());
    }

    @Test
    public void testDeleteTopicSelectEntities() {
        TopicEvent event = new TopicEvent(Collections.singleton(new TopicEvent.AuditedTopic(TEST_TOPIC_NAME1)),
                TopicEvent.EventType.DELETE);

        Set<Auditor.AuditInformation> auditInformation = createSampleAuditInformation();

        List<TopicEntity> topicsToBeDeleted = entitySelector.selectEntitiesToBeDeleted(event, auditInformation);

        Assert.assertEquals(topicsToBeDeleted.size(), 1);
        Assert.assertEquals(topicsToBeDeleted.get(0).getQualifiedName().getQualifiedName(),
                TEST_TOPIC_NAME1 + "@" + LazyAtlasConfigStore.getInstance().getTopicMetadataNamespace());
    }

    @Test
    public void testDeleteTopicSelectEntities_NothingToDelete() {
        TopicEvent event = new TopicEvent(Collections.singleton(new TopicEvent.AuditedTopic("test-3")),
                TopicEvent.EventType.CREATE);

        Set<Auditor.AuditInformation> auditInformation = createSampleAuditInformation();

        List<TopicEntity> topicsToBeDeleted = entitySelector.selectEntitiesToBeDeleted(event, auditInformation);

        Assert.assertTrue(topicsToBeDeleted.isEmpty());
    }

    @Test
    public void testSelectTopicsForProducer_SelectAll() {
        Set<Auditor.AuditInformation> auditInformation = createSampleAuditInformation();

        List<TopicEntity> topicsToProduceIn = entitySelector.selectAllowedTopics(auditInformation);
        Assert.assertEquals(topicsToProduceIn.size(), 2);
        Assert.assertEquals(topicsToProduceIn.toArray(),
                auditInformation.stream().map(audit -> new TopicEntity(audit.resource().name())).toArray());
    }

    @Test
    public void testSelectTopicsForProducer_SelectError() {
        ResourcePattern resource = new ResourcePattern(ResourceType.TOPIC, TEST_TOPIC_NAME1, PatternType.LITERAL);
        List<TopicEntity> topicsToProduceIn = entitySelector.selectAllowedTopics(
                Collections.singleton(new Auditor.AuditInformation(resource, 3))); // 3 = unknown topics or partitions

        Assert.assertTrue(topicsToProduceIn.isEmpty());
    }

    @Test
    public void testSelectTopicsForConsumer() {
        ResourcePattern resource1 = new ResourcePattern(ResourceType.TOPIC, TEST_TOPIC_NAME1, PatternType.LITERAL);

        Set<Auditor.AuditInformation> auditInfo = new HashSet<>(Collections
                .singletonList(new Auditor.AuditInformation(resource1, 0)));

        List<TopicEntity> topicsToConsumeFrom = entitySelector.selectAllowedTopics(auditInfo);

        Assert.assertEquals(topicsToConsumeFrom.size(), 1);
        Assert.assertEquals(topicsToConsumeFrom.get(0).getQualifiedName().getQualifiedName(),
                TEST_TOPIC_NAME1 + "@" + LazyAtlasConfigStore.getInstance().getTopicMetadataNamespace());
    }

    @Test
    public void testSelectTopicsToUpdateReplicationFactor() {
        Set<Auditor.AuditInformation> auditInfo = new HashSet<>(Arrays.asList(
            new Auditor.AuditInformation(
                new ResourcePattern(ResourceType.TOPIC, TEST_TOPIC_NAME1,
                    PatternType.LITERAL),
                0
            ),
            new Auditor.AuditInformation(
                new ResourcePattern(ResourceType.TOPIC, TEST_TOPIC_NAME2,
                    PatternType.LITERAL),
                -1
            )
        ));

        List<TopicEntity> expected = Collections.singletonList(
                TopicEntity.withReplicationFactor(TEST_TOPIC_NAME1, 3)
        );
        Set<TopicEvent.AuditedTopic> auditedTopics = new HashSet<>();
        auditedTopics.add(TopicEvent.AuditedTopic.withReplicationFactor(TEST_TOPIC_NAME1, 3));
        auditedTopics.add(TopicEvent.AuditedTopic.withReplicationFactor(TEST_TOPIC_NAME2, 3));
        TopicEvent event = new TopicEvent(auditedTopics, TopicEvent.EventType.REPLICATION_FACTOR_CHANGE);

        List<TopicEntity> actual = entitySelector.selectTopicsToUpdateReplicationFactor(event, auditInfo)
            .stream()
            .sorted(Comparator.comparing(o -> o.getQualifiedName().getQualifiedName()))
            .collect(Collectors.toList());

        Assert.assertEquals(expected, actual);
    }

    @Test
    public void selectTopicsForCreatePartition_returnsAllowedTopics() {
        Set<Auditor.AuditInformation> auditInformation = new HashSet<>();
        auditInformation.add(new Auditor.AuditInformation(topicResource(TEST_TOPIC_NAME1), 0));
        auditInformation.add(new Auditor.AuditInformation(topicResource(TEST_TOPIC_NAME2), 29)); // topic authorization failed
        auditInformation.add(new Auditor.AuditInformation(unknownResource("unknown-1"), 0));

        List<TopicEntity> expectedTopics = new ArrayList<>();
        expectedTopics.add(TopicEntity.withPartitionCount(TEST_TOPIC_NAME1, 7));
        Set<TopicEvent.AuditedTopic> auditedTopics = new HashSet<>();
        auditedTopics.add(TopicEvent.AuditedTopic.withPartitionNumber(TEST_TOPIC_NAME1, 7));
        auditedTopics.add(TopicEvent.AuditedTopic.withPartitionNumber(TEST_TOPIC_NAME2, 7));
        auditedTopics.add(TopicEvent.AuditedTopic.withPartitionNumber("unknown-1", 7));
        TopicEvent event = new TopicEvent(auditedTopics, TopicEvent.EventType.PARTITION_CHANGE);

        List<TopicEntity> actualTopics = entitySelector.selectTopicsForCreatePartition(event, auditInformation);

        Assert.assertEquals(expectedTopics, actualTopics);
    }


    private Set<Auditor.AuditInformation> createSampleAuditInformation() {
        ResourcePattern resource1 = new ResourcePattern(ResourceType.TOPIC, TEST_TOPIC_NAME1, PatternType.LITERAL);
        ResourcePattern resource2 = new ResourcePattern(ResourceType.TOPIC, TEST_TOPIC_NAME2, PatternType.LITERAL);

        return new HashSet<>(Arrays.asList(new Auditor.AuditInformation(resource1, 0),
                new Auditor.AuditInformation(resource2, 0)));
    }

    private ResourcePattern topicResource(String name) {
        return new ResourcePattern(ResourceType.TOPIC, name, PatternType.LITERAL);
    }

    private ResourcePattern unknownResource(String name) {
        return new ResourcePattern(ResourceType.UNKNOWN, name, PatternType.LITERAL);
    }
}
