/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.utils;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

public class XMLConfigLoader {
    private static final Logger LOG                 = LoggerFactory.getLogger(XMLConfigLoader.class);
    private static final String XML_CONFIG_PROPERTY = "property";
    private static final String XML_CONFIG_NAME     = "name";
    private static final String XML_CONFIG_VALUE    = "value";
    private static final String PATH_SEPARATOR      = "/";

    public static void loadConfig(String configFileName, Map<Object, Object> properties) throws AtlasBaseException {
        try (InputStream input = getFileInputStream(configFileName)) {
            loadConfig(input, properties);
        } catch (FileNotFoundException e) {
            String msg = String.format("The xml configuration file %s not found", configFileName);

            if (LOG.isDebugEnabled()) {
                LOG.debug(msg, e);
            }
            throw new AtlasBaseException(msg, e);
        } catch (IOException | IllegalArgumentException e) {
            String msg = String.format("Unable to load the config file %s", configFileName);

            if (LOG.isDebugEnabled()) {
                LOG.debug(msg, e);
            }
            throw new AtlasBaseException(msg, e);
        } catch (Exception e) {
            String msg = String.format("Unable to load the config file %s due to parsing error", configFileName);

            if (LOG.isDebugEnabled()) {
                LOG.debug(msg, e);
            }
            throw new AtlasBaseException(msg, e);
        }
    }

    public static void loadConfig(InputStream input, Map<Object, Object> properties) throws AtlasBaseException {
        try {
            DocumentBuilderFactory xmlDocumentBuilderFactory = DocumentBuilderFactory.newInstance();

            xmlDocumentBuilderFactory.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);
            xmlDocumentBuilderFactory.setFeature("http://xml.org/sax/features/external-general-entities", false);
            xmlDocumentBuilderFactory.setFeature("http://xml.org/sax/features/external-parameter-entities", false);
            xmlDocumentBuilderFactory.setIgnoringComments(true);
            xmlDocumentBuilderFactory.setNamespaceAware(true);

            DocumentBuilder xmlDocumentBuilder = xmlDocumentBuilderFactory.newDocumentBuilder();
            Document        xmlDocument        = xmlDocumentBuilder.parse(input);
            NodeList        nList              = xmlDocument.getElementsByTagName(XML_CONFIG_PROPERTY);
            xmlDocument.getDocumentElement().normalize();


            for (int temp = 0; temp < nList.getLength(); temp++) {
                Node nNode = nList.item(temp);

                if (nNode.getNodeType() == Node.ELEMENT_NODE) {
                    Element eElement = (Element) nNode;

                    String propertyName = "";
                    String propertyValue = "";

                    if (eElement.getElementsByTagName(XML_CONFIG_NAME).item(0) != null) {
                        propertyName = eElement.getElementsByTagName(XML_CONFIG_NAME).item(0).getTextContent().trim();
                    }

                    if (eElement.getElementsByTagName(XML_CONFIG_VALUE).item(0) != null) {
                        propertyValue = eElement.getElementsByTagName(XML_CONFIG_VALUE).item(0).getTextContent().trim();
                    }

                    if (properties.get(propertyName) != null) {
                        properties.remove(propertyName);
                    }

                    properties.put(propertyName, propertyValue);
                }
            }

        } catch (ParserConfigurationException | IOException | SAXException e) {
            LOG.error("Error loading the xml configuration file", e);

            throw new AtlasBaseException("Error loading the xml configuration file", e);
        }
    }

    private static InputStream getFileInputStream(String path) throws FileNotFoundException {
        InputStream ret           = null;
        String      sanitizedPath = new File(path).getName(); // Guard against path traversal attacks

        if (StringUtils.isNotBlank(sanitizedPath)) {
            File f = new File(sanitizedPath);

            if (f.exists()) {
                ret = new FileInputStream(f);
            } else {
                ret = XMLConfigLoader.class.getResourceAsStream(path);

                if (ret == null && !path.startsWith(PATH_SEPARATOR)) {
                    ret = XMLConfigLoader.class.getResourceAsStream(PATH_SEPARATOR + path);
                }

                if (ret == null) {
                    ret = ClassLoader.getSystemClassLoader().getResourceAsStream(path);

                    if (ret == null && !path.startsWith(PATH_SEPARATOR)) {
                        ret = ClassLoader.getSystemResourceAsStream(PATH_SEPARATOR + path);
                    }
                }
            }
        }

        return ret;
    }
}