/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.BackingStore;

import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.mapdb.DB;
import org.mapdb.DBMaker;
import org.mapdb.Serializer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.nio.file.Paths;
import java.util.UUID;
import java.util.concurrent.ConcurrentMap;

public class BackingStoreProvider {
    private static final Logger LOG = LoggerFactory.getLogger(BackingStoreProvider.class);

    private DB backingStore;
    private String backingStorePath;

    public BackingStoreProvider() {
        this(false, UUID.randomUUID().toString());
    }

    public BackingStoreProvider(boolean useBackingStoreDirectory, String tempFileName) {
        if (backingStore != null) {
            return;
        }

        backingStore = createBackingStore(useBackingStoreDirectory, tempFileName);
    }

    public ConcurrentMap createMap(String name) {
        LOG.info(String.format("BackingStoreProvider.createMap: %s", name));
        return this.backingStore.hashMap(name).keySerializer(Serializer.STRING).valueSerializer(Serializer.STRING).createOrOpen();
    }

    public DB.TreeMapSink<String, String> createMapSink(String collectionName) {
        LOG.debug(String.format("BackingStoreProvider.createMapSink: %s", collectionName));
        return this.backingStore.treeMap(collectionName, Serializer.STRING, Serializer.STRING).createFromSink();
    }

    private DB createBackingStore(boolean useBackingStoreDirectory, String fileToWrite) {
        this.backingStorePath = getTemporaryPath(fileToWrite);
        if (!useBackingStoreDirectory || StringUtils.isEmpty(backingStorePath)) {
            LOG.info("Backing Store: Using in-memory.");
            return DBMaker
                    .memoryDirectDB()
                    .concurrencyScale(PropertiesUtils.getNumWorkers())
                    .closeOnJvmShutdown()
                    .make();
        } else {
            LOG.info("Backing Store: Using backing store: {}", backingStorePath);
            return DBMaker
                    .fileDB(new File(backingStorePath))
                    .fileMmapEnable()
                    .concurrencyScale(PropertiesUtils.getNumWorkers())
                    .fileDeleteAfterClose()
                    .closeOnJvmShutdown()
                    .make();
        }
    }

    private String getTemporaryPath(String fileToWrite) {
        String tempFilePath = Paths.get(PropertiesUtils.getTempDirectory(), String.format("%s.db", fileToWrite)).toString();
        File f = new File(tempFilePath);
        if (f.exists()) {
            LOG.info("BackingStore: Found: {}. Will delete and continue.", tempFilePath);
            close();
        }

        return tempFilePath;
    }

    public void close() {
        if (this.backingStore != null) {
            this.backingStore.close();
            this.backingStore = null;
        }

        FileUtils.deleteQuietly(new File(this.backingStorePath));
    }
}
