/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.BackingStore;

import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.atlas.tools.utils.Spawner;
import org.apache.atlas.tools.utils.MapStoringJsonList;
import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;

public class BulkLoader {
    private static final Logger LOG = LoggerFactory.getLogger(BulkLoader.class);
    private static final String FORMAT_TEMP_FILE_COLLECTION = "%s-%s.csv";

    private Session session;
    private List<String> separatedFileNames;

    public BulkLoader(Session session) {
        this.session = session;
    }

    public void load() throws IOException {
        sortIndividualFilesById(separatedFileNames);
        loadSortedCollections();
    }

    private void loadSortedCollections() {
        SinkLoader slChildParent = SinkLoader.create(session, RelationCollections.CHILD_PARENT);
        SinkLoader slParentChild = SinkLoader.create(session, RelationCollections.PARENT_CHILD);
        SinkLoader slTemplateInstances = SinkLoader.create(session, RelationCollections.TEMPLATE_INSTANCES);
        SinkLoader slInstanceOf = SinkLoader.create(session, RelationCollections.INSTANCE_OF);
        SinkLoader slDataFlowInputs = SinkLoader.create(session, RelationCollections.DATA_FLOW_INPUTS);
        SinkLoader slDataFlowOutputs = SinkLoader.create(session, RelationCollections.DATA_FLOW_OUTPUTS);
        SinkLoader slControlFlowInputs = SinkLoader.create(session, RelationCollections.CONTROL_FLOW_INPUTS);
        SinkLoader slControlFlowOutputs = SinkLoader.create(session, RelationCollections.CONTROL_FLOW_OUTPUTS);
        SinkLoader slIdentityGuidMap = SinkLoader.create(session, RelationCollections.IDENTITY_GUID_MAP);
        SinkLoader slGuidTypeMap = SinkLoader.create(session, RelationCollections.GUID_TYPE_MAP);

        session.getRelationsRepository().load(
                slInstanceOf.getResult(),
                new MapStoringJsonList(slDataFlowInputs.getResult()),
                new MapStoringJsonList(slDataFlowOutputs.getResult()),
                new MapStoringJsonList(slControlFlowInputs.getResult()),
                new MapStoringJsonList(slControlFlowOutputs.getResult()),
                slChildParent.getResult(),
                new MapStoringJsonList(slParentChild.getResult()),
                new MapStoringJsonList(slTemplateInstances.getResult()),
                slIdentityGuidMap.getResult(),
                slGuidTypeMap.getResult());

        deleteTemporarySortedCsvFiles();
    }

    private void deleteTemporarySortedCsvFiles() {
        try {
            LOG.info("Sorted temporary files: Delete: Starting...");

            SinkLoader.deleteTemporarySortedCsvFiles(session, new String[] {
                            RelationCollections.PARENT_CHILD,
                            RelationCollections.CHILD_PARENT,
                            RelationCollections.TEMPLATE_INSTANCES,
                            RelationCollections.CONTROL_FLOW_OUTPUTS,
                            RelationCollections.CONTROL_FLOW_INPUTS,
                            RelationCollections.DATA_FLOW_OUTPUTS,
                            RelationCollections.DATA_FLOW_INPUTS,
                            RelationCollections.INSTANCE_OF,
                            RelationCollections.IDENTITY_GUID_MAP,
                            RelationCollections.GUID_TYPE_MAP
                    }
            );
        }
        finally {
            System.gc();
            LOG.info("Sorted temporary files: Delete: Done!");
        }
    }

    public void separate(InputStream inputStream) {
        try {
            separatedFileNames = separateRelationsByType(inputStream);
        } catch (IOException e) {
            LOG.error("BulkLoader: IOException!", e);
        } catch (Exception e) {
            LOG.error("BulkLoader: Exception!", e);
        }
    }

    private void sortIndividualFilesById(List<String> csvFilesWithCollections) {
        try {
            if (PropertiesUtils.useSpawner()) {
                Spawner.exec(FileSortAndMerge.class, csvFilesWithCollections);
            }
            else {
                new FileSortAndMerge().run(csvFilesWithCollections);
            }
        } catch (IOException exception) {
            LOG.error("Error: {}", csvFilesWithCollections, exception);
        } catch (InterruptedException e) {
            LOG.error("Error: {}", csvFilesWithCollections, e);
        }
    }

    private List<String> separateRelationsByType(InputStream inputStream) throws IOException {
        RelationsTypeSeparator relationsReader = new RelationsTypeSeparator(session.getTemporaryFileToWriteWithoutExtension());
        relationsReader.read(inputStream);

        return relationsReader.getSeparatedFilesPath();
    }

    public void writeToTemporaryLocation(String collectionName, InputStream inputStream) throws IOException {
        String tempFileName = String.format(FORMAT_TEMP_FILE_COLLECTION, session.getTemporaryFileToWriteWithoutExtension(), collectionName);
        IOUtils.copy(inputStream, new FileOutputStream(new File(tempFileName)));
        separatedFileNames.add(tempFileName);
    }
}
