/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.BackingStore;

import com.google.code.externalsorting.csv.CsvExternalSort;
import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.commons.csv.CSVRecord;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

public class FileSortAndMerge {
    private static final Logger LOG = LoggerFactory.getLogger(FileSortAndMerge.class);
    private static final String FILE_EXTENSION_CSV = ".csv";
    private static final String FILE_EXTENSION_SORTED_CSV = "-sorted.csv";

    private List<String> filesSorted = new ArrayList<>();

    public static void main(String[] args) {
        FileSortAndMerge fileSortAndMerge = new FileSortAndMerge();
        fileSortAndMerge.run(Arrays.asList(args));
    }

    public void run(List<String> filesToSort) {
        try {
            LOG.info("Sorting: {} files: Starting...", filesToSort.size());

            String temporaryDirectory = PropertiesUtils.getTempDirectory();

            AtomicLong idx = new AtomicLong(0);
            filesToSort.parallelStream().forEach(f -> {
                File tmpDirectory = new File(String.format("%s/%s", temporaryDirectory, idx.getAndIncrement()));
                String fileSorted = sortAndMerge(tmpDirectory, f);
                if (StringUtils.isEmpty(fileSorted)) {
                    return;
                }

                filesSorted.add(fileSorted);
                FileUtils.deleteQuietly(new File(f));
            });
        }
        finally {
            System.gc();
            LOG.info("Sorting: {} files: Done!", filesToSort.size());
        }
    }

    public List<String> getFilesSorted() {
        return filesSorted;
    }

    private String sortAndMerge(File tmpDirectory, String file) {
        try {
            LOG.info("Sorting: {}: Starting...", file);

            if (!tmpDirectory.exists()) {
                FileUtils.forceMkdir(tmpDirectory);
            }

            File fileSorted = new File(file.replace(FILE_EXTENSION_CSV, FILE_EXTENSION_SORTED_CSV));
            File inputFile = new File(file);
            Comparator<CSVRecord> comparator = (op1, op2) -> op1.get(0).compareTo(op2.get(0));
            List<File> sortInBatch = CsvExternalSort.sortInBatch(inputFile, comparator,
                    CsvExternalSort.DEFAULTMAXTEMPFILES,
                    Charset.defaultCharset(),
                    tmpDirectory, false, 0);

            CsvExternalSort.mergeSortedFiles(sortInBatch, fileSorted, comparator, Charset.defaultCharset(), false, true);
            FileUtils.deleteDirectory(tmpDirectory);
            return fileSorted.toString();
        } catch (Exception ex) {
            LOG.error("Sorting: {}: Exception!", file, ex);
        } finally {
            LOG.info("Sorting: {}: Done!", file);
        }

        return StringUtils.EMPTY;
    }
}
