/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.BackingStore;

import org.apache.atlas.tools.converter.EntityRelationsParser;
import org.apache.atlas.tools.models.Entity;
import org.apache.atlas.tools.models.Relation;
import org.apache.atlas.tools.pc.AsyncFileWriter;
import org.apache.atlas.tools.utils.JsonUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class RelationsTypeSeparator extends EntityRelationsParser {
    private static final Logger LOG = LoggerFactory.getLogger(RelationsTypeSeparator.class);

    private Map<String, AsyncFileWriter> files = new HashMap<>();
    private List<String> separatedFilesPath = new ArrayList<>();

    public RelationsTypeSeparator(String inputFile) {
        try {
            LOG.info("RelationsTypeSeparator: Init: Starting...");
            initBufferedWriter(inputFile, RelationCollections.INSTANCE_OF);
            initBufferedWriter(inputFile, RelationCollections.DATA_FLOW_INPUTS);
            initBufferedWriter(inputFile, RelationCollections.DATA_FLOW_OUTPUTS);
            initBufferedWriter(inputFile, RelationCollections.CONTROL_FLOW_INPUTS);
            initBufferedWriter(inputFile, RelationCollections.CONTROL_FLOW_OUTPUTS);
            initBufferedWriter(inputFile, RelationCollections.CHILD_PARENT);
            initBufferedWriter(inputFile, RelationCollections.PARENT_CHILD);
            initBufferedWriter(inputFile, RelationCollections.TEMPLATE_INSTANCES);
        } finally {
            LOG.info("RelationsTypeSeparator: Init: Done!");
        }
    }

    private void initBufferedWriter(String inputFile, String collectionName) {
        String tempFileName = String.format("%s-%s.csv", inputFile, collectionName);
        File f = new File(tempFileName);
        try {
            this.separatedFilesPath.add(tempFileName);
            AsyncFileWriter asyncFileWriter = new AsyncFileWriter(collectionName, tempFileName, new BufferedWriter(new FileWriter(f)));
            this.files.put(collectionName, asyncFileWriter);
        } catch (IOException e) {
            e.printStackTrace();
            LOG.error("RelationsTypeSeparator: Error creating: {} - {} - {}", collectionName, inputFile, tempFileName, e);
        }
    }

    @Override
    protected void endEntities() {

    }

    @Override
    protected void startEntities() {

    }

    @Override
    protected void endRelations() {
        try {
            for (Map.Entry<String, AsyncFileWriter> entry : this.files.entrySet()) {
                entry.getValue().flush();
                entry.getValue().close();
            }
        } finally {
            LOG.info("RelationsTypeSeparator: Processing: Done!");
        }
    }

    @Override
    protected void startRelations() {
        LOG.info("RelationsTypeSeparator: Processing: Starting...");
    }

    @Override
    protected void processEntity(Entity entity) {

    }

    @Override
    protected void processRelation(Relation relation) {
        if (relation.size() == 0) {
            return;
        }

        try {
            separate(relation);
        } finally {
            relation.get().clear();
            relation.set(null);
        }
    }

    private void separate(Relation relation) {
        switch (relation.getType()) {
            case Relation.RELATION_TYPE_INSTANCE_OF: {
                List instances = Relation.InstanceOf.getInstances(relation);
                String template = Relation.InstanceOf.getTemplate(relation);

                addKeysValue(RelationCollections.INSTANCE_OF, instances, template);
                addKeyValues(RelationCollections.TEMPLATE_INSTANCES, template, instances);

                instances.clear();
                break;
            }

            case Relation.RELATION_TYPE_PARENT_CHILD: {
                List children = Relation.ParentChild.getChildren(relation);
                String parent = Relation.ParentChild.getParent(relation);

                addKeysValue(RelationCollections.CHILD_PARENT, children, parent);
                addKeyValues(RelationCollections.PARENT_CHILD, parent, children);

                children.clear();
                break;
            }

            case Relation.RELATION_TYPE_DATA_FLOW: {
                List targets = Relation.DataFlow.getTargets(relation);
                List sources = Relation.DataFlow.getSources(relation);

                addKeysValues(RelationCollections.DATA_FLOW_INPUTS, targets, sources);
                addKeysValues(RelationCollections.DATA_FLOW_OUTPUTS, sources, targets);

                targets.clear();
                sources.clear();
                break;
            }

            case Relation.RELATION_TYPE_CONTROL_FLOW: {
                List targets = Relation.ControlFlow.getTargets(relation);
                List sources = Relation.ControlFlow.getSources(relation);

                addKeysValues(RelationCollections.CONTROL_FLOW_INPUTS, targets, sources);
                addKeysValues(RelationCollections.CONTROL_FLOW_OUTPUTS, sources, targets);

                targets.clear();
                sources.clear();
                break;
            }
        }
    }

    private void addKeysValues(String collectionName, List keys, List values) {
        for (Object key : keys) {
            addKeyValues(collectionName, (String) key, values);
        }
    }

    private void addKeysValue(String collectionName, List keys, String value) {
        AsyncFileWriter bufferedWriter = this.files.get(collectionName);
        for (Object key : keys) {
            addToFlatFile(bufferedWriter, (String) key, value);
        }
    }

    private void addKeyValues(String collectionName, String key, List values) {
        AsyncFileWriter bufferedWriter = this.files.get(collectionName);
        addToFlatFile(bufferedWriter, key, values.toString().replace(',', ' '));
    }

    private void addToFlatFile(AsyncFileWriter asyncFileWriter, String key, String value) {
        String line = getKeyValue(key, value);
        asyncFileWriter.write(line);
    }

    private String getKeyValue(String key, String value) {
        return String.format("%s,\"%s\"%n", key, value);
    }

    public List<String> getSeparatedFilesPath() {
        return this.separatedFilesPath;
    }
}
