/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.BackingStore;

import org.apache.atlas.tools.models.Session;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVRecord;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.mapdb.DB;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.LinkedList;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

public class SinkLoader {
    private static final Logger LOG = LoggerFactory.getLogger(SinkLoader.class);
    private static final String SORTED_FILE_NAME_FORMAT = "%s-%s-sorted.csv";
    private static final int CSV_RECORD_KEY_INDEX = 0;
    private static final int CSV_RECORD_VALUE_INDEX = 1;
    private static final String STR_CLOSE_OPEN_BRACKET = "][";

    private final SinkEntryWriter sinkEntryWriter;
    private final String fileName;
    private final Session session;
    private ConcurrentMap<String, String> result;
    private Thread thread;

    private SinkLoader(Session session, boolean keysOnly) {
        this.session = session;
        this.fileName = session.getTemporaryFileToWriteWithoutExtension();
        this.sinkEntryWriter = keysOnly
                ? new SinkKeyWriter()
                : new SinkKeyValueWriter();
    }

    public void loadAsync(String collection, String collectionInDB) {
        thread = new Thread(() -> {
            try {
                result = loadInternal(fileName, collection, collectionInDB);
            } catch (IOException exception) {
                LOG.error("Error loading: {}.", exception);
            }
        });

        thread.start();
    }

    public ConcurrentMap<String, String> getResult() {
        try {
            thread.join();
        } catch (InterruptedException e) {
            return new ConcurrentHashMap<>();
        }

        return result;
    }

    private ConcurrentMap<String, String> loadInternal(String fileName, String collectionName, String collectionNameForDB) throws IOException {
        String sortedCsvInput = String.format(SORTED_FILE_NAME_FORMAT, fileName, collectionName);
        BufferedReader bufferedReader = null;
        DB.TreeMapSink sink = null;
        try {
            LOG.info("{} - {}: Load: Starting...", collectionName, collectionNameForDB);

            bufferedReader = new BufferedReader(new FileReader(sortedCsvInput));

             sink = populateSink(collectionNameForDB, sortedCsvInput, bufferedReader);

            return (ConcurrentMap<String, String>) sink.create();
        } catch (Exception ex) {
            LOG.error("Error loading collection: {}", collectionName, ex);
            return new ConcurrentHashMap<>();
        } finally {
            if (bufferedReader != null) {
                bufferedReader.close();
            }
            LOG.info("{} - {}: Load: Done!", collectionName, collectionNameForDB);
        }
    }

    private DB.TreeMapSink<String, String> populateSink(String collectionNameForDB, String sortedCsvInput, BufferedReader bufferedReader) throws IOException {
        DB.TreeMapSink<String, String> sink = session.getBackingStoreProvider().createMapSink(collectionNameForDB);
        CSVParser csvParser = new CSVParser(bufferedReader, CSVFormat.DEFAULT);

        try {
            int lineNumber = 0;
            Queue<Object[]> peekQueue = new LinkedList<>();
            for (CSVRecord csvRecord : csvParser) {
                try {
                    if (csvRecord == null) {
                        continue;
                    }

                    String key = csvRecord.get(CSV_RECORD_KEY_INDEX);
                    StringBuilder sbValue = new StringBuilder(csvRecord.get(CSV_RECORD_VALUE_INDEX));

                    peekQueue.add(new Object[]{key, sbValue});
                    if (peekQueue.size() <= 1) {
                        continue;
                    }

                    Object[] current = peekQueue.poll();
                    Object[] next = peekQueue.peek();
                    if (next[CSV_RECORD_KEY_INDEX].equals(current[CSV_RECORD_KEY_INDEX])) {
                        current[CSV_RECORD_VALUE_INDEX] = mergeValue(current[CSV_RECORD_VALUE_INDEX], peekQueue.poll()[CSV_RECORD_VALUE_INDEX]);
                        peekQueue.add(current);
                        continue;
                    }

                    lineNumber++;
                    addToSink(sink, current);
                } catch (org.mapdb.DBException.NotSorted e) {
                    LOG.error("BulkLoader: Keys not sorted: {} - {}: line number: {}", collectionNameForDB, sortedCsvInput, lineNumber, e);
                }
            }

            Object[] temp;
            while ((temp = peekQueue.poll()) != null) {
                addToSink(sink, temp);
            }
        }
        finally {
            if (csvParser != null) {
                csvParser.close();
            }
        }

        return sink;
    }

    private void addToSink(DB.TreeMapSink<String, String> sink, Object[] current) {
        sinkEntryWriter.write(sink, current);
    }

    private Object mergeValue(Object current, Object next) {
        return ((StringBuilder) current).append(next.toString());
    }

    public static void deleteTemporarySortedCsvFiles(Session session, String[] collectionNames) {
        String baseFilePath = session.getTemporaryFileToWriteWithoutExtension();
        for (String collectionName : collectionNames) {
            String sortedCsvInput = String.format(SinkLoader.SORTED_FILE_NAME_FORMAT, baseFilePath, collectionName);
            FileUtils.deleteQuietly(new File(sortedCsvInput));
        }
    }

    private interface SinkEntryWriter {
        void write(DB.TreeMapSink<String, String> sink, Object[] current);
    }
    private class SinkKeyWriter implements SinkEntryWriter {
        @Override
        public void write(DB.TreeMapSink<String, String> sink, Object[] current) {
            String key = (String) current[CSV_RECORD_KEY_INDEX];
            if (StringUtils.isEmpty(key)) {
                return;
            }

            sink.put(key, StringUtils.EMPTY);
        }
    }

    private class SinkKeyValueWriter implements SinkEntryWriter {
        @Override
        public void write(DB.TreeMapSink<String, String> sink, Object[] current) {
            String key = (String) current[CSV_RECORD_KEY_INDEX];
            if (StringUtils.isEmpty(key)) {
                return;
            }

            String value = ((StringBuilder) current[CSV_RECORD_VALUE_INDEX]).toString();
            if (StringUtils.isEmpty(value)) {
                return;
            }

            sink.put(key, StringUtils.replace(value, STR_CLOSE_OPEN_BRACKET, " "));
        }
    }

    public static SinkLoader create(Session session, String collection) {
        SinkLoader sl = new SinkLoader(session, false);
        sl.loadAsync(collection, collection);
        return sl;
    }

    public static SinkLoader create(Session session, String collection, String collectionInDB) {
        SinkLoader sl = new SinkLoader(session, true);
        sl.loadAsync(collection, collectionInDB);
        return sl;
    }
}
