/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.BackingStore;

import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentMap;

public class UniqueAttributeManager {
    private static final Logger LOG = LoggerFactory.getLogger(UniqueAttributeManager.class);
    private static final boolean PERFORM_QUALIFIED_NAME_VALIDATION = PropertiesUtils.getPerformQualifiedNameValidation();
    private static Set<String> TYPES_TO_CONSIDER_FOR_QUALIFIED_NAME_VALIDATION_CHECK = PropertiesUtils.getTypesToConsiderForPerformQualifiedNameValidation();
    private static Set<String> TYPES_TO_CONSIDER_FOR_STORING = PropertiesUtils.getTypesToConsiderForReferenceStore();

    protected ConcurrentMap<String, String> typeQualifiedNameGuid;

    protected BackingStoreProvider backingStoreProvider;
    private Session session;

    public UniqueAttributeManager(Session session) {
        this.session = session;
        this.backingStoreProvider = session.getBackingStoreProvider();

        this.typeQualifiedNameGuid = this.backingStoreProvider.createMap("typeQualifiedNameGuid");
    }

    public void register(String guid, String typeName, String qualifiedName, String qualifiedNameCreateTime) {
        addToGuidQualifiedNameCreateTime(guid, typeName, qualifiedNameCreateTime);
        addToQualifiedNameId(guid, typeName, qualifiedName);
    }

    public boolean validate(String guid, String typeName, String qualifiedName) {
        if (PERFORM_QUALIFIED_NAME_VALIDATION == false || !TYPES_TO_CONSIDER_FOR_QUALIFIED_NAME_VALIDATION_CHECK.contains(typeName)) {
            return true;
        }

        if (this.typeQualifiedNameGuid.containsKey(qualifiedName)) {
            LOG.warn("Entity: {}: Duplicate qualifiedName: {} - {}", guid, this.typeQualifiedNameGuid.get(qualifiedName), qualifiedName);
            return false;
        }

        return true;
    }

    public String getQualifiedNameTimestamp(String guid) {
        if (StringUtils.isEmpty(guid) || !this.session.getRelationsRepository().getGuidQualifiedNameCreatedTime().containsKey(guid)) {
            LOG.debug("UniqueAttributeManager: Entry not found for: {}", guid);
            return StringUtils.EMPTY;
        }

        return this.session.getRelationsRepository().getGuidQualifiedNameCreatedTime().get(guid);
    }

    protected void addToQualifiedNameId(String guid, String typeName, String qualifiedName) {
        if (!TYPES_TO_CONSIDER_FOR_QUALIFIED_NAME_VALIDATION_CHECK.contains(typeName)) {
            return;
        }

        this.typeQualifiedNameGuid.put(qualifiedName, guid);
    }

    protected void addToGuidQualifiedNameCreateTime(String guid, String typeName, String qualifiedNameCreateTime) {
        if (!TYPES_TO_CONSIDER_FOR_STORING.contains(typeName)) {
            return;
        }

        this.session.getRelationsRepository().getGuidQualifiedNameCreatedTime().put(guid, qualifiedNameCreateTime);
    }

    public void clearTypeQualifiedName() {
        this.typeQualifiedNameGuid.clear();
    }
}
