/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools;

import org.apache.atlas.ApplicationProperties;
import org.apache.atlas.AtlasException;
import org.apache.atlas.tools.converter.ConversionException;
import org.apache.atlas.tools.converter.Converter;
import org.apache.atlas.tools.converter.EntityRelationsWriter;
import org.apache.atlas.tools.models.ZipEntryMembersInput;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.atlas.tools.utils.ResourceUtils;
import org.apache.atlas.tools.writers.ZipWriterSplitter;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URISyntaxException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.UUID;
import java.util.function.Consumer;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

public class NavigatorBridge {
    private static final Logger LOG = LoggerFactory.getLogger(NavigatorBridge.class);

    private static final String LAUNCHER_SHELL_SCRIPT_NAME = "nav2atlas.sh";
    private static final String DISPLAY_PREFIX = "NavigatorBridge";
    private static final Pattern CLUSTER_NAME_VALIDATION_REGEX = Pattern.compile("^[a-zA-Z0-9_]*$");
    private static final String APPLICATION_PROPERTY_CLUSTER_NAME = "atlas.nav2atlas.cluster.name";

    private static final String CMD_LINE_SHORT_OPTION_CLUSTER_NAME = "cn";
    private static final String CMD_LINE_SHORT_OPTION_INPUT_FILE = "f";
    private static final String CMD_LINE_SHORT_OPTION_OUTPUT_FILE = "o";

    public static void main(String[] args) {
        CommandLine cmd = getCommandLine(args);
        if (cmd == null) {
            System.exit(-1);
            displayCrlf("Done!");
            LOG.info("NavigatorBridge: Exiting!");
        }

        String fileToConvert = getStringValueWithDefault(cmd, CMD_LINE_SHORT_OPTION_INPUT_FILE);
        String fileToWrite = getStringValueWithDefault(cmd, CMD_LINE_SHORT_OPTION_OUTPUT_FILE);
        String guid1 = getStringValueWithDefault(cmd, "guid1");
        String guid2 = getStringValueWithDefault(cmd, "guid2");
        String clusterName = getClusterNameFromConfig(cmd);
        String currentWorkingDirectory = getCurrentWorkingDirectory();

        try {
            String clusterNameForConverter = clusterName;
            if (!validateClusterName(clusterNameForConverter)) {
                String errorMessage = String.format("'%s' should be valid! Valid name should have only alphanumeric characters and '_'.", clusterNameForConverter);
                LOG.error(errorMessage);
                System.err.println(errorMessage);
                System.exit(-1);
                return;
            }

            if (StringUtils.isEmpty(fileToConvert)) {
                displayCrlf("File: %s was not found!", fileToConvert);
                return;
            }

            if (fileToConvert.equalsIgnoreCase(fileToWrite)) {
                displayCrlf("File: Input and output files are same: %s!", fileToConvert);
                return;
            }

            String fileToWriteWithAbsolutePath = new File(fileToWrite).getAbsolutePath();
            if (StringUtils.isNotEmpty(fileToWrite) && FileUtils.getFile(fileToWriteWithAbsolutePath).exists()) {
                FileUtils.deleteQuietly(new File(fileToWrite));
            }

            Session session = new Session(currentWorkingDirectory, Paths.get(fileToWrite), clusterName, guid1, guid2, null, 0);
            performConversion(session, fileToConvert, fileToWriteWithAbsolutePath);
        } catch (IOException e) {
            LOG.error( String.format("Error locating: %s", fileToConvert), e);
            if (FileUtils.getFile(fileToWrite).exists()) {
                FileUtils.getFile(fileToWrite).delete();
            }

            e.printStackTrace();
        } catch (ConversionException e) {
            LOG.error( "Conversion Exception!", e);
        } catch (Exception e) {
            LOG.error( "General error!", e);
        }

        displayCrlf("Done!");
        LOG.info("NavigatorBridge: Exiting!");
        System.exit(0);
    }

    private static void performConversion(Session session,
                                          String fileToConvert,
                                          String fileToWriteWithAbsolutePath) throws IOException, ConversionException {

        String temporaryFile = getTemporaryFile(fileToWriteWithAbsolutePath);
        session.setTemporaryFileToWrite(temporaryFile);

        ZipWriterSplitter zipWriterSplitter = new ZipWriterSplitter(temporaryFile);
        convertToFile(zipWriterSplitter, fileToConvert, temporaryFile, session);

        displayCrlf("Conversion: Moving temporary file (s)...");
        zipWriterSplitter.moveFiles(fileToWriteWithAbsolutePath);
    }

    private static String getTemporaryFile(String originalFile) {
        Path path = Paths.get(originalFile);
        String tempFileToWrite = Paths.get(PropertiesUtils.getTempDirectory(), path.getFileName().toString()).toAbsolutePath().toString();
        LOG.info("Conversion: Temporary file: {}...", tempFileToWrite);
        return tempFileToWrite;
    }

    private static String getCurrentWorkingDirectory() {
        String cwd = null;
        try {
            String jarFilePath = NavigatorBridge.class.getProtectionDomain().getCodeSource().getLocation().toURI().getPath();
            File jarFile = new File(jarFilePath);
            cwd = jarFile.getParent();
        } catch (URISyntaxException e) {
            LOG.error( "Error fetching current working directory.", e);
        }
        LOG.info( "Current working directory: " + cwd);
        return cwd;
    }

    private static boolean validateClusterName(String cn) {
        System.out.println(String.format("Validating cluster name: '%s' ...", cn));
        return StringUtils.isNotEmpty(cn) && CLUSTER_NAME_VALIDATION_REGEX.matcher(cn).matches();
    }

    private static String getStringValueWithDefault(CommandLine cmd, String optionName) {
        String val = cmd.getOptionValue(optionName);
        if (StringUtils.isEmpty(val)) {
            return UUID.randomUUID().toString();
        }

        return val.trim();
    }

    private static void convertToFile(EntityRelationsWriter entityRelationsWriter,
                                      String fileToConvert, String fileToWrite,
                                      Session session) throws IOException {
        displayCrlf("Conversion: Starting: %s...", fileToConvert);
        LOG.info("Conversion: Starting: {}...", fileToConvert);

        Converter converter = new Converter(session);

        displayCrlf("Conversion: Reading: %s...", fileToConvert);
        converter.setInputSummary(ResourceUtils.getZipFileComment(fileToConvert));
        converter.convertNamespaces(
                ResourceUtils.getStringFromStream(getInputStreamFromZip(fileToConvert, ZipEntryMembersInput.ZIP_ENTRY_NAMESPACES.toString())),
                ResourceUtils.getStringFromStream(getInputStreamFromZip(fileToConvert, ZipEntryMembersInput.ZIP_ENTRY_PROPERTIES_MAPPING.toString())));

        InputStream inputStreamFromZip = getInputStreamFromZip(fileToConvert, ZipEntryMembersInput.ZIP_ENTRY_RELATIONS.toString());
        converter.processRelations(getInputStreamFromZip(fileToConvert, ZipEntryMembersInput.ZIP_ENTRY_RELATIONS.toString()));

        displayCrlf("Conversion: Generating: %s...", fileToWrite);

        LOG.info("Copying: identity-guid map...");
        ((ZipInputStream) inputStreamFromZip).getNextEntry(); // expecting ZipEntryMembersInput.ZIP_ENTRY_IDENTITY_GUID_MAP
        converter.loadIdentityGuid(inputStreamFromZip);

        LOG.info("Copying: guid-type map...");
        ((ZipInputStream) inputStreamFromZip).getNextEntry();
        converter.loadGuidTypeMap(inputStreamFromZip);

        LOG.info("Conversion: Generating: {}...", fileToWrite);
        ((ZipInputStream) inputStreamFromZip).getNextEntry();
        converter.convertEntities(inputStreamFromZip, entityRelationsWriter);

        entityRelationsWriter.flushAndClose();
        converter.clear();
    }

    private static String getClusterNameFromConfig(CommandLine cmd) {
        try {
            String[] valuesFromCommandline = cmd.getOptionValues(CMD_LINE_SHORT_OPTION_CLUSTER_NAME);
            String clusterNameFromCommandLine = StringUtils.join(valuesFromCommandline, " ");
            if (StringUtils.isNotEmpty(clusterNameFromCommandLine)) {
                return clusterNameFromCommandLine;
            }

            return ApplicationProperties.get().getString(APPLICATION_PROPERTY_CLUSTER_NAME);
        } catch (AtlasException e) {
            LOG.error( String.format("Error fetching clusterName from atlas-application.properties: {}", APPLICATION_PROPERTY_CLUSTER_NAME), e);
            return null;
        }
    }

    private static InputStream getInputStreamFromZip(String fileToConvert, String zipEntryName) throws IOException {
        InputStream fileOutputStream = new FileInputStream(new File(fileToConvert));
        ZipInputStream zipInputStream = new ZipInputStream(fileOutputStream);

        ZipEntry zs = null;
        while((zs = zipInputStream.getNextEntry()) != null) {
            if (zs.getName().equals(zipEntryName)) {
                return zipInputStream;
            }
        }

        zipInputStream.close();
        fileOutputStream.close();
        return null;
    }

    private static OutputStream getOutputStream(String fileToWrite) throws IOException {
        return FileUtils.openOutputStream(new File(fileToWrite));
    }

    private static void displayCrlf(String... formatMessage) {
        displayFn(System.out::println, formatMessage);
    }

    private static void displayFn(Consumer<String> fn, String... formatMessage) {
        String prefix = DISPLAY_PREFIX + ": ";
        if (formatMessage.length == 1) {
            fn.accept(prefix + formatMessage[0]);
        } else {
            fn.accept(prefix + String.format(formatMessage[0], formatMessage[1]));
        }
    }

    private static CommandLine getCommandLine(String[] args) {
        Options options = new Options();

        Option inputFile = new Option(CMD_LINE_SHORT_OPTION_INPUT_FILE, "file", true, "Input: Navigator file to be converted.");
        inputFile.setRequired(true);
        options.addOption(inputFile);

        Option outputFile = new Option(CMD_LINE_SHORT_OPTION_OUTPUT_FILE, "output", true, "Output: Atlas ZIP file after conversion.");
        outputFile.setRequired(true);
        options.addOption(outputFile);

        Option clusterName = new Option(CMD_LINE_SHORT_OPTION_CLUSTER_NAME, "clusterName", true, "Cluster Name for Atlas Entities.");
        clusterName.setRequired(true);
        clusterName.setArgs(Option.UNLIMITED_VALUES);
        options.addOption(clusterName);

        try {
            return new DefaultParser().parse(options, args);
        } catch (ParseException e) {
            printUsage(options);
            return null;
        }
    }

    private static void printUsage(Options options) {
        System.out.println();
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(LAUNCHER_SHELL_SCRIPT_NAME, options);
        System.out.println();
    }
}
