/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.tools.models.MultiNodeContainer;
import org.apache.atlas.type.AtlasType;
import org.apache.atlas.type.AtlasTypeRegistry;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.LinkedHashMap;
import java.util.Map;

public class AtlasTypeFactory {
    private static final Logger LOG = LoggerFactory.getLogger(AtlasTypeFactory.class);
    private static final String ATLAS_TYPE_FS_PERMISSIONS = "fs_permissions";
    public static final String ATTR_RELATIONSHIP_ATTRIBUTES = "relationshipAttributes";

    private AtlasTypeRegistry atlasTypeRegistry;
    private static final ThreadLocal<Map<String, String>> defaultEntityTypeJsonMap = ThreadLocal.withInitial(() -> new LinkedHashMap<>());

    public AtlasTypeFactory(AtlasTypeRegistry atlasTypeRegistry) {
        this.atlasTypeRegistry = atlasTypeRegistry;
    }

    public MultiNodeContainer create(String atlasEntityTypeName) {
        if (!defaultEntityTypeJsonMap.get().containsKey(atlasEntityTypeName)) {
            AtlasType atlasType = getAtlasTypeWithValidation(atlasEntityTypeName);
            if (atlasType == null) {
                return null;
            }

            defaultEntityTypeJsonMap.get().put(atlasEntityTypeName, AtlasType.toJson(atlasType.createDefaultValue()));
        }

        MultiNodeContainer mnc = AtlasType.fromJson(defaultEntityTypeJsonMap.get().get(atlasEntityTypeName), MultiNodeContainer.class);
        mnc.setValue(ATTR_RELATIONSHIP_ATTRIBUTES, new LinkedHashMap<String, Object>());
        return mnc;
    }

    private AtlasType getAtlasTypeWithValidation(String atlasEntityTypeName) {
        AtlasType atlasType = null;
        try {
            atlasType = getAtlasType(atlasEntityTypeName);
        } catch (AtlasBaseException e) {
            LOG.error(String.format("Error creating: %s. This type will be ignored.", atlasEntityTypeName));
        }

        return atlasType;
    }

    public MultiNodeContainer createFsPermissions() {
        return create(ATLAS_TYPE_FS_PERMISSIONS);
    }

    private AtlasType getAtlasType(String atlasEntityTypeName) throws AtlasBaseException {
        return atlasTypeRegistry.getType(atlasEntityTypeName);
    }
}
