/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import com.google.common.annotations.VisibleForTesting;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.instance.AtlasRelatedObjectId;
import org.apache.atlas.model.instance.AtlasStruct;
import org.apache.atlas.model.typedef.AtlasEnumDef;
import org.apache.atlas.tools.models.Conversion.Constants;
import org.apache.atlas.tools.models.MultiNodeContainer;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.JsonUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

public class ConversionMethods {
    private static final Logger LOG = LoggerFactory.getLogger(ConversionMethods.class);
    private static final String QUALIFIED_NAME_AT_SIGN = "@";
    private static Map<String, String> propertyTypeAtlasAttributeTypeMap = new HashMap<String, String>() {{
        put("TEXT", "string");
        put("INTEGER", "int");
        put("BOOLEAN", "boolean");
        put("DATE", "date");
        put("DOUBLE", "double");
    }};

    public static String prefix(String params, String s) {
        return String.format("%s%s", params, s);
    }

    public static AtlasObjectId toAtlasObjectIdByGuid(String id) {
        return new AtlasObjectId(id);
    }

    public static AtlasRelatedObjectId toAtlasRelatedObjectIdByGuid(String id, String typeName, String relationshipTypeName) {
        AtlasObjectId atlasObjectId = toAtlasObjectIdByGuid(id);
        AtlasRelatedObjectId relatedObjectId = new AtlasRelatedObjectId(atlasObjectId);
        relatedObjectId.setTypeName(typeName);
        relatedObjectId.setRelationshipType(relationshipTypeName);

        return relatedObjectId;
    }

    public static AtlasObjectId toAtlasObjectId(String typeName, String id) {
        return new AtlasObjectId(id, typeName);
    }

    public static Object toQualifiedName(String[] splits, String suffix) {
        int length = splits.length;
        return String.format("%s@%s%s",
                StringUtils.join(splits, ".", 0, length - 1),
                splits[length - 1], Objects.toString(suffix, ""));
    }

    public static Object toQualifiedName2(String[] splits, String suffix) {
        for (int i = 0; i < splits.length; i++) {
            splits[i] = splits[i].replace('/', ' ').trim();
        }

        return toQualifiedName(splits, suffix);
    }

    public static Object toDate(String rawValue) {
        return new Date(getTimestamp(rawValue));
    }

    public static long getTimestamp(String str) {
        try {
            if (StringUtils.isEmpty(str)) {
                return 0;
            }

            return ZonedDateTime.parse(str, DateTimeFormatter.ISO_ZONED_DATE_TIME).toInstant().toEpochMilli();
        } catch (Exception e) {
            LOG.error(String.format("Unable to parse date-time string: %s", str), e);
            return -1;
        }
    }

    public static Object strComp(String params, String rawValue) {
        return params.equals(rawValue);
    }

    public static AtlasEntity.Status toStatus(Object o) {
        if (o == null) {
            return AtlasEntity.Status.ACTIVE;
        }

        if (o instanceof Boolean) {
            Boolean b = (Boolean) o;
            return !b ? AtlasEntity.Status.ACTIVE : AtlasEntity.Status.DELETED;
        }

        return AtlasEntity.Status.valueOf(o.toString());
    }

    public static String removeStartsWith(String params, String rawValue) {
        return StringUtils.removeStart(rawValue, params);
    }

    public static String fixParentPath(String rawValue) {
        if (StringUtils.isEmpty(rawValue)) {
            return rawValue;
        }

        return StringUtils.removeStart(rawValue, "/").replace('/', '.');
    }

    static final int USER_INDEX = 3;
    static final int GROUP_INDEX = 6;
    static final int OTHER_INDEX = 9;

    public static void toFsPermissions(Object rawValue, MultiNodeContainer fsPermissions) {
        if (!(rawValue instanceof String) || StringUtils.isEmpty((String) rawValue)) {
            return;
        }

        String permissionsStr = (String) rawValue;
        int length = permissionsStr.length();

        if (length >= USER_INDEX) {
            fsPermissions.setValue("attributes.user", getPermissionOrdinal(StringUtils.substring(permissionsStr, 0, USER_INDEX)));
        }

        if (length >= GROUP_INDEX) {
            fsPermissions.setValue("attributes.group", getPermissionOrdinal(StringUtils.substring(permissionsStr, 3, GROUP_INDEX)));
        }

        if (length >= OTHER_INDEX) {
            fsPermissions.setValue("attributes.others", getPermissionOrdinal(StringUtils.substring(permissionsStr, 6, OTHER_INDEX)));
        }
    }

    @VisibleForTesting
    static String getPermissionOrdinal(String s) {
        switch (s) {
            case "--x":
                return "EXECUTE";
            case "r--":
                return "READ";
            case "rw-":
                return "READ_WRITE";
            case "r-x":
                return "READ_EXECUTE";
            case "-wx":
                return "WRITE_EXECUTE";
            case "rwx":
                return "ALL";
            case "---":
            default:
                return "NONE";
        }
    }


    public static List<AtlasStruct> toHiveOrderList(String typeName, List<String> items) {
        List<AtlasStruct> sortCols = new ArrayList<>();
        for (int i = 0, itemsSize = items.size(); i < itemsSize; i++) {
            sortCols.add(createHiveOrderType(typeName, items.get(i), i));
        }

        return sortCols;
    }

    private static AtlasStruct createHiveOrderType(String typeName, String item, int order) {
        final String ATTR_NAME_COL = "col";
        final String ATTR_NAME_ORDER = "order";

        AtlasStruct struct = new AtlasStruct(typeName);
        struct.setAttribute(ATTR_NAME_COL, item);
        struct.setAttribute(ATTR_NAME_ORDER, order++);

        return struct;
    }

    public static Object toColumnQualifiedNameFromTable(String currentQualifiedName, Object rawValue) {
        if (rawValue == null) {
            return rawValue;
        }

        int indexOfAt = StringUtils.lastIndexOf(currentQualifiedName, QUALIFIED_NAME_AT_SIGN);
        return insertString(currentQualifiedName, String.format(".%s", rawValue.toString()), indexOfAt - 1);
    }

    private static String insertString(String originalString, String stringToBeInserted, int index) {
        String newString = new String();
        for (int i = 0; i < originalString.length(); i++) {
            newString += originalString.charAt(i);

            if (i == index) {
                newString += stringToBeInserted;
            }
        }

        return newString;
    }

    public static Object getInputs(Session session, String identity) {
        return session.getRelationsRepository().getInputs(identity);
    }

    public static Object getOutputs(Session session, String identity) {
        return session.getRelationsRepository().getOutputs(identity);
    }

    public static Object getInferredInputs(Session session, String identity) {
        return session.getRelationsRepository().getInferredInputs(identity);
    }

    public static Object getInferredOutputs(Session session, String identity) {
        return session.getRelationsRepository().getInferredOutputs(identity);
    }

    public static Object getTemplateId(Session session, String identity) {
        return session.getRelationsRepository().getTemplateId(identity);
    }

    public static Object getAllChildrenInputs(Session session, String identity) {
        return session.getRelationsRepository().getAllTemplateInstanceInputs(identity);
    }

    public static Object getAllChildrenOutputs(Session session, String identity) {
        return session.getRelationsRepository().getAllTemplateInstanceOutputs(identity);
    }

    public static Object getParentsOfInputs(Session session, String identity) {
        return session.getRelationsRepository().getParentsOfInputs(identity);
    }

    public static Object getParentsOfOutputs(Session session, String identity) {
        return session.getRelationsRepository().getParentsOfOutputs(identity);
    }

    public static Object getParent(Session session, String identity) {
        return session.getRelationsRepository().getParent(identity);
    }

    public static String urlPrefix(String s) {
        return StringUtils.substringBefore(s, "://");
    }

    public static Object toAtlasAttributeType(String propertyType, Map<String, Object> tempVars, String tempVarsKey) {
        String atlasType = getAtlasAttributeType(propertyType, tempVars, tempVarsKey);
        if (StringUtils.isNotEmpty(atlasType) && tempVars.containsKey("multiValued") && (boolean) tempVars.get("multiValued") == true) {
            return String.format("array<%s>", atlasType);
        }

        return atlasType;
    }

    private static String getAtlasAttributeType(String typeName, Map<String, Object> tempVars, String tempVarsKey) {
        if (propertyTypeAtlasAttributeTypeMap.containsKey(typeName)) {
            return propertyTypeAtlasAttributeTypeMap.get(typeName);
        }

        if (typeName.equals("ENUM") && tempVars.get(tempVarsKey) != null) {
            String atlasType = (String) tempVars.get(tempVarsKey);
            return atlasType;
        }

        LOG.warn("Could not find appropriate mapping for property type: {}", typeName);
        return typeName;
    }

    public static AtlasEnumDef createNewEnumDef(String name, Object rawValue) {
        if (rawValue == null && !(rawValue instanceof List)) {
            return null;
        }

        Integer ordinal = 0;
        List list = (List) rawValue;
        AtlasEnumDef enumDef = new AtlasEnumDef(String.format("%s_enum", name));
        for (Object o : list) {
            enumDef.addElement(new AtlasEnumDef.AtlasEnumElementDef(o.toString(), StringUtils.EMPTY, ordinal++));
        }

        return enumDef;
    }

    public static Object padLeft(Object rawValue) {
        return JsonUtils.getPaddedString((String) rawValue);
    }

    public static String getParentQualifiedName(Session session, String identity) {
        String parentId = (String) ConversionMethods.getParent(session, identity);
        if (StringUtils.isEmpty(parentId)) {
            LOG.debug("Reference not found: Parent not found: Identity: {}", identity);
            return getEntityGuid(session, identity);
        }

        return getQualifiedNameFromSession(session, identity, parentId);
    }

    public static Object getTemplateQualifiedName(Session session, String identity) {
        String parentId = (String) ConversionMethods.getTemplateId(session, identity);
        if (StringUtils.isEmpty(parentId)) {
            LOG.debug("Reference not found: Template not found: Identity: {}", identity);
            return getEntityGuid(session, identity);
        }

        return getQualifiedNameFromSession(session, identity, parentId);
    }

    private static String getQualifiedNameFromSession(Session session, String identity, String parentId) {
        String qnFromSession = session.getUniqueNameValidator().getQualifiedNameTimestamp(session.getGuidProvider().get(parentId));
        if (StringUtils.isEmpty(qnFromSession)) {
            LOG.debug("Parent qualifiedName not found!: Identity: {}", identity);
            return getEntityGuid(session, identity);
        }

        return removeTimestamp(qnFromSession);
    }

    private static String removeTimestamp(String qnFromSession) {
        return StringUtils.substringBeforeLast(qnFromSession, ":");
    }

    public static String getEntityGuid(Session session, String id) {
        if (StringUtils.isEmpty(id)) {
            return null;
        }

        return session.getGuidProvider().get(StringUtils.leftPad(id, Constants.PADDING_SIZE, "0"));
    }

    public static String getEntityReferredGuid(Session session, String id) {
        return session.getGuidProvider().getReferred((String) ConversionMethods.padLeft(id));
    }

    public static List<String> getChildren(Session session, String parentId) {
        return session.getRelationsRepository().getChildren(parentId);
    }

    public static String getParentTimestamp(Session session, String identity) {
        String parentId = (String) ConversionMethods.getParent(session, identity);
        if (StringUtils.isEmpty(parentId)) {
            LOG.debug("Reference not found: Parent not found: Identity: {}", identity);
            return null;
        }

        String qnFromSession = session.getUniqueNameValidator().getQualifiedNameTimestamp(session.getGuidProvider().get(parentId));
        if (StringUtils.isEmpty(qnFromSession)) {
            LOG.debug("Parent qualifiedName not found!: Identity: {}", identity);
            return null;
        }

        return removeQualifiedName(qnFromSession);
    }

    private static String removeQualifiedName(String qnFromSession) {
        return StringUtils.substringAfterLast(qnFromSession, ":");
    }
}
