/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import com.google.common.annotations.VisibleForTesting;
import org.apache.atlas.model.impexp.AtlasExportRequest;
import org.apache.atlas.model.impexp.AtlasExportResult;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.tools.BackingStore.BulkLoader;
import org.apache.atlas.tools.BackingStore.RelationCollections;
import org.apache.atlas.tools.models.Conversion.Constants;
import org.apache.atlas.tools.models.Entity;
import org.apache.atlas.tools.models.MultiNodeContainer;
import org.apache.atlas.tools.models.NodeContainer;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.models.ZipEntryMembers;
import org.apache.atlas.type.AtlasType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

public class Converter {
    private static final Logger LOG = LoggerFactory.getLogger(Converter.class);

    private final SpecsLoader specsLoader;
    private final String IMPORT_OPTIONS_KEY_ZIP_SOURCE_FORMAT = "format";
    private final String IMPORT_OPTIONS_VALUE_ZIP_SOURCE_FORMAT_VALUE = "zipSourceDirect";

    private EntityConverter entityConverter;
    private Session session;
    private RelationsRepository relationsRepository;
    private final AtlasExportResult atlasExportResult;
    private BulkLoader bulkLoader;

    public Converter(Session session) {
        this.session = session;
        relationsRepository = new RelationsRepository(session);
        this.specsLoader = new SpecsLoader();
        this.atlasExportResult = new AtlasExportResult();
        this.session.setupUniqueNameValidator();
    }

    public void setInputSummary(String json) {
        try {
            LOG.info("Summary: Set: Starting...");

            MultiNodeContainer mnc = AtlasType.fromJson(json, MultiNodeContainer.class);
            AtlasExportRequest request = new AtlasExportRequest();
            mnc.set(IMPORT_OPTIONS_KEY_ZIP_SOURCE_FORMAT, IMPORT_OPTIONS_VALUE_ZIP_SOURCE_FORMAT_VALUE);
            request.setOptions(mnc.get());
            this.atlasExportResult.setRequest(request);
            LOG.info("{}", json);
        } finally {
            LOG.info("Summary: Set: Done!");
        }
    }

    public void processRelations(InputStream inputStream) {
        try {
            LOG.info("Processing relations: Starting...");
            if (this.session.isBulkLoading()) {
                bulkLoader = new BulkLoader(session);
                bulkLoader.separate(inputStream);
            } else {
                RelationsReadConvertWrite relationsReadConvertWrite = new RelationsReadConvertWrite(session);
                relationsReadConvertWrite.convert(inputStream);
            }
        } catch (Exception exception) {
            LOG.error("Processing relations: Errors. Generated file may be unusable.", exception);
        } finally {
            LOG.info("Processing relations: Done!");
        }
    }

    public void convertNamespaces(String namespacesJson, String propertyMappingJson) {
        try {
            LOG.info("Converting: Namespaces: Starting...");
            NamespaceConverter namespaceConverter = specsLoader.createNamespaceConverter(session);
            MultiNodeContainer[] namespaces = AtlasType.fromJson(namespacesJson, MultiNodeContainer[].class);

            MultiNodeContainer propertyMapping = AtlasType.fromJson(propertyMappingJson, MultiNodeContainer.class);
            namespaceConverter.convert(namespaces, propertyMapping);
        } catch (Exception e) {
            LOG.error("Converting: Namespaces: Error!", e);
        } finally {
            LOG.info("Converting: Namespaces: Done!");
        }
    }

    public void convertEntities(InputStream inputStream, EntityRelationsWriter writer) {
        try {
            LOG.info("Entities: Starting...");
            if (session.isBulkLoading()) {
                this.bulkLoader.load();
            }

            this.entityConverter = specsLoader.createEntityConverter(this.session);
            writeTypesDef(writer);
            convertEntitiesInternal(inputStream, writer);
            writeExportOrder(writer, this.session.getOrder().size());
            writeConversionSummary(writer);
        } catch (Exception exception) {
            LOG.error("Entities: Errors. Generated file may be unusable.", exception);
        } finally {
            LOG.info("Entities: Done!");
            writer.flush();
        }
    }

    public void loadIdentityGuid(InputStream inputStream) {
        loadCollection(inputStream, RelationCollections.IDENTITY_GUID_MAP);
    }

    public void loadGuidTypeMap(InputStream inputStream) {
        loadCollection(inputStream, RelationCollections.GUID_TYPE_MAP);
    }

    public void loadCollection(InputStream inputStream, String collectionName) {
        try {
            LOG.info("Writing: {}: Starting...", collectionName);

            if (session.isBulkLoading()) {
                bulkLoader.writeToTemporaryLocation(collectionName, inputStream);
            }
            else {
                session.getRelationsRepository().loadMap(collectionName, inputStream);
            }
        } catch (IOException exception) {
            LOG.error("Error copying collection: {}", collectionName, exception);
        }
        finally {
            LOG.info("Writing: {}: Done!", collectionName);
        }
    }

    public void writeTypesDef(EntityRelationsWriter writer) throws IOException {
        try {
            LOG.info("Writing to ZIP: Type Definitions: Starting...");

            writer.writeEntry(ZipEntryMembers.ENTRY_ATLAS_TYPESDEF.toString(), AtlasType.toJson(session.getAtlasTypesDef()));
        } catch (IOException e) {
            LOG.error("Writing to ZIP: Errors. Generated file may be unusable.", e);
        } finally {
            LOG.info("Writing to ZIP: Type Definitions: Done!");
            writer.flush();
        }
    }

    private void convertEntitiesInternal(InputStream inputStream, EntityRelationsWriter entityRelationsWriter) throws IOException {
        entityConverter.convert(inputStream, entityRelationsWriter);
        this.atlasExportResult.setMetrics(entityConverter.getJobSummary().getMetrics());
    }

    private void writeConversionSummary(EntityRelationsWriter writer) {
        try {
            LOG.info("Writing to ZIP: Summary: Starting...");
            atlasExportResult.setOperationStatus(AtlasExportResult.OperationStatus.SUCCESS);
            String payload = AtlasType.toJson(atlasExportResult);
            writer.writeEntry(ZipEntryMembers.ENTRY_ATLAS_EXPORT_INFO.toString(), payload);
        } catch (Exception e) {
            LOG.error("Writing to ZIP: Errors. Generated file may be unusable.", e);
        } finally {
            LOG.info("Writing to ZIP: Summary: Done!");
            writer.flush();
        }
    }

    public void writeExportOrder(EntityRelationsWriter writer, int size) throws IOException {
        LOG.info("Writing to ZIP: Order: Starting...");
        try {
            int streamSize = size;
            writer.writeEntry(ZipEntryMembers.ENTRY_ATLAS_EXPORT_ORDER.toString(), "[]");
            this.atlasExportResult.incrementMeticsCounter("streamSize", streamSize);
            writer.setStreamSize(streamSize);
        } finally {
            LOG.info("Writing to ZIP: Order: Done!");
            writer.flush();
        }
    }

    @VisibleForTesting
    public List<AtlasEntity.AtlasEntityWithExtInfo> process(String inputJson) throws ConversionException {
        if (entityConverter == null) {
            entityConverter = specsLoader.createEntityConverter(this.session);
        }

        return convertFromJson(inputJson);
    }

    @VisibleForTesting
    List<AtlasEntity.AtlasEntityWithExtInfo> convertFromJson(String inputJson) throws ConversionException {
        NodeContainer nodeContainer = AtlasType.fromJson(inputJson, NodeContainer.class);
        if (nodeContainer == null) {
            throw new ConversionException("Empty input encountered: " + inputJson);
        }

        List<AtlasEntity.AtlasEntityWithExtInfo> ret = new ArrayList<>();
        for (Entity incomingEntity : nodeContainer.getEntities()) {
            if (!incomingEntity.get().containsKey(Constants.INCOMING_FIELD_NAME_INTERNAL_TYPE)) {
                LOG.warn(String.format("Incoming entity: Incorrect format: %s", AtlasType.toJson(incomingEntity)));
                continue;
            }

            final List<AtlasEntity.AtlasEntityWithExtInfo> convertedEntities = entityConverter.convert(incomingEntity);
            if (convertedEntities == null) {
                LOG.warn(String.format("Error converting: %s", AtlasType.toJson(incomingEntity)));
            }

            this.session.getUniqueNameValidator().register(convertedEntities);
            ret.addAll(convertedEntities);
        }

        return ret;
    }

    public void clear() {
        try {
            LOG.info("Clean-up: Starting...");
            if (entityConverter == null) {
                return;
            }

            entityConverter.clearValidationErrors();
            session.clear();
        }
        catch (Exception exception) {
            LOG.error("Clean-up: Error!", exception);
        }
        finally {
            LOG.info("Clean-up: Done!");
        }
    }
}
