/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.tools.models.Entity;
import org.apache.atlas.tools.models.JobSummary;
import org.apache.atlas.tools.models.Relation;
import org.apache.atlas.tools.pc.ConversionManager;
import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.atlas.type.AtlasType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.Set;

public class EntityReadConvertWrite {
    private static final Logger LOG = LoggerFactory.getLogger(EntityReadConvertWrite.class);
    private static final boolean PERFORM_QUALIFIED_NAME_VALIDATION = PropertiesUtils.getPerformQualifiedNameValidation();
    private static final Set<String> TYPES_TO_CONSIDER_FOR_REFERENCE_STORING = PropertiesUtils.getTypesToConsiderForReferenceStore();

    private RelationsRepository relationsRepository;
    private final EntityRelationsWriter writer;
    private final Reader reader;
    private NodesToAtlasElementConverter converter;
    private ConversionManager manager;
    private final JobSummary jobSummary;

    private class Reader extends EntityRelationsParser {

        private String currentEntityType;

        @Override
        protected void startEntities() {
            try {
                manager = new ConversionManager(relationsRepository, converter);
                writer.startEntities();
            } catch (IOException e) {
                LOG.error( String.format("Error: startEntities: %s", e.getMessage()), e);
            }
        }

        @Override
        protected void processEntity(Entity incoming) {
            Entity entity = incoming;
            if (entity.get().size() == 0) {
                return;
            }

            String incomingEntityType = entity.getType();
            if (this.currentEntityType != null
                    && incomingEntityType != null
                    && !this.currentEntityType.equals(incomingEntityType)) {
                LOG.info("Waiting for: {} to complete...", this.currentEntityType);
                manager.drain();
                extractResults();
                writer.flush();

                converter.session.getUniqueNameValidator().clear();
                converter.session.getGuidProvider().setType(incomingEntityType);
            }

            this.currentEntityType = incomingEntityType;
            manager.checkProduce(entity);
            jobSummary.incrementProduced(currentEntityType);
            extractResults();
        }

        @Override
        protected void endEntities() {
            try {
                LOG.info(String.format("Waiting to end..."));
                manager.drain();
                extractResults();
                shutdownConversionManager();
                writer.endEntities();
            } catch (IOException e) {
                LOG.error( String.format("Error: endEntities: %s", e.getMessage()), e);
            } catch (InterruptedException e) {
                LOG.error( "manager.shutdown: Interrupted!", e);
            }
        }

        private void extractResults() {
            try {
                Object result;
                while (((result = manager.getResults().poll())) != null) {
                    if (!(result instanceof AtlasEntity.AtlasEntityWithExtInfo)) {
                        LOG.error( "Results contain unexpected item.");
                        continue;
                    }

                    AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo = (AtlasEntity.AtlasEntityWithExtInfo) result;
                    String entityGuid = entityWithExtInfo.getEntity().getGuid();
                    String typeName = entityWithExtInfo.getEntity().getTypeName();
                    jobSummary.incrementProcessed(typeName);

                    if (PERFORM_QUALIFIED_NAME_VALIDATION && !converter.session.getUniqueNameValidator().validate(entityWithExtInfo)) {
                        LOG.warn("Duplicate qualifiedName found: {} - {} - {}", entityGuid,
                                this.currentEntityType,
                                AtlasType.toJson(entityWithExtInfo));
                        continue;
                    }

                    converter.session.getUniqueNameValidator().register(entityWithExtInfo);
                    jobSummary.incrementCommitted(typeName);
                    writeEntityInfo(typeName, entityWithExtInfo, entityGuid);
                }
            } catch (Exception e) {
                LOG.error( "Error extracting results.", e);
            }
        }

        @Override
        protected void startRelations() {
        }

        @Override
        protected void endRelations() {
        }

        private void shutdownConversionManager() throws InterruptedException {
            LOG.info(String.format("Conversion Manager: Shutdown: Started..."));
            manager.shutdown();
            converter.clear();
            LOG.info(String.format("Conversion Manager: Shutdown: Done!"));
        }

        private void writeEntityInfo(String entityType, AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo, String entityGuid) throws IOException {
            writer.writeEntity(entityGuid, entityWithExtInfo);

            if (!TYPES_TO_CONSIDER_FOR_REFERENCE_STORING.contains(entityType)) {
                return;
            }

            converter.session.addOrder(entityGuid, entityType);
        }

        @Override
        protected void processRelation(Relation relation) {
        }
    }

    public EntityReadConvertWrite(RelationsRepository relationsRepository, EntityRelationsWriter writer, NodesToAtlasElementConverter converter) {
        this.relationsRepository = relationsRepository;
        this.converter = converter;
        this.writer = writer;
        this.reader = new Reader();
        jobSummary = new JobSummary();
    }

    public void start(InputStream inputStream) throws IOException {
        reader.read(inputStream);
        writer.close();
    }

    public JobSummary getJobSummary() {
        return this.jobSummary;
    }
}
