/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonToken;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.atlas.tools.models.Entity;
import org.apache.atlas.tools.models.NameValuePairEntity;
import org.apache.atlas.tools.models.Relation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.Map;
import java.util.function.Consumer;


public abstract class EntityRelationsParser {
    private static final Logger LOG = LoggerFactory.getLogger(EntityRelationsParser.class);

    private final Entity entity = new Entity();
    private final Relation relation = new Relation();
    private static final String FIELD_ENTITIES = "entities";
    private static final String FIELD_RELATIONS = "relations";

    public void read(final InputStream inputStream) throws IOException {
        ObjectMapper mapper = new ObjectMapper();
        final JsonFactory factory = mapper.getFactory();

        try (JsonParser parser = factory.createParser(inputStream)) {
            if (parser.nextToken() != JsonToken.START_OBJECT) {
                throw new IOException("Expected data to start with an Object");
            }

            while (parser.nextToken() != JsonToken.END_OBJECT) {
                final String fieldName = parser.getCurrentName() == null ? "" : parser.getCurrentName();

                switch (fieldName) {
                    case FIELD_ENTITIES:
                        startEntities();
                        processNameValueEntity(mapper, parser, entity, x -> processEntity(x));
                        endEntities();
                        break;

                    case FIELD_RELATIONS:
                        startRelations();
                        processNameValueEntity(mapper, parser, relation, x -> processRelation(x));
                        endRelations();
                        break;

                    default:
                        throw new IllegalStateException(String.format("Unexpected token - %s", fieldName));
                }
            }

        } catch (Exception ex) {
            throw new IOException(ex);
        }
    }

    protected abstract void endEntities();

    protected abstract void startEntities();

    protected abstract void endRelations();

    protected abstract void startRelations();

    protected abstract void processEntity(Entity entity);

    protected abstract void processRelation(Relation relation);

    private <T extends NameValuePairEntity> void processNameValueEntity(ObjectMapper mapper, JsonParser parser, T nameValuePairEntity, Consumer<T> f) throws IOException {
        String simpleClassName = nameValuePairEntity.getClass().getSimpleName();
        parser.nextToken();
        while (parser.nextToken() != JsonToken.END_ARRAY) {
            f.accept(createDerivedNameValuePairEntity(simpleClassName, getPropertiesMap(mapper, parser)));
        }
    }

    private <T extends NameValuePairEntity> T createDerivedNameValuePairEntity(String entityType, Map<String, Object> propertiesMap) {
        switch (entityType) {
            case "Entity": return (T) new Entity(propertiesMap);
            case "Relation": return (T) new Relation(propertiesMap);

            default:
                throw new RuntimeException(String.format("Creation of entityType: {}, not supported!", entityType));
        }
    }

    private Map<String, Object> getPropertiesMap(ObjectMapper mapper, JsonParser parser) throws IOException {
        try {
            final JsonNode node = parser.readValueAsTree();
            return mapper.convertValue(node, Map.class);
        }
        catch (Exception ex) {
            LOG.error("Error parsing: {}", ex);
        }

        return null;
    }
}
