/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.typedef.AtlasBusinessMetadataDef;
import org.apache.atlas.model.typedef.AtlasEnumDef;
import org.apache.atlas.model.typedef.AtlasStructDef;
import org.apache.atlas.tools.models.Conversion.Spec;
import org.apache.atlas.tools.models.Conversion.Specs;
import org.apache.atlas.tools.models.MultiNodeContainer;
import org.apache.atlas.tools.models.NameValuePairEntity;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.JsonUtils;
import org.apache.atlas.type.AtlasType;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class NamespaceConverter extends NodeToAtlasElement {
    private static final Logger LOG = LoggerFactory.getLogger( NamespaceConverter.class);
    private static final String KEY_PROPERTIES = "properties";
    private static final int ATTRIBUTE_SEARCH_WEIGHT = 10;

    private final String businessMetadataDefJson;
    private final String attributeDefJson;
    private final Map<String, String> sourceTypeToAtlasTypeMap;
    private final NodeToAtlasElement propertiesConverter;
    private final String ATTR_APPLICABLE_ENTITY_TYPES = "applicableEntityTypes";
    private ApplicableTypes applicableTypes;

    public NamespaceConverter(Session session, Specs specs, Map<String, String> sourceTypeToAtlasTypeMap, NodeToAtlasElement propertiesConverter) {
        super(session, specs, null);
        this.sourceTypeToAtlasTypeMap = sourceTypeToAtlasTypeMap;
        this.propertiesConverter = propertiesConverter;
        this.businessMetadataDefJson = AtlasType.toJson(new AtlasBusinessMetadataDef());
        AtlasStructDef.AtlasAttributeDef atlasAttributeDef = new AtlasStructDef.AtlasAttributeDef();
        atlasAttributeDef.setOption("applicableEntityTypes", StringUtils.EMPTY);
        this.attributeDefJson = AtlasType.toJson(atlasAttributeDef);
    }

    public void convert(MultiNodeContainer[] namespaces, MultiNodeContainer propertyMapping) {
        convertPropertyMappings(propertyMapping, sourceTypeToAtlasTypeMap);

        for (MultiNodeContainer namespace : namespaces) {
            AtlasBusinessMetadataDef atlasBusinessMetadataDef = convertNamespaceToAtlasBusinessMetadata(namespace);

            if (namespace.get(KEY_PROPERTIES) == null || !(namespace.get(KEY_PROPERTIES) instanceof List)) {
                continue;
            }

            List propertiesList = (List) namespace.get(KEY_PROPERTIES);
            for (Object propertyObject : propertiesList) {
                if (propertyObject == null || !(propertyObject instanceof LinkedHashMap)) {
                    continue;
                }

                AtlasStructDef.AtlasAttributeDef atlasAttributeDef =
                        convertPropertyToAtlasAttributeDef(atlasBusinessMetadataDef, (LinkedHashMap) propertyObject);
                atlasBusinessMetadataDef.getAttributeDefs().add(atlasAttributeDef);
            }

            session.getAtlasTypesDef().getBusinessMetadataDefs().add(atlasBusinessMetadataDef);
        }
    }

    private AtlasBusinessMetadataDef convertNamespaceToAtlasBusinessMetadata(MultiNodeContainer namespace) {
        MultiNodeContainer namespaceDef = AtlasType.fromJson(businessMetadataDefJson, MultiNodeContainer.class);
        for (String key : specs.getSpecs().keySet()) {
            for (Spec spec : specs.getSpec(key)) {
                NameValuePairEntity nameValuePairEntity = new NameValuePairEntity(namespace.get());

                final Object processedValue = getProcessedValue(nameValuePairEntity, spec);
                assignValues(namespaceDef, spec, processedValue);
            }
        }
        return JsonUtils.fromLinkedHashMap(namespaceDef.get(), AtlasBusinessMetadataDef.class);
    }

    private AtlasStructDef.AtlasAttributeDef convertPropertyToAtlasAttributeDef(AtlasBusinessMetadataDef atlasBusinessMetadataDef, LinkedHashMap propertyObject) {
        MultiNodeContainer attributeDef = AtlasType.fromJson(attributeDefJson, MultiNodeContainer.class);
        MultiNodeContainer processedProperty = new MultiNodeContainer(propertyObject);
        processedProperty.set("options", "{}");
        for (String key : propertiesConverter.specs.getSpecs().keySet()) {
            for (Spec spec : propertiesConverter.specs.getSpec(key)) {
                NameValuePairEntity nameValuePairEntity = new NameValuePairEntity(processedProperty.get());

                final Object processedValue = propertiesConverter.getProcessedValue(nameValuePairEntity, spec);
                propertiesConverter.assignValues(attributeDef, spec, processedValue);
                if (propertiesConverter.getTempVars().get(TEMP_VARIABLES_ENUM_DEF) != null) {
                    AtlasEnumDef enumDef = (AtlasEnumDef) propertiesConverter.getTempVars().get(TEMP_VARIABLES_ENUM_DEF);
                    session.getAtlasTypesDef().getEnumDefs().add(enumDef);
                    propertiesConverter.getTempVars().remove(TEMP_VARIABLES_ENUM_DEF);
                }
            }
        }

        AtlasStructDef.AtlasAttributeDef atlasAttributeDef = JsonUtils.fromLinkedHashMap(attributeDef.get(), AtlasStructDef.AtlasAttributeDef.class);
        atlasAttributeDef.setOption(ATTR_APPLICABLE_ENTITY_TYPES, applicableTypes.get(atlasBusinessMetadataDef.getName(), atlasAttributeDef.getName()));

        if (atlasAttributeDef.getTypeName().equals("string")) {
            atlasAttributeDef.setSearchWeight(ATTRIBUTE_SEARCH_WEIGHT);
        }

        return atlasAttributeDef;
    }

    public void convertPropertyMappings(MultiNodeContainer propertyMapping, Map<String, String> sourceTypeToAtlasTypeMap) {
         this.applicableTypes = new ApplicableTypes(propertyMapping, sourceTypeToAtlasTypeMap);
    }

    static class ApplicableTypes {
        private static final String FIELD_NAMESPACE = "namespace";
        private static final String FIELD_NAME = "name";
        private final Map<String, String> sourceTypeToAtlasTypeMap;
        private Map<String, List<String>> namespacePropertyCache = new LinkedHashMap<>();

        public ApplicableTypes(MultiNodeContainer multiNodeContainer, Map<String, String> sourceTypeToAtlasTypeMap) {
            this.sourceTypeToAtlasTypeMap = sourceTypeToAtlasTypeMap;
            cacheMappings(multiNodeContainer, namespacePropertyCache);
        }

        public String get(String namespace, String property) {
            List<String> applicableTypes = namespacePropertyCache.get(String.format("%s:%s", namespace, property));
            return JsonUtils.to(applicableTypes);
        }

        private void cacheMappings(MultiNodeContainer multiNodeContainer, Map<String, List<String>> namespacePropertyCache) {
            Object root = multiNodeContainer.get("nav");
            if (!(root instanceof LinkedHashMap)) {
                return;
            }

            LinkedHashMap<String, Object> linkedHashMap = (LinkedHashMap) root;
            for (Map.Entry<String, Object> entry : linkedHashMap.entrySet()) {
                String key = entry.getKey();
                List<String> namespacePropertyKeys = getNamespacePropertyKeys(entry.getValue());

                String atlasType = this.sourceTypeToAtlasTypeMap.get(key);
                if (StringUtils.isEmpty(atlasType)) {
                    continue;
                }

                addToCache(namespacePropertyCache, namespacePropertyKeys, atlasType);
            }
        }

        private void addToCache(Map<String, List<String>> namespacePropertyCache, List<String> namespacePropertyKeys, String atlasType) {
            for (String cacheKey : namespacePropertyKeys) {
                if (!namespacePropertyCache.containsKey(cacheKey)) {
                    namespacePropertyCache.put(cacheKey, new ArrayList<>());
                }

                namespacePropertyCache.get(cacheKey).add(atlasType);
            }
        }

        private List<String> getNamespacePropertyKeys(Object x) {
            List<String> ret = new ArrayList<>();

            if (x == null || !(x instanceof List)) {
                return ret;
            }

            List list = (List) x;
            for (Object o : list) {
                if (o == null || !(o instanceof LinkedHashMap)) {
                    continue;
                }

                LinkedHashMap h = (LinkedHashMap) o;
                ret.add(String.format("%s:%s", h.get(FIELD_NAMESPACE), h.get(FIELD_NAME)));
            }

            return ret;
        }
    }
}
