/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.typedef.AtlasEnumDef;
import org.apache.atlas.tools.models.Conversion.Method;
import org.apache.atlas.tools.models.Conversion.Spec;
import org.apache.atlas.tools.models.Conversion.Specs;
import org.apache.atlas.tools.models.MultiNodeContainer;
import org.apache.atlas.tools.models.NameValuePairEntity;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.JsonUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.collections.CollectionUtils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class NodeToAtlasElement {
    private static final String TEMP_VARIABLES_KEY_IDENTITY = "identity";
    public static final String TEMP_VARIABLES_ENUM_DEF = "ENUM_DEF";
    public static final String TEMP_VARIABLE_ENUM_NAME = "ENUM";
    private static final String TEMP_VAR_CURRENT_QUALIFIED_NAME = "currentQualifiedName";

    private static final String SPEC_KEY_SEPARATOR = ",";
    private static final String FIELD_CONSTANT = "#";
    private static final String FIELD_METHOD = "!";
    private static final String FIELD_UNPROCESSED_VALUE = "~";
    private static final String FIELD_VARIABLE_VALUE = "$";
    private static final String FIELD_CONSTANT_EMPTY_ARRAY = "#[]";
    private static final String PROPERTY_PREFIX_NAV = "__nav_";

    protected Session session;
    protected final Specs specs;
    protected final AtlasTypeFactory atlasTypeFactory;
    private static final ThreadLocal<Map<String, Object>> tempVars = ThreadLocal.withInitial(() -> new LinkedHashMap<>());

    public NodeToAtlasElement(Session session, Specs specs, AtlasTypeFactory atlasTypeFactory) {
        this.session = session;
        this.specs = specs;
        this.atlasTypeFactory = atlasTypeFactory;
    }

    protected Object getProcessedValue(NameValuePairEntity incoming, Spec spec) {
        Object val = getFromValue(incoming, spec);
        if (spec.getMethods() == null) {
            return val;
        }

        return applyMethods(incoming, spec, val);
    }

    private Object applyMethods(NameValuePairEntity incoming, Spec spec, Object val) {
        for (Method method : spec.getMethods()) {
            if (method.getFirstParam() != null && method.getFirstParam().equals("all")) {
                val = processMethod(spec.getFrom(), method, incoming, val);
            } else if (val instanceof List) {
                List list = (List) val;
                for (int i = 0; i < list.size(); i++) {
                    final Object element = processMethod(spec.getFrom(), method, incoming, list.get(i));
                    list.set(i, element);
                }
            } else if (val instanceof Map) {
                Map map = (Map) val;
                for (Object item : map.entrySet()) {
                    Map.Entry entry = (Map.Entry) item;
                    final Object element = processMethod(spec.getFrom(), method, incoming, entry);
                    map.put(entry.getKey(), element);
                }
            } else {
                val = processMethod(spec.getFrom(), method, incoming, val);
            }
        }

        return val;
    }

    private Object getFromValue(NameValuePairEntity incoming, Spec spec) {
        Object val = null;
        if (spec.getFrom().startsWith(FIELD_UNPROCESSED_VALUE)) {
            val = incoming.get(spec.getFrom().substring(1));
        } else if (spec.getFrom().startsWith(FIELD_METHOD)) {
            val = processMethod(null, Method.from(spec.getFrom()), null, null);
        } else if (spec.getFrom().equals(FIELD_CONSTANT_EMPTY_ARRAY)) {
            val = new ArrayList<>();
        } else if (spec.getFrom().startsWith(FIELD_CONSTANT)) {
            val = spec.getFromConstant();
        } else if (spec.getFrom().startsWith(FIELD_VARIABLE_VALUE)) {
            val = this.getTempVars().get(spec.getFromVariable());
        } else {
            val = incoming.get().get(spec.getFrom());
        }
        return val;
    }

    protected void assignValues(MultiNodeContainer mnc, Spec spec, Object val) {
        if (spec.isVar()) {
            tempVars.get().put(spec.getVarName(), val);
        } else {
            if (StringUtils.isEmpty(spec.getTo())) {
                return;
            }

            mnc.setValue(spec.isAppend(), spec.getTo(), val);
        }
    }

    protected Object processMethod(String from, Method method, NameValuePairEntity incoming, Object rawValue) {
        final String[] params = method.getParams();
        return processMethod(from, method.getName(), params, incoming, rawValue);
    }

    protected Object processMethod(String from, String methodName, String[] methodParams, NameValuePairEntity incoming, Object rawValue) {
        String identity = (String) tempVars.get().get(TEMP_VARIABLES_KEY_IDENTITY);
        switch (methodName) {
            case "strComp":
                return ConversionMethods.strComp(methodParams[0], (String) rawValue);

            case "removeStartsWith":
                return ConversionMethods.removeStartsWith(methodParams[0], (String) rawValue);

            case "lowerCase":
                return StringUtils.lowerCase((String) rawValue);

            case "fixParentPath":
                return ConversionMethods.fixParentPath((String) rawValue);

            case "join": {
                String[] values = getConvertedKeys2(this.specs, from, incoming);
                return StringUtils.join(values);
            }

            case "join2": {
                String[] values = getConvertedKeys2(this.specs, from, incoming);
                return StringUtils.join(values, methodParams[0]);
            }

            case "sanitize": {
                String ret = StringUtils.remove((String) rawValue, "\n");
                if (methodParams == null || methodParams.length == 0) {
                    return ret;
                }

                for (String mp : methodParams) {
                    ret = StringUtils.remove((String) rawValue, mp);
                }

                return ret;
            }

            case "toStatus":
                return ConversionMethods.toStatus(rawValue);

            case "padLeft": {
                return ConversionMethods.padLeft(rawValue);
            }

            case "prefixGuid":
                if (rawValue == null) {
                    return null;
                }

                if (!(rawValue instanceof String)) {
                    return null;
                }

                return ConversionMethods.getEntityGuid(session, (String) rawValue);

            case "referredGuid": {
                if (rawValue == null) {
                    return null;
                }

                return ConversionMethods.getEntityReferredGuid(session, (String) rawValue);
            }

            case "prefix":
                if (rawValue == null || !(rawValue instanceof String) || StringUtils.isEmpty((String) rawValue)) {
                    return null;
                }

                return ConversionMethods.prefix(methodParams[0], (String) rawValue);

            case "toDate":
                return ConversionMethods.toDate((String) rawValue);

            case "toTimeStamp":
                return ConversionMethods.getTimestamp((String) rawValue);

            case "toAtlasObjectId": {
                if (rawValue == null || !(rawValue instanceof String)) {
                    return null;
                }

                return ConversionMethods.toAtlasObjectId(methodParams[0], (String) rawValue);
            }

            case "toAtlasObjectIdByGuid": {
                if (rawValue == null || !(rawValue instanceof String) || StringUtils.isEmpty((String) rawValue)) {
                    return null;
                }

                return ConversionMethods.toAtlasObjectIdByGuid((String) rawValue);
            }

            case "toHiveOrderList": {
                if (rawValue == null || !(rawValue instanceof List)) {
                    return null;
                }

                return ConversionMethods.toHiveOrderList(methodParams[0], (List<String>) rawValue);
            }

            case "toAtlasObjectIdWithUniqueAttrs": {
                if (rawValue == null || !(rawValue instanceof String)) {
                    return null;
                }

                String uniqueVal = (String) rawValue;
                String typeName = methodParams[0];
                Map<String, Object> map = new HashMap<String, Object>() {{
                    put(methodParams[1], uniqueVal);
                }};

                return new AtlasObjectId(typeName, map);
            }

            case "createType": {
                return atlasTypeFactory.create(methodParams[0]);
            }

            case "setAttribute": {
                MultiNodeContainer en = (MultiNodeContainer) rawValue;
                en.setValue(methodParams[1], incoming.get().get(from));
                return en;
            }

            case "toPosixPermissions": {
                if (rawValue == null) {
                    break;
                }

                MultiNodeContainer fs_permissions = atlasTypeFactory.createFsPermissions();
                ConversionMethods.toFsPermissions(rawValue, fs_permissions);
                return fs_permissions.get();
            }

            case "nullToBooleanFalse": {
                if (rawValue == null) {
                    return false;
                }

                return rawValue;
            }

            case "urlPrefix": {
                return ConversionMethods.urlPrefix((String) rawValue);
            }

            case "toJsonStr": {
                return (rawValue != null) ? JsonUtils.createObject(from, rawValue) : null;
            }

            case "stringifyJson": {
                if (rawValue != null) {
                    LinkedHashMap<String, String> map = new LinkedHashMap<>();
                    map.put(String.format("%s%s", PROPERTY_PREFIX_NAV, from), JsonUtils.to(rawValue));
                    return map;
                } else {
                    return null;
                }
            }

            case "inferFieldIndex": {
                if (rawValue != null) {
                    return rawValue;
                }

                String parentId = (String) ConversionMethods.getParent(session, identity);
                List<String> children = ConversionMethods.getChildren(session, parentId);
                if (CollectionUtils.isEmpty(children)) {
                    return 0;
                }

                int index = children.indexOf(identity);
                if (index == -1) {
                    return 0;
                }

                return index;
            }

            case "toColumnQualifiedNameFromTable": {
                return ConversionMethods.toColumnQualifiedNameFromTable(tempVars.get().get(TEMP_VAR_CURRENT_QUALIFIED_NAME).toString(), rawValue);
            }

            case "getClusterName": {
                return session.getClusterName();
            }

            case "getParent": {
                return ConversionMethods.getParent(session, identity);
            }

            case "getTemplateQualifiedName": {
                return ConversionMethods.getTemplateQualifiedName(session, identity);
            }

            case "getInputs": {
                return ConversionMethods.getInputs(session, identity);
            }

            case "getOutputs": {
                return ConversionMethods.getOutputs(session, identity);
            }

            case "getInferredInputs": {
                return ConversionMethods.getInferredInputs(session, identity);
            }

            case "getParentsOfInputs": {
                return ConversionMethods.getParentsOfInputs(session, identity);
            }

            case "getInferredOutputs": {
                return ConversionMethods.getInferredOutputs(session, identity);
            }

            case "getParentsOfOutputs": {
                return ConversionMethods.getParentsOfOutputs(session, identity);
            }

            case "getAllChildrenInputs": {
                return ConversionMethods.getAllChildrenInputs(session, identity);
            }

            case "getAllChildrenOutputs": {
                return ConversionMethods.getAllChildrenOutputs(session, identity);
            }

            case "getTemplateId": {
                return ConversionMethods.getTemplateId(session, identity);
            }

            case "global:partitionKeys": {
                this.session.addPartitionKeys(rawValue);
                return rawValue;
            }

            case "toAtlasObjectIdByGuidWithParitionKeyLookup": {
                if (this.session.lookupPartitionKey(this.tempVars.get().get(TEMP_VAR_CURRENT_QUALIFIED_NAME))) {
                    if (rawValue instanceof String) {
                        return ConversionMethods.toAtlasRelatedObjectIdByGuid((String) rawValue, methodParams[0], methodParams[1]);
                    }
                }
                return ConversionMethods.toAtlasObjectIdByGuid((String) rawValue);
            }

            case "toAtlasAttributeType": {
                return ConversionMethods.toAtlasAttributeType((String) rawValue, this.tempVars.get(), methodParams[0]);
            }

            case "createNewEnumType": {
                AtlasEnumDef enumDef = ConversionMethods.createNewEnumDef((String) this.tempVars.get().get("name"), rawValue);
                if (enumDef != null) {
                    this.getTempVars().put(TEMP_VARIABLES_ENUM_DEF, enumDef);
                    this.getTempVars().put(TEMP_VARIABLE_ENUM_NAME, enumDef.getName());
                }

                return rawValue;
            }

            case "toProcessQualifiedName": {
                List inputs = (List) ConversionMethods.getInferredInputs(session, identity);
                List outputs = (List) ConversionMethods.getInferredOutputs(session, identity);

                String entityGuid = ConversionMethods.getEntityGuid(session, identity);
                String qn = ProcessQualifiedNameGenerator.toProcessQualifiedName(session, entityGuid, inputs, outputs,
                        (String) incoming.get(methodParams[0]));
                this.tempVars.get().put(TEMP_VAR_CURRENT_QUALIFIED_NAME, qn);
                return qn;
            }

            case "toOperationExecQualifiedName": {
                String[] values = getConvertedKeys2(specs, from, incoming);
                String entityGuid = ConversionMethods.getEntityGuid(session, identity);
                String qn = ProcessQualifiedNameGenerator.toOperationExecQualifiedName(session, entityGuid, values);
                this.tempVars.get().put(TEMP_VAR_CURRENT_QUALIFIED_NAME, qn);
                return qn;
            }

            case "toColumnLineageQualifiedName": {
                String entityGuid = ConversionMethods.getEntityGuid(session, identity);
                List outputs = (List) ConversionMethods.getOutputs(session, identity);
                String parentQualifiedName = ConversionMethods.getParentQualifiedName(session, identity);

                String qn = ProcessQualifiedNameGenerator.toColumnLineageQualifiedName(session, entityGuid, parentQualifiedName, outputs);
                this.tempVars.get().put(TEMP_VAR_CURRENT_QUALIFIED_NAME, qn);
                return qn;
            }

            case "parentTimeStamp": {
                String timestamp = ConversionMethods.getParentTimestamp(session, identity);
                if (StringUtils.isEmpty(timestamp)) {
                    return 0l;
                }

                return Long.valueOf(timestamp);
            }
        }

        return rawValue;
    }

    private String[] getConvertedKeys2(Specs specs, String from, NameValuePairEntity incoming) {
        String[] keys = from.split(SPEC_KEY_SEPARATOR);
        String[] values = from.split(SPEC_KEY_SEPARATOR);

        for (int i = 0; i < keys.length; i++) {
            final Object processedValue = getProcessedValue(incoming, specs.getSpecWithDefault(keys[i]).get(0));
            values[i] = (processedValue == null) ? StringUtils.EMPTY : processedValue.toString();
        }

        return values;
    }

    public boolean shouldCombineConvertedEntities() {
        return this.specs.getShouldCombineEntities();
    }

    public Map<String, Object> getTempVars() {
        return tempVars.get();
    }
}
