/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.typedef.AtlasTypesDef;
import org.apache.atlas.tools.models.Conversion.Constants;
import org.apache.atlas.tools.models.Conversion.Specs;
import org.apache.atlas.tools.models.Entity;
import org.apache.atlas.tools.models.JobSummary;
import org.apache.atlas.tools.models.Relation;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.AtlasTypeRegistryUtils;
import org.apache.atlas.type.AtlasType;
import org.apache.atlas.type.AtlasTypeRegistry;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

public class NodesToAtlasElementConverter implements EntityConverter {
    private static final Logger LOG = LoggerFactory.getLogger(NodesToAtlasElementConverter.class);

    private final Map<String, List<NodeToAtlasEntity>> nodeConvertersMap = new ConcurrentHashMap<>();
    protected Session session;
    private AtlasTypeRegistry atlasTypeRegistry;
    private final AtlasTypeFactory atlasTypeFactory;
    private final ReferenceValidator referenceValidator;
    private JobSummary jobSummary;

    public NodesToAtlasElementConverter(Session session, AtlasTypesDef typesdef) {
        this.session = session;
        atlasTypeRegistry = AtlasTypeRegistryUtils.getAtlasTypeRegistry(typesdef);
        this.atlasTypeFactory = new AtlasTypeFactory(atlasTypeRegistry);
        this.referenceValidator = new ReferenceValidator(session);
    }

    @Override
    public void addSpecs(Specs specs) {
        String key = String.format("%s", specs.getSourceType());
        if (!this.nodeConvertersMap.containsKey(key)) {
            this.nodeConvertersMap.put(key, new ArrayList<>());
        }

        this.nodeConvertersMap.get(key).add(new NodeToAtlasEntity(this.session, atlasTypeFactory, specs));
    }

    public boolean hasConverter(Entity incomingEntity) {
        if (MapUtils.isEmpty(incomingEntity.get())) {
            return false;
        }

        String converterKey = getConverterKey(incomingEntity);
        if (StringUtils.isEmpty(converterKey)) {
            return false;
        }

        return nodeConvertersMap.containsKey(converterKey);
    }


    public List<AtlasEntity.AtlasEntityWithExtInfo> convert(Entity incomingEntity) {
        if (MapUtils.isEmpty(incomingEntity.get())) {
            LOG.warn( String.format("Incoming entity: Empty: %s", AtlasType.toJson(incomingEntity)));
            return null;
        }

        String converterKey = getConverterKey(incomingEntity);
        if (StringUtils.isEmpty(converterKey)) {
            return null;
        }

        AtlasEntity.AtlasEntitiesWithExtInfo entitiesWithExtInfo = new AtlasEntity.AtlasEntitiesWithExtInfo();
        boolean shouldCombineEntities = true;
        for (NodeToAtlasElement nodeToElement : nodeConvertersMap.get(converterKey)) {
            if (nodeToElement == null) {
                LOG.error( String.format("No appropriate converter available for type: %s", converterKey));
            }

            NodeToAtlasEntity nodeToEntity = (NodeToAtlasEntity) nodeToElement;
            shouldCombineEntities = nodeToElement.shouldCombineConvertedEntities();
            final AtlasEntity entity = nodeToEntity.convert(incomingEntity);
            entitiesWithExtInfo.addEntity(entity);
        }

        return getAtlasEntityWithExtInfo(shouldCombineEntities, entitiesWithExtInfo);
    }

    private String getConverterKey(Entity incomingEntity) {
        if (!incomingEntity.get().containsKey(Constants.INCOMING_FIELD_NAME_INTERNAL_TYPE)) {
            LOG.error( String.format("Entity: Does not have %s: %s", Constants.INCOMING_FIELD_NAME_INTERNAL_TYPE, AtlasType.toJson(incomingEntity)));
            return StringUtils.EMPTY;
        }

        String incomingEntityInternalType = (String) incomingEntity.get().get(Constants.INCOMING_FIELD_NAME_INTERNAL_TYPE);
        String key = incomingEntityInternalType;
        if (nodeConvertersMap.containsKey(key)) {
            return key;
        }

        if (!incomingEntity.get().containsKey(Constants.INCOMING_FIELD_NAME_TYPE)) {
            LOG.error( String.format("Attempting to retrieve 2nd level key. Entity: Does not have %s: %s",
                                    Constants.INCOMING_FIELD_NAME_TYPE, AtlasType.toJson(incomingEntity)));
            return StringUtils.EMPTY;
        }

        String incomingEntityType = (String) incomingEntity.get().get(Constants.INCOMING_FIELD_NAME_TYPE);
        key = String.format("%s:%s", incomingEntityInternalType, incomingEntityType);
        if (nodeConvertersMap.containsKey(key)) {
            return key;
        }

        LOG.debug(String.format("No converter found for type: %s.", incomingEntityInternalType));
        return StringUtils.EMPTY;
    }

    @Override
    public void convert(Relation relation) throws ConversionException {
    }

    private List<AtlasEntity.AtlasEntityWithExtInfo> getAtlasEntityWithExtInfo(boolean shouldCombineEntities,
                                                                               AtlasEntity.AtlasEntitiesWithExtInfo entitiesWithExtInfo) {
        if (shouldCombineEntities) {
            AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo = new AtlasEntity.AtlasEntityWithExtInfo(entitiesWithExtInfo.getEntities().get(0));
            if (entityWithExtInfo == null || entityWithExtInfo.getEntity() == null) {
                return Collections.emptyList();
            }

            for (int i = 1; i < entitiesWithExtInfo.getEntities().size(); i++) {
                entityWithExtInfo.addReferredEntity(entitiesWithExtInfo.getEntities().get(i));
            }

            return Collections.singletonList(entityWithExtInfo);
        } else {
            List<AtlasEntity.AtlasEntityWithExtInfo> entityWithExtInfos = new ArrayList<>();
            for (AtlasEntity entity : entitiesWithExtInfo.getEntities()) {
                if (entity == null) {
                    continue;
                }

                entityWithExtInfos.add(new AtlasEntity.AtlasEntityWithExtInfo(entity));
            }

            return entityWithExtInfos;
        }
    }

    @Override
    public void convert(InputStream inputStream, EntityRelationsWriter writer) throws IOException {
        EntityReadConvertWrite entityReadConvertWrite = new EntityReadConvertWrite(session.getRelationsRepository(), writer, this);
        entityReadConvertWrite.start(inputStream);

        this.setJobSummary(entityReadConvertWrite.getJobSummary());
    }

    private void setJobSummary(JobSummary jobSummary) {
        this.jobSummary = jobSummary;
    }

    @Override
    public JobSummary getJobSummary() {
        return this.jobSummary;
    }

    @Override
    public String getTypesDef() {
        AtlasTypesDef typesDef = new AtlasTypesDef(
                new ArrayList<>(atlasTypeRegistry.getAllEnumDefs()),
                new ArrayList<>(atlasTypeRegistry.getAllStructDefs()),
                new ArrayList<>(atlasTypeRegistry.getAllClassificationDefs()),
                new ArrayList<>(atlasTypeRegistry.getAllEntityDefs()),
                new ArrayList<>(atlasTypeRegistry.getAllRelationshipDefs()));

        return AtlasType.toJson(typesDef);
    }

    @Override
    public Set<String> getOrder() {
        return this.session.getOrder().keySet();
    }

    @Override
    public void clearValidationErrors() {
    }

    public boolean isValid(AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo) {
        return this.referenceValidator.isValid(entityWithExtInfo);
    }

    public void clear() {
    }
}
