/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.converter;

import org.apache.atlas.tools.models.Session;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.StringUtils;

import java.util.List;

public class ProcessQualifiedNameGenerator {
    private static String SEPARATOR = "->";
    private static String STR_QUALIFIED_NAME_SEPARATOR = "@";

    public static String toProcessQualifiedName(Session session, String entityGuid, List inputs, List outputs, String salt) {
        String prefix = getQnPrefix(session.getClusterName(), salt);
        String inputQn = getQualifiedNameForEntityIds(session, inputs);
        String outputQn = getQualifiedNameForEntityIds(session, outputs);

        return  (StringUtils.isEmpty(inputQn) && StringUtils.isEmpty(outputQn))
                ? sanitizeForQn(String.format("%s:%s%s", prefix, entityGuid, session.getClusterName()), entityGuid)
                : sanitizeForQn(String.format("%s:%s%s%s", prefix, inputQn,
                ((StringUtils.isEmpty(inputQn) || StringUtils.isEmpty(outputQn)) ? StringUtils.EMPTY : SEPARATOR), outputQn), entityGuid);
    }

    private static String getQualifiedNameForEntityIds(Session session, List list) {
        if (CollectionUtils.isEmpty(list)) {
            return StringUtils.EMPTY;
        }

        StringBuilder qualifiedNames = new StringBuilder();
        for (int i = 0; i < list.size(); i++) {
            Object s = list.get(i);
            String guid = ConversionMethods.getEntityGuid(session, (String) s);
            String qualifiedNameTimestamp = session.getUniqueNameValidator().getQualifiedNameTimestamp(guid);
            if (StringUtils.isEmpty(qualifiedNameTimestamp)) {
                if (StringUtils.isEmpty(guid)) {
                    continue;
                }

                qualifiedNameTimestamp = guid;
            }

            if (i > 0) {
                qualifiedNames.append(":");
            }

            qualifiedNames.append(qualifiedNameTimestamp.toLowerCase());
        }

        return qualifiedNames.toString();
    }

    private static String getQnPrefix(String clusterName, String salt) {
        return String.format("%s:%s", clusterName, salt);
    }

    public static String toColumnLineageQualifiedName(Session session,
                                                      String fallbackId,
                                                      String parentQualifiedName,
                                                      List outputs) {
        String outputPartQn = StringUtils.EMPTY;

        if (!CollectionUtils.isEmpty(outputs)) {
            String guid = ConversionMethods.getEntityGuid(session, (String) outputs.get(0));
            String outputQn = session.getUniqueNameValidator().getQualifiedNameTimestamp(guid);
            outputPartQn = StringUtils.substringBefore(outputQn, STR_QUALIFIED_NAME_SEPARATOR);

            if (StringUtils.isEmpty(outputPartQn)) {
                outputPartQn = StringUtils.substringAfterLast(outputPartQn, ".");
            }
        }

        if (StringUtils.isEmpty(outputPartQn)) {
            outputPartQn = fallbackId;
        }

        return sanitizeForQueryQn(String.format("%s:%s", parentQualifiedName, outputPartQn), fallbackId);
    }

    private static String sanitizeForQn(String s, String fallbackId) {
        String qn = s.replaceAll("/", StringUtils.EMPTY);
        return qn.length() > Integer.MAX_VALUE
                ? fallbackId
                : qn;
    }

    private static String sanitizeForQueryQn(String s, String fallbackId) {
        return sanitizeForQn(s.replace(' ', '_'), fallbackId);
    }

    public static String toOperationExecQualifiedName(Session session, String entityGuid, String[] values) {
        String qn = StringUtils.join(values, ":");
        if (StringUtils.isEmpty(qn)) {
            qn = entityGuid;
        }

        return sanitizeForQn(qn, entityGuid);
    }
}
