/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.tools.models.Session;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import static org.apache.atlas.model.instance.AtlasObjectId.KEY_UNIQUE_ATTRIBUTES;
import static org.apache.atlas.tools.models.Conversion.Constants.ATTRIBUTE_NAME_GUID;

public class ReferenceValidator {
    private static final Logger LOG = LoggerFactory.getLogger(ReferenceValidator.class);
    public static final String QUALIFIED_NAME = "qualifiedName";

    private Session session;

    public ReferenceValidator(Session session) {
        this.session = session;
    }

    public boolean isValid(AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo) {
        if (entityWithExtInfo == null || entityWithExtInfo.getEntity() == null) {
            return true;
        }

        AtlasEntity entity = entityWithExtInfo.getEntity();
        return isValid(entity, entity.getTypeName(), entityWithExtInfo.getReferredEntities(), entity.getRelationshipAttributes());
    }

    private boolean isValid(AtlasEntity entity, String entityTypeName, Map internalReferences, Map externalReferences) {
        boolean ret = true;
        Iterator<Map.Entry<Object, Object> > iterator = externalReferences.entrySet().iterator();
        while(iterator.hasNext()) {
            Map.Entry<Object, Object> entry = iterator.next();
            Object key = entry.getKey();
            Object value = entry.getValue();

            if (value != null && value instanceof List) {
                List list = (List) value;
                ret = checkList(entity.getGuid(), entityTypeName, internalReferences, (String) key, list);
            } else if (value != null && value instanceof LinkedHashMap) {
                ret = checkMap(entity, internalReferences, (String) key, (LinkedHashMap) value);
            }
        }

        return ret;
    }

    private boolean checkMap(AtlasEntity entity, Map internalReferences, String key, LinkedHashMap objectId) {
        boolean ret = isValidObjectId(internalReferences, objectId);
        if (ret) {
            return ret;
        }

        objectId.remove(ATTRIBUTE_NAME_GUID);
        objectId.put(KEY_UNIQUE_ATTRIBUTES, Collections.singletonMap(QUALIFIED_NAME, getQualifiedNameForNonExistentEntity(entity)));

        LOG.warn("Entity: {} - {}: Invalid reference found. Updated with: Attribute: {}: Reference: {}",
                entity.getGuid(), entity.getTypeName(), key, objectId);
        return ret;
    }

    private String getQualifiedNameForNonExistentEntity(AtlasEntity entity) {
        String qn;
        if (entity.getAttribute(QUALIFIED_NAME) != null) {
            qn = (String) entity.getAttribute(QUALIFIED_NAME);
            qn = StringUtils.replace(qn, "@", String.format(".%s@", entity.getGuid()));
        }
        else {
            qn = entity.getGuid();
        }

        return qn;
    }

    private boolean checkList(String entityGuid, String entityTypeName, Map internalReferences, String key, List list) {
        Boolean ret = true;
        for (int i = 0; i < list.size(); i++) {
            Object item = list.get(i);
            if (item instanceof LinkedHashMap) {
                ret = isValidObjectId(internalReferences, (LinkedHashMap) item);
                if (ret) {
                    continue;
                }

                LOG.warn("Entity: {} - {}: Invalid reference found. Removing: Attribute: {}: Reference: {}",
                        entityGuid, entityTypeName, key, list.get(i));

                list.remove(i);
            }
        }

        return ret;
    }

    private boolean isValidObjectId(Map internalReferences, LinkedHashMap objectId) {
        if (!objectId.containsKey(ATTRIBUTE_NAME_GUID)) {
            return true;
        }

        Object guid = objectId.get(ATTRIBUTE_NAME_GUID);
        if ((internalReferences != null && internalReferences.containsKey(guid)) || session.containsGuid((String) guid)) {
            return true;
        }

        return false;
    }
}
