/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.tools.BackingStore.BackingStoreProvider;
import org.apache.atlas.tools.BackingStore.RelationCollections;
import org.apache.atlas.tools.models.Relation;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.MapStoringList;
import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVRecord;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentMap;
import java.util.function.Function;

public class RelationsRepository {
    private static final Logger LOG = LoggerFactory.getLogger(RelationsRepository.class);

    private static final Set<String> TYPES_TO_NOT_SEEK_PARENTS = PropertiesUtils.getTypesInLineageDoNotSeekParent();
    private static final Set<String> TYPES_INCOMING_PROCESS = PropertiesUtils.getIncomingOperationTypes();

    private final BackingStoreProvider backingStoreProvider;
    private final Session session;

    private Map<String, String> instanceOf;
    private MapStoringList dataFlowInputs;
    private MapStoringList dataFlowOutputs;
    private MapStoringList controlFlowInputs;
    private MapStoringList controlFlowOutputs;
    private Map<String, String> childParent;
    private MapStoringList parentChild;
    private MapStoringList templateInstances;
    private Map<String, String> identityGuidMap;
    private Map<String, String> guidTypeMap;
    private Map<String, String> guidQualifiedNameCreatedTime;
    private Map<String, String> order;

    public RelationsRepository(Session session) {
        if (session != null) {
            session.setRelationsRepository(this);
        }

        this.session = session;
        this.backingStoreProvider = session.getBackingStoreProvider();

        initCollections(session);
    }

    private void initCollections(Session session) {
        guidQualifiedNameCreatedTime = this.backingStoreProvider.createMap(RelationCollections.GUID_QUALIFIED_NAME_CREATE_TIME);
        order = backingStoreProvider.createMap(RelationCollections.ORDER);

        if (session != null && session.isBulkLoading()) {
            return;
        }

        instanceOf = this.backingStoreProvider.createMap(RelationCollections.INSTANCE_OF);
        dataFlowInputs = new MapStoringList(this.backingStoreProvider.createMap(RelationCollections.DATA_FLOW_INPUTS));
        dataFlowOutputs = new MapStoringList(this.backingStoreProvider.createMap(RelationCollections.DATA_FLOW_OUTPUTS));
        controlFlowInputs = new MapStoringList(this.backingStoreProvider.createMap(RelationCollections.CONTROL_FLOW_INPUTS));
        controlFlowOutputs = new MapStoringList(this.backingStoreProvider.createMap(RelationCollections.CONTROL_FLOW_OUTPUTS));
        childParent = this.backingStoreProvider.createMap(RelationCollections.CHILD_PARENT);
        parentChild = new MapStoringList(this.backingStoreProvider.createMap(RelationCollections.PARENT_CHILD));
        templateInstances = new MapStoringList(this.backingStoreProvider.createMap(RelationCollections.TEMPLATE_INSTANCES));
        identityGuidMap = this.backingStoreProvider.createMap(RelationCollections.IDENTITY_GUID_MAP);
        guidTypeMap = this.backingStoreProvider.createMap(RelationCollections.GUID_TYPE_MAP);
    }

    public void load(ConcurrentMap<String, String> instanceOf,
                     MapStoringList dataFlowInputs,
                     MapStoringList dataFlowOutputs,
                     MapStoringList controlFlowInputs,
                     MapStoringList controlFlowOutputs,
                     ConcurrentMap<String, String> childParent,
                     MapStoringList parentChild,
                     MapStoringList templateInstances,
                     ConcurrentMap<String, String> identityGuidMap,
                     ConcurrentMap<String, String> guidTypeMap) {
        this.instanceOf = instanceOf;
        this.dataFlowInputs = dataFlowInputs;
        this.dataFlowOutputs = dataFlowOutputs;
        this.controlFlowInputs = controlFlowInputs;
        this.controlFlowOutputs = controlFlowOutputs;
        this.childParent = childParent;
        this.parentChild = parentChild;
        this.templateInstances = templateInstances;
        this.identityGuidMap = identityGuidMap;
        this.guidTypeMap = guidTypeMap;

        LOG.info("RelationsRepository: Load: Done!");
    }

    public void add(Relation relation) {
        if (StringUtils.isEmpty(relation.getType())) {
            return;
        }

        final String type = relation.getType();
        if (type.equals(Relation.RELATION_TYPE_INSTANCE_OF)) {
            addInstanceOf(Relation.InstanceOf.getInstances(relation), Relation.InstanceOf.getTemplate(relation));
        } else if (type.equals(Relation.RELATION_TYPE_DATA_FLOW)) {
            addDataFlow(Relation.DataFlow.getTargets(relation), Relation.DataFlow.getSources(relation));
        } else if (type.equals(Relation.RELATION_TYPE_PARENT_CHILD)) {
            addParentChild(Relation.ParentChild.getChildren(relation), Relation.ParentChild.getParent(relation));
        } else if (type.equals(Relation.RELATION_TYPE_CONTROL_FLOW)) {
            addControlFlow(Relation.ControlFlow.getTargets(relation), Relation.ControlFlow.getSources(relation));
        }
    }

    public void addInstanceOf(List instanceIds, String templateId) {
        for (Object id : instanceIds) {
            this.instanceOf.put((String) id, templateId);
        }

        this.templateInstances.addList(templateId, instanceIds);
    }

    public int getInstancesOfSize() {
        return instanceOf.size();
    }

    public void addDataFlow(List targets, List sources) {
        addElementsToMap(targets, this.dataFlowInputs, sources);
        addElementsToMap(sources, this.dataFlowOutputs, targets);
    }

    public void addControlFlow(List targets, List sources) {
        addElementsToMap(targets, this.controlFlowInputs, sources);
        addElementsToMap(sources, this.controlFlowOutputs, targets);
    }

    public int getDataFlowInputsSize() {
        return this.dataFlowInputs.size();
    }

    public int getDataFlowOutputsSize() {
        return this.dataFlowOutputs.size();
    }

    public int getControlFlowInputsSize() {
        return this.controlFlowInputs.size();
    }

    public int getControlFlowOutputsSize() {
        return this.controlFlowOutputs.size();
    }

    public void addParentChild(List children, String parent) {
        for (Object id : children) {
            this.childParent.put((String) id, parent);
        }

        this.parentChild.addList(parent, children);
    }

    public int getChildParentSize() {
        return this.childParent.size();
    }

    public int getParentChildSize() {
        return this.parentChild.size();
    }

    public String getTemplateId(String id) {
        return this.instanceOf.get(id);
    }

    public List getInputs(String id) {
        return this.dataFlowInputs.get(id);
    }

    public List getOutputs(String id) {
        return this.dataFlowOutputs.get(id);
    }

    public List getInferredInputs(String id) {
        return getInferredList(id, this.dataFlowInputs, this.controlFlowInputs);
    }

    public List getInferredOutputs(String id) {
        return getInferredList(id, this.dataFlowOutputs, this.controlFlowOutputs);
    }

    public List getParentsOfInputs(String id) {
        return getParentsOfList(id, this.dataFlowInputs, this.controlFlowInputs);
    }

    public List getParentsOfOutputs(String id) {
        return getParentsOfList(id, this.dataFlowOutputs, this.controlFlowOutputs);
    }

    public List getAllTemplateInstanceInputs(String id) {
        return getAllTemplateInstanceList(id, this::getParentsOfInputs);
    }


    public List getAllTemplateInstanceOutputs(String id) {
        return getAllTemplateInstanceList(id, this::getParentsOfOutputs);
    }

    public List getAllTemplateInstanceList(String id, Function<String, List> fn) {
        if (!this.templateInstances.containsKey(id)) {
            return null;
        }

        Set<String> ids = new HashSet<>();
        for (Object o : this.templateInstances.get(id)) {
            final List cids = fn.apply((String) o);
            if (cids == null) {
                continue;
            }

            ids.addAll(cids);
        }

        return new ArrayList(ids);
    }

    private List getInferredList(String id, MapStoringList... flowCollections) {
        List children = this.parentChild.get(id);
        if (children == null) {
            return null;
        }

        Set<String> parents = new HashSet<>();
        for (MapStoringList flowCollection : flowCollections) {
            for (Object cid : children) {
                final List dataFlowOutputs = flowCollection.get((String) cid);
                if (CollectionUtils.isEmpty(dataFlowOutputs)) {
                    continue;
                }

                for (Object cidx : dataFlowOutputs) {
                    final String parent = this.getParent((String) cidx);
                    if (StringUtils.isEmpty(parent)) {
                        continue;
                    }

                    parents.add(parent);
                }
            }
        }

        return new ArrayList(parents);
    }

    private List getParentsOfList(String id, MapStoringList... flowCollections) {
        Set<String> parents = new HashSet<>();
        for (MapStoringList flowCollection : flowCollections) {
            final List list = flowCollection.get(id);
            if (CollectionUtils.isEmpty(list)) {
                continue;
            }

            for (Object cidx : list) {
                String guid = ConversionMethods.getEntityGuid(session, (String) cidx);
                if (StringUtils.isNotEmpty(guid) && session.getOrder().containsKey(guid) &&
                        TYPES_TO_NOT_SEEK_PARENTS.contains(session.getOrder().get(guid))) {
                    parents.add((String) cidx);
                } else {
                    final String parent = this.childParent.containsKey(cidx)
                            ? getParent((String) cidx)
                            : (String) cidx;
                    if (StringUtils.isEmpty(parent)) {
                        continue;
                    }

                    parents.add(parent);
                }
            }
        }

        return new ArrayList(parents);
    }

    public String getParent(String childId) {
        return this.childParent.get(childId);
    }

    public List<String> getChildren(String parentId) {
        return this.parentChild.get(parentId);
    }

    private void addElementsToMap(List<String> entriesToBeAffected, MapStoringList map, List<String> elementsToBeAdded) {
        for (String entry : entriesToBeAffected) {
            String key = entry;
            map.addList(key, elementsToBeAdded);
        }
    }

    private void addElementToListWithinMap(Map<String, List> map, String key, List<String> elementsToBeAdded) {
        if (!map.containsKey(key)) {
            map.put(key, new ArrayList<String>());
        }

        List<String> list = map.get(key);
        list.addAll(elementsToBeAdded);
        map.put(key, list);
    }

    public boolean containsInLineage(String id) {
        try {
            return (controlFlowInputs.containsKey(id) && CollectionUtils.isNotEmpty(controlFlowInputs.get(id)) && containsInLineageList(id, controlFlowInputs))
                    || (controlFlowOutputs.containsKey(id) && CollectionUtils.isNotEmpty(controlFlowOutputs.get(id)) && containsInLineageList(id, controlFlowOutputs))
                    || (dataFlowInputs.containsKey(id) && CollectionUtils.isNotEmpty(dataFlowInputs.get(id)) && containsInLineageList(id, dataFlowInputs))
                    || (dataFlowOutputs.containsKey(id) && CollectionUtils.isNotEmpty(dataFlowOutputs.get(id)) && containsInLineageList(id, dataFlowOutputs));
        } catch (Exception ex) {
            LOG.warn("containsInLineage: {}: Exception!", id, ex);
            return false;
        }
    }

    private boolean containsInLineageList(String identity, MapStoringList flowList) {
        if (!flowList.containsKey(identity)) {
            return false;
        }

        List list = flowList.get(identity);
        if (CollectionUtils.isEmpty(list)) {
            return false;
        }

        for (Object cid : list) {
            String s = (String) cid;
            if (this.identityGuidMap.containsKey(s)) {
                String guid = this.identityGuidMap.get(s);
                if (StringUtils.isNotEmpty(guid) && this.guidTypeMap.containsKey(guid)) {
                    String type = this.guidTypeMap.get(guid);
                    return TYPES_INCOMING_PROCESS.contains(type);
                }
            }
        }

        return false;
    }

    public void clear() {
        this.templateInstances.clear();
        this.parentChild.clear();
        this.childParent.clear();
        this.controlFlowOutputs.clear();
        this.controlFlowInputs.clear();
        this.dataFlowOutputs.clear();
        this.dataFlowInputs.clear();
        this.instanceOf.clear();

        this.identityGuidMap.clear();
        this.guidQualifiedNameCreatedTime.clear();
        this.order.clear();
    }

    public void loadMap(String collectionName, InputStream inputStream) throws IOException {
        Map<String, String> map = (collectionName.equals(RelationCollections.IDENTITY_GUID_MAP))
                ? this.identityGuidMap
                : this.guidTypeMap;

        parseCsv(inputStream, map);
    }

    private void parseCsv(InputStream inputStream, Map<String, String> map) throws IOException {
        CSVParser csvParser = new CSVParser(new InputStreamReader(inputStream), CSVFormat.DEFAULT);
        for (CSVRecord csvRecord : csvParser) {
            map.put(csvRecord.get(0), csvRecord.get(1));
        }
    }

    public Map<String, String> getIdentityGuidMap() {
        return this.identityGuidMap;
    }

    public Map<String, String> getGuidQualifiedNameCreatedTime() {
        return this.guidQualifiedNameCreatedTime;
    }

    public Map<String, String> getOrder() {
        return this.order;
    }
}
