/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.typedef.AtlasTypesDef;
import org.apache.atlas.tools.models.Conversion.Constants;
import org.apache.atlas.tools.models.Conversion.Specs;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.utils.JsonUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.LinkedHashMap;
import java.util.Map;

public class SpecsLoader {
    private static final Logger LOG = LoggerFactory.getLogger(SpecsLoader.class);

    public NamespaceConverter createNamespaceConverter(Session session) {
        String currentWorkingDirectory = session.getCurrentWorkingDirectory();
        Specs specs = createSpecs(currentWorkingDirectory, "namespace");
        Specs specsProperties = createSpecs(currentWorkingDirectory, "namespace-properties");
        session.setAtlasTypesDef(getAtlasTypesDef(currentWorkingDirectory));

        Map<String, String> typeMap = getSourceTypeToAtlasTypeMap(session);

        return new NamespaceConverter(session, specs, typeMap, new NodeToAtlasElement(session, specsProperties, null));
    }

    private Map<String, String> getSourceTypeToAtlasTypeMap(Session session) {
        Map<String, String> typeMap = new LinkedHashMap<>();
        Specs[] allSpecs = createAllSpecs(session);
        for (Specs sp : allSpecs) {
            String key = sp.getSourceTypeOnly();
            if (typeMap.containsKey(key)) {
                continue;
            }

            typeMap.put(key, sp.getAtlasType());
        }

        return typeMap;
    }

    public EntityConverter createEntityConverter(Session session) {
        Specs[] allSpecs = createAllSpecs(session);
        NodesToAtlasElementConverter entityConverter = new NodesToAtlasElementConverter(session, session.getAtlasTypesDef());
        for (Specs specs : allSpecs) {
            entityConverter.addSpecs(specs);
        }

        return entityConverter;
    }

    private String[] getSpecIndex(String currentWorkingDirectory) {
        return getObjectFromJsonViaOverride(currentWorkingDirectory,
                Constants.SPECS_DIRECTORY_NAME,
                Constants.SPEC_INDEX + Constants.EXT_JSON,
                String[].class);
    }

    private Specs[] createAllSpecs(Session session) {
        String currentWorkingDirectory = session.getCurrentWorkingDirectory();
        String[] specFiles = getSpecIndex(currentWorkingDirectory);
        if (ArrayUtils.isEmpty(specFiles)) {
            throw new RuntimeException(String.format("'%s' is empty!", Constants.SPEC_INDEX));
        }

        if (session.getAtlasTypesDef() == null) {
            session.setAtlasTypesDef(getAtlasTypesDef(currentWorkingDirectory));
        }

        Specs[] specsAll = new Specs[specFiles.length];
        for (int i = 0; i < specFiles.length; i++) {
            String specFile = specFiles[i];
            specsAll[i] = createSpecs(currentWorkingDirectory, specFile);
        }

        return specsAll;
    }

    public Specs createSpecs(String currentWorkingDirectory, String specFile) {
        Specs specsParent = new Specs();
        Specs specs = getSpecFromFile(currentWorkingDirectory, specFile);
        if (specs == null) {
            throw new RuntimeException("Error loading spec file: " + specFile);
        }

        if (specs.requires()) {
            for (String s : specs.getRequires()) {
                specsParent.merge(getSpecFromFile(currentWorkingDirectory, s));
            }
        }

        specsParent.merge(specs);
        specsParent.setAtlasType(specs.getAtlasType());
        specsParent.setSourceType(specs.getSourceType());
        specsParent.setShouldCombineEntities(specs.getShouldCombineEntities());

        return specsParent;
    }

    public static String[] getWritingOrder(Session session) {
        return getObjectFromJsonViaOverride(session.getCurrentWorkingDirectory(),
                Constants.SPECS_DIRECTORY_NAME,
                Constants.ENTITY_TYPE_WRITING_ORDER + Constants.EXT_JSON,
                String[].class);
    }

    private static AtlasTypesDef getAtlasTypesDef(String currentWorkingDirectory) {
        return getObjectFromJsonViaOverride(currentWorkingDirectory,
                StringUtils.EMPTY,
                Constants.ATLAS_TYPES_DEF_JSON + Constants.EXT_JSON,
                AtlasTypesDef.class);
    }

    private Specs getSpecFromFile(String currentWorkingDirectory, String specFile) {
        String defaultPath = Constants.SPECS_DIRECTORY_NAME + specFile + Constants.EXT_JSON;
        String customizedPath = getCustomizedPath(currentWorkingDirectory, defaultPath);

        Specs fromResource = JsonUtils.fromResourceFileWithExt(defaultPath, Specs.class);
        if (FileUtils.getFile(customizedPath).exists()) {
            Specs customized = JsonUtils.fromLib(customizedPath, Specs.class);

            if (customized != null) {
                fromResource.merge(customized);
                LOG.info(String.format("Customizations loaded from: %s", customizedPath));
            } else {
                LOG.error(String.format("Customizations could not be loaded from: %s.", customizedPath));
            }
        }

        return fromResource;
    }

    private static <T> T getObjectFromJsonViaOverride(String currentWorkingDirectory, String subDir, String fileName, Class<T> clazz) {
        String defaultPath = subDir + fileName;
        String customizedPath = getCustomizedPath(currentWorkingDirectory, defaultPath);
        if (FileUtils.getFile(customizedPath).exists()) {
            T o = JsonUtils.fromLib(customizedPath, clazz);
            if (o == null) {
                LOG.info(String.format("Customizations could not be loaded from: %s", customizedPath));
                return null;
            }
            else {
                LOG.info(String.format("Customizations loaded from: %s", customizedPath));
                return o;
            }
        }

        return JsonUtils.fromResourceFileWithExt(defaultPath, clazz);
    }

    private static String getCustomizedPath(String currentWorkingDirectory, String defaultPath) {
        return currentWorkingDirectory + "/" + defaultPath;
    }
}
