/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.tools.BackingStore.UniqueAttributeManager;
import org.apache.atlas.tools.models.Session;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;

public class UniqueNameValidator {
    private static final Logger LOG = LoggerFactory.getLogger(UniqueNameValidator.class);
    private static final String ATTR_ATTRIBUTE_QUALIFIED_NAME = "qualifiedName";

    private final UniqueAttributeManager manager;

    public UniqueNameValidator(Session session) {
        this.manager = new UniqueAttributeManager(session);
    }

    public void register(List<AtlasEntity.AtlasEntityWithExtInfo> convertedEntities) {
        for (AtlasEntity.AtlasEntityWithExtInfo e : convertedEntities) {
            register(e);
        }
    }

    public void register(AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo) {
        register(entityWithExtInfo.getEntity());
        if (MapUtils.isEmpty(entityWithExtInfo.getReferredEntities())) {
            return;
        }

        for (AtlasEntity e : entityWithExtInfo.getReferredEntities().values()) {
            register(e);
        }
    }

    public void register(AtlasEntity entity) {
        if (entity == null) {
            return;
        }

        manager.register(entity.getGuid(), entity.getTypeName(), getTypeQualifiedName(entity), getQualifiedNameCreatedTime(entity));
    }

    public boolean validate(AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo) {
        AtlasEntity entity = entityWithExtInfo.getEntity();
        return manager.validate(entity.getGuid(), entity.getTypeName(), getTypeQualifiedName(entity));
    }

    public void close() {
        this.manager.clearTypeQualifiedName();
    }

    public String getQualifiedNameTimestamp(String guid) {
        return this.manager.getQualifiedNameTimestamp(guid);
    }

    private String getTypeQualifiedName(AtlasEntity entity) {
        if (!entity.hasAttribute(ATTR_ATTRIBUTE_QUALIFIED_NAME)) {
            return StringUtils.EMPTY;
        }

        return (String) entity.getAttribute(ATTR_ATTRIBUTE_QUALIFIED_NAME);
    }

    private String getQualifiedNameCreatedTime(AtlasEntity entity) {
        if (!entity.hasAttribute(ATTR_ATTRIBUTE_QUALIFIED_NAME)) {
            return StringUtils.EMPTY;
        }

        return String.format("%s:%s", entity.getAttribute(ATTR_ATTRIBUTE_QUALIFIED_NAME), entity.getCreateTime().getTime());
    }

    public void clear() {
        this.manager.clearTypeQualifiedName();
    }
}
