/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.models.Conversion;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import org.apache.commons.lang.StringUtils;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlRootElement;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

@JsonAutoDetect(getterVisibility = PUBLIC_ONLY, setterVisibility = PUBLIC_ONLY, fieldVisibility = NONE)
@JsonSerialize(include = JsonSerialize.Inclusion.NON_NULL)
@JsonIgnoreProperties(ignoreUnknown = true)
@XmlRootElement
@XmlAccessorType(XmlAccessType.PROPERTY)
public class Spec {
    private static final String COLON_SEPARATOR = ":";
    private static final String APPEND_HINT = "+";
    private static final String SUB_ATTR_SEPARATOR = ".";
    private static final String ATTR_SEPARATOR = "attributes" + SUB_ATTR_SEPARATOR;
    private static final String VAR_SEPARATOR = "var" + COLON_SEPARATOR;
    private static final String REL_SEPARATOR = "rel" + SUB_ATTR_SEPARATOR;

    private String from;
    private String to;
    private Method[] methods;
    private boolean isAttr;
    private boolean isRel;
    private boolean isSubAttr;
    private boolean isVar;
    private String varName;
    private boolean isAppend;

    public Spec(String key) {
        this.setFrom(key);
    }

    public Spec(String key, String value) {
        this.setFrom(key);
        this.setTo(value);
    }

    public Spec(String key, Map value) {
        this.setFrom(key);
        this.setTo((String) value.get("to"));
        this.methods = getMethodsFrom(value);
    }

    private Method[] getMethodsFrom(Map value) {
        Object o = value.get("method");
        if (o == null) {
            return null;
        }

        if (o instanceof List) {
            List list = (List) o;
            Method[] methods = new Method[list.size()];
            for (int i = 0; i < list.size(); i++) {
                methods[i] = new Method((Map) list.get(i));
            }

            return methods;
        }

        return new Method[]{new Method((Map) o)};
    }

    public String getFrom() {
        return this.from;
    }

    public List<String> getFromList() {
        if (!isMultiSource(this.from, ",")) {
            return Collections.singletonList(this.from);
        }

        return Arrays.asList(this.from.split(","));
    }

    private boolean isMultiSource(String p, String s) {
        return p.contains(s);
    }

    public void setFrom(String from) {
        int index = from.indexOf(COLON_SEPARATOR);
        this.from = index == -1
                ? from
                : from.substring(0, index);
    }

    public String getTo() {
        return to;
    }

    public void setTo(String to) {
        this.isVar = isMultiSource(to, VAR_SEPARATOR);
        this.isAttr = to.contains(ATTR_SEPARATOR);
        this.isRel = to.contains(REL_SEPARATOR);
        this.isSubAttr = to.contains(SUB_ATTR_SEPARATOR);
        this.isAppend = to.contains(APPEND_HINT);
        if (this.isVar) {
            this.varName = StringUtils.substringAfter(to, VAR_SEPARATOR);
            if (StringUtils.isEmpty(this.varName)) {
                this.varName = this.from;
            }
        }

        if (this.isAppend) {
            this.to = StringUtils.substringBefore(to, APPEND_HINT);
        } else {
            this.to = to;
        }
    }

    public Method[] getMethods() {
        return methods;
    }

    public void setMethods(Method[] method) {
        this.methods = method;
    }

    public boolean isAttr() {
        return isAttr;
    }

    public boolean isRel() {
        return isRel;
    }

    public boolean isSubAttr() {
        return this.isSubAttr;
    }

    public boolean isVar() {
        return this.isVar;
    }

    public Object getFromConstant() {
        return StringUtils.substring(this.from, 1);
    }

    public Object getFromVariable() {
        return StringUtils.substring(this.from, 1);
    }

    public String getVarName() {
        return this.varName;
    }

    public boolean isAppend() {
        return this.isAppend;
    }

    @JsonIgnore
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();

        sb.append("{");
        sb.append("from: " + from);
        sb.append(", to: " + to);
        sb.append(", isAttr: " + isAttr);
        sb.append(", isRel: " + isRel);
        sb.append(", isVar: " + isVar);
        sb.append("}");

        return sb.toString();
    }
}
