/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.models.Conversion;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlRootElement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

@JsonAutoDetect(getterVisibility=PUBLIC_ONLY, setterVisibility=PUBLIC_ONLY, fieldVisibility=NONE)
@JsonIgnoreProperties(ignoreUnknown=true)
@XmlRootElement
@XmlAccessorType(XmlAccessType.PROPERTY)
public class Specs {
    private static final Logger LOG = LoggerFactory.getLogger(Specs.class);

    private Map<String, List<Spec>> specs;

    private String sourceType;
    private String atlasType;
    private String[] requires;
    private Map<String, String> typeMap;
    private boolean shouldCombineEntities = true;

    public Specs() {
        this.specs = new LinkedHashMap<>();
    }

    public Map<String, List<Spec>> getSpecs() {
        return specs;
    }

    public void setSpecs(Map<String, Object> specs) {
        for (Map.Entry<String, Object> entry : specs.entrySet()) {
            String key = entry.getKey();
            Object value = entry.getValue();

            if (!this.specs.containsKey(key)) {
                this.specs.put(key, new ArrayList<>());
            }

            if (value instanceof String) {
                this.specs.get(key).add(new Spec(key, (String) value));
            } else if (value instanceof HashMap) {
                this.specs.get(key).add(new Spec(key, (Map) value));
            } else if (value instanceof List) {
                for (Object item : (List) value) {
                    this.specs.get(key).add(new Spec(key, (Map) item));
                }

                continue;
            } else {
                LOG.error( String.format("Cannot process: %s - %s", key, value));
            }
        }
    }

    public String getAtlasType() {
        return atlasType;
    }

    public void setAtlasType(String atlasType) {
        this.atlasType = atlasType;
    }

    public int size() {
        return specs.size();
    }

    public List<Spec> getSpec(String key) {
        return specs.get(key);
    }

    public List<Spec> getSpecWithDefault(String key) {
        if (specs.containsKey(key)) {
            return specs.get(key);
        }

        return new ArrayList<Spec>() {{
            add(new Spec(key));
        }};
    }

    public String getSourceType() {
        return sourceType;
    }

    public String getSourceTypeOnly() {
        return StringUtils.substringBefore(this.sourceType, ":");
    }

    public void setSourceType(String sourceType) {
        this.sourceType = sourceType;
    }

    public String[] getRequires() {
        return this.requires;
    }

    public void setRequires(String[] requires) {
        this.requires = requires;
    }

    public boolean requires() {
        return this.requires != null && this.requires.length > 0;
    }

    public void setTypeMap(Map<String, String> map) {
        this.typeMap = map;
    }

    public Map<String, String> getTypeMap() {
        return this.typeMap;
    }

    public boolean getShouldCombineEntities() {
        return shouldCombineEntities;
    }

    public void setShouldCombineEntities(boolean shouldCombine) {
        this.shouldCombineEntities = shouldCombine;
    }

    public void merge(Specs requiresSpecs) {
        for (Map.Entry<String, List<Spec>> spec : requiresSpecs.getSpecs().entrySet()) {
            this.specs.put(spec.getKey(), spec.getValue());
        }
    }
}
