/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.models;

import com.fasterxml.jackson.annotation.JsonAnySetter;
import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import org.apache.commons.lang.StringUtils;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlRootElement;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

@JsonAutoDetect(getterVisibility=PUBLIC_ONLY, setterVisibility=PUBLIC_ONLY, fieldVisibility=NONE)
@JsonSerialize(include=JsonSerialize.Inclusion.NON_NULL)
@JsonIgnoreProperties(ignoreUnknown=true)
@XmlRootElement
@XmlAccessorType(XmlAccessType.PROPERTY)
public class MultiNodeContainer {
    private static final String PROPERTY_SEPARATOR = ".";

    private Map<String, Object> multiLevelKVPairs;

    public MultiNodeContainer() {
        this(new HashMap<>());
    }

    public MultiNodeContainer(Map<String, Object> map) {
        this.multiLevelKVPairs = map;
    }

    @JsonAnySetter
    public void set(String key, Object value) {
        if (value instanceof List) {
            this.multiLevelKVPairs.put(key, (List<Object>) value);
        } else if (value instanceof Map) {
            this.multiLevelKVPairs.put(key, (Map<String, Object>) value);
        } else {
            this.multiLevelKVPairs.put(key, value);
        }
    }

    public Map<String, Object> get() {
        return multiLevelKVPairs;
    }

    public Object get(String multiLevelKey) {
        if (!multiLevelKey.contains(PROPERTY_SEPARATOR)) {
            return multiLevelKVPairs.get(multiLevelKey);
        }

        String[] keys = StringUtils.split(multiLevelKey, PROPERTY_SEPARATOR);
        if (multiLevelKVPairs.containsKey(keys[0])) {
            Object level1Collection = multiLevelKVPairs.get(keys[0]);
            if (multiLevelKVPairs.get(keys[0]) instanceof Map) {
                Map<String, Object> innerMap = (Map<String, Object>) level1Collection;
                return innerMap.get(keys[1]);
            } else {
                if (multiLevelKVPairs.get(keys[0]) instanceof List) {
                    return ((List<Object>) level1Collection).get(Integer.valueOf(keys[1]));
                }
            }
        }

        return null;
    }
    public void setValue(String multiLevelKey, Object value) {
        setValue(false, multiLevelKey, value);
    }

    public void setValue(boolean append, String multiLevelKey, Object value) {
        if (append) {
            if (value == null) {
                return;
            }

            Object existingValue = get(multiLevelKey);
            if ((existingValue instanceof LinkedHashMap) && value instanceof LinkedHashMap) {
                ((LinkedHashMap) existingValue).putAll(((LinkedHashMap) value));
                setMapValue(multiLevelKVPairs, multiLevelKey, existingValue);
                return;
            }

            if ((existingValue instanceof LinkedHashMap) && value instanceof Map.Entry) {
                ((LinkedHashMap) existingValue).put(((Map.Entry) value).getKey(), ((Map.Entry) value).getValue());
                setMapValue(multiLevelKVPairs, multiLevelKey, existingValue);
                return;
            }

            if (existingValue == null && value instanceof LinkedHashMap && ((LinkedHashMap) value).size() > 0) {
                setMapValue(multiLevelKVPairs, multiLevelKey, value);
                return;
            }

            if (existingValue == null && value instanceof Map.Entry) {
                Map map = new LinkedHashMap();
                map.put(((Map.Entry) value).getKey(), ((Map.Entry) value).getValue());
                setMapValue(multiLevelKVPairs, multiLevelKey, value);
                return;
            }
         } else {
            setMapValue(multiLevelKVPairs, multiLevelKey, value);
        }
    }

    private void setMapValue(Map<String, Object> map, String multiLevelKey, Object value) {
        if (!multiLevelKey.contains(PROPERTY_SEPARATOR)) {
            map.put(multiLevelKey, value);
            return;
        }

        String key = StringUtils.substringBefore(multiLevelKey, PROPERTY_SEPARATOR);
        Object level1Collection = map.get(key);
        if (level1Collection == null) {
            return;
        }

        if (level1Collection instanceof Map) {
            Map<String, Object> innerMap = (Map<String, Object>) level1Collection;
            setMapValue(innerMap, StringUtils.substringAfter(multiLevelKey, PROPERTY_SEPARATOR), value);
            return;
        } else if (level1Collection instanceof List) {
            List<Object> innerMap = (List<Object>) level1Collection;
            innerMap.add(value);
        }
    }
}
