/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.models;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import org.apache.atlas.tools.utils.JsonUtils;
import org.apache.commons.lang.StringUtils;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlRootElement;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;
import static org.apache.atlas.tools.utils.JsonUtils.getPaddedString;

@JsonAutoDetect(getterVisibility = PUBLIC_ONLY, setterVisibility = PUBLIC_ONLY, fieldVisibility = NONE)
@JsonSerialize
@JsonIgnoreProperties(ignoreUnknown = true)
@XmlRootElement
@XmlAccessorType(XmlAccessType.PROPERTY)
public class Relation extends NameValuePairEntity {
    public static final String ATTR_ENTITY_IDS = "entityIds";
    public static final String ATTR_ENTITY_ID = "entityId";
    public static final String ATTR_SOURCES = "sources";
    public static final String ATTR_TARGETS = "targets";

    public static final String RELATION_TYPE_INSTANCE_OF = "INSTANCE_OF";
    public static final String RELATION_TYPE_DATA_FLOW = "DATA_FLOW";
    public static final String RELATION_TYPE_PARENT_CHILD = "PARENT_CHILD";
    public static final String RELATION_TYPE_CONTROL_FLOW = "CONTROL_FLOW";

    private static final String ATTR_TYPE = "type";

    public Relation() {
    }

    public Relation(String a1, String x1) {
        if (map == null) {
            map = new LinkedHashMap<>();
        }

        map.put(a1, x1);
    }

    public Relation(Map<String, Object> entry) {
        set(entry);
    }

    public String getType() {
        return (String) map.get(ATTR_TYPE);
    }

    public static class ControlFlow {
        private static final String ATTR_SOURCES = "sources";
        private static final String ATTR_TARGETS = "targets";

        public static List getSources(Relation r) {
            return JsonUtils.getPaddedList((List) ((Map) r.get(ATTR_SOURCES)).get(ATTR_ENTITY_IDS));
        }

        public static List getTargets(Relation r) {
            return JsonUtils.getPaddedList((List) ((Map) r.get(ATTR_TARGETS)).get(ATTR_ENTITY_IDS));
        }
    }

    public static class DataFlow {
        public static List getSources(Relation r) {
            return JsonUtils.getPaddedList((List) ((Map) r.get(ATTR_SOURCES)).get(ATTR_ENTITY_IDS));
        }

        public static List getTargets(Relation r) {
            return JsonUtils.getPaddedList((List) ((Map) r.get(ATTR_TARGETS)).get(ATTR_ENTITY_IDS));
        }
    }

    public static class InstanceOf {
        private static final String ATTR_INSTANCES = "instances";
        private static final String ATTR_TEMPLATE = "template";

        public static List getInstances(Relation r) {
            return JsonUtils.getPaddedList((List) ((Map) r.get(ATTR_INSTANCES)).get(ATTR_ENTITY_IDS));
        }

        public static String getTemplate(Relation r) {
            return JsonUtils.getPaddedString(((Map) r.get(ATTR_TEMPLATE)).get(ATTR_ENTITY_ID).toString());
        }
    }

    public static class ParentChild {
        private static final String ATTR_NAME_CHILDREN = "children";
        private static String ATTR_NAME_PARENT = "parent";

        public static String getParent(Relation r) {
            if (!r.containsKey(ATTR_NAME_PARENT)) {
                return StringUtils.EMPTY;
            }

            return JsonUtils.getPaddedString(((Map) r.get(ATTR_NAME_PARENT)).get(Relation.ATTR_ENTITY_ID).toString());
        }

        public static List getChildren(Relation r) {
            if (!r.containsKey(ATTR_NAME_CHILDREN)) {
                return Collections.emptyList();
            }

            return JsonUtils.getPaddedList((List) ((Map) r.get(ATTR_NAME_CHILDREN)).get(Relation.ATTR_ENTITY_IDS));
        }
    }
}
