/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.models;

import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.typedef.AtlasTypesDef;
import org.apache.atlas.tools.BackingStore.BackingStoreProvider;
import org.apache.atlas.tools.converter.RelationsRepository;
import org.apache.atlas.tools.converter.UniqueNameValidator;
import org.apache.atlas.tools.utils.GuidProvider;
import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.commons.lang.StringUtils;

import java.nio.file.Path;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

public class Session {
    private static final String CLUSTER_NAME_DEFAULT = "cm";

    private final BackingStoreProvider backingStoreProvider;
    private final GuidProvider guidProvider;
    private String currentWorkingDirectory;
    private String clusterName;
    private final String guidPrefix;
    private final String referredEntitiesGuidPrefix;
    private AtlasTypesDef atlasTypesDef;
    private long timestamp;
    private RelationsRepository relationsRepository;
    private Set<String> partitionKeysUniqueAttributes = new HashSet<>();
    private Path fileToWrite;
    private String temporaryFileToWrite;
    private UniqueNameValidator uniqueNameValidator;

    public Session() {
        this(StringUtils.EMPTY, null, StringUtils.EMPTY, UUID.randomUUID().toString(), UUID.randomUUID().toString(), null, 0);
    }
    public Session(String currentWorkingDirectory, String clusterName, String guidPrefix, String referredEntitiesGuidPrefix, long timestamp) {
        this(currentWorkingDirectory, null, clusterName, guidPrefix, referredEntitiesGuidPrefix, null, timestamp);
    }

    public Session(String currentWorkingDirectory, Path fileToWrite, String clusterName, String guidPrefix, String referredEntitiesGuidPrefix, AtlasTypesDef atlasTypesDef, long timestamp) {
        this.fileToWrite = fileToWrite;

        this.backingStoreProvider = new BackingStoreProvider(PropertiesUtils.getBackingStoreProviderConfig(), getTempFileName());
        this.currentWorkingDirectory = currentWorkingDirectory;
        this.clusterName = StringUtils.isEmpty(clusterName) ? CLUSTER_NAME_DEFAULT : clusterName;
        this.guidPrefix = createPrefix(guidPrefix);
        this.referredEntitiesGuidPrefix = createPrefix2(referredEntitiesGuidPrefix);
        this.atlasTypesDef = atlasTypesDef;
        this.timestamp = timestamp;
        this.guidProvider = new GuidProvider(this);
    }

    private String getTempFileName() {
        if (this.fileToWrite == null) {
            return UUID.randomUUID().toString();
        }

        return this.getFileToWrite().getFileName().toString();
    }

    private String createPrefix(String guidPrefix) {
        int lastIndexOfDash = guidPrefix.lastIndexOf("-") + 1;
        return StringUtils.substring(guidPrefix, 0, lastIndexOfDash);
    }

    private String createPrefix2(String guidPrefix) {
        String ret = createPrefix(guidPrefix);
        while(ret.equals(this.guidPrefix)) {
            ret = createPrefix(UUID.randomUUID().toString());
        }

        return ret;
    }

    public String getEntityGuidPrefix() {
        return this.guidPrefix;
    }

    public String getReferredEntityGuidPrefix() {
        return this.referredEntitiesGuidPrefix;
    }

    public String getClusterName() {
        return this.clusterName;
    }

    public long getTimestamp() {
        return timestamp == 0L ? System.currentTimeMillis() : this.timestamp;
    }

    public void setRelationsRepository(RelationsRepository relationsRepository) {
        this.relationsRepository = relationsRepository;
    }

    public RelationsRepository getRelationsRepository() {
        return this.relationsRepository;
    }

    public String getCurrentWorkingDirectory() {
        return this.currentWorkingDirectory;
    }

    public AtlasTypesDef getAtlasTypesDef() {
        return atlasTypesDef;
    }

    public void setAtlasTypesDef(AtlasTypesDef atlasTypesDef) {
        this.atlasTypesDef = atlasTypesDef;
    }

    public void addPartitionKeys(Object rawValue) {
        if (rawValue instanceof AtlasObjectId) {
            AtlasObjectId objectId = (AtlasObjectId) rawValue;

            for (Object o : objectId.getUniqueAttributes().values()) {
                partitionKeysUniqueAttributes.add((String) o);
            }
        } else if (rawValue instanceof String) {
            partitionKeysUniqueAttributes.add((String) rawValue);
        }
    }

    public int getPartitionKeysSize() {
        return this.partitionKeysUniqueAttributes.size();
    }

    public boolean lookupPartitionKey(Object rawValue) {
        return this.partitionKeysUniqueAttributes.contains((String) rawValue);
    }

    public Map<String, String> getOrder() {
        return this.getRelationsRepository().getOrder();
    }

    public void addOrder(String guid, String entityType) {
        this.getRelationsRepository().getOrder().put(guid, entityType);
    }

    public boolean containsGuid(String guid) {
        return getRelationsRepository().getOrder().containsKey(guid);
    }

    public void setFileToWrite(Path fileToWrite) {
        this.fileToWrite = fileToWrite;
    }

    public Path getFileToWrite() {
        return fileToWrite;
    }

    public String getFileToWriteName() {
        return fileToWrite.getFileName().toString();
    }

    public boolean isBulkLoading() {
        return PropertiesUtils.getBulkLoadRelations();
    }

    public String getFileToWriteNameWithNoExtension() {
        return StringUtils.substringBeforeLast(this.getFileToWrite().toAbsolutePath().toString(), ".");
    }

    public void setTemporaryFileToWrite(String temporaryFileToWrite) {
        this.temporaryFileToWrite = temporaryFileToWrite;
    }

    public String getTemporaryFileToWrite() {
        return this.temporaryFileToWrite;
    }

    public String getTemporaryFileToWriteWithoutExtension() {
        return StringUtils.substringBeforeLast(this.getTemporaryFileToWrite(), ".");
    }

    public void clear() {
        this.backingStoreProvider.close();
    }

    public UniqueNameValidator getUniqueNameValidator() {
        return uniqueNameValidator;
    }

    public void setupUniqueNameValidator() {
        this.uniqueNameValidator = new UniqueNameValidator(this);
    }

    public GuidProvider getGuidProvider() {
        return this.guidProvider;
    }

    public BackingStoreProvider getBackingStoreProvider() {
        return backingStoreProvider;
    }
}
