/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.pc;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.Writer;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

public class AsyncFileWriter {
    private static final Logger LOG = LoggerFactory.getLogger(AsyncFileWriter.class);
    private static final String THREAD_STOP_TOKEN = "{}";

    private final Runner runner;
    private final Thread runnerTh;

    private BlockingQueue<String> queue;

    public AsyncFileWriter(String collectionName, String fileName, Writer writer) {
        this.queue = new LinkedBlockingQueue<>();

        this.runner = new Runner(queue, fileName, writer);
        this.runnerTh = new Thread(this.runner, collectionName);

        startRunner();
    }

    private void startRunner() {
        try {
            runnerTh.start();
        }
        catch (IllegalThreadStateException e) {
            LOG.info("Runner: Error: {}: Start failed!", this.runnerTh.getName(), e);
        }
        finally {
            LOG.info("Runner: {}: Starting...", this.runnerTh.getName());
        }
    }

    public void write(String item) {
        this.queue.add(item);
    }

    public void flush() {
        try {
            write(THREAD_STOP_TOKEN);

            this.runnerTh.join();
            this.runner.flush();
        } catch (InterruptedException e) {
            LOG.error("Interrupted: {}", this.runner.getFileName());
        }
        finally {
            LOG.info("Runner: Flush: {}: Done!", this.runner.getFileName());
        }
    }

    public void close() {
        this.runner.close();
    }

    private static class Runner implements Runnable {
        private static final int BATCH_SIZE = 1000000;

        private BlockingQueue<String> queue;

        private String fileName;
        private Writer writer;
        private long counter;
        private long batches;

        public Runner(BlockingQueue<String> queue, String fileName, Writer writer) {
            this.queue = queue;
            this.fileName = fileName;
            this.writer = writer;
        }

        @Override
        public void run() {
            String item = StringUtils.EMPTY;
            try {
                while (!Thread.currentThread().isInterrupted()) {
                    item = queue.poll(5, TimeUnit.DAYS);
                    if (item.equals(THREAD_STOP_TOKEN)) {
                        LOG.info("{}: Exiting!",  Thread.currentThread().getName());
                        return;
                    }

                    writer.write(item);

                    if (++counter >= BATCH_SIZE) {
                        LOG.info("Written: {} (million)!", ++batches);
                        counter = 0;
                    }
                }
            } catch (IOException e) {
                LOG.error("Error writing: {}", item);

            } catch (InterruptedException e) {
                LOG.error("AsyncFileWriter: {}: Interrupted!", Thread.currentThread().getName(), e);
            } finally {
                LOG.info("AsyncFileWriter: {}: Done!", Thread.currentThread().getName());
            }
        }

        public void flush() {
            try {
                LOG.info("Flush: {}: Starting...", this.fileName);
                while (!this.queue.isEmpty()) {
                    writer.write(this.queue.poll());
                }

                writer.flush();
            } catch (IOException ex) {
                LOG.error("Error writing: {}...", this.fileName, ex);
            } finally {
                LOG.info("Flush: {}: Done!", this.fileName);
            }
        }

        public void close() {
            try {
                this.writer.close();
            } catch (IOException exception) {
                LOG.error("Error: Closing: {}", this.fileName);
            }
        }

        public String getFileName() {
            return this.fileName;
        }
    }
}
