/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.pc;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.pc.WorkItemConsumer;
import org.apache.atlas.tools.converter.ConversionException;
import org.apache.atlas.tools.converter.ConversionMethods;
import org.apache.atlas.tools.converter.EntityConverter;
import org.apache.atlas.tools.converter.RelationsRepository;
import org.apache.atlas.tools.models.Entity;
import org.apache.atlas.type.AtlasType;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.atomic.AtomicLong;

public class ConversionWorkItemConsumer extends WorkItemConsumer<Entity> {
    private static final Logger LOG = LoggerFactory.getLogger(ConversionWorkItemConsumer.class);
    private final static String ELEMENT_TYPE_FILE = "FILE";
    private final static String ELEMENT_TYPE_DIRECTORY = "DIRECTORY";

    private final RelationsRepository relationshipRepository;
    private final EntityConverter converter;
    private int currentBatch;
    private final int batchSize;
    private AtomicLong currentIndex = new AtomicLong();
    private String currentEntityType;
    private boolean INCLUDE_ALL_FS_ELEMENT;

    public ConversionWorkItemConsumer(RelationsRepository relationshipRepository, EntityConverter converter, BlockingQueue<Entity> queue, boolean includeAllFsElement, int batchSize) {
        super(queue);
        this.relationshipRepository = relationshipRepository;
        this.converter = converter;
        this.batchSize = batchSize;
        INCLUDE_ALL_FS_ELEMENT = includeAllFsElement;
        this.currentBatch = 0;
    }

    @Override
    protected void doCommit() {
        if (currentBatch >= batchSize) {
            LOG.info(String.format("Processed: %s: Total: %,d", this.currentEntityType, currentIndex.get()));
            currentBatch = 0;
        }
    }

    @Override
    protected void processItem(Entity entity) {
        try {
            currentIndex.incrementAndGet();
            currentBatch++;
            if (MapUtils.isEmpty(entity.get())) {
                LOG.info( "Skipped: Empty entity");
                return;
            }

            if (!converter.hasConverter(entity)) {
                return;
            }

            currentEntityType = entity.getType();
            if (!shouldIncludeFsElement(entity)) {
                commit();
                return;
            }

            List<AtlasEntity.AtlasEntityWithExtInfo> list = converter.convert(entity);
            if (CollectionUtils.isEmpty(list)) {
                return;
            }

            commit();
            for (AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo : list) {
                converter.isValid(entityWithExtInfo);
                addResult(entityWithExtInfo);
            }
        } catch (ConversionException ex) {
            LOG.error( String.format("Error converting: %s", AtlasType.toJson(entity)), ex);
        }
        finally {
            entity.get().clear();
            entity.set(null);
        }
    }

    private boolean shouldIncludeFsElement(Entity entity) {
        if (INCLUDE_ALL_FS_ELEMENT) {
            return true;
        }

        if (!currentEntityType.equals(ELEMENT_TYPE_FILE) && !currentEntityType.equals(ELEMENT_TYPE_DIRECTORY)) {
            return true;
        }

        String identity = entity.getIdentity();
        return this.relationshipRepository.containsInLineage((String) ConversionMethods.padLeft(identity));
    }
}
