/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.utils;

import org.apache.atlas.tools.BackingStore.BackingStoreProvider;
import org.apache.atlas.tools.converter.ConversionMethods;
import org.apache.atlas.tools.models.Session;

import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentMap;

public class GuidProvider {
    private static boolean USE_TRUE_GUID = PropertiesUtils.getGuidProviderStrategy();
    private static Set<String> INCOMING_TYPES_NOT_TO_CONSIDER_FOR_STORING = PropertiesUtils.getIncomingTypesNotConsideredForStoring();

    private final GuidProviderStrategy guidProviderStrategy;

    public GuidProvider(Session session) {
        this.guidProviderStrategy = (USE_TRUE_GUID)
                ? new TrueGuid(session)
                : new PrefixBasedGuid(session);
    }

    public String get(String identity) {
        return this.guidProviderStrategy.get(identity);
    }

    public String getReferred(String identity) {
        return this.guidProviderStrategy.getReferred(identity);
    }

    public void setType(String incomingEntityType) {
        this.guidProviderStrategy.setPreserveValuesForLookup(!INCOMING_TYPES_NOT_TO_CONSIDER_FOR_STORING.contains(incomingEntityType));
    }

    private interface GuidProviderStrategy {
        String get(String identity);

        String getReferred(String identity);

        void setPreserveValuesForLookup(boolean preserve);
    }

    private static class PrefixBasedGuid implements GuidProviderStrategy {

        private Session session;

        public PrefixBasedGuid(Session session) {
            this.session = session;
        }

        @Override
        public String get(String identity) {
            return ConversionMethods.prefix(this.session.getEntityGuidPrefix(), identity);
        }

        @Override
        public String getReferred(String identity) {
            return ConversionMethods.prefix(this.session.getReferredEntityGuidPrefix(), identity);
        }

        @Override
        public void setPreserveValuesForLookup(boolean preserve) {

        }
    }

    private static class TrueGuid implements GuidProviderStrategy {
        private boolean preserve;
        private Session session;

        public TrueGuid(Session session) {
            this.session = session;
            this.preserve = true;
        }

        @Override
        public String get(String identity) {
            String key = identity;

            return getGuid(key);
        }

        @Override
        public String getReferred(String identity) {
            String key = String.format("%s:1", identity);
            return getGuid(key);
        }

        @Override
        public void setPreserveValuesForLookup(boolean preserve) {
            this.preserve = preserve;
        }

        private String getGuid(String key) {
            if (!session.getRelationsRepository().getIdentityGuidMap().containsKey(key)) {
                String value = UUID.randomUUID().toString();
                if (preserve) {
                    session.getRelationsRepository().getIdentityGuidMap().put(key, value);
                } else {
                    return value;
                }
            }

            return session.getRelationsRepository().getIdentityGuidMap().get(key);
        }
    }
}
