/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.utils;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import org.apache.atlas.tools.models.Conversion.Constants;
import org.apache.atlas.tools.models.MultiNodeContainer;
import org.apache.atlas.type.AtlasType;
import org.apache.atlas.utils.AtlasJson;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class JsonUtils {
    private static final Logger LOG = LoggerFactory.getLogger(JsonUtils.class);

    private static final ObjectMapper mapper = new ObjectMapper().enable(SerializationFeature.INDENT_OUTPUT);

    public static <T> T fromLinkedHashMap(Object linkedHashMap, Class<?> type) {
        try {
            return (T) mapper.convertValue(linkedHashMap, type);
        } catch (IllegalArgumentException ex) {
            LOG.error(String.format("Converting: {}", AtlasType.toJson(linkedHashMap)));
            return null;
        }
    }

    public static Object createObject(String key, Object value) {
        if (value instanceof LinkedHashMap) {
            return value;
        }

        if (value instanceof Map.Entry) {
            return value;
        }

        String json = String.format("{\"%s\": \"%s\"}", key, value);
        try {
            return mapper.readValue(json, new TypeReference<Map<String, Object>>() {
            });
        } catch (IOException e) {
            return null;
        }
    }

    public static String to(MultiNodeContainer o) {
        return to(o.get());
    }

    public static String to(Object o) {
        try {
            return mapper.writeValueAsString(o);
        } catch (JsonProcessingException e) {
            LOG.error("Error converting object to string.", e);
            return null;
        }
    }

    public static <T> T fromResourceFile(String fileName, Class<T> clazz) {
        return fromResourceFile(fileName + Constants.EXT_JSON, clazz);
    }

    public static <T> T fromResourceFileWithExt(String fileName, Class<T> clazz) {
        final String fileWithExt = fileName;
        try {
            final InputStream inputStream = ResourceUtils.getStringFromResources(fileWithExt);
            return AtlasJson.fromJson(inputStream, clazz);
        } catch (IOException e) {
            return null;
        }
    }

    public static <T> T fromLib(String fileName, Class<T> clazz) {
        if (!FileUtils.getFile(fileName).exists()) {
            return null;
        }

        String json = ResourceUtils.getStringFromFile(fileName);
        return AtlasJson.fromJson(json, clazz);
    }

    public static String getPaddedString(String s) {
        return StringUtils.leftPad(s, Constants.PADDING_SIZE, "0");
    }

    public static List getPaddedList(List list) {
        List<String> ret = new ArrayList<>();
        for (int i = 0; i < list.size(); i++) {
            String s = (String) list.get(i);
            if (StringUtils.isEmpty(s)) {
                continue;
            }

            ret.add(getPaddedString(s));
        }

        return ret;
    }
}
