/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.utils;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicInteger;

public class MapStoringList {
    private static final Logger LOG = LoggerFactory.getLogger(MapStoringList.class);

    private static final String LIST_ELEMENT_KEY_FORMAT = "%s#%s";
    protected final ConcurrentMap<String, String> map;
    private AtomicInteger entryCount = new AtomicInteger();

    public MapStoringList(ConcurrentMap<String, String> map) {
        this.map = map;
    }

    public void addList(String key, List<String> elementsToBeAdded) {
        entryCount.incrementAndGet();
        storeElements(map, key, elementsToBeAdded);
    }

    public boolean containsKey(String key) {
        if (StringUtils.isEmpty(key)) {
            return false;
        }

        return this.map.containsKey(key);
    }

    private void storeElements(ConcurrentMap<String, String> map, String key, List<String> elementsToBeAdded) {
        map.putIfAbsent(key, "0");
        int startIndex;
        synchronized (map.get(key)) {
            startIndex = Integer.valueOf(map.get(key));
            map.put(key, Integer.toString(startIndex + elementsToBeAdded.size()));

            for (int i = 0, index = startIndex; i < elementsToBeAdded.size(); i++, index++) {
                String currentKey = String.format(LIST_ELEMENT_KEY_FORMAT, key, index);
                map.put(currentKey, elementsToBeAdded.get(i));
            }
        }
    }

    public List<String> get(String key) {
        if (StringUtils.isEmpty(key) || !map.containsKey(key)) {
            return null;
        }

        List<String> ret = new ArrayList<>();
        int size = Integer.valueOf(map.get(key));
        for (int i = 0; i < size; i++) {
            String currentKey = String.format(LIST_ELEMENT_KEY_FORMAT, key, i);
            ret.add(map.get(currentKey));
        }

        return ret;
    }

    public int size() {
        return entryCount.get();
    }

    public void clear() {
        this.map.clear();
    }
}
