/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.utils;

import org.apache.atlas.ApplicationProperties;
import org.apache.atlas.AtlasException;
import org.apache.atlas.tools.models.Conversion.Constants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

public class PropertiesUtils {
    private static final Logger LOG = LoggerFactory.getLogger(PropertiesUtils.class);

    private static int CONVERSION_NUM_WORKERS_DEFAULT = 8;
    private static int CONVERSION_BATCH_SIZE_DEFAULT = 500;
    private static int APPLICATION_PROPERTY_CONVERSION_BATCH_SIZE_DEFAULT = 10000000; // 10 M
    private static final boolean CONVERSION_INCLUDE_ALL_FS_ELEMENT = false;
    private static final boolean APPLICATION_PROPERTY_BULK_LOAD_RELATIONS_DEFAULT = false;
    private static final boolean APPLICATION_PROPERTY_QUALIFIED_NAME_VALIDATION_DEFAULT = true;
    private static final boolean APPLICATION_PROPERTY_USE_TRUE_GUID_DEFAULT = false;
    private static final boolean APPLICATION_PROPERTY_USE_SPAWNER_DEFAULT = true;
    private static String APPLICATION_PROPERTY_TYPES_FOR_QUALIFIED_NAME_VALIDATION_CHECK_DEFAULT =
            "hdfs_path,aws_s3_v2_bucket,aws_s3_v2_object,aws_s3_v2_directory,hive_db,hive_table,hive_column";

    private static String APPLICATION_PROPERTY_TYPES_FOR_REFERENCE_STORING_DEFAULT =
        "hdfs_path,aws_s3_v2_bucket,aws_s3_v2_object,aws_s3_v2_directory,hive_db,hive_table,hive_storagedesc,hive_column,hive_process,impala_process,spark_process";

    private static String TYPES_IN_LINEAGE_DO_NOT_NEED_PARENT =
            "hdfs_path,aws_s3_v2_bucket,aws_s3_v2_object,aws_s3_v2_directory,hive_table,hive_db";

    private static Set<String> INCOMING_TYPES_NOT_TO_CONSIDER_FOR_STORING = new HashSet<String>() {{
        add("OPERATION_EXECUTION");
        add("SUB_OPERATION");
    }};

    private static Set<String> INCOMING_TYPES_PROCESS = new HashSet<String>() {{
        add("hv_query");
        add("hv_query_exec");
        add("impala_query");
        add("impala_query_exec");
        add("spark_op");
        add("spark_operation_execution");
    }};

    public static int getBatchSize() {
        try {
            return ApplicationProperties.get().getInt(Constants.APPLICATION_PROPERTY_CONVERSION_BATCH_SIZE, CONVERSION_BATCH_SIZE_DEFAULT);
        } catch (AtlasException exception) {
            LOG.error(String.format("Error fetching: %s. Using default: %d",
                    Constants.APPLICATION_PROPERTY_CONVERSION_BATCH_SIZE, CONVERSION_BATCH_SIZE_DEFAULT),
                    exception);
            return CONVERSION_BATCH_SIZE_DEFAULT;
        }
    }

    public static int getNumWorkers() {
        try {
            return ApplicationProperties.get().getInt(Constants.APPLICATION_PROPERTY_CONVERSION_NUM_WORKERS, CONVERSION_NUM_WORKERS_DEFAULT);
        } catch (AtlasException exception) {
            LOG.error(String.format("Error fetching: %s. Using default: %d.",
                    Constants.APPLICATION_PROPERTY_CONVERSION_NUM_WORKERS, CONVERSION_NUM_WORKERS_DEFAULT),
                    exception);
            return CONVERSION_NUM_WORKERS_DEFAULT;
        }
    }

    public static boolean getAllFSElement() {
        try {
            return ApplicationProperties.get().getBoolean(Constants.APPLICATION_PROPERTY_INCLUDE_ALL_FS_ELEMENT, CONVERSION_INCLUDE_ALL_FS_ELEMENT);
        } catch (AtlasException exception) {
            LOG.error(String.format("Error fetching: %s. Using default: %d.",
                    Constants.APPLICATION_PROPERTY_INCLUDE_ALL_FS_ELEMENT, CONVERSION_NUM_WORKERS_DEFAULT),
                    exception);
            return CONVERSION_INCLUDE_ALL_FS_ELEMENT;
        }
    }

    public static int getEntitiesPerFileThreshold() {
        try {
            return ApplicationProperties.get().getInt(Constants.APPLICATION_PROPERTY_ENTITY_COUNT_PER_FILE, APPLICATION_PROPERTY_CONVERSION_BATCH_SIZE_DEFAULT);
        } catch (AtlasException exception) {
            LOG.error(String.format("Error fetching: %s. Using default: %d",
                    Constants.APPLICATION_PROPERTY_CONVERSION_BATCH_SIZE, APPLICATION_PROPERTY_CONVERSION_BATCH_SIZE_DEFAULT),
                    exception);
            return APPLICATION_PROPERTY_CONVERSION_BATCH_SIZE_DEFAULT;
        }
    }

    public static String getTempDirectory() {
        String defaultTempDirectory = System.getProperty("java.io.tmpdir");
        try {
            return ApplicationProperties.get().getString(Constants.APPLICATION_PROPERTY_TEMP_DIRECTORY, defaultTempDirectory);
        } catch (AtlasException exception) {
            LOG.error(String.format("Error fetching: %s. Using default: %s",
                    Constants.APPLICATION_PROPERTY_TEMP_DIRECTORY, defaultTempDirectory),
                    exception);
            return defaultTempDirectory;
        }
    }

    public static boolean getBackingStoreProviderConfig() {
        try {
            return ApplicationProperties.get().getBoolean(Constants.APPLICATION_PROPERTY_BACKING_STORE, false);
        } catch (AtlasException e) {
            LOG.error(String.format("Error fetching: %s. Using in-memory provider.", Constants.APPLICATION_PROPERTY_BACKING_STORE), e);
            return false;
        }
    }

    public static boolean getBulkLoadRelations() {
        final String applicationPropertyBulkLoadRelations = Constants.APPLICATION_PROPERTY_BULK_LOAD_RELATIONS;

        try {
            return ApplicationProperties.get().getBoolean(applicationPropertyBulkLoadRelations, APPLICATION_PROPERTY_BULK_LOAD_RELATIONS_DEFAULT);
        } catch (AtlasException e) {
            LOG.error("Error fetching: %s. Bulk-load: {}: {}}!",
                    applicationPropertyBulkLoadRelations,
                    APPLICATION_PROPERTY_BULK_LOAD_RELATIONS_DEFAULT, e);
            return APPLICATION_PROPERTY_BULK_LOAD_RELATIONS_DEFAULT;
        }
    }

    public static boolean getPerformQualifiedNameValidation() {
        final String applicationProperty = Constants.APPLICATION_PROPERTY_QUALIFIED_NAME_VALIDATION;

        try {
            return ApplicationProperties.get().getBoolean(applicationProperty, APPLICATION_PROPERTY_QUALIFIED_NAME_VALIDATION_DEFAULT);
        } catch (AtlasException e) {
            LOG.error("Error fetching: %s. {}: {}}!",
                    applicationProperty,
                    APPLICATION_PROPERTY_QUALIFIED_NAME_VALIDATION_DEFAULT, e);
            return APPLICATION_PROPERTY_QUALIFIED_NAME_VALIDATION_DEFAULT;
        }
    }

    public static Set<String> getTypesToConsiderForPerformQualifiedNameValidation() {
        String str = getTypesToConsiderForPerformQualifiedNameValidationStr();
        String[] types = str.split(",");

        return new HashSet<String>(Arrays.asList(types));
    }

    private static String getTypesToConsiderForPerformQualifiedNameValidationStr() {
        final String applicationProperty = Constants.APPLICATION_PROPERTY_TYPES_FOR_QUALIFIED_NAME_VALIDATION_CHECK;

        try {
            return ApplicationProperties.get().getString(applicationProperty, APPLICATION_PROPERTY_TYPES_FOR_QUALIFIED_NAME_VALIDATION_CHECK_DEFAULT);
        } catch (AtlasException e) {
            LOG.error("Error fetching: %s. {}: {}}!",
                    applicationProperty,
                    APPLICATION_PROPERTY_TYPES_FOR_QUALIFIED_NAME_VALIDATION_CHECK_DEFAULT, e);
            return APPLICATION_PROPERTY_TYPES_FOR_QUALIFIED_NAME_VALIDATION_CHECK_DEFAULT;
        }
    }

    public static Set<String> getTypesToConsiderForReferenceStore() {
        String str = getTypesToConsiderForReferenceStoreStr();
        String[] types = str.split(",");

        return new HashSet<String>(Arrays.asList(types));
    }

    public static Set<String> getTypesInLineageDoNotSeekParent() {
        String str = TYPES_IN_LINEAGE_DO_NOT_NEED_PARENT;
        String[] types = str.split(",");

        return new HashSet<String>(Arrays.asList(types));
    }

    public static String getTypesToConsiderForReferenceStoreStr() {
        final String applicationProperty = Constants.APPLICATION_PROPERTY_TYPES_FOR_REFERENCE_STORING;

        try {
            return ApplicationProperties.get().getString(applicationProperty, APPLICATION_PROPERTY_TYPES_FOR_REFERENCE_STORING_DEFAULT);
        } catch (AtlasException e) {
            LOG.error("Error fetching: %s. {}: {}}!",
                    applicationProperty,
                    APPLICATION_PROPERTY_TYPES_FOR_REFERENCE_STORING_DEFAULT, e);
            return APPLICATION_PROPERTY_TYPES_FOR_REFERENCE_STORING_DEFAULT;
        }
    }

    public static boolean getGuidProviderStrategy() {
        final String applicationProperty = Constants.APPLICATION_PROPERTY_USE_TRUE_GUID;
        try {
            return ApplicationProperties.get().getBoolean(applicationProperty, APPLICATION_PROPERTY_USE_TRUE_GUID_DEFAULT);
        } catch (AtlasException e) {
            LOG.error("Error fetching: %s. {}: {}}!",
                    applicationProperty,
                    APPLICATION_PROPERTY_USE_TRUE_GUID_DEFAULT, e);
            return APPLICATION_PROPERTY_USE_TRUE_GUID_DEFAULT;
        }
    }

    public static boolean useSpawner() {
        final String applicationProperty = Constants.APPLICATION_PROPERTY_USE_SPAWNER;
        try {
            return ApplicationProperties.get().getBoolean(applicationProperty, APPLICATION_PROPERTY_USE_SPAWNER_DEFAULT);
        } catch (AtlasException e) {
            LOG.error("Error fetching: %s. {}: {}}!",
                    applicationProperty,
                    APPLICATION_PROPERTY_USE_SPAWNER_DEFAULT, e);
            return APPLICATION_PROPERTY_USE_SPAWNER_DEFAULT;
        }
    }

    public static Set<String> getIncomingTypesNotConsideredForStoring() {
        return INCOMING_TYPES_NOT_TO_CONSIDER_FOR_STORING;
    }

    public static Set<String> getIncomingOperationTypes() {
        return INCOMING_TYPES_PROCESS;
    }
}
