/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.utils;

import org.apache.atlas.tools.NavigatorBridge;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.zip.ZipFile;

public class ResourceUtils {
    private static final Logger LOG = LoggerFactory.getLogger(ResourceUtils.class);
    private static final String JSON_FILE_ENCODING = "UTF-8";

    public static InputStream getStringFromResources(String fileName) {
        final InputStream resourceAsStream = getResourceAsStream(fileName);
        if (resourceAsStream == null) {
            LOG.error( String.format("Error fetching file: %s from resources.", fileName));
            return null;
        }

        return resourceAsStream;
    }

    private static InputStream getResourceAsStream(String fileName) {
        final String fileWithPath = "/" + fileName;
        InputStream is = NavigatorBridge.class.getResourceAsStream(fileWithPath);
        if (is == null) {
            is = Thread.currentThread().getContextClassLoader().getResourceAsStream(fileWithPath);
        }

        return is;
    }

    public static String getStringFromFile(String fileName) {
        StringWriter writer = new StringWriter();
        try {
            IOUtils.copy(new FileInputStream(new File(fileName)), writer, JSON_FILE_ENCODING);
        } catch (IOException e) {
            LOG.error( String.format("Error loading converter spec: %s", fileName), e);
        }
        return writer.toString();
    }

    public static String getStringFromStream(InputStream stream) {
        StringWriter writer = new StringWriter();
        try {
            IOUtils.copy(stream, writer, JSON_FILE_ENCODING);
        } catch (IOException e) {
            LOG.error( "Error fetching string from stream.", e);
        }
        return writer.toString();
    }

    public static InputStream getInputStream(String subDir, String filename) {
        String filePath = String.format("%s/%s", subDir, filename);
        return Thread.currentThread().getContextClassLoader().getResourceAsStream(filePath);
    }

    public static String getZipFileComment(String fileToConvert) {
        final String DEFAULT_COMMENT = "{}";
        try {
            ZipFile zipFile = new ZipFile(new File(fileToConvert));
            String ret = zipFile.getComment();
            return (StringUtils.isEmpty(ret) ? DEFAULT_COMMENT : ret);
        } catch (IOException exception) {
            LOG.error("Error accessing: {}", fileToConvert, exception);
            return DEFAULT_COMMENT;
        }
    }
}
