/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.utils;

import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

public final class Spawner {
    private static final Logger LOG = LoggerFactory.getLogger(Spawner.class);

    private static final String SYSTEM_PROPERTY_JAVA_HOME = "java.home";
    private static final String SYSTEM_PROPERTY_CLASS_PATH = "java.class.path";
    private static final String STR_JAVA = "java";
    private static final String STR_BIN = "bin";
    private static final String ARGUMENT_CLASS_PATH = "-cp";

    public Spawner() {}

    public static int exec(Class klass, List<String> args) throws IOException, InterruptedException {
        List<String> command;
        try {
            command = getCommand(klass, args);
            if (LOG.isDebugEnabled()) {
                LOG.debug("JavaProcess: {}: Starting...", command);
            }

            LOG.info("JavaProcess: {}: {}: Starting...", klass.getSimpleName(), args.toString());
            ProcessBuilder builder = new ProcessBuilder(command);

            Process process = builder.inheritIO().start();
            process.waitFor();
            return process.exitValue();
        }
        finally {
            LOG.info("JavaProcess: {}: {}: Done!", klass.getSimpleName(), args.toString());
        }
    }

    private static List<String> getCommand(Class clazz, List<String> args) {
        List<String> command = new LinkedList<String>();
        String javaHome = System.getProperty(SYSTEM_PROPERTY_JAVA_HOME);
        String javaBin = new StringBuilder(javaHome)
                .append(File.separator)
                .append(STR_BIN)
                .append(File.separator)
                .append(STR_JAVA).toString();

        String classpath = System.getProperty(SYSTEM_PROPERTY_CLASS_PATH);
        String className = clazz.getName();

        command.add(javaBin);
        command.add(ARGUMENT_CLASS_PATH);
        command.add(classpath);
        command.add(className);
        if (CollectionUtils.isNotEmpty(args)) {
            command.addAll(args);
        }

        return command;
    }
}
