/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.writers;

import org.apache.atlas.tools.converter.EntityRelationsWriter;
import org.apache.atlas.type.AtlasType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.OutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

public class AtlasZipWriter implements EntityRelationsWriter {
    private static final Logger LOG = LoggerFactory.getLogger(AtlasZipWriter.class);
    private static final String JSON_ARRAY_START = "[";
    private static final String JSON_COMMA = ",";
    private static final String EXT_JSON = ".json";
    private static final String JSON_EMPTY_OBJECT = "{}";
    private static final String JSON_ARRAY_END = "]";

    private final ZipOutputStream zipOutputStream;
    private String ZIP_FILE_COMMENT_FORMAT = "{\"entitiesCount\":%d, \"total\":%d}";

    public AtlasZipWriter(OutputStream outputStream) {
        zipOutputStream = new ZipOutputStream(outputStream);
    }

    @Override
    public void startEntities() throws IOException {
        zipOutputStream.putNextEntry(new ZipEntry("entities.json"));
        writeBytes(JSON_ARRAY_START);
    }

    @Override
    public void writeEntity(String id, Object o) throws IOException {
        writeBytes(AtlasType.toJson(o) + JSON_COMMA);
    }

    @Override
    public void writeEntry(String name, String payload) throws IOException {
        zipOutputStream.putNextEntry(new ZipEntry(name + EXT_JSON));
        writeBytes(payload);
        zipOutputStream.closeEntry();
    }

    private void writeBytes(String payload) throws IOException {
        zipOutputStream.write(payload.getBytes());
    }

    @Override
    public void endEntities() throws IOException {
        writeBytes(JSON_EMPTY_OBJECT);
        writeBytes(JSON_ARRAY_END);
        close();
    }

    @Override
    public void startRelations() throws IOException {

    }

    @Override
    public void writeRelation(String id, Object o) throws IOException {
        writeEntry(id, AtlasType.toJson(o));
    }

    @Override
    public void endRelations() throws IOException {

    }

    @Override
    public void close() throws IOException {
        zipOutputStream.closeEntry();
    }

    @Override
    public void flush() {
        try {
            zipOutputStream.flush();
        } catch (IOException e) {
            LOG.error("Error: Flush: ", e);
        }
    }

    public void setStreamSize(long size) {
        zipOutputStream.setComment(String.format(ZIP_FILE_COMMENT_FORMAT, size, -1));
    }

    public void flushAndClose() throws IOException {
        LOG.info("Flush: Starting...");

        zipOutputStream.flush();
        zipOutputStream.close();

        LOG.info("Flush: Done!");
    }

    public void setStreamSizeAndTotal(int streamSize, long total) {
        zipOutputStream.setComment(String.format(ZIP_FILE_COMMENT_FORMAT, streamSize, total));
    }
}
