/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.writers;

import org.apache.atlas.tools.converter.EntityRelationsWriter;
import org.apache.atlas.tools.utils.PropertiesUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class ZipWriterSplitter implements EntityRelationsWriter {
    private static final Logger LOG = LoggerFactory.getLogger(ZipWriterSplitter.class);
    private static final int ENTITIES_PER_FILE_THRESHOLD = PropertiesUtils.getEntitiesPerFileThreshold();

    private static final int    FILE_INDEX_SIZE = 4;
    private static final String FILE_INDEX_LEFT_PAD = "0";

    private static final String FILE_SEPARATOR = ".";

    private final int threshold;
    private final String mainFileName;
    private final List<String> fileSeries;
    private String currentFileName;

    AtlasZipWriter currentZipWriter;
    private int currentIndex;
    private Map<String, String> commonEntries;
    private final String SPLIT_FILE_NAME_FORMAT = "%s-%s-of-%s.%s";

    public ZipWriterSplitter(String mainFileName) {
        this(mainFileName, ENTITIES_PER_FILE_THRESHOLD);
    }

    public ZipWriterSplitter(String mainFileName, int threshold) {
        this.threshold = threshold;
        this.mainFileName = mainFileName;
        this.currentFileName = this.mainFileName;
        this.fileSeries = new ArrayList<>();
        this.commonEntries = new HashMap<>();

        setupNextFile();
    }

    @Override
    public void startEntities() throws IOException {
        resetEntryIndex();
        currentZipWriter.startEntities();
    }

    @Override
    public void writeEntity(String id, Object o) throws IOException {
        if (hasReachedThreshold()) {
            LOG.info("ZipWriterSplitter: File size threshold reached: {}", currentIndex);
            endEntities();
            this.setStreamSizeInternal(currentIndex);
            setupNextFile();
        }

        currentZipWriter.writeEntity(id, o);
        incrementEntryIndex();
    }

    private void incrementEntryIndex() {
        currentIndex++;
    }

    private void setupNextFile() {
        try {
            closeAndFlushCurrentFile();

            String nextFileName = getNextFileName(fileSeries.size());
            fileSeries.add(nextFileName);
            this.currentFileName = nextFileName;
            this.currentZipWriter = new AtlasZipWriter(getOutputStream(this.currentFileName));

            addCommonEntriesToCurrentFile();
            startEntitiesForCurrentFile();
            LOG.info("ZipWriterSplitter: Current file: {}", this.currentFileName);
        } catch (IOException e) {
            LOG.error("Error closing: {}", this.currentFileName);
        }
    }

    private void startEntitiesForCurrentFile() throws IOException {
        if (fileSeries.size() > 1) {
            startEntities();
        }
    }

    private void addCommonEntriesToCurrentFile() throws IOException {
        for (String key : this.commonEntries.keySet()) {
            writeEntry(key, this.commonEntries.get(key));
        }
    }

    private void closeAndFlushCurrentFile() throws IOException {
        if (currentZipWriter != null) {
            currentZipWriter.flushAndClose();
            setStreamSizeInternal(currentIndex);
        }
    }

    private void resetEntryIndex() {
        currentIndex = 0;
    }

    private String getNextFileName(int index) {
        if (index == 0) {
            return this.mainFileName;
        }

        return StringUtils.replace(mainFileName, ".zip", "-" + index + ".zip");
    }

    @Override
    public void endEntities() throws IOException {
        currentZipWriter.endEntities();
    }

    @Override
    public void startRelations() throws IOException {

    }

    @Override
    public void writeRelation(String id, Object o) throws IOException {

    }

    @Override
    public void endRelations() throws IOException {

    }

    @Override
    public void close() throws IOException {

    }

    @Override
    public void writeEntry(String name, String payload) throws IOException {
        this.currentZipWriter.writeEntry(name, payload);
        if (!this.commonEntries.containsKey(name)) {
            this.commonEntries.put(name, payload);
        }
    }

    @Override
    public void flush() {

    }

    @Override
    public void flushAndClose() throws IOException {
        closeAndFlushCurrentFile();
    }

    private void setStreamSizeInternal(long streamSize) {
        currentZipWriter.setStreamSize(streamSize);
    }

    @Override
    public void setStreamSize(long total) {
        currentZipWriter.setStreamSizeAndTotal(currentIndex, total);
    }

    private static OutputStream getOutputStream(String fileToWrite) throws IOException {
        return FileUtils.openOutputStream(new File(fileToWrite));
    }

    private boolean hasReachedThreshold() {
        return currentIndex >= this.threshold;
    }

    public List<String> getGeneratedFiles() {
        return this.fileSeries;
    }

    public void moveFiles(String fileToWrite) throws IOException {
        try {
            int fileSeriesSize = fileSeries.size();
            LOG.info("Conversion: Moving temporary file (s): Count: {}: Starting...", fileSeriesSize);

            if (fileSeriesSize == 1) {
                moveSingleFile(fileSeries.get(0), fileToWrite);
                return;
            }

            String fileToWriteWithoutExtension = StringUtils.substringBeforeLast(fileToWrite, FILE_SEPARATOR);
            String fileExtension = StringUtils.substringAfterLast(fileToWrite, FILE_SEPARATOR);
            for (int index = 0; index < fileSeriesSize; index++) {
                moveSingleFile(fileSeries.get(index),
                        getNumberedFileName(fileToWriteWithoutExtension, fileExtension, (index + 1), fileSeriesSize));
            }
        } finally {
            LOG.info("Conversion: Moving temporary file (s): Done!");
        }
    }

    private String getNumberedFileName(String prefix, String extension, int index, int fileSeriesSize) {
        String paddedIndex = StringUtils.leftPad(Integer.toString(index), FILE_INDEX_SIZE, FILE_INDEX_LEFT_PAD);
        return String.format(SPLIT_FILE_NAME_FORMAT, prefix, paddedIndex, fileSeriesSize, extension);
    }

    private void moveSingleFile(String src, String dest) throws IOException {
        LOG.info("Moving: {} to {}", src, dest);
        FileUtils.moveFile(new File(src), new File(dest));
    }
}
