/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.typedef.AtlasTypesDef;
import org.apache.atlas.tools.converter.ConversionException;
import org.apache.atlas.tools.converter.Converter;
import org.apache.atlas.tools.converter.RelationsReadConvertWrite;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.type.AtlasType;
import org.apache.commons.io.IOUtils;
import org.json.JSONException;
import org.skyscreamer.jsonassert.JSONAssert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.AfterTest;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.List;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.fail;

public class BaseTest {
    private static final String EXPECTED_ATLAS_FILE_NAME_FORMAT = "%s-atlas";

    final static String guid1 = "e7d578c9-5978-43d9-aacd-000000146600";
    final static String guid2 = "e7d578c9-5978-43d9-dddd-000000146600";

    protected final Session session = new Session(null, null, null, guid1, guid2, getDefaultTypesDef(), 1571175700001L);
    protected final Converter converter = new Converter(session);

    @AfterTest
    public void clearConverter() {
        if (this.converter == null) {
            return;
        }

        this.converter.clear();
    }

    protected void assertEntityWithFileInput(String subdir, String inputFile) throws JSONException, ConversionException, IOException {
        assertEntity(subdir, inputFile, String.format(EXPECTED_ATLAS_FILE_NAME_FORMAT, inputFile));
    }

    public void assertEntity(String subDir, String inputJsonFile, String expectedJsonFile) throws IOException, ConversionException, JSONException {
        AtlasEntity.AtlasEntityWithExtInfo expected = readObjectFromJson(subDir, expectedJsonFile, AtlasEntity.AtlasEntityWithExtInfo.class);
        List<AtlasEntity.AtlasEntityWithExtInfo> actual = convert(subDir, inputJsonFile);

        assertNotNull(expected);
        assertNotNull(actual);
        assertEqualsEntity(actual.get(0), expected);
    }

    public void assertEqualsEntity(Object actual, Object expected) throws JSONException {
        JSONAssert.assertEquals(AtlasType.toJson(expected), AtlasType.toJson(actual), true);
    }

    public void assertAttributes(AtlasEntity actual, AtlasEntity expected, String attributeName) {
        assertEquals(actual.getAttribute(attributeName), expected.getAttribute(attributeName), attributeName);
    }

    public static <T> T readObjectFromJson(String subDir, String filename, Class<T> objectClass) throws IOException {
        String s = readString(subDir, filename);
        return AtlasType.fromJson(s, objectClass);
    }

    public static String readString(String subDir, String filename) throws IOException {
        InputStream stream = getInputStream(subDir, filename);
        StringWriter writer = new StringWriter();
        IOUtils.copy(stream, writer, "UTF-8");
        return writer.toString();
    }

    public static InputStream getInputStream(String subDir, String filename) {
        String filePath = String.format("%s/%s.json", subDir, filename);
        return Thread.currentThread().getContextClassLoader().getResourceAsStream(filePath);
    }

    protected AtlasEntity getEntityAt(List<AtlasEntity.AtlasEntityWithExtInfo> actualArray, int i) {
        return actualArray.get(i).getEntity();
    }

    protected AtlasEntity getFirstEntity(List<AtlasEntity.AtlasEntityWithExtInfo> actualArray) {
        return actualArray.get(0).getEntity();
    }

    protected AtlasTypesDef getDefaultTypesDef() {
        try {
            return readObjectFromJson(".", "typesdef", AtlasTypesDef.class);
        } catch (IOException e) {
            fail("Error loading typesedef.");
        }

        return null;
    }

    protected List<AtlasEntity.AtlasEntityWithExtInfo> convert(String subDir, String fileName) throws IOException, ConversionException {
        return converter.process(readString(subDir, fileName));
    }

    protected void setupRelations(String subDir, String filename) throws IOException {
        RelationsReadConvertWrite relationsReadConvertWrite = new RelationsReadConvertWrite(this.session);
        relationsReadConvertWrite.convert(getInputStream(subDir, filename));
    }
}
