/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.typedef.AtlasBusinessMetadataDef;
import org.apache.atlas.tools.BaseTest;
import org.apache.atlas.tools.models.MultiNodeContainer;
import org.jetbrains.annotations.NotNull;
import org.json.JSONException;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.Map;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

public class BusinessMetadataTest extends BaseTest {
    @Test
    public void namespaceConversion() throws IOException, JSONException {
        String expectedNamespaceName = "NS1";

        AtlasBusinessMetadataDef[] expectedBusinessMetadataDef = readObjectFromJson("namespaces", "namespace-ns1-atlas", AtlasBusinessMetadataDef[].class);
        MultiNodeContainer[] namespaces = readObjectFromJson("namespaces", "namespace-ns1", MultiNodeContainer[].class);
        MultiNodeContainer propertyMapping = readObjectFromJson("namespaces", "property-mapping", MultiNodeContainer.class);

        SpecsLoader specsLoader = new SpecsLoader();
        NamespaceConverter namespaceConverter = specsLoader.createNamespaceConverter(session);
        namespaceConverter.convert(namespaces, propertyMapping);

        assertTrue(session.getAtlasTypesDef().getBusinessMetadataDefs().size() > 0);
        assertTrue(session.getAtlasTypesDef().getBusinessMetadataDefs().get(0).getAttributeDefs().size() > 0);

        for (AtlasBusinessMetadataDef namespaceDef : session.getAtlasTypesDef().getBusinessMetadataDefs()) {
            if (namespaceDef.getName().equals(expectedNamespaceName)) {
                continue;
            }

            assertEqualsEntity(namespaceDef, expectedBusinessMetadataDef[0]);
        }
    }

    @Test
    public void spidermanNamespace() throws IOException, JSONException {
        String expectedNamespaceName = "namespace_spiderman";
        int enumDefCountPreConversion = session.getAtlasTypesDef().getEnumDefs().size();
        AtlasBusinessMetadataDef[] expectedBusinessMetadataDef = readObjectFromJson("namespaces", "spiderman-atlas", AtlasBusinessMetadataDef[].class);
        MultiNodeContainer[] namespaces = readObjectFromJson("namespaces", "spiderman", MultiNodeContainer[].class);
        MultiNodeContainer propertyMapping = readObjectFromJson("namespaces", "property-mapping", MultiNodeContainer.class);

        SpecsLoader specsLoader = new SpecsLoader();
        NamespaceConverter namespaceConverter = specsLoader.createNamespaceConverter(session);
        namespaceConverter.convert(namespaces, propertyMapping);

        assertTrue(session.getAtlasTypesDef().getBusinessMetadataDefs().size() > 0);
        assertEquals(session.getAtlasTypesDef().getEnumDefs().size(),  enumDefCountPreConversion + 1);

        for (AtlasBusinessMetadataDef namespaceDef : session.getAtlasTypesDef().getBusinessMetadataDefs()) {
            if (!namespaceDef.getName().equals(expectedNamespaceName)) {
                continue;
            }

            assertEqualsEntity(namespaceDef, expectedBusinessMetadataDef[0]);
        }
    }

    @Test
    public void applicableTypes() throws IOException {
        MultiNodeContainer propertyMapping = readObjectFromJson("namespaces", "property-mapping", MultiNodeContainer.class);

        assertNotNull(propertyMapping);

        Map<String, String> sourceTypeToAtlasTypeMap = getSourceTypeAtlasTypeMap();
        NamespaceConverter.ApplicableTypes applicableTypes = new NamespaceConverter.ApplicableTypes(propertyMapping, sourceTypeToAtlasTypeMap);
        assertEquals(applicableTypes.get("NS1", "P1"), "[ \"hive_database\" ]");
    }

    @NotNull
    private Map<String, String> getSourceTypeAtlasTypeMap() {
        return new LinkedHashMap<String, String>() {{
                put("hv_database", "hive_database");
            }};
    }
}
