/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.tools.BaseTest;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.tools.models.ZipEntryMembers;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

public class ConverterTest extends BaseTest {
    private static final String ATLAS_TYPE_HIVE_DB = "hive_db";
    private static final String ATLAS_TYPE_HIVE_TABLE = "hive_table";
    private static final String ATLAS_TYPE_HIVE_COLUMN = "hive_column";
    private static final String ATLAS_TYPE_HDFS_PATH = "hdfs_path";

    private static class WriterSpy implements EntityRelationsWriter {
        Map<String, Object> entries = new ConcurrentHashMap<>();
        Map<String, Integer> typeIdMap = new ConcurrentHashMap<>();

        public WriterSpy() {
            entries.clear();
            typeIdMap.clear();
        }

        @Override
        public void startEntities() throws IOException {

        }

        @Override
        public void writeEntity(String id, Object o) throws IOException {
            entries.put(id, o);
            if (o instanceof AtlasEntity.AtlasEntityWithExtInfo) {
                AtlasEntity.AtlasEntityWithExtInfo e = (AtlasEntity.AtlasEntityWithExtInfo) o;
                String typeName = e.getEntity().getTypeName();

                if (!typeIdMap.containsKey(typeName)) {
                    typeIdMap.put(typeName, 0);
                }

                Integer count = typeIdMap.get(typeName);
                typeIdMap.put(typeName, ++count);
            }
        }

        @Override
        public void endEntities() throws IOException {

        }

        @Override
        public void startRelations() throws IOException {

        }

        @Override
        public void writeRelation(String id, Object o) throws IOException {

        }

        @Override
        public void endRelations() throws IOException {

        }

        @Override
        public void close() throws IOException {

        }

        @Override
        public void writeEntry(String name, String payload) throws IOException {
            entries.put(name, payload);
        }

        @Override
        public void flush() {

        }

        @Override
        public void flushAndClose() throws IOException {

        }

        @Override
        public void setStreamSize(long streamSize) {

        }

        public Map<String, Object> getEntries() {
            return entries;
        }

        public Integer getTypeIdMap(String typeName) {
            if (!this.typeIdMap.containsKey(typeName)) {
                return null;
            }

            return this.typeIdMap.get(typeName);
        }
    }
    @Test
    public void dbTableColumn() throws IOException {
        WriterSpy writerSpy = new WriterSpy();

        String guid1 = "e7d578c9-5978-43d9-aaaa-000000146600";
        String guid2 = "e7d578c9-5978-43d9-bbbb-000000146600";
        Session session = new Session(null, null, null, guid1, guid2, getDefaultTypesDef(), 1571175700001L);
        Converter converter = new Converter(session);
        converter.convertEntities(getInputStream("mixed", "default-db"), writerSpy);

        assertNotNull(writerSpy.getEntries());
        assertEquals(writerSpy.getEntries().size(), 49);
        assertNotNull(writerSpy.getEntries().get(ZipEntryMembers.ENTRY_ATLAS_TYPESDEF.toString()));
        assertNull(writerSpy.getTypeIdMap(ATLAS_TYPE_HDFS_PATH), "hdfs_path elements should be filtered out.");
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_DB).intValue(), 1);
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_TABLE).intValue(), 4);
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_COLUMN).intValue(), 41);
    }

    @Test
    public void stocksInvalidReferenceSkipsEntities() throws IOException, ConversionException {
        WriterSpy writerSpy = new WriterSpy();

        String guid1 = "e7d578c9-5978-43d9-aacd-000000146600";
        String guid2 = "e7d578c9-5978-43d9-dddd-000000146600";
        Session session = new Session(null,null, null, guid1, guid2, getDefaultTypesDef(), 1571175700001L);

        Converter converter = new Converter(session);
        converter.processRelations(getInputStream("relationships", "stocks-relations"));
        converter.convertEntities(getInputStream("mixed", "stocks"), writerSpy);

        assertNotNull(writerSpy.getEntries());
        assertEquals(writerSpy.getEntries().size(), 16 + 3);
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_DB).intValue(), 1);
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_TABLE).intValue(), 1);
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_COLUMN).intValue(), 14);
    }

    @Test
    public void partitionTest() throws IOException, ConversionException {
        final String COLUMN_GUID_VOLUME = "e7d578c9-5978-43d9-aacd-000000000007";

        WriterSpy writerSpy = new WriterSpy();

        String guid1 = "e7d578c9-5978-43d9-aacd-000000000000";
        String guid2 = "e7d578c9-5978-43d9-dddd-000000000000";
        Session session = new Session(null,null, null, guid1, guid2, getDefaultTypesDef(), 1571175700001L);
        Converter converter = new Converter(session);
        converter.convertEntities(getInputStream("mixed", "db-partitions"), writerSpy);

        assertNotNull(writerSpy.getEntries());
        assertEquals(writerSpy.getEntries().size(), 10);
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_DB).intValue(), 1);
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_TABLE).intValue(), 1);
        assertEquals(writerSpy.getTypeIdMap(ATLAS_TYPE_HIVE_COLUMN).intValue(), 5);

        assertTrue(session.getPartitionKeysSize() > 0);
        assertNotNull(writerSpy.getEntries().get(COLUMN_GUID_VOLUME));

        AtlasEntity.AtlasEntityWithExtInfo entityWithExtInfo = (AtlasEntity.AtlasEntityWithExtInfo) writerSpy.getEntries().get(COLUMN_GUID_VOLUME);
        assertNotNull(entityWithExtInfo.getEntity().getRelationshipAttribute("table"));

        LinkedHashMap objectId = (LinkedHashMap) entityWithExtInfo.getEntity().getRelationshipAttribute("table");
        assertNotNull(objectId);
        assertEquals(objectId.get("typeName"), "hive_table");
        assertEquals(objectId.get("relationshipType"), "hive_table_partitionkeys");
    }
}
