/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.tools.BackingStore.BackingStoreProvider;
import org.apache.atlas.tools.BaseTest;
import org.apache.atlas.tools.models.Entity;
import org.apache.atlas.tools.models.Relation;
import org.apache.atlas.tools.utils.JsonUtils;
import org.apache.atlas.tools.utils.MapStoringList;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

public class EntityRelationsReaderTest extends BaseTest {
    private static final Logger LOG = LoggerFactory.getLogger(EntityRelationsReaderTest.class);


    public static class ERReader extends EntityRelationsParser {

        private final Map<String, String> childParent;
        private final MapStoringList parentChild;
        private int entityCount;
        private int relationshipCount;
        private boolean endEntitiesCalled;
        private boolean startEntitiesCalled;
        private boolean endRelationsCalled;
        private boolean startRelationsCalled;

        public ERReader() {
            this.childParent = new ConcurrentHashMap<>();
            this.parentChild = new MapStoringList(new ConcurrentHashMap<>());
        }

        public ERReader(ConcurrentMap<String, String> map1, ConcurrentMap<String, String> map2) {
            this.childParent = map1;
            parentChild = new MapStoringList(map2);
        }

        @Override
        protected void endEntities() {
            endEntitiesCalled = true;
        }

        @Override
        protected void startEntities() {
            startEntitiesCalled = true;
        }

        @Override
        protected void endRelations() {
            endRelationsCalled = true;
        }

        @Override
        protected void startRelations() {
            startRelationsCalled = true;
        }

        @Override
        protected void processEntity(Entity entity) {
            entityCount++;
            assertNotNull(entity);
            assertTrue(entity.size() > 0);
        }

        @Override
        protected void processRelation(Relation relation) {
            try {
                if (relation == null || !relation.containsKey("identity")) {
                    LOG.error("relation is null");
                    return;
                }
                relationshipCount++;
                add(relation);
            } catch (Exception e) {
                LOG.error("Error reading: {}", e);
            }
        }

        public void addParentChild(List children, String parent) {
            for (Object id : children) {
                this.childParent.put((String) id, parent);
            }

            this.parentChild.addList(parent, children);
        }

        public void add(Relation relation) {
            if (StringUtils.isEmpty(relation.getType())) {
                return;
            }

            final String type = relation.getType();
            switch (type) {
                case Relation.RELATION_TYPE_PARENT_CHILD:
                    addParentChild(Relation.ParentChild.getChildren(relation), Relation.ParentChild.getParent(relation));
                    break;
            }
        }

        public List<String> getChildren(String parentId) {
            return this.parentChild.get(parentId);
        }
    }

    @Test
    public void verifyRelationsRead() throws IOException {
        InputStream inputStream = getInputStream("mixed", "default-db");
        ERReader reader = new ERReader();

        reader.read(inputStream);
        assertEquals(reader.entityCount, 49);
        assertEquals(reader.relationshipCount, 57);
        assertTrue(reader.startEntitiesCalled);
        assertTrue(reader.endEntitiesCalled);
        assertTrue(reader.startRelationsCalled);
        assertTrue(reader.endRelationsCalled);
    }

    @Test
    public void verifyStoringRelations() throws IOException {
        BackingStoreProvider backingStoreProvider = new BackingStoreProvider();

        ConcurrentMap<String, String> map1 = backingStoreProvider.createMap("childParent");
        ConcurrentMap<String, String> map2 = backingStoreProvider.createMap("parentChild");
        InputStream inputStream = getInputStream("relationships", "stocks-relations");
        ERReader reader = new ERReader(map1, map2);

        reader.read(inputStream);

        assertTrue(reader.relationshipCount > 0);
        assertEquals(reader.getChildren(JsonUtils.getPaddedString("2")).toArray(new String[]{}),
                new String[]{
                        JsonUtils.getPaddedString("4"),
                        JsonUtils.getPaddedString("5"),
                        JsonUtils.getPaddedString("6"),
                        JsonUtils.getPaddedString("9"),
                        JsonUtils.getPaddedString("10"),
                        JsonUtils.getPaddedString("11")});

        LOG.info("processed: {}", reader.relationshipCount);
    }
}
