/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.tools.converter;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.tools.BaseTest;
import org.apache.atlas.tools.writers.AtlasZipWriter;
import org.json.JSONException;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

public class HiveConversionTest extends BaseTest {
    @BeforeClass
    public void setup() throws IOException {
        setupRelations("hive", "hive-relations");
    }

    @Test
    public void hiveDB() throws Exception {
        assertEntityWithFileInput("hive", "hive-db");
    }

    @Test
    public void hiveDBs() throws Exception {
        AtlasEntity.AtlasEntitiesWithExtInfo expectedArray = readObjectFromJson("hive", "hive-db-multiple-atlas", AtlasEntity.AtlasEntitiesWithExtInfo.class);
        List<AtlasEntity.AtlasEntityWithExtInfo> actualArray = converter.process(readString("hive", "hive-db-multiple"));

        assertEquals(actualArray.size(), expectedArray.getEntities().size());

        for (int i = 0; i < expectedArray.getEntities().size(); i++) {
            AtlasEntity expected = expectedArray.getEntities().get(i);
            AtlasEntity actual = getEntityAt(actualArray, i);

            assertDB(actual, expected);
        }
    }

    private void assertDB(AtlasEntity actual, AtlasEntity expected) {
        final String qualifiedNameAttr = "qualifiedName";
        assertNotNull(actual);
        assertAttributes(actual, expected, "name");
        assertEquals(actual.getTypeName(), expected.getTypeName(), "typeName");
        assertEquals(actual.getGuid(), expected.getGuid(), "guid");
        assertEquals(actual.getAttribute(qualifiedNameAttr), expected.getAttribute(qualifiedNameAttr), qualifiedNameAttr);
    }

    @Test
    public void hiveTable() throws Exception {
        AtlasEntity expected = readObjectFromJson("hive", "hive-table-atlas", AtlasEntity.class);
        List<AtlasEntity.AtlasEntityWithExtInfo> actualArray = convert("hive", "hive-table");
        AtlasEntity actual = getFirstEntity(actualArray);

        assertNotNull(actual);
        assertEqualsEntity(actual, expected);
    }

    @Test(enabled = false)
    public void hiveTableWithTags() throws Exception {
        final String EXPECTED_TAG_NAME = "TEMPORARY";
        AtlasEntity expected = readObjectFromJson("hive", "hive-table-tags-atlas", AtlasEntity.class);
        List<AtlasEntity.AtlasEntityWithExtInfo> actualArray = convert("hive", "hive-table-tags");
        AtlasEntity actual = getFirstEntity(actualArray);

        assertNotNull(actual);
        assertEqualsEntity(actual, expected);
        assertEquals(actual.getClassifications().get(0).getTypeName(), EXPECTED_TAG_NAME);
    }

    @Test
    public void hiveDBTableRelated() throws Exception {
        AtlasEntity.AtlasEntitiesWithExtInfo expected = readObjectFromJson("hive", "hive-db-table-atlas", AtlasEntity.AtlasEntitiesWithExtInfo.class);
        List<AtlasEntity.AtlasEntityWithExtInfo> actual = convert("hive", "hive-db-table");

        assertNotNull(expected);
        assertNotNull(actual);
        assertEqualsEntity(expected.getEntities().get(0), actual.get(0).getEntity());
        assertEqualsEntity(expected.getEntities().get(1), actual.get(1).getEntity());
    }

    @Test
    public void hiveColumn() throws IOException, ConversionException, JSONException {
        assertEntityWithFileInput("hive", "hive-column");
    }

    @Test
    public void hiveSD() throws IOException, ConversionException, JSONException {
        AtlasEntity.AtlasEntityWithExtInfo expected = readObjectFromJson("hive", "hive-table-sd-atlas", AtlasEntity.AtlasEntityWithExtInfo.class);
        List<AtlasEntity.AtlasEntityWithExtInfo> actual = convert("hive", "hive-table");

        assertNotNull(actual);
        assertEqualsEntity(getEntityAt(actual, 0), expected.getEntity());
        final String sdGuid = "e7d578c9-5978-43d9-dddd-000000000031";
        assertEqualsEntity(actual.get(0).getReferredEntities().get(sdGuid), expected.getReferredEntity(sdGuid));
    }

    @Test
    public void tableCreatesSDReferredEntity() throws IOException, ConversionException {
        List<AtlasEntity.AtlasEntityWithExtInfo> actualArray = convert("hive", "hive-table");
        AtlasEntity.AtlasEntityWithExtInfo actual = actualArray.get(0);

        assertNotNull(actual);
        assertTrue(!actual.getReferredEntities().isEmpty());
    }

    @Test
    public void tablePartition() throws IOException, ConversionException, JSONException {
        assertEntityWithFileInput("hive", "hive-table-partition-SD");
        assertTrue(this.session.getPartitionKeysSize() > 0);
    }

    @DataProvider(name = "data")
    private Object[][] comparisonQueriesProvider() {
        return new Object[][]{
                {"hive-query"},
                {"hive-query-part"},
                {"hive-query-execution"}
        };
    }

    @Test(dataProvider = "data")
    public void verify(String file) throws ConversionException, JSONException, IOException {
        assertEntityWithFileInput("hive", file);
    }
}
