/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.converter;

import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.tools.BaseTest;
import org.apache.atlas.tools.models.Session;
import org.apache.atlas.type.AtlasType;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.List;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

public class ReferenceValidatorTest extends BaseTest  {
    @Test
    public void nullEntityIsValidatedAsValid() throws IOException {
        Session session = new Session();
        ReferenceValidator referenceValidator = new ReferenceValidator(session);

        assertTrue(referenceValidator.isValid(null));
    }

    @Test
    public void noExternalReferencesAreValidated() throws IOException {
        ReferenceValidator referenceValidator = new ReferenceValidator(session);

        String json = readString("hdfs", "file-atlas");
        AtlasEntity entity = AtlasType.fromJson(json, AtlasEntity.class);

        assertTrue(referenceValidator.isValid(new AtlasEntity.AtlasEntityWithExtInfo(entity)));
    }

    @Test
    public void externallyReferencedIfNotFoundAreUpdated() throws IOException {
        ReferenceValidator referenceValidator = new ReferenceValidator(session);

        String json = readString("hive", "hive-table-partition-SD-atlas");
        AtlasEntity.AtlasEntityWithExtInfo entityExtInfo = AtlasType.fromJson(json, AtlasEntity.AtlasEntityWithExtInfo.class);

        assertNotNull(entityExtInfo.getEntity().getRelationshipAttribute("db"));
        assertTrue(referenceValidator.isValid(entityExtInfo));
        assertNotNull(entityExtInfo.getEntity().getRelationshipAttribute("db"));

        LinkedHashMap objectId = (LinkedHashMap) entityExtInfo.getEntity().getRelationshipAttribute("db");
        assertNotNull(objectId.get("uniqueAttributes"));
    }

    @Test
    public void externallyReferencedListsAreValidated() throws IOException {
        final String REF_INPUTS = "inputs";
        final String REF_OUTPUTS = "outputs";

        ReferenceValidator referenceValidator = new ReferenceValidator(session);

        String json = readString("hive", "hive-query-atlas");
        AtlasEntity.AtlasEntityWithExtInfo entityExtInfo = AtlasType.fromJson(json, AtlasEntity.AtlasEntityWithExtInfo.class);

        assertEquals(((List) entityExtInfo.getEntity().getRelationshipAttribute(REF_INPUTS)).size(), 1);
        assertEquals(((List) entityExtInfo.getEntity().getRelationshipAttribute(REF_OUTPUTS)).size(), 1);

        Object outputRef = ((List) entityExtInfo.getEntity().getRelationshipAttribute(REF_OUTPUTS)).get(0);
        session.getOrder().put((String) ((LinkedHashMap) outputRef).get("guid"), "test");

        assertTrue(referenceValidator.isValid(entityExtInfo));

        assertNotNull(entityExtInfo.getEntity().getRelationshipAttribute(REF_INPUTS));
        assertNotNull(entityExtInfo.getEntity().getRelationshipAttribute(REF_OUTPUTS));
        assertEquals(((List) entityExtInfo.getEntity().getRelationshipAttribute(REF_INPUTS)).size(), 0);
        assertEquals(((List) entityExtInfo.getEntity().getRelationshipAttribute(REF_OUTPUTS)).size(), 1);
    }
}
