/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.tools.utils;

import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import static org.testng.Assert.assertEquals;

public class MapStoringListTest {
    @Test
    public void simple() {
        MapStoringList map = new MapStoringList(new ConcurrentHashMap<>());
        map.addList("a", getList("1", "2", "3"));
        map.addList("b", getList("1", "2", "3"));
        map.addList("a", getList("4", "5", "6"));
        map.addList("a", getList("7"));

        assertEquals(map.get("a"), Arrays.asList("1", "2", "3", "4", "5", "6", "7"));
        assertEquals(map.get("b"), Arrays.asList("1", "2", "3"));
    }

    @Test
    public void concurrentAddition() throws InterruptedException {
        final int MAX_ELEMENTS_A = 1500;
        final int MAX_ELEMENTS_B = 3000;
        final int MAX_ELEMENTS_C = 4000;
        final int MAX_THREADS = 12;
        MapStoringList map = new MapStoringList(new ConcurrentHashMap<>());
        Runnable runnable = () -> {
            for (int i = 0; i < MAX_ELEMENTS_A; i++) {
                map.addList("a", getList(Integer.toString(i)));
            }

            for (int i = 0; i < MAX_ELEMENTS_B; i++) {
                map.addList("b", getList( Integer.toString(i)));
            }

            for (int i = 0; i < MAX_ELEMENTS_C; i++) {
                map.addList("c", getList(Integer.toString(i)));
            }
        };

        startJoinThread(MAX_THREADS, runnable);
        startJoinThread(MAX_THREADS, runnable);
        startJoinThread(MAX_THREADS, runnable);

        assertEquals(map.size(), (MAX_ELEMENTS_A + MAX_ELEMENTS_B + MAX_ELEMENTS_C) * MAX_THREADS * 3);
        assertEquals(new HashSet<>(map.get("a")), new HashSet<>(getStringArray(MAX_ELEMENTS_A)));
        assertEquals(new HashSet<>(map.get("b")), new HashSet<>(getStringArray(MAX_ELEMENTS_B)));
        assertEquals(new HashSet<>(map.get("c")), new HashSet<>(getStringArray(MAX_ELEMENTS_C)));
    }

    private void startJoinThread(int MAX_THREADS, Runnable runnable) throws InterruptedException {
        Thread[] thread = new Thread[MAX_THREADS];
        for (int i = 0; i < MAX_THREADS; i++) {
            thread[i] = new Thread(runnable);
            thread[i].start();
        }

        for (int i = 0; i < MAX_THREADS; i++) {
            thread[i].join();
        }
    }

    private List<String> getList(String... items) {
        return Arrays.asList(items);
    }

    private List<String> getList2(String prefix, String item) {
        return Arrays.asList(String.format("%s-%s", prefix, item));
    }

    private List<String> getStringArray(int max) {
        List<String> ret = new ArrayList<>();
        for (int i = 0; i < max; i++) {
            ret.add(Integer.toString(i));
        }

        return ret;
    }
}
